////////////////////////////////////////////////////////////////////////////////
// chaos.cpp : main

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos.h"
#include "chaos_file_serializer.h"
#include "chaos_image_realizer.h"
#include "chaos_message_imager.h"
#include "chaos_sound_manager.h"
#include "chaos_cd_player.h"
#include "chaos_wave_player.h"
#include "chaos_resource.h"
#include "chaos_exception.h"
#include "chaos_start_dlg.h"
#include "chaos_config_dlg.h"
#include "chaos_message_box.h"
#include "chaos_auto_array.h"
#include "chaos_utility.h"

#include "resource.h"

////////////////////////////////////////////////////////////////////////////////
// chaos : main class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos::chaos() :
	system(NULL),
	image_manager(NULL),
	last_time(0),
	special_mode(spmode_none),
	special_current_selection(0),
	history_offset(0),
	snapshot_page(0),
	preview_snapno(-1),
	postpone_preview_snapno_update(0),
	enable_appreciate(true),
	mouse_scroll_remainder(0),
	idling_stop(false),
	full_screen(false),
	app_quit(false),
	main_wnd(NULL),
	log_wnd(NULL)
{

#ifdef CHAOS_VIDEO_DIRECTX

	ddraw_module = NULL;
	ddraw = NULL;

#elif defined(CHAOS_VIDEO_SDL)

	screen = NULL;

#endif // CHAOS_VIDEO

#ifdef CHAOS_CDAUDIO_SDL

	audio_notify_counter = 0;

#endif // CHAOS_CDAUDIO

	system = new chaos_system;
	image_manager = new chaos_image_manager;
	snapshots.resize(snapshot_row_count, NULL);
}

// destructor

chaos::~chaos() {
	delete system;
	delete image_manager;
	clear_all_snapshots();
}

// get singleton instance

chaos* chaos::instance() {
	static chaos mainobj;
	return &mainobj;
}

////////////////////////////////////////////////////////////////////////////////
// get & set

// set arguments

void chaos::set_args(int argc, char* argv[]) {
	args.clear();
	args.reserve(argc);
	for (int i = 0; i < argc; i++) {
		args.push_back(argv[i]);
	}
}

////////////////////////////////////////////////////////////////////////////////
// main window & screen

#ifdef CHAOS_SCREEN_WIN32

// main window procedure

LRESULT CALLBACK chaos::main_wnd_proc(
	HWND wnd, UINT message, WPARAM wparam, LPARAM lparam)
{
	bool dwp = false;
	LRESULT result = 0;
	switch (message) {
	case WM_CREATE:
		instance()->set_last_time(get_ticks());
		SetTimer(wnd, 1, 10, NULL);
		chaos_sound_manager::instance()->set_main_wnd(wnd);
		break;
	case WM_DESTROY:
		if (instance()->is_full_screen()) {
			instance()->set_full_screen(false, false);
		}
		KillTimer(wnd, 1);
		if (instance()->get_log_wnd() != NULL) {
			DestroyWindow(instance()->get_log_wnd());
		}
		instance()->set_main_wnd(NULL);
		chaos_cd_player::instance()->set_notify_wnd(NULL);
		chaos_cd_player::instance()->stop();
		chaos_wave_player::instance()->stop();
		chaos_sound_manager::instance()->dispose();
		PostQuitMessage(0);
		break;
	case WM_TIMER:
		instance()->idle();
		break;
	case WM_LBUTTONDOWN:
		instance()->on_mouse_lbutton_pressed(
			(int)LOWORD(lparam), (int)HIWORD(lparam));
		break;
	case WM_RBUTTONDOWN:
		instance()->on_mouse_rbutton_pressed(
			(int)LOWORD(lparam), (int)HIWORD(lparam));
		break;
	case WM_MOUSEMOVE:
		instance()->on_mouse_moved(
			(int)LOWORD(lparam), (int)HIWORD(lparam));
		break;
	case WM_MOUSEWHEEL:
		instance()->on_mouse_scrolled(
			(int)LOWORD(lparam), (int)HIWORD(lparam),
			(short)HIWORD(wparam));
		break;
	case WM_KEYDOWN:
		switch (wparam) {
		case VK_RETURN:
			instance()->on_key_enter_pressed();
			break;
		case VK_SPACE:
			instance()->on_key_enter_pressed();
			break;
		case VK_ESCAPE:
			instance()->on_key_escape_pressed();
			break;
		case VK_BACK:
			instance()->on_key_backspace_pressed();
			break;
		case VK_UP:
			instance()->on_key_arrow_pressed(up_arrow);
			break;
		case VK_DOWN:
			instance()->on_key_arrow_pressed(down_arrow);
			break;
		case VK_LEFT:
			instance()->on_key_arrow_pressed(left_arrow);
			break;
		case VK_RIGHT:
			instance()->on_key_arrow_pressed(right_arrow);
			break;
		case VK_NEXT:
			instance()->on_key_next_pressed();
			break;
		case VK_PRIOR:
			instance()->on_key_prior_pressed();
			break;
		case VK_HOME:
			instance()->on_key_home_pressed();
			break;
		case VK_END:
			instance()->on_key_end_pressed();
			break;
		case VK_DELETE:
			instance()->on_key_delete_pressed();
			break;
		case VK_TAB:
			instance()->on_key_tab_pressed(GetKeyState(VK_SHIFT) < 0);
			break;
		case VK_CONTROL:
			instance()->on_key_control_pressed();
			break;
		}
		break;
	case WM_KEYUP:
		switch (wparam) {
		case VK_CONTROL:
			instance()->on_key_control_released();
			break;
		}
		break;
	case WM_SYSKEYDOWN:
		switch (wparam) {
		case VK_RETURN:
			instance()->set_full_screen(
				!instance()->is_full_screen(), true);
			instance()->update_screen(false);
			break;
		default:
			dwp = true;
			break;
		}
		break;
	case WM_PAINT:
		{ // dummy block
			PAINTSTRUCT ps;
			HDC dc = BeginPaint(wnd, &ps);
 			chaos_image* image =
				instance()->image_manager->get_main_image();
			if (image->is_valid()) {
				BITMAPINFO bmi;
				memset(&bmi, 0, sizeof(bmi));
				bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				bmi.bmiHeader.biWidth = image->get_width();
				bmi.bmiHeader.biHeight = -image->get_height();
				bmi.bmiHeader.biBitCount = 32;
				bmi.bmiHeader.biPlanes = 1;
				bmi.bmiHeader.biCompression = BI_RGB;
				SetDIBitsToDevice(
					dc,
					0, 0,
					image->get_width(), image->get_height(),
					0, 0,
					0, image->get_height(),
					image->get_bits(),
					&bmi, DIB_RGB_COLORS);
			}
			EndPaint(wnd, &ps);
		}
		break;
	case MM_MCINOTIFY:
		if (wparam == MCI_NOTIFY_SUCCESSFUL) {
			chaos_cd_player* cd_player = chaos_cd_player::instance();
			if (cd_player->get_cd_device_id() == (MCIDEVICEID)lparam) {
				cd_player->notify();
			}
		}
		break;
	case chaos_sound_manager::WM_SOUND_LOADED:
		((chaos_sound*)lparam)->start();
		break;
	case chaos_sound_manager::WM_SOUND_EVENT:
		((chaos_sound*)lparam)->process_event();
		break;
	default:
		dwp = true;
		break;
	}
	if (dwp) {
		result = DefWindowProc(wnd, message, wparam, lparam);
	}
	return result;
}

// create main window

HWND chaos::create_main_wnd() {
	static char classname[] = "MainWnd";
	HINSTANCE process_instance = (HINSTANCE)GetModuleHandle(NULL);
	WNDCLASSEX wc;
	memset(&wc, 0, sizeof(wc));
	wc.cbSize = sizeof(wc);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = main_wnd_proc;
	wc.hInstance = process_instance;
	wc.hIcon = LoadIcon(process_instance, MAKEINTRESOURCE(IDI_CHAOS));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.lpszClassName = classname;
	RegisterClassEx(&wc);
	chaos_dim sys_dim(system->get_sys_width(), system->get_sys_height());
	RECT wnd_rect;
	wnd_rect.left = 0;
	wnd_rect.top = 0;
	wnd_rect.right = sys_dim.width;
	wnd_rect.bottom = sys_dim.height;
	DWORD style = WS_OVERLAPPED | WS_BORDER |
		WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX |
		WS_VISIBLE;
	AdjustWindowRect(&wnd_rect, style, FALSE);
	wnd_rect.right -= wnd_rect.left;
	wnd_rect.bottom -= wnd_rect.top;
	wnd_rect.left = 0;
	wnd_rect.top = 0;
	if (system->is_window_center()) {
		RECT workarea_rect;
		SystemParametersInfo(
			SPI_GETWORKAREA, 0, &workarea_rect, 0);
		int x_ofs = (workarea_rect.right-workarea_rect.left)/2-
				(wnd_rect.right-wnd_rect.left)/2,
			y_ofs = (workarea_rect.bottom-workarea_rect.top)/2-
				(wnd_rect.bottom-wnd_rect.top)/2;
		if (x_ofs < workarea_rect.left) {
			x_ofs = workarea_rect.left;
		}
		if (y_ofs < workarea_rect.top) {
			y_ofs = workarea_rect.top;
		}
		wnd_rect.left += x_ofs;
		wnd_rect.right += x_ofs;
		wnd_rect.top += y_ofs;
		wnd_rect.bottom += y_ofs;
	}
	HWND main_wnd = CreateWindow(
		classname,
		chaos_resource::instance()->get_resource_text(
			chaos_resource::app_name).c_str(),
		style,
		wnd_rect.left, wnd_rect.top,
		wnd_rect.right-wnd_rect.left, wnd_rect.bottom-wnd_rect.top,
		NULL,
		NULL,
		process_instance,
		NULL);
	if (main_wnd != NULL) {
		set_main_wnd(main_wnd);
		chaos_cd_player::instance()->set_notify_wnd(main_wnd);
		if (system->is_full_screen()) {
			PostMessage(main_wnd, WM_SYSKEYDOWN, VK_RETURN, 0);
		}
	}
	return main_wnd;
}

// close main window

void chaos::close_main_wnd() {
	ASSERT(get_main_wnd() != NULL);
	PostMessage(get_main_wnd(), WM_CLOSE, 0, 0);
}

#elif defined(CHAOS_SCREEN_SDL)

// sdl event

void chaos::process_event(SDL_Event* ev) {
	switch (ev->type) {
	case SDL_QUIT:
		app_quit = true;
		break;
	case SDL_KEYDOWN:
		switch (ev->key.keysym.sym) {
		case SDLK_RETURN:
			if ((ev->key.keysym.mod & KMOD_ALT) == 0) {
				on_key_enter_pressed();
			} else {
				set_full_screen(!is_full_screen(), true);
				update_screen(false);
			}
			break;
		case SDLK_SPACE:
			on_key_enter_pressed();
			break;
		case SDLK_ESCAPE:
			on_key_escape_pressed();
			break;
		case SDLK_BACKSPACE:
			on_key_backspace_pressed();
			break;
		case SDLK_UP:
			on_key_arrow_pressed(up_arrow);
			break;
		case SDLK_DOWN:
			on_key_arrow_pressed(down_arrow);
			break;
		case SDLK_LEFT:
			on_key_arrow_pressed(left_arrow);
			break;
		case SDLK_RIGHT:
			on_key_arrow_pressed(right_arrow);
			break;
		case SDLK_PAGEDOWN:
			on_key_next_pressed();
			break;
		case SDLK_PAGEUP:
			on_key_prior_pressed();
			break;
		case SDLK_HOME:
			on_key_home_pressed();
			break;
		case SDLK_END:
			on_key_end_pressed();
			break;
		case SDLK_DELETE:
			on_key_delete_pressed();
			break;
		case SDLK_TAB:
			on_key_tab_pressed(
				(ev->key.keysym.mod & KMOD_SHIFT) != 0);
			break;
		case SDLK_LCTRL:
		case SDLK_RCTRL:
			on_key_control_pressed();
			break;
		}
		break;
	case SDL_KEYUP:
		switch (ev->key.keysym.sym) {
		case SDLK_LCTRL:
		case SDLK_RCTRL:
			on_key_control_released();
			break;
		}
		break;
	case SDL_MOUSEBUTTONDOWN:
		switch (ev->button.button) {
		case SDL_BUTTON_LEFT:
			on_mouse_lbutton_pressed(
				ev->button.x, ev->button.y);
			break;
		case SDL_BUTTON_RIGHT:
			on_mouse_rbutton_pressed(
				ev->button.x, ev->button.y);
			break;
		case 4:	// wheel down
			on_mouse_scrolled(
				ev->button.x, ev->button.y,
				(int)scroll_delta);
			break;
		case 5:	// wheel up
			on_mouse_scrolled(
				ev->button.x, ev->button.y,
				-(int)scroll_delta);
			break;
		}
		break;
	case SDL_MOUSEMOTION:
		on_mouse_moved(
			ev->motion.x, ev->motion.y);
		break;
	}
}

// clear events

void chaos::clear_events() {
	SDL_Event ev;
	while (SDL_PollEvent(&ev)) {
	}
}

// create screen

bool chaos::create_screen() {
	chaos_dim sys_dim(system->get_sys_width(), system->get_sys_height());
	full_screen = true;
	set_full_screen(false, false);
	if (screen == NULL) {
		return false;
	}
	SDL_WM_SetCaption(
		chaos_resource::instance()->get_resource_text(
			chaos_resource::app_name).c_str(),
		NULL);
	if (system->is_full_screen()) {
		set_full_screen(true, false);
	}
	return true;
}

#endif // CHAOS_SCREEN

// update screen(multi rect)

void chaos::update_screen(
	const std::list<chaos_rect>& updates,
	bool update_now)
{

#ifdef CHAOS_SCREEN_WIN32

	RECT invalidate_rect;
	for (
		std::list<chaos_rect>::const_iterator it = updates.begin();
		it != updates.end();
		it++)
	{
		invalidate_rect.left = (*it).get_left();
		invalidate_rect.top = (*it).get_top();
		invalidate_rect.right = (*it).get_right();
		invalidate_rect.bottom = (*it).get_bottom();
		InvalidateRect(get_main_wnd(), &invalidate_rect, TRUE);
	}
	if (update_now) {
		UpdateWindow(get_main_wnd());
	}

#elif defined(CHAOS_SCREEN_SDL)

	ASSERT(screen != NULL);
	if (SDL_MUSTLOCK(screen)) {
		if (SDL_LockSurface(screen) < 0) {
			return;
		}
	}
	chaos_image* main_image = image_manager->get_main_image();
	SDL_PixelFormat* screen_format = screen->format;
	chaos_auto_array<SDL_Rect> draws(new SDL_Rect[updates.size()]);
	int draw_count = 0;
	for (
		std::list<chaos_rect>::const_iterator it = updates.begin();
		it != updates.end();
		it++)
	{
		chaos_rect update_rect = *it;
		if (update_rect.empty()) {
			continue;
		}
		draws[draw_count].x = (Sint16)update_rect.x;
		draws[draw_count].y = (Sint16)update_rect.y;
		draws[draw_count].w = (Uint16)update_rect.width;
		draws[draw_count].h = (Uint16)update_rect.height;
		draw_count++;
		chaos_uint32* src = main_image->get_bits(
			update_rect.x, update_rect.y);
		int src_step = main_image->get_scanline_length()-update_rect.width;
		chaos_uint8* dst = (chaos_uint8*)screen->pixels+
			update_rect.y*screen->pitch+
			update_rect.x*sizeof(chaos_uint32);
		int dst_step = screen->pitch;
		if (screen_format->BytesPerPixel != sizeof(chaos_uint32)) {
			// unknown format
		} else if (
			(screen_format->Rloss != 0) ||
			(screen_format->Gloss != 0) ||
			(screen_format->Bloss != 0) ||
			(screen_format->Rshift != CHAOS_RGBA_RED_SHIFT) ||
			(screen_format->Gshift != CHAOS_RGBA_GREEN_SHIFT) ||
			(screen_format->Bshift != CHAOS_RGBA_BLUE_SHIFT))
		{
			int y = update_rect.height;
			do {
				chaos_uint32* dst2 = (chaos_uint32*)dst;
				int x = update_rect.width;
				do {
					chaos_uint32 col = *(src++);
					*(dst2++) = SDL_MapRGB(
						screen_format,
						(chaos_uint8)CHAOS_RGBA_RED(col),
						(chaos_uint8)CHAOS_RGBA_GREEN(col),
						(chaos_uint8)CHAOS_RGBA_BLUE(col));
				} while (--x != 0);
				src += src_step;
				dst += dst_step;
			} while (--y != 0);
		} else {
			int copys = update_rect.width*sizeof(chaos_uint32);
			src_step += update_rect.width;
			if (((int)(src_step*sizeof(chaos_uint32)) == copys) &&
				(dst_step == copys))
			{
				memcpy(dst, src, copys*update_rect.height);
			} else {
				int y = update_rect.height;
				do {
					memcpy(dst, src, copys);
					src += src_step;
					dst += dst_step;
				} while (--y != 0);
			}
		}
	}
	if (SDL_MUSTLOCK(screen)) {
		SDL_UnlockSurface(screen);
	}
	if (draw_count > 0) {
		SDL_UpdateRects(screen, draw_count, draws.get());
	}
	update_now;

#endif // CHAOS_SCREEN

}

// update screen(single rect)

void chaos::update_screen(
	const chaos_rect& update_rect,
	bool update_now)
{
	std::list<chaos_rect> updates;
	updates.push_back(update_rect);
	update_screen(updates, update_now);
}

// update screen(all)

void chaos::update_screen(
	bool update_now)
{
	std::list<chaos_rect> updates;
	updates.push_back(
		chaos_rect(
			0, 0,
			system->get_sys_width(), system->get_sys_height()));
	update_screen(updates, update_now);
}

// is full screen

bool chaos::is_full_screen() const {
	return full_screen;
}

// set full screen

bool chaos::set_full_screen(bool new_full_screen, bool prop_set) {
	if (new_full_screen == is_full_screen()) {
		return true;
	}
	chaos_dim sys_dim(system->get_sys_width(), system->get_sys_height());
	bool success = true;

#ifdef CHAOS_VIDEO_DIRECTX

	ASSERT(get_main_wnd() != NULL);
	RECT wnd_rect;
	DWORD style;
	if (new_full_screen) {
		wnd_rect.left = 0;
		wnd_rect.top = 0;
		wnd_rect.right = sys_dim.width;
		wnd_rect.bottom = sys_dim.height;
		style = GetWindowLong(get_main_wnd(), GWL_STYLE) &
			~(WS_BORDER | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX);
		AdjustWindowRect(&wnd_rect, style, FALSE);
		SetWindowLong(get_main_wnd(), GWL_STYLE, style);
		SetWindowPos(
			get_main_wnd(), NULL,
			0, 0,
			wnd_rect.right-wnd_rect.left, wnd_rect.bottom-wnd_rect.top,
			SWP_NOZORDER | SWP_NOACTIVATE);
	}
	try {
		if (new_full_screen) {
			if (ddraw != NULL) {
				return true;
			}
			if (ddraw_module == NULL) {
				ddraw_module = LoadLibrary("ddraw.dll");
			}
			if (ddraw_module == NULL) {
				throw chaos_exception(chaos_exception::generic);
			}
			typedef HRESULT WINAPI DirectDrawCreateProc(
				GUID*, IDirectDraw**, IUnknown*);
			DirectDrawCreateProc* pDirectDrawCreate =
				(DirectDrawCreateProc*)
					GetProcAddress(ddraw_module, "DirectDrawCreate");
			if (pDirectDrawCreate == NULL) {
				throw chaos_exception(chaos_exception::generic);
			}
			HRESULT result = pDirectDrawCreate(NULL, &ddraw, NULL);
			if (result != DD_OK) {
				throw chaos_exception(chaos_exception::generic);
			}
			result = ddraw->SetCooperativeLevel(
				get_main_wnd(),
				DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE | DDSCL_ALLOWREBOOT);
			if (result != DD_OK) {
				throw chaos_exception(chaos_exception::generic);
			}
			result = ddraw->SetDisplayMode(
				sys_dim.width, sys_dim.height, 32);
			if (result != DD_OK) {
				result = ddraw->SetDisplayMode(
					sys_dim.width, sys_dim.height, 24);
			}
			if (result != DD_OK) {
				result = ddraw->SetDisplayMode(
					sys_dim.width, sys_dim.height, 16);
			}
			if (result != DD_OK) {
				throw chaos_exception(chaos_exception::generic);
			}
			full_screen = true;
		} else {
			if (ddraw == NULL) {
				return true;
			}
			HRESULT result;
			if (ddraw != NULL) {
				result = ddraw->SetCooperativeLevel(
					NULL,
					DDSCL_NORMAL);
				if (result != DD_OK) {
					throw chaos_exception(chaos_exception::generic);
				}
				result = ddraw->RestoreDisplayMode();
				if (result != DD_OK) {
					throw chaos_exception(chaos_exception::generic);
				}
				ddraw->Release();
				ddraw = NULL;
			}
			full_screen = false;
		}
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : main::full screen - "
			"%s\n",
			e.what());
		if (ddraw != NULL) {
			ddraw->Release();
			ddraw = NULL;
		}
		full_screen = false;
		success = false;
	}
	if (!is_full_screen()) {
		wnd_rect.left = 0;
		wnd_rect.top = 0;
		wnd_rect.right = sys_dim.width;
		wnd_rect.bottom = sys_dim.height;
		style = GetWindowLong(get_main_wnd(), GWL_STYLE) |
			WS_BORDER | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;
		AdjustWindowRect(&wnd_rect, style, FALSE);
		wnd_rect.right -= wnd_rect.left;
		wnd_rect.bottom -= wnd_rect.top;
		wnd_rect.left = 0;
		wnd_rect.top = 0;
		if (system->is_window_center()) {
			RECT workarea_rect;
			SystemParametersInfo(
				SPI_GETWORKAREA, 0, &workarea_rect, 0);
			int x_ofs = (workarea_rect.right-workarea_rect.left)/2-
					(wnd_rect.right-wnd_rect.left)/2,
				y_ofs = (workarea_rect.bottom-workarea_rect.top)/2-
					(wnd_rect.bottom-wnd_rect.top)/2;
			if (x_ofs < workarea_rect.left) {
				x_ofs = workarea_rect.left;
			}
			if (y_ofs < workarea_rect.top) {
				y_ofs = workarea_rect.top;
			}
			wnd_rect.left += x_ofs;
			wnd_rect.right += x_ofs;
			wnd_rect.top += y_ofs;
			wnd_rect.bottom += y_ofs;
		}
		SetWindowLong(get_main_wnd(), GWL_STYLE, style);
		SetWindowPos(
			get_main_wnd(), NULL,
			wnd_rect.left, wnd_rect.top,
			wnd_rect.right-wnd_rect.left, wnd_rect.bottom-wnd_rect.top,
			SWP_NOZORDER | SWP_NOACTIVATE);
	}

#elif defined(CHAOS_VIDEO_SDL)

	bool hardware_surface = false;
	if (SDL_VideoModeOK(
			sys_dim.width, sys_dim.height, 32,
			(new_full_screen? SDL_FULLSCREEN: 0) |
			SDL_HWSURFACE) == 32)
	{
		hardware_surface = true;
	}
	screen = SDL_SetVideoMode(
		sys_dim.width, sys_dim.height, 32,
		(new_full_screen? SDL_FULLSCREEN: 0) |
			(hardware_surface? SDL_HWSURFACE: SDL_SWSURFACE));
	if (screen == NULL) {
		LOG_ERR(
			"error : main::full screen - "
			"set video mode\n");
		exit(0);
	}
	if (hardware_surface) {
		SDL_PixelFormat* screen_format = screen->format;
		if ((screen_format->BytesPerPixel != sizeof(chaos_uint32)) ||
			(screen_format->Rloss != 0) ||
			(screen_format->Gloss != 0) ||
			(screen_format->Bloss != 0) ||
			(screen_format->Rshift != CHAOS_RGBA_RED_SHIFT) ||
			(screen_format->Gshift != CHAOS_RGBA_GREEN_SHIFT) ||
			(screen_format->Bshift != CHAOS_RGBA_BLUE_SHIFT))
		{
			LOG_WRN(
				"warning : main::full screen - "
				"bad hardware surface format\n");
			screen = SDL_SetVideoMode(
				sys_dim.width, sys_dim.height, 32,
				(new_full_screen? SDL_FULLSCREEN: 0) |
					SDL_SWSURFACE);
			if (screen == NULL) {
				LOG_ERR(
					"error : main::full screen - "
					"set video mode\n");
				exit(0);
			}
		}
	}
	full_screen = new_full_screen;

#endif // CHAOS_VIDEO

	if (prop_set &&
		success &&
		(new_full_screen != system->is_full_screen()))
	{
		system->set_full_screen(new_full_screen);
	}
	return success;
}

////////////////////////////////////////////////////////////////////////////////
// log window

#ifdef CHAOS_GUI_WIN32

// log window procedure

LRESULT CALLBACK chaos::log_wnd_proc(
	HWND wnd, UINT message, WPARAM wparam, LPARAM lparam)
{
	bool dwp = false;
	LRESULT result = 0;
	switch (message) {
	case WM_CREATE:
		CreateWindow(
			"RICHEDIT",
			"",
			WS_CHILD | WS_VISIBLE | WS_VSCROLL |
				ES_MULTILINE | ES_AUTOVSCROLL,
			0, 0, 0, 0,
			wnd,
			(HMENU)1,
			(HINSTANCE)GetWindowLong(wnd, GWL_HINSTANCE),
			NULL);
		break;
	case WM_DESTROY:
		instance()->set_log_wnd(NULL);
		chaos_log::instance()->set_log_wnd(NULL);
		break;
	case WM_SIZE:
		{ // dummy block
			HWND rich_edit = GetDlgItem(wnd, 1);
			if (rich_edit != NULL) {
				SetWindowPos(
					rich_edit, NULL,
					0, 0, LOWORD(lparam), HIWORD(lparam),
					SWP_NOZORDER | SWP_NOACTIVATE);
			}
		}
		break;
	case WM_SETFOCUS:
		{ // dummy block
			HWND rich_edit = GetDlgItem(wnd, 1);
			if (rich_edit != NULL) {
				SetFocus(rich_edit);
			}
		}
		break;
	case WM_LOGOUT:
		{ // dummy block
			HWND rich_edit = GetDlgItem(wnd, 1);
			if (rich_edit != NULL) {
				int length = GetWindowTextLength(rich_edit);
				SendMessage(rich_edit, EM_SETSEL, length, length);
				SendMessage(rich_edit, EM_REPLACESEL, 0, lparam);
				SendMessage(rich_edit, EM_SCROLLCARET, 0, 0);
			}
		}
		break;
	default:
		dwp = true;
		break;
	}
	if (dwp) {
		result = DefWindowProc(wnd, message, wparam, lparam);
	}
	return result;
}

// create log window

HWND chaos::create_log_wnd() {
	enum {
		WND_WIDTH  = 400,
		WND_HEIGHT = 300
	};
	static char classname[] = "LogWnd";
	LoadLibrary("riched32.dll");
	HINSTANCE process_instance = (HINSTANCE)GetModuleHandle(NULL);
	WNDCLASSEX wc;
	memset(&wc, 0, sizeof(wc));
	wc.cbSize = sizeof(wc);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = log_wnd_proc;
	wc.hInstance = process_instance;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)COLOR_WINDOW;
	wc.lpszClassName = classname;
	RegisterClassEx(&wc);
	HWND log_wnd = CreateWindow(
		classname,
		"Log",
		WS_OVERLAPPED | WS_THICKFRAME |
			WS_CAPTION | WS_SYSMENU |
			WS_MINIMIZEBOX | WS_MAXIMIZEBOX,
		0, 0,
		GetSystemMetrics(SM_CXSCREEN)/2, GetSystemMetrics(SM_CYSCREEN)/2,
		NULL,
		NULL,
		process_instance,
		NULL);
	if (log_wnd != NULL) {
		set_log_wnd(log_wnd);
		chaos_log::instance()->set_log_wnd(log_wnd);
		ShowWindow(log_wnd, SW_SHOWMINNOACTIVE);
		HMENU sys_menu = GetSystemMenu(log_wnd, FALSE);
		if (sys_menu != NULL) {
			EnableMenuItem(
				sys_menu, SC_CLOSE,
				MF_BYCOMMAND | MF_DISABLED);
		}
	}
	return log_wnd;
}

// close log window

void chaos::close_log_wnd() {
	ASSERT(get_log_wnd() != NULL);
	PostMessage(get_log_wnd(), WM_CLOSE, 0, 0);
}

#endif // CHAOS_GUI_WIN32

////////////////////////////////////////////////////////////////////////////////
// special mode

// change special mode

void chaos::change_special_mode(special_mode_type new_mode) {
	special_mode_type old_mode = get_special_mode();
	set_special_mode(new_mode);
	system->set_dirty_image(true);
	system->set_dirty_message(true);
	system->change_skip_mode(chaos_system::skip_none);
	special_current_selection = -1;
	preview_snapno = -1;
	postpone_preview_snapno_update = 0;
	chaos_string font_name;
	if (new_mode == spmode_none) {

#ifdef CHAOS_FONT_WIN32

		font_name = system->get_font_name();

#elif defined(CHAOS_FONT_GTK)

		font_name = system->get_font_xlfd();

#elif defined(CHAOS_FONT_PANGO)

		font_name = system->get_font_pango();

#endif // CHAOS_FONT

		image_manager->change_font(
			font_name,
			system->get_font_size(),
			system->is_font_bold(),
			system->is_font_anti_alias());
	} else {

#ifdef CHAOS_FONT_WIN32

		font_name = system->get_sys_font_name();

#elif defined(CHAOS_FONT_GTK)

		font_name = system->get_sys_font_xlfd();

#elif defined(CHAOS_FONT_PANGO)

		font_name = system->get_sys_font_pango();

#endif // CHAOS_FONT

		image_manager->change_font(
			font_name,
			system->get_sys_medium_font_size(),
			system->is_sys_font_bold(),
			system->is_sys_font_anti_alias());
	}
	image_manager->clear_message_image();
	switch (old_mode) {
	case spmode_save:
	case spmode_load:
		clear_all_snapshots();
		break;
	}
	switch (new_mode) {
	case spmode_main_menu:
		{ // dummy block
			chaos_string oinfo = system->get_image_info(0),
				rinfo = remove_message_window_info(oinfo);
			enable_appreciate = (rinfo != oinfo);
		}
		break;
	case spmode_recollect:
		system->flush_next_message_text();
		history_offset = 0;
		break;
	case spmode_save:
	case spmode_load:
		load_all_snapshots();
		break;
	}
}

////////////////////////////////////////////////////////////////////////////////
// selection

// get selection count

int chaos::get_selection_count() const {
	int count = 0;
	switch (get_special_mode()) {
	case spmode_none:
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_selmessage:
		case chaos_system::sysmode_selevent:
			count = system->get_selection_count();
			break;
		}
		break;
	case spmode_main_menu:
		count = chaos_resource::instance()->get_main_menu_count();
		break;
	case spmode_save:
	case spmode_load:
		count = (int)snapshot_row_count+(int)snapshot_page_count;
		break;
	case spmode_recollect:
		count = (int)recollect_row+2;
		break;
	case spmode_save_confirm:
	case spmode_load_confirm:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		break;
	}
	return count;
}

// get current selection

int chaos::get_current_selection() const {
	int selno = -1;
	switch (get_special_mode()) {
	case spmode_none:
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_selmessage:
		case chaos_system::sysmode_selevent:
			selno = system->get_current_selection();
			break;
		}
		break;
	case spmode_main_menu:
	case spmode_recollect:
	case spmode_save:
	case spmode_load:
	case spmode_save_confirm:
	case spmode_load_confirm:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		selno = special_current_selection;
		break;
	}
	return selno;
}

// set current selection

void chaos::set_current_selection(int new_selno) {
	int old_selno = get_current_selection();
	if (new_selno == old_selno) {
		return;
	}
	switch (get_special_mode()) {
	case spmode_none:
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_selmessage:
		case chaos_system::sysmode_selevent:
			system->set_current_selection(new_selno);
			break;
		}
		break;
	case spmode_main_menu:
	case spmode_recollect:
	case spmode_save_confirm:
	case spmode_load_confirm:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		special_current_selection = new_selno;
		break;
	case spmode_save:
	case spmode_load:
		special_current_selection = new_selno;
		if (is_snapshot_valid(old_selno) ||
			is_snapshot_valid(new_selno))
		{
			postpone_preview_snapno_update = 100;
		}
		break;
	}
	if (system->is_dirty_image() ||
		system->is_dirty_message())
	{
		system->set_dirty_image(true);
	} else {
		std::list<chaos_rect> updates;
		chaos_rect old_sel_rect = get_selection_rect(old_selno),
			new_sel_rect = get_selection_rect(new_selno);
		if (!old_sel_rect.empty()) {
			draw_image(NULL, &old_sel_rect);
			updates.push_back(old_sel_rect);
		}
		if (!new_sel_rect.empty() && (new_sel_rect != old_sel_rect)) {
			draw_image(NULL, &new_sel_rect);
			updates.push_back(new_sel_rect);
		}
		update_screen(updates, true);
	}
}

// is enable selection

bool chaos::is_enable_selection(int selno) const {
	bool enable = true;
	switch (get_special_mode()) {
	case spmode_none:
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_selmessage:
		case chaos_system::sysmode_selevent:
			enable = true;
			break;
		}
		break;
	case spmode_main_menu:
		switch (chaos_resource::instance()->get_main_menu_item(selno)) {
		case chaos_resource::main_menu_skip_read:
			if (!system->is_enable_skip_mode(chaos_system::skip_read)) {
				enable = false;
			}
			break;
		case chaos_resource::main_menu_appreciate:
			enable = enable_appreciate;
			break;
		case chaos_resource::main_menu_save:
			if (!system->is_enable_save()) {
				enable = false;
			}
			break;
		case chaos_resource::main_menu_load:
			if (!system->is_enable_load()) {
				enable = false;
			}
			break;
		case chaos_resource::main_menu_recollect:
			if (system->get_history_count() <= 0) {
				enable = false;
			}
			break;
		}
		break;
	case spmode_load:
		if ((selno >= 0) &&
			(selno < (int)snapshot_row_count))
		{
			enable = is_snapshot_valid(selno);
		}
		break;
	case spmode_recollect:
		if (selno < (int)recollect_row) {
			enable = (history_offset+selno < system->get_history_count());
		} else if (selno == (int)recollect_row) {
			enable = (history_offset > 0);
		} else {
			enable = (history_offset+(int)recollect_row <
				system->get_history_count());
		}
		break;
	}
	return enable;
}

// get selection rect

chaos_rect chaos::get_selection_rect(int selno) const {
	chaos_rect sel_rect;
	chaos_dim sys_dim(system->get_sys_width(), system->get_sys_height());
	switch (get_special_mode()) {
	case spmode_none:
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_selmessage:
		case chaos_system::sysmode_selevent:
			if ((selno >= 0) ||
				(selno == (int)chaos_system::select_invalid))
			{
				int selno2 = selno;
				if (selno2 < 0) {
					selno2 = 0;
				}
				chaos_string info = system->get_selection_info(selno2);
				int types,
					x1, y1, x2, y2,
					unk1, unk2, unk3;
				chaos_color sel_color;
				if (chaos_system::decode_selection_info(
						info,
						types,
						x1, y1, x2, y2,
						unk1, unk2, unk3,
						sel_color))
				{
					sel_rect = chaos_rect(x1, y1, x2-x1, y2-y1);
				}
			}
			break;
		}
		break;
	case spmode_main_menu:
		if (selno >= 0) {
			int count = get_selection_count();
			int font_size = system->get_sys_medium_font_size();
			sel_rect = chaos_rect(
				sys_dim.width/4,
				(sys_dim.height-count*font_size)/2+selno*font_size,
				sys_dim.width/2,
				font_size);
		}
		break;
	case spmode_save:
	case spmode_load:
		if ((selno >= 0) &&
			(selno < (int)snapshot_row_count))
		{
			int font_size = system->get_sys_small_font_size();
			sel_rect = chaos_rect(
				font_size/2,
				(selno+8)*font_size,
				font_size*14,
				font_size);
		} else if (
			(selno >= (int)snapshot_row_count) &&
			(selno < (int)snapshot_row_count+(int)snapshot_page_count))
		{
			int page = selno-(int)snapshot_row_count;
			int font_size = system->get_sys_small_font_size();
			int left = (sys_dim.width*page)/(int)snapshot_page_count+2,
				right = (sys_dim.width*(page+1))/(int)snapshot_page_count-2;
			sel_rect = chaos_rect(
				left,
				font_size*4,
				right-left,
				font_size+font_size/2);
		}
		break;
	case spmode_recollect:
		if (selno >= 0) {
			int font_size = system->get_sys_small_font_size();
			int recollect_line =
				sys_dim.height/((int)recollect_row*font_size);
			if (selno < (int)recollect_row) {
				sel_rect = chaos_rect(
					font_size,
					((int)recollect_row-selno-1)*recollect_line*
						font_size,
					sys_dim.width-font_size*2,
					font_size*recollect_line);
			} else if (selno == (int)recollect_row) {
				sel_rect = chaos_rect(
					sys_dim.width-font_size,
					((int)recollect_row*recollect_line-1)*
						font_size,
					font_size,
					font_size);
			} else {
				sel_rect = chaos_rect(
					sys_dim.width-font_size,
					0,
					font_size,
					font_size);
			}
		}
		break;
	case spmode_save_confirm:
	case spmode_load_confirm:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		break;
	}
	if (!is_enable_selection(selno)) {
		sel_rect.clear();
	}
	return sel_rect;
}

// get selection color

chaos_color chaos::get_selection_color(int selno) const {
	chaos_color sel_color;
	switch (get_special_mode()) {
	case spmode_none:
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_selmessage:
		case chaos_system::sysmode_selevent:
			if (selno >= 0) {
				chaos_string info = system->get_selection_info(selno);
				int types,
					x1, y1, x2, y2,
					unk1, unk2, unk3;
				if (chaos_system::decode_selection_info(
						info,
						types,
						x1, y1, x2, y2,
						unk1, unk2, unk3,
						sel_color))
				{
				}
			} else if (selno == (int)chaos_system::select_invalid) {
				sel_color = chaos_color(48, 48, 48);
			}
			break;
		}
		break;
	case spmode_main_menu:
		if (selno >= 0) {
			sel_color = chaos_color(0, 32, 96);
		}
		break;
	case spmode_save:
	case spmode_load:
		if (selno >= 0) {
			sel_color = chaos_color(64, 64, 64);
		}
		break;
	case spmode_recollect:
		if (selno >= 0) {
			sel_color = chaos_color(96, 32, 0);
		}
		break;
	case spmode_save_confirm:
	case spmode_load_confirm:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		break;
	}
	return sel_color;
}

// get hit selection

int chaos::get_hit_selection(chaos_pos pos) const {
	int count = get_selection_count();
	int selno;
	for (selno = 0; selno < count; selno++) {
		chaos_rect sel_rect = get_selection_rect(selno);
		if ((pos.x >= sel_rect.get_left()) &&
			(pos.x < sel_rect.get_right()) &&
			(pos.y >= sel_rect.get_top()) &&
			(pos.y < sel_rect.get_bottom()))
		{
			break;
		}
	}
	if (selno >= count) {
		selno = -1;
	}
	return selno;
}

////////////////////////////////////////////////////////////////////////////////
// background

// remove message window info

chaos_string chaos::remove_message_window_info(
	const chaos_string& oinfo) const
{
	chaos_string rinfo = oinfo;
	int count = system->get_message_window_info_count();
	for (int i = 0; i < count; i++) {
		rinfo = chaos_image_realizer::remove_special_info(
			rinfo,
			system->get_message_window_info(i));
	}
	return rinfo;
}

// get background info

chaos_string chaos::get_background_info() const {
	chaos_string info = system->get_image_info(0);
	chaos_dim sys_dim(system->get_sys_width(), system->get_sys_height());
	bool hide_message = false,
		dark = false,
		load_save = false;
	chaos_color color;
	switch (get_special_mode()) {
	case spmode_main_menu:
	case spmode_recollect:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		hide_message = true;
		dark = true;
		break;
	case spmode_appreciate:
		hide_message = true;
		break;
	case spmode_save:
	case spmode_load:
		info = "";
		if (is_snapshot_valid(preview_snapno)) {
			info = snapshots[preview_snapno]->get_values()->
				get_value("image_info0").get_string();
		}
		hide_message = true;
		load_save = true;
		if (get_special_mode() == spmode_save) {
			color = chaos_color(128, 64, 64);
		} else {
			color = chaos_color(64, 64, 128);
		}
		break;
	case spmode_save_confirm:
	case spmode_load_confirm:
		break;
	}
	if (hide_message) {
		info = remove_message_window_info(info);
	}
	if (dark) {
		info = chaos_image_realizer::encode_fade_fill_info(
			128,
			0, 0, sys_dim.width, sys_dim.height,
			chaos_color(0, 0, 0),
			info);
	}
	if (load_save) {
		chaos_color hilight = chaos_color(128, 128, 128),
			shadow = chaos_color(32, 32, 32),
			gray = chaos_color(
				(color.get_red()+96)/2,
				(color.get_green()+96)/2,
				(color.get_blue()+96)/2);
		int font_size = system->get_sys_small_font_size();
		chaos_string info2;
		info2 = chaos_image_realizer::encode_fill_clr_info(
			color);
		int top = font_size*4,
			bottom = top+font_size+font_size/2;
		info2 = chaos_image_realizer::encode_fill_info(
			0, 0, sys_dim.width, bottom,
			(color-chaos_color(32, 32, 32)).normalize(),
			info2);
		info2 = chaos_image_realizer::encode_fill_info(
			0, bottom, sys_dim.width, 1,
			hilight,
			info2);
		for (int page = 0; page < (int)snapshot_page_count; page++) {
			int left = (sys_dim.width*page)/(int)snapshot_page_count+2,
				right = (sys_dim.width*(page+1))/(int)snapshot_page_count-2;
			chaos_rect tag_rect(left, top, right-left, bottom-top);
			chaos_color tag_color;
			if (page == snapshot_page) {
				tag_color = color;
				tag_rect.height++;
			} else {
				tag_color = gray;
			}
			info2 = chaos_image_realizer::encode_fill_info(
				tag_rect.x-1, tag_rect.y-1,
				tag_rect.width+2, tag_rect.height+1,
				hilight,
				info2);
			info2 = chaos_image_realizer::encode_fill_info(
				tag_rect.x+tag_rect.width, tag_rect.y,
				1, tag_rect.height,
				shadow,
				info2);
			info2 = chaos_image_realizer::encode_fill_info(
				tag_rect.x, tag_rect.y,
				tag_rect.width, tag_rect.height,
				tag_color,
				info2);
		}
		chaos_rect pv_rect(
			sys_dim.width/2-font_size, font_size*8,
			sys_dim.width/2, sys_dim.height/2);
		info = chaos_image_realizer::encode_stretch_blt_info(
			pv_rect.x, pv_rect.y,
			pv_rect.width, pv_rect.height,
			info2,
			info);
	}
	return info;
}

////////////////////////////////////////////////////////////////////////////////
// message

// get message text

chaos_string chaos::get_message_text() const {
	chaos_string message_text;
	chaos_dim sys_dim(system->get_sys_width(), system->get_sys_height());
	switch (get_special_mode()) {
	case spmode_none:
		{ // dummy block
			int font_size = system->get_font_size();
			int shape = system->get_message_shape();
			if (shape != 3) {
				message_text = chaos_message_imager::encode_shape_info(
					shape);
			}
			chaos_color subcolor = system->get_message_subcolor();
			if (subcolor != chaos_color(0, 0, 0)) {
				message_text += chaos_message_imager::encode_subcolor_info(
					subcolor);
			}
			int count = system->get_message_frame_count();
			for (int frame = 0; frame < count; frame++) {
				chaos_string text = system->get_message_text(frame);
				if (text.empty()) {
					continue;
				}
				chaos_string info = system->get_message_info(frame);
				int x, y,
					length;
				chaos_color color;
				chaos_system::decode_message_info(
					info,
					x, y,
					length,
					color);
				message_text += chaos_message_imager::encode_position_info(
					x+font_size, y);
				message_text += chaos_message_imager::encode_length_info(
					length);
				message_text += chaos_message_imager::encode_defcolor_info(
					color);
				message_text += text;
			}
		}
		break;
	case spmode_main_menu:
		{ // dummy block
			int font_size = system->get_sys_medium_font_size();
			message_text = chaos_message_imager::encode_font_size_info(
				font_size);
			int count = chaos_resource::instance()->get_main_menu_count();
			for (int frame = 0; frame < count; frame++) {
				chaos_string menu_text =
					chaos_resource::instance()->get_main_menu_text(frame);
				message_text += chaos_message_imager::encode_position_info(
					(sys_dim.width-(menu_text.length()*font_size)/2)/2,
					(sys_dim.height-count*font_size)/2+frame*font_size);
				message_text += chaos_message_imager::encode_curcolor_info(
					(chaos_color)(
						is_enable_selection(frame)?
							chaos_color(255, 255, 255):
							chaos_color(192, 192, 192)));
				message_text += menu_text;
			}
			font_size = system->get_sys_small_font_size();
			chaos_string version_text =
				chaos_resource::instance()->get_resource_text(
						chaos_resource::app_name)+
				" Ver. "+
				chaos_resource::instance()->get_resource_text(
						chaos_resource::app_version);
			message_text += chaos_message_imager::encode_font_size_info(
				font_size);
			message_text += chaos_message_imager::encode_position_info(
				sys_dim.width-((version_text.length()+1)*font_size)/2,
				sys_dim.height-(font_size+font_size/2));
			message_text += chaos_message_imager::encode_curcolor_info(
				chaos_color(128, 192, 255));
			message_text += version_text;
		}
		break;
	case spmode_appreciate:
		break;
	case spmode_save:
	case spmode_load:
		{ // dummy block
			int font_size = system->get_sys_large_font_size();
			chaos_string text;
			chaos_color color;
			if (get_special_mode() == spmode_save) {
				text = chaos_resource::instance()->get_resource_text(
					chaos_resource::main_menu_save);
				color = chaos_color(128, 64, 64);
			} else {
				text = chaos_resource::instance()->get_resource_text(
					chaos_resource::main_menu_load);
				color = chaos_color(64, 64, 128);
			}
			color += chaos_color(128, 128, 128);
			message_text =
				chaos_message_imager::encode_shape_info(0);
			message_text +=
				chaos_message_imager::encode_font_size_info(
					font_size);
			message_text += 
				chaos_message_imager::encode_position_info(
					font_size/2, font_size/2);
			message_text += chaos_message_imager::encode_curcolor_info(
				color);
			message_text += text;
			font_size = system->get_sys_small_font_size();
			message_text +=
				chaos_message_imager::encode_shape_info(3);
			message_text +=
				chaos_message_imager::encode_font_size_info(
					font_size);
			chaos_color white(255, 255, 255),
				gray(128, 128, 128);
			message_text += chaos_message_imager::encode_curcolor_info(
				white);
			for (int snapno = 0; snapno < (int)snapshot_row_count; snapno++) {
				message_text +=
					chaos_message_imager::encode_position_info(
						font_size,
						(snapno+8)*font_size);
				color = white;
				if (is_snapshot_valid(snapno)) {
					text = snapshots[snapno]->get_date()+
						" "+
						snapshots[snapno]->get_time();
				} else {
					text = "-------------------";
					if (get_special_mode() == spmode_load) {
						color = gray;
					}
				}
				message_text +=
					chaos_utility::format_str(
						"%3d: ",
						snapshot_page*(int)snapshot_row_count+
							snapno+1);
				if (color != white) {
					message_text +=
						chaos_message_imager::encode_curcolor_info(
							color);
				}
				message_text += text;
				if (color != white) {
					message_text +=
						chaos_message_imager::encode_curcolor_info(
							white);
				}
			}
			for (int page = 0; page < (int)snapshot_page_count; page++) {
				text = chaos_utility::format_str("%d", page+1);
				message_text +=
					chaos_message_imager::encode_position_info(
						(sys_dim.width*(page*2+1))/
							((int)snapshot_page_count*2)-
								(text.length()*font_size)/4,
						font_size*4+font_size/4);
				message_text += text;
			}
			if (is_snapshot_valid(preview_snapno)) {
				text = snapshots[preview_snapno]->get_values()->
					get_value("next_message_text").
						get_string();
				message_text += chaos_message_imager::encode_length_info(
					sys_dim.width/font_size-4);
				message_text +=
					chaos_message_imager::encode_position_info(
						font_size,
						font_size*25);
				message_text += text;
			}
		}
		break;
	case spmode_recollect:
		{ // dummy block
			int font_size = system->get_sys_small_font_size();
			int recollect_line =
				sys_dim.height/((int)recollect_row*font_size);
			message_text = chaos_message_imager::encode_font_size_info(
				font_size);
			message_text += chaos_message_imager::encode_length_info(
				sys_dim.width/font_size-4);
			int count = (int)recollect_row+2;
			for (int frame = 0; frame < count; frame++) {
				if (frame < count-2) {
					message_text +=
						chaos_message_imager::encode_position_info(
							font_size*2,
							((int)recollect_row-frame-1)*recollect_line*
								font_size);
					message_text += system->get_history_message_text(
						history_offset+frame);
				} else {
					message_text +=
						chaos_message_imager::encode_curcolor_info(
							(chaos_color)(
								is_enable_selection(frame)?
									chaos_color(255, 255, 128):
									chaos_color(128, 128, 128)));
					message_text +=
						chaos_message_imager::encode_position_info(
							sys_dim.width-font_size,
							(frame == count-2)?
								(sys_dim.height-font_size): 0);
					message_text +=
						chaos_resource::instance()->get_resource_text(
							(frame == count-2)?
								chaos_resource::recollect_pagedown:
								chaos_resource::recollect_pageup);
				}
			}
		}
		break;
	case spmode_save_confirm:
	case spmode_load_confirm:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		break;
	}
	return message_text;
}

////////////////////////////////////////////////////////////////////////////////
// image

// draw image

void chaos::draw_image(
	chaos_rect* update_rect,
	const chaos_rect* clip_rect)
{
	chaos_string info = get_background_info();
	int selno = get_current_selection();
	chaos_rect sel_rect = get_selection_rect(selno);
	chaos_color sel_color = get_selection_color(selno);
	bool prompt_visible = false, prompt_place = false;
	int prompt_x = 0, prompt_y = 0;
	switch (get_special_mode()) {
	case spmode_none:
		prompt_visible = system->is_prompt_visible();
		if (prompt_visible) {
			chaos_string pinfo = system->get_prompt_info();
			chaos_system::decode_prompt_info(
				pinfo,
				prompt_place,
				prompt_x, prompt_y);
		}
		break;
	}
	image_manager->draw_main_image(
		info,
		&sel_rect,
		sel_color,
		prompt_visible, prompt_place,
		chaos_pos(prompt_x, prompt_y),
		update_rect,
		clip_rect);
}

////////////////////////////////////////////////////////////////////////////////
// fix

// set fix values

void chaos::set_fixes() const {
	chaos_value_manager<chaos_string> fixes;
	chaos_resource::instance()->get_fixes(&fixes);
	for (
		chaos_value_manager<chaos_string>::iterator
			it = fixes.begin();
		it != fixes.end();
		it++)
	{
		system->set_fix((*it).first, (*it).second);
	}
}

////////////////////////////////////////////////////////////////////////////////
// xml

// set xml date

void chaos::set_xml_date_time(chaos_xml* xml) {
	int year, month, day, hour, minute, second;

#ifdef CHAOS_PLATFORM_WIN32

	SYSTEMTIME systime;
	GetLocalTime(&systime);
	year = systime.wYear;
	month = systime.wMonth;
	day = systime.wDay;
	hour = systime.wHour;
	minute = systime.wMinute;
	second = systime.wSecond;

#elif defined(CHAOS_PLATFORM_UNIX)

	time_t now;
	time(&now);
	struct tm* now_time = localtime(&now);
	year = now_time->tm_year+1900;
	month = now_time->tm_mon+1;
	day = now_time->tm_mday;
	hour = now_time->tm_hour;
	minute = now_time->tm_min;
	second = now_time->tm_sec;

#endif // CHAOS_PLATFORM

	xml->set_date(
		chaos_utility::format_str(
			"%04d-%02d-%02d",
			year, month, day));
	xml->set_time(
		chaos_utility::format_str(
			"%02d:%02d:%02d",
			hour, minute, second));
}

// set xml info

void chaos::set_xml_info(chaos_xml* xml) {
	chaos_string info = chaos_file_serializer::instance()->get_pack_info();
	xml->set_info(info);
}

////////////////////////////////////////////////////////////////////////////////
// snapshot

// get snapshot filename

chaos_string chaos::get_snapshot_filename(int datano) {
	ASSERT(
		(datano >= 1) &&
		(datano <= (int)snapshot_page_count*(int)snapshot_row_count));
	return chaos_utility::format_str(
		"chaos_sv%03d.dat", datano);
}

// is snapshot valid

bool chaos::is_snapshot_valid(int snapno) const {
	return (snapno >= 0) &&
		(snapno < (int)snapshot_row_count) &&
		(snapshots[snapno] != NULL);
}

// load snapshot from file

bool chaos::load_snapshot_from_file(
	const chaos_string& filename,
	chaos_xml* snapshot) const
{
	bool exist = chaos_file_serializer::instance()->exist_file(
		filename,
		chaos_file_serializer::fpos_chaos_dir);
	if (!exist) {
		return false;
	}
	chaos_file_data filedata;
	if (!chaos_file_serializer::instance()->load_file_data(
			&filedata,
			filename,
			chaos_file_serializer::fpos_chaos_dir) ||
		!snapshot->load(&filedata))
	{
		return false;
	}
	return true;
}

// save snapshot to file

bool chaos::save_snapshot_to_file(
	const chaos_string& filename,
	chaos_xml* snapshot) const
{
	set_xml_date_time(snapshot);
	set_xml_info(snapshot);
	chaos_file_data filedata;
	if (!snapshot->save(&filedata)) {
		return false;
	}
	if (!chaos_file_serializer::instance()->save_file_data(
			&filedata,
			filename,
			chaos_file_serializer::fpos_chaos_dir))
	{
		return false;
	}
	return true;
}

// load snapshots

void chaos::load_all_snapshots() {
	ASSERT(
		(snapshot_page >= 0) &&
		(snapshot_page < (int)snapshot_page_count));
	clear_all_snapshots();
	for (int snapno = 0; snapno < (int)snapshot_row_count; snapno++) {
		chaos_string filename = get_snapshot_filename(
			snapshot_page*(int)snapshot_row_count+snapno+1);
		if (chaos_file_serializer::instance()->exist_file(
				filename,
				chaos_file_serializer::fpos_chaos_dir))
		{
			snapshots[snapno] = new chaos_xml;
			if (!load_snapshot_from_file(
					filename,
					snapshots[snapno]))
			{
				delete snapshots[snapno];
				snapshots[snapno] = NULL;
			}
		}
	}
}

// clear all snapshots

void chaos::clear_all_snapshots() {
	for (int snapno = 0; snapno < (int)snapshot_row_count; snapno++) {
		if (snapshots[snapno] != NULL) {
			delete snapshots[snapno];
			snapshots[snapno] = NULL;
		}
	}
}

// update preview snapno

void chaos::update_preview_snapno() {
	int old_snapno = preview_snapno;
	preview_snapno = -1;
	int selno = get_current_selection();
	switch (get_special_mode()) {
	case spmode_save:
	case spmode_load:
		if ((selno >= 0) &&
			(selno < (int)snapshot_row_count))
		{
			preview_snapno = selno;
		}
		break;
	}
	if ((preview_snapno != old_snapno) &&
		(is_snapshot_valid(old_snapno) ||
			is_snapshot_valid(preview_snapno)))
	{
		system->set_dirty_image(true);
		system->set_dirty_message(true);
	}
}

////////////////////////////////////////////////////////////////////////////////
// profile

// get profile filename

chaos_string chaos::get_profile_filename() {
	return "chaos_profile.dat";
}

// load profile from file

bool chaos::load_profile_from_file(
	const chaos_string& filename,
	chaos_xml* profile) const
{
	bool exist = chaos_file_serializer::instance()->exist_file(
		filename,
		chaos_file_serializer::fpos_chaos_dir);
	if (exist) {
		chaos_file_data filedata;
		if (!chaos_file_serializer::instance()->load_file_data(
				&filedata,
				filename,
				chaos_file_serializer::fpos_chaos_dir) ||
			!profile->load(&filedata))
		{
			exist = false;
		}
	}
	if (!exist) {
		if (!system->save_profile(profile)) {
			return false;
		}
	}
	return true;
}

// save profile to file

bool chaos::save_profile_to_file(
	const chaos_string& filename,
	chaos_xml* profile) const
{
	set_xml_date_time(profile);
	set_xml_info(profile);
	chaos_file_data filedata;
	if (!profile->save(&filedata)) {
		return false;
	}
	if (!chaos_file_serializer::instance()->save_file_data(
			&filedata,
			filename,
			chaos_file_serializer::fpos_chaos_dir))
	{
		return false;
	}
	return true;
}

// set default profile

bool chaos::set_default_profile(
	chaos_xml* profile) const
{
	chaos_value_manager<chaos_string>* prof_values = profile->get_values();
	chaos_value_manager<chaos_string> sysproperties;
	chaos_resource::instance()->get_def_sysproperties(&sysproperties);
	bool dirty = false;
	for (
		chaos_value_manager<chaos_string>::iterator
			it = sysproperties.begin();
		it != sysproperties.end();
		it++)
	{
		if (!prof_values->exist_value((*it).first)) {
			prof_values->set_value((*it).first, (*it).second);
			dirty = true;
		}
	}
	return dirty;
}

// config profile

void chaos::config_profile(
	const chaos_xml* profile)
{
	chaos_system::log_mode_type old_log_mode = system->get_log_mode();
	chaos_string old_bgm_sound_type = system->get_bgm_sound_type(),
		old_cd_drive = system->get_cd_drive();
	system->load_profile(profile);
	image_manager->change_image_buffer_count(
		system->get_sys_image_buffer_count());
	if (system->is_full_screen() != is_full_screen()) {
		set_full_screen(system->is_full_screen(), false);
		update_screen(false);
	}
	chaos_system::log_mode_type new_log_mode = system->get_log_mode();
	if (new_log_mode != old_log_mode) {
		if (new_log_mode != chaos_system::logmode_window) {

#ifdef CHAOS_GUI_WIN32

			close_log_wnd();

#else // CHAOS_GUI_WIN32

			chaos_log::instance()->set_output_stderr(false);

#endif // CHAOS_GUI_WIN32

		}
		if (new_log_mode == chaos_system::logmode_window) {

#ifdef CHAOS_GUI_WIN32

			create_log_wnd();

#else // CHAOS_GUI_WIN32

			chaos_log::instance()->set_output_stderr(true);

#endif // CHAOS_GUI_WIN32

		}
	}
	chaos_log::instance()->set_log_level(
		(chaos_log::log_level_type)system->get_log_level());
	chaos_string new_bgm_sound_type = system->get_bgm_sound_type(),
		new_cd_drive = system->get_cd_drive();
	if ((new_bgm_sound_type != old_bgm_sound_type) ||
		(new_cd_drive != old_cd_drive))
	{
		chaos_cd_player::instance()->open_match_device(
			new_bgm_sound_type,
			new_cd_drive,
			system->get_cd_info());
		system->play_cd_audio(system->get_cd_audio_info());
	}
	chaos_cd_player::instance()->set_sound_volume(
		system->get_bgm_sound_volume());
	chaos_wave_player::instance()->set_sound_volume(
		system->get_effect_sound_volume());
}

////////////////////////////////////////////////////////////////////////////////
// readflag

// get readflag filename

chaos_string chaos::get_readflag_filename() {
	return "chaos_readflag.dat";
}

// load readflag from file

bool chaos::load_readflag_from_file(
	const chaos_string& filename,
	chaos_xml* readflag) const
{
	bool exist = chaos_file_serializer::instance()->exist_file(
		filename,
		chaos_file_serializer::fpos_chaos_dir);
	if (exist) {
		chaos_file_data filedata;
		if (!chaos_file_serializer::instance()->load_file_data(
				&filedata,
				filename,
				chaos_file_serializer::fpos_chaos_dir) ||
			!readflag->load(&filedata))
		{
			exist = false;
		}
	}
	if (exist) {
		if (!system->load_readflag(readflag)) {
			return false;
		}
	}
	return true;
}

// save readflag to file

bool chaos::save_readflag_to_file(
	const chaos_string& filename,
	chaos_xml* readflag) const
{
	set_xml_date_time(readflag);
	set_xml_info(readflag);
	chaos_file_data filedata;
	if (!readflag->save(&filedata)) {
		return false;
	}
	if (!chaos_file_serializer::instance()->save_file_data(
			&filedata,
			filename,
			chaos_file_serializer::fpos_chaos_dir))
	{
		return false;
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// user input

// on mouse left-button pressed

void chaos::on_mouse_lbutton_pressed(int x, int y) {
	bool enter = false, selcheck = false;
	switch (get_special_mode()) {
	case spmode_none:
		if (system->get_skip_mode() != chaos_system::skip_none) {
			system->change_skip_mode(
				chaos_system::skip_none);
		} else {
			switch (system->get_system_mode()) {
			case chaos_system::sysmode_message:
			case chaos_system::sysmode_userprompt:
			case chaos_system::sysmode_userclick:
				enter = true;
				break;
			case chaos_system::sysmode_selmessage:
			case chaos_system::sysmode_selevent:
				selcheck = true;
				break;
			}
		}
		break;
	case spmode_appreciate:
		enter = true;
		break;
	case spmode_main_menu:
	case spmode_save:
	case spmode_load:
	case spmode_recollect:
	case spmode_save_confirm:
	case spmode_load_confirm:
	case spmode_rewind_confirm:
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		selcheck = true;
		break;
	}
	if (selcheck) {
		int selno = get_hit_selection(chaos_pos(x, y));
		if ((selno >= 0) && (selno == get_current_selection())) {
			enter = true;
		}
	}
	if (enter) {
		on_key_enter_pressed();
	}
}

// on mouse right-button pressed

void chaos::on_mouse_rbutton_pressed(int /*x*/, int /*y*/) {
	on_key_escape_pressed();
}

// on mouse moved

void chaos::on_mouse_moved(int x, int y) {
	int count = get_selection_count();
	if (count > 0) {
		int selno = get_hit_selection(chaos_pos(x, y));
		if ((selno >= 0) &&
			(selno != get_current_selection()))
		{
			set_current_selection(selno);
		}
	}
}

// on mouse scrolled

void chaos::on_mouse_scrolled(int /*x*/, int /*y*/, int scroll) {
	mouse_scroll_remainder += scroll;
	int move = abs(mouse_scroll_remainder)/(int)scroll_delta;
	if (mouse_scroll_remainder < 0) {
		move = -move;
	}
	if (move == 0) {
		return;
	}
	mouse_scroll_remainder -= move*(int)scroll_delta;
	switch (get_special_mode()) {
	case spmode_none:
		if (system->get_skip_mode() == chaos_system::skip_none) {
			if (move < 0) {
				switch (system->get_system_mode()) {
				case chaos_system::sysmode_message:
					system->change_skip_mode(
						chaos_system::skip_cur_message);
					break;
				case chaos_system::sysmode_userprompt:
					system->change_system_mode(
						chaos_system::sysmode_interpret);
					break;
				}
			} else {
				if (system->get_history_count() > 0) {
					change_special_mode(spmode_recollect);
				}
			}
		}
		break;
	case spmode_recollect:
		{ // dummy block
			int count = system->get_history_count();
			if (count > (int)recollect_row) {
				int new_history_offset = history_offset+move;
				if (new_history_offset < 0) {
					new_history_offset = 0;
				} else if (new_history_offset > count-(int)recollect_row) {
					new_history_offset = count-(int)recollect_row;
				}
				if (new_history_offset != history_offset) {
					history_offset = new_history_offset;
					system->set_dirty_message(true);
					int selno = get_current_selection();
					if ((selno >= 0) && !is_enable_selection(selno)) {
						set_current_selection(-1);
					}
				}
			}
		}
		break;
	}
}

// on keyboard enter pressed

void chaos::on_key_enter_pressed() {
	int selno = get_current_selection();
	switch (get_special_mode()) {
	case spmode_none:
		if (system->get_skip_mode() != chaos_system::skip_none) {
			system->change_skip_mode(
				chaos_system::skip_none);
		} else {
			switch (system->get_system_mode()) {
			case chaos_system::sysmode_message:
				system->change_skip_mode(
					chaos_system::skip_cur_message);
				break;
			case chaos_system::sysmode_userprompt:
			case chaos_system::sysmode_userclick:
				system->change_system_mode(
					chaos_system::sysmode_interpret);
				break;
			case chaos_system::sysmode_selmessage:
			case chaos_system::sysmode_selevent:
				if (selno >= 0) {
					system->change_system_mode(
						chaos_system::sysmode_interpret);
				}
				break;
			}
		}
		break;
	case spmode_main_menu:
		switch (chaos_resource::instance()->get_main_menu_item(selno)) {
		case chaos_resource::main_menu_skip_read:
			change_special_mode(spmode_none);
			system->change_skip_mode(chaos_system::skip_read);
			break;
		case chaos_resource::main_menu_appreciate:
			change_special_mode(spmode_appreciate);
			break;
		case chaos_resource::main_menu_save:
			change_special_mode(spmode_save);
			break;
		case chaos_resource::main_menu_load:
			change_special_mode(spmode_load);
			break;
		case chaos_resource::main_menu_recollect:
			if (system->get_history_count() > 0) {
				change_special_mode(spmode_recollect);
			}
			break;
		case chaos_resource::main_menu_config:
			{ // dummy block
				idling_stop = true;
				chaos_xml profile;
				system->save_profile(&profile);
				chaos_config_dlg* dlg = chaos_config_dlg::instance();
				dlg->set_profile(&profile);

#ifdef CHAOS_SCREEN_SDL

				if (is_full_screen()) {
					set_full_screen(false, false);
					update_screen(false);
				}

#endif // CHAOS_SCREEN

				if (dlg->do_modal(get_main_wnd()) &&
					dlg->is_dirty())
				{
					config_profile(&profile);
					save_profile_to_file(
						get_profile_filename(),
						&profile);
					system->set_dirty_profile(false);
				} else {

#ifdef CHAOS_SCREEN_SDL

				if (system->is_full_screen()) {
					set_full_screen(true, false);
					update_screen(false);
				}

#endif // CHAOS_SCREEN

				}
				dlg->set_profile(NULL);
				idling_stop = false;
				set_last_time(get_ticks());
				change_special_mode(spmode_none);

#ifdef CHAOS_SCREEN_SDL

				clear_events();

#endif // CHAOS_SCREEN

			}
			break;
		case chaos_resource::main_menu_restart:
			system->restart();
			change_special_mode(spmode_none);
			break;
		case chaos_resource::main_menu_exit:
			close_app();
			break;
		}
		break;
	case spmode_appreciate:
		change_special_mode(spmode_none);
		break;
	case spmode_save:
	case spmode_load:
		if ((selno >= 0) &&
			(selno < (int)snapshot_row_count))
		{
			if (postpone_preview_snapno_update > 0) {
				postpone_preview_snapno_update = 0;
				update_preview_snapno();
				system->set_dirty_image(false);
				system->set_dirty_message(false);
				chaos_string message_text = get_message_text();
				image_manager->update_message_image(
					message_text,
					NULL);
				draw_image(NULL, NULL);
				update_screen(true);
			}
			if (get_special_mode() == spmode_save) {
				if (snapshots[selno] != NULL) {
					snapshots[selno]->clear();
				} else {
					snapshots[selno] = new chaos_xml;
				}
				if (system->save_snapshot(snapshots[selno])) {
					save_snapshot_to_file(
						get_snapshot_filename(
							snapshot_page*(int)snapshot_row_count+selno+1),
						snapshots[selno]);
				}
			} else {
				if (snapshots[selno] != NULL) {
					system->load_snapshot(snapshots[selno]);
				}
			}
			if (system->is_dirty_readflag()) {
				system->set_dirty_readflag(false);
				chaos_xml readflag;
				if (system->save_readflag(&readflag)) {
					save_readflag_to_file(
						get_readflag_filename(),
						&readflag);
				}
			}
			change_special_mode(spmode_none);
		} else if (
			(selno >= (int)snapshot_row_count) &&
			(selno < (int)snapshot_row_count+(int)snapshot_page_count))
		{
			int page = selno-(int)snapshot_row_count;
			if (page != snapshot_page) {
				snapshot_page = page;
				preview_snapno = -1;
				postpone_preview_snapno_update = 0;
				load_all_snapshots();
				system->set_dirty_image(true);
				system->set_dirty_message(true);
			}
		}
		break;
	case spmode_recollect:
		if (selno >= 0) {
			int count = system->get_history_count();
			if (selno < (int)recollect_row) {
				int step = history_offset+selno;
				system->rewind_history(step);
				change_special_mode(spmode_none);
			} else if (count > (int)recollect_row) {
				int new_history_offset = history_offset+
					((selno == (int)recollect_row)?
						(-(int)recollect_row):
						(int)recollect_row);
				if (new_history_offset < 0) {
					new_history_offset = 0;
				} else if (new_history_offset > count-(int)recollect_row) {
					new_history_offset = count-(int)recollect_row;
				}
				if (new_history_offset != history_offset) {
					history_offset = new_history_offset;
					system->set_dirty_message(true);
					if (!is_enable_selection(selno)) {
						set_current_selection(-1);
					}
				}
			}
		}
		break;
	}
}

// on keyboard escape pressed

void chaos::on_key_escape_pressed() {
	switch (get_special_mode()) {
	case spmode_none:
		if (system->get_skip_mode() != chaos_system::skip_none) {
			system->change_skip_mode(
				chaos_system::skip_none);
		} else {
			if (system->is_dirty_profile()) {
				system->set_dirty_profile(false);
				chaos_xml profile;
				if (system->save_profile(&profile)) {
					save_profile_to_file(
						get_profile_filename(),
						&profile);
				}
			}
			change_special_mode(spmode_main_menu);
		}
		break;
	case spmode_main_menu:
	case spmode_appreciate:
	case spmode_recollect:
		change_special_mode(spmode_none);
		break;
	case spmode_save:
	case spmode_load:
		change_special_mode(spmode_none);
		clear_all_snapshots();
		break;
	case spmode_restart_confirm:
	case spmode_exit_confirm:
		change_special_mode(spmode_main_menu);
		break;
	case spmode_save_confirm:
		change_special_mode(spmode_save);
		break;
	case spmode_load_confirm:
		change_special_mode(spmode_load);
		break;
	case spmode_rewind_confirm:
		change_special_mode(spmode_recollect);
		break;
	}
}

// on keyboard backspace pressed

void chaos::on_key_backspace_pressed() {
	switch (get_special_mode()) {
	case spmode_none:
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_userprompt:
		case chaos_system::sysmode_selmessage:
			system->rewind_history(1);
			break;
		}
		break;
	}
}

// on keyboard arrow pressed

void chaos::on_key_arrow_pressed(arrow_type arrow) {
	int count = get_selection_count();
	if (count <= 0) {
		return;
	}
	int old_selno = get_current_selection(),
		new_selno = -1;
	chaos_rect old_rect;
	if (old_selno >= 0) {
		old_rect = get_selection_rect(old_selno);
	}
	int new_distance = -1;
	for (int selno = 0; selno < count; selno++) {
		if ((selno == old_selno) ||
			!is_enable_selection(selno))
		{
			continue;
		}
		if (old_selno < 0) {
			new_selno = selno;
			break;
		}
		chaos_rect sel_rect = get_selection_rect(selno);
		bool moveable = false;
		switch (arrow) {
		case up_arrow:
			if (sel_rect.y+sel_rect.height <= old_rect.y) {
				moveable = true;
			}
			break;
		case down_arrow:
			if (sel_rect.y >= old_rect.y+old_rect.height) {
				moveable = true;
			}
			break;
		case left_arrow:
			if (sel_rect.x+sel_rect.width <= old_rect.x) {
				moveable = true;
			}
			break;
		case right_arrow:
			if (sel_rect.x >= old_rect.x+old_rect.width) {
				moveable = true;
			}
			break;
		}
		if (moveable) {
			int distance =
					abs((sel_rect.x+sel_rect.width/2)-
						(old_rect.x+old_rect.width/2))+
					abs((sel_rect.y+sel_rect.height/2)-
						(old_rect.y+old_rect.height/2));
			if ((new_selno < 0) ||
				(distance < new_distance))
			{
				new_selno = selno;
				new_distance = distance;
			}
		}
	}
	if (new_selno >= 0) {
		switch (get_special_mode()) {
		case spmode_save:
		case spmode_load:
			if ((old_selno < (int)snapshot_row_count) &&
				(new_selno >= (int)snapshot_row_count))
			{
				new_selno = (int)snapshot_row_count+snapshot_page;
			}
			break;
		}
		if (new_selno != old_selno) {
			set_current_selection(new_selno);
		}
	} else {
		switch (get_special_mode()) {
		case spmode_recollect:
			{ // dummy block
				int count = system->get_history_count();
				if ((count > (int)recollect_row) &&
					((arrow == up_arrow) || (arrow == down_arrow)))
				{
					history_offset += (arrow == down_arrow)?
						(-1): 1;
					if (history_offset < 0) {
						history_offset = 0;
					} else if (history_offset > count-(int)recollect_row) {
						history_offset = count-(int)recollect_row;
					}
					system->set_dirty_message(true);
					if (!is_enable_selection(old_selno)) {
						set_current_selection(-1);
					}
				}
			}
			break;
		}
	}
}

// on keyboard next pressed

void chaos::on_key_next_pressed() {
	switch (get_special_mode()) {
	case spmode_save:
	case spmode_load:
		if (snapshot_page < (int)snapshot_page_count-1) {
			snapshot_page++;
			preview_snapno = -1;
			postpone_preview_snapno_update = 0;
			load_all_snapshots();
			system->set_dirty_image(true);
			system->set_dirty_message(true);
			set_current_selection((int)snapshot_row_count+snapshot_page);
		}
		break;
	case spmode_recollect:
		if (is_enable_selection((int)recollect_row)) {
			set_current_selection((int)recollect_row);
			on_key_enter_pressed();
		}
		break;
	}
}

// on keyboard prior pressed

void chaos::on_key_prior_pressed() {
	switch (get_special_mode()) {
	case spmode_save:
	case spmode_load:
		if (snapshot_page > 0) {
			snapshot_page--;
			preview_snapno = -1;
			postpone_preview_snapno_update = 0;
			load_all_snapshots();
			system->set_dirty_image(true);
			system->set_dirty_message(true);
			set_current_selection((int)snapshot_row_count+snapshot_page);
		}
		break;
	case spmode_recollect:
		if (is_enable_selection((int)recollect_row+1)) {
			set_current_selection((int)recollect_row+1);
			on_key_enter_pressed();
		}
		break;
	}
}

// on keyboard home pressed

void chaos::on_key_home_pressed() {
	int count = get_selection_count();
	if (count <= 0) {
		return;
	}
	switch (get_special_mode()) {
	case spmode_save:
	case spmode_load:
		if (count > (int)snapshot_row_count) {
			count = (int)snapshot_row_count;
		}
		break;
	case spmode_recollect:
		if (count > (int)recollect_row) {
			count = (int)recollect_row;
		}
		break;
	}
	int old_selno = get_current_selection(),
		new_selno;
	for (new_selno = 0; new_selno < count; new_selno++) {
		if (is_enable_selection(new_selno)) {
			break;
		}
	}
	if ((new_selno < count) && (new_selno != old_selno)) {
		set_current_selection(new_selno);
	}
}

// on keyboard end pressed

void chaos::on_key_end_pressed() {
	int count = get_selection_count();
	switch (get_special_mode()) {
	case spmode_save:
	case spmode_load:
		if (count > (int)snapshot_row_count) {
			count = (int)snapshot_row_count;
		}
		break;
	case spmode_recollect:
		if (count > (int)recollect_row) {
			count = (int)recollect_row;
		}
		break;
	}
	if (count <= 0) {
		return;
	}
	int old_selno = get_current_selection(),
		new_selno;
	for (new_selno = count-1; new_selno >= 0; new_selno--) {
		if (is_enable_selection(new_selno)) {
			break;
		}
	}
	if ((new_selno >= 0) && (new_selno != old_selno)) {
		set_current_selection(new_selno);
	}
}

// on keyboard delete pressed

void chaos::on_key_delete_pressed() {
	switch (get_special_mode()) {
	case spmode_none:
		{ // dummy block
			chaos_string oinfo = system->get_image_info(0),
				rinfo = remove_message_window_info(oinfo);
			if (rinfo != oinfo) {
				change_special_mode(spmode_appreciate);
			}
		}
		break;
	case spmode_appreciate:
		change_special_mode(spmode_none);
		break;
	}
}

// on keyboard tab pressed

void chaos::on_key_tab_pressed(bool shift) {
	bool sel_move = false;
	switch (get_special_mode()) {
	case spmode_none:
		if (system->get_skip_mode() != chaos_system::skip_none) {
			system->change_skip_mode(chaos_system::skip_none);
		} else if (system->is_enable_skip_mode(chaos_system::skip_read)) {
			system->change_skip_mode(chaos_system::skip_read);
		} else {
			sel_move = true;
		}
		break;
	default:
		sel_move = true;
		break;
	}
	if (!sel_move) {
		return;
	}
	int dir = shift? -1: 1;
	int count = get_selection_count();
	if (count <= 0) {
		return;
	}
	int old_selno = get_current_selection(),
		new_selno,
		start_selno,
		end_selno;
	if (old_selno >= 0) {
		if (count <= 1) {
			return;
		}
		start_selno = (old_selno+count+dir)%count;
		end_selno = old_selno;
	} else if (dir > 0) {
		start_selno = 0;
		end_selno = count;
	} else {
		start_selno = count-1;
		end_selno = -1;
	}
	for (
		new_selno = start_selno;
		new_selno != end_selno;
		new_selno = (new_selno+count+dir)%count)
	{
		if (is_enable_selection(new_selno)) {
			break;
		}
	}
	if ((new_selno != end_selno) && (new_selno != old_selno)) {
		set_current_selection(new_selno);
	}
}

// on keyboard control pressed

void chaos::on_key_control_pressed() {
	switch (get_special_mode()) {
	case spmode_none:
		if (system->get_skip_mode() == chaos_system::skip_none) {
			system->change_skip_mode(chaos_system::skip_force);
		}
		break;
	}
}

// on keyboard control released

void chaos::on_key_control_released() {
	switch (get_special_mode()) {
	case spmode_none:
		system->change_skip_mode(chaos_system::skip_none);
		break;
	}
}

////////////////////////////////////////////////////////////////////////////////
// execute

// initialize

bool chaos::initialize() {
	// environment setting

#ifdef CHAOS_PLATFORM_WIN32

	CoInitialize(NULL);
	timeBeginPeriod(1);

#endif // CHAOS_PLATFORM_WIN32

#ifdef CHAOS_GUI_WIN32

	InitCommonControls();

#endif // CHAOS_GUI_WIN32

#ifdef CHAOS_API_SDL

	Uint32 sdl_flags = 0;

#ifdef CHAOS_VIDEO_SDL

	sdl_flags |= SDL_INIT_VIDEO;

#endif // CHAOS_VIDEO_SDL

#ifdef CHAOS_SOUND_SDL

	sdl_flags |= SDL_INIT_AUDIO;

#endif // CHAOS_SOUND

#ifdef CHAOS_CDAUDIO_SDL

	sdl_flags |= SDL_INIT_CDROM;

#endif // CHAOS_CDAUDIO_SDL

	SDL_Init(sdl_flags);

#endif // CHAOS_API_SDL

	// program directory
	chaos_string program_dir;
	{ // dummy block

#ifdef CHAOS_PLATFORM_WIN32

		char module_path[_MAX_PATH];
		memset(module_path, 0, sizeof(module_path));
		GetModuleFileName(GetModuleHandle(NULL), module_path, _MAX_PATH);
		char module_drive[_MAX_DRIVE], module_dir[_MAX_DIR];
		_splitpath(module_path, module_drive, module_dir, NULL, NULL);
		program_dir = module_drive;
		program_dir += module_dir;

#elif defined(CHAOS_PLATFORM_UNIX)

		char start_dir[_POSIX_PATH_MAX+1],
			module_dir[_POSIX_PATH_MAX+1];
		if (getcwd(start_dir, sizeof(start_dir)) == NULL) {
			strcpy(start_dir, ".");
		}
		if (args.size() > 0) {
			strcpy(module_dir, args[0].c_str());
		} else {
			module_dir[0] = '\0';
		}
		char* separator = strrchr(module_dir, '/');
		if (separator != NULL) {
			*separator = '\0';
			chdir(module_dir);
			if (getcwd(module_dir, sizeof(module_dir)) == NULL) {
				strcpy(module_dir, ".");
			}
			chdir(start_dir);
		} else {
			strcpy(module_dir, start_dir);
		}
		program_dir = chaos_utility::filename_to_sjis(module_dir)+"/";

#endif // CHAOS_PLATFORM_UNIX

	}
	// parse args
	chaos_file_serializer* file_serializer =
		chaos_file_serializer::instance();
	file_serializer->set_base_dir(program_dir);
	file_serializer->set_chaos_dir(program_dir);
	bool fource_startup = false;
	for (size_t i = 1; i < args.size(); i++) {
		bool base_dir = false,
			chaos_dir = false;
		chaos_string arg = chaos_utility::locale_to_sjis(args[i]);
		if (chaos_utility::icompare(arg, "--base_dir") == 0) {
			base_dir = true;
		} else if (chaos_utility::icompare(arg, "--chaos_dir") == 0) {
			chaos_dir = true;
		} else if (chaos_utility::icompare(arg, "--startup") == 0) {
			fource_startup = true;
		}
		if (base_dir || chaos_dir) {
			if (i+1 >= args.size()) {
				continue;
			}
			i++;
			chaos_string dir = chaos_utility::locale_to_sjis(args[i]);
			if (dir.size() > 2) {
				char front = dir[0],
					back = dir[dir.length()-1];
				if (((front == '"') && (back == '"')) ||
					((front == '\'') && (back == '\'')))
				{
					dir = dir.substr(1, dir.length()-2);
				}
			}
			if (!dir.empty()) {
				if (chaos_utility::is_str_sjistrail(
						dir, dir.length()-1) ||
					(dir[dir.length()-1] !=
						chaos_file_serializer::get_dir_separator()))
				{
					dir += chaos_file_serializer::get_dir_separator();
				}
				if (base_dir) {
					file_serializer->set_base_dir(dir);
				}
				if (chaos_dir) {
					file_serializer->set_chaos_dir(dir);
				}
			}
		}
	}
	// pack data block
	for (int packno = 1; true; packno++) {
		chaos_string pack_index;
		if (packno == 1) {
			pack_index = "Lump.cpn";
		} else {
			pack_index = chaos_utility::format_str("Lump%d.cpn", packno);
		}
		if (!file_serializer->exist_file(
				pack_index,
				chaos_file_serializer::fpos_base_dir))
		{
			break;
		}
		if (!file_serializer->add_pack_block(pack_index)) {
			throw chaos_exception(chaos_exception::generic);
		}
	}
	// unpacked data
	if (file_serializer->get_pack_block_count() <= 0) {
		if (!file_serializer->add_pack_block(
			chaos_string("unpack")+
				chaos_file_serializer::get_dir_separator()))
		{
			throw chaos_exception(chaos_exception::generic);
		}
		if (!file_serializer->add_pack_block(
			chaos_string("unpack2")+
				chaos_file_serializer::get_dir_separator()))
		{
			throw chaos_exception(chaos_exception::generic);
		}
	}
	// fix
	set_fixes();
	// profile
	bool dirty_profile;
	chaos_xml profile;
	if (!load_profile_from_file(
			get_profile_filename(),
			&profile))
	{
		throw chaos_exception(chaos_exception::generic);
	}
	dirty_profile = set_default_profile(&profile);
	// start dialog
	bool show_startup =
			profile.get_values()->get_value("show_startup").get_boolean();
	if (show_startup || fource_startup) {
		idling_stop = true;
		chaos_config_dlg* dlg = chaos_config_dlg::instance();
		dlg->set_profile(&profile);
		chaos_start_dlg* start_dlg =
			chaos_start_dlg::instance();
		if (!start_dlg->do_modal()) {
			return false;
		}
		dirty_profile |= start_dlg->is_dirty();
		dlg->set_profile(NULL);
		idling_stop = false;
	}
	system->load_profile(&profile);
	if (dirty_profile) {
		system->set_dirty_profile(dirty_profile);
	}
	// log
	chaos_string log_filename =
		chaos_file_serializer::instance()->get_base_dir()+
			"chaos.log";
	chaos_log::instance()->set_log_filename(log_filename.c_str());
	chaos_log::instance()->set_log_level(
		(chaos_log::log_level_type)
			system->get_log_level());
	// start script check
	bool exist_script = false;
	if (file_serializer->get_pack_block_count() > 0) {
		exist_script = file_serializer->exist_file(
			chaos_file_serializer::get_pure_filename(
				system->get_start_script_filename()),
			chaos_file_serializer::fpos_packed);
	}
	if (!exist_script) {
		chaos_message_box::show_message(
			chaos_resource::instance()->get_resource_text(
				chaos_resource::err_cant_find_script));
		return false;
	}
	// readflag
	{ // dummy block
		chaos_xml readflag;
		if (!load_readflag_from_file(
				get_readflag_filename(),
				&readflag))
		{
			throw chaos_exception(chaos_exception::generic);
		}
		system->load_readflag(&readflag);
	}
	// cd check
	while(
		!chaos_cd_player::instance()->open_match_device(
			system->get_bgm_sound_type(),
			system->get_cd_drive(),
			system->get_cd_info()))
	{
		chaos_message_box::show_message(
			chaos_resource::instance()->get_resource_text(
				chaos_resource::err_cant_find_cd_rom));
		idling_stop = true;
		chaos_config_dlg* dlg = chaos_config_dlg::instance();
		dlg->set_profile(&profile);
		if (!dlg->do_modal(NULL)) {
			return false;
		}
		if (dlg->is_dirty()) {
			system->load_profile(&profile);
			system->set_dirty_profile(true);
		}
		dlg->set_profile(NULL);
		idling_stop = false;
	}
	// sound volume
	chaos_cd_player::instance()->set_sound_volume(
		system->get_bgm_sound_volume());
	chaos_wave_player::instance()->set_sound_volume(
		system->get_effect_sound_volume());
	// main window create

#ifdef CHAOS_SCREEN_WIN32

	if (create_main_wnd() == NULL) {
		return false;
	}

#elif defined(CHAOS_SCREEN_SDL)

	if (!create_screen()) {
		return false;
	}

#endif // CHAOS_SCREEN

	// log window create
	if (system->get_log_mode() == chaos_system::logmode_window) {

#ifdef CHAOS_GUI_WIN32

		if (create_log_wnd() == NULL) {
			return false;
		}

#else // CHAOS_GUI_WIN32

		chaos_log::instance()->set_output_stderr(true);

#endif // CHAOS_GUI_WIN32

	}
	// sound manager
	chaos_sound_manager::instance()->initialize();
	// system restart
	if (!system->restart()) {
		throw chaos_exception(chaos_exception::generic);
	}
	// image
	chaos_string font_name;

#ifdef CHAOS_FONT_WIN32

		font_name = system->get_font_name();

#elif defined(CHAOS_FONT_GTK)

		font_name = system->get_font_xlfd();

#elif defined(CHAOS_FONT_PANGO)

		font_name = system->get_font_pango();

#endif // CHAOS_FONT

	if (!image_manager->initialize(
			chaos_dim(system->get_sys_width(), system->get_sys_height()),
			font_name,
			system->get_font_size(),
			system->is_font_bold(),
			system->is_font_anti_alias(),
			system->get_sys_image_buffer_count()))
	{
		throw chaos_exception(chaos_exception::generic);
	}
	return true;
}

// finalize

bool chaos::finalize() {
	// cd audio
	chaos_cd_player::instance()->dispose();
	// wave audio
	chaos_wave_player::instance()->dispose();
	// sound manager
	chaos_sound_manager::instance()->dispose();
	// image manager
	image_manager->dispose();
	// profile
	if (system->is_dirty_profile()) {
		system->set_dirty_profile(false);
		chaos_xml profile;
		if (system->save_profile(&profile)) {
			save_profile_to_file(
				get_profile_filename(),
				&profile);
		}
	}
	// readflag
	if (system->is_dirty_readflag()) {
		system->set_dirty_readflag(false);
		chaos_xml readflag;
		if (system->save_readflag(&readflag)) {
			save_readflag_to_file(
				get_readflag_filename(),
				&readflag);
		}
	}
	// environment

#ifdef CHAOS_PLATFORM_WIN32

	timeEndPeriod(1);
	CoUninitialize();

#endif // CHAOS_PLATFORM_WIN32

#ifdef CHAOS_API_SDL

	SDL_Quit();

#endif // CHAOS_API_SDL

	return true;
}

// message loop

int chaos::message_loop() {
	int result;

#ifdef CHAOS_SCREEN_WIN32

	MSG msg;
	bool quit = false;
	do {
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if (GetMessage(&msg, NULL, 0, 0)) {
				if ((msg.message != WM_SYSKEYDOWN) ||
					(msg.wParam != VK_RETURN) ||
					(msg.hwnd != get_main_wnd()))
				{
					TranslateMessage(&msg);
				}
				DispatchMessage(&msg);
			} else {
				quit = true;
			}
		} else {
			bool work = idle();
			Sleep(work? 0: 1);
		}
	} while (!quit);
	result = (int)msg.wParam;

#elif defined(CHAOS_SCREEN_SDL)

	SDL_Event ev;
	while (!app_quit) {
		while (SDL_PollEvent(&ev)) {
			process_event(&ev);
		}
		bool work = idle();
		if (!work) {
			SDL_Delay(1);
		}
	}
	result = 0;

#endif // CHAOS_SCREEN

	return result;
}

// close application

void chaos::close_app() {
	app_quit = true;

#ifdef CHAOS_SCREEN_WIN32

	close_main_wnd();

#endif // CHAOS_SCREEN_WIN32

}

// idle

bool chaos::idle() {

#ifdef CHAOS_DEBUG

	if (chaos_debug::is_asserting()) {
		return false;
	}

#endif // CHAOS_DEBUG

	if (idling_stop) {
		return false;
	}
	bool work = false;
	chaos_uint32 old_time = last_time;
	last_time = get_ticks();
	int pass_time = (int)(last_time-old_time);

#ifdef CHAOS_CDAUDIO_SDL

	audio_notify_counter += pass_time;
	if (audio_notify_counter >= (int)audio_notify_elapse) {
		audio_notify_counter = 0;
		chaos_cd_player::instance()->notify();
	}

#endif // CHAOS_CDAUDIO

	if (postpone_preview_snapno_update > 0) {
		postpone_preview_snapno_update -= pass_time;
		if (postpone_preview_snapno_update <= 0) {
			postpone_preview_snapno_update = 0;
			update_preview_snapno();
		}
	}
	if (get_special_mode() == spmode_none) {
		work = system->execute(pass_time);
		switch (system->get_system_mode()) {
		case chaos_system::sysmode_breakapp:
			close_app();
			break;
		case chaos_system::sysmode_load:
			system->change_system_mode(chaos_system::sysmode_interpret);
			change_special_mode(spmode_load);
			break;
		}
	}
	bool dirty_image = system->is_dirty_image(),
		dirty_message = system->is_dirty_message();
	if (dirty_image || dirty_message) {
		system->set_dirty_image(false);
		system->set_dirty_message(false);
		chaos_rect clip_rect;
		if (dirty_message) {
			chaos_string message_text = get_message_text();
			image_manager->update_message_image(
				message_text,
				&clip_rect);
		}
		chaos_rect update_rect;
		draw_image(
			&update_rect,
			dirty_image? NULL: &clip_rect);
		update_screen(update_rect, false);
		work = true;
	}
	return work;
}

// get ticks

chaos_uint32 chaos::get_ticks() {
	chaos_uint32 tick;

#ifdef CHAOS_API_WIN32

	tick = (chaos_uint32)timeGetTime();

#elif defined(CHAOS_API_SDL)

	tick = (chaos_uint32)SDL_GetTicks();

#endif // CHAOS_API

	return tick;
}

// execute chaos

int chaos::execute() {
	int result;
	if (!initialize()) {
		result = 1;
	} else {
		result = message_loop();
	}
	finalize();
	return result;
}

////////////////////////////////////////////////////////////////////////////////
// main

#ifdef CHAOS_GUI_WIN32

// windows main procedure

int WINAPI WinMain(
	HINSTANCE /*process_instance*/, HINSTANCE /*prev_process_instance*/,
	LPSTR /*cmd_line*/, int /*show_cmd*/)
{

#ifdef CHAOS_DEBUG

	_CrtSetDbgFlag(
		_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_LEAK_CHECK_DF);

#endif // CHAOS_DEBUG

	int result;
	chaos* mainobj = chaos::instance();
	try {
		mainobj->set_args(__argc, __argv);
		result = mainobj->execute();
	} catch (std::exception& e) {
		mainobj->set_idling_stop(true);
		static char report[256+1];
		snprintf(report, sizeof(report)-1, "exception : %s", e.what());
		report[sizeof(report)-1] = '\0';
		MessageBox(NULL, report, "system error", MB_OK | MB_ICONERROR);
		result = 1;
	}
	return result;
}

#elif defined(CHAOS_GUI_GTK)

// main procedure

int main(int argc, char* argv[]) {
	gtk_set_locale();
	gtk_init(&argc, &argv);
	gtk_rc_parse("./gtkrc");
	int result;
	chaos* mainobj = chaos::instance();
	try {
		mainobj->set_args(argc, argv);
		result = mainobj->execute();
	} catch (std::exception& e) {
		mainobj->set_idling_stop(true);
		fprintf(stderr, "exception : %s", e.what());
		result = 1;
	}
	return result;
}

#endif
