////////////////////////////////////////////////////////////////////////////////
// chaos_bmp_image_loader.cpp : bmp image loader

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_bmp_image_loader.h"
#include "chaos_buffer.h"
#include "chaos_utility.h"
#include "chaos_exception.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_bmp_image_loader : bmp image loader class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// constructor

chaos_bmp_image_loader::chaos_bmp_image_loader(
		chaos_file_data* org_file_data,
		chaos_image* org_image) :
	chaos_image_loader(
		org_file_data,
		org_image)
{
}

// destructor

chaos_bmp_image_loader::~chaos_bmp_image_loader() {
}

////////////////////////////////////////////////////////////////////////////////
// implementation

// load image

bool chaos_bmp_image_loader::load() {
	bool result = false;
	try {
		chaos_buffer buf;
		// bitmap file header
		buf.resize(2+4+2+2+4);
		if (!read_file(buf.data(), buf.size())) {
			throw chaos_exception(chaos_exception::io_error);
		}
		if (((char)buf[0] != 'B') ||
			((char)buf[1] != 'M'))
		{
			throw chaos_exception(chaos_exception::bad_format);
		}
		// bitmap info header
		buf.resize(4+4+4+2+2+4+4+4+4+4+4);
		if (!read_file(buf.data(), buf.size())) {
			throw chaos_exception(chaos_exception::io_error);
		}
		size_t info_size = (size_t)(chaos_uint32)(
				buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24));
		if (info_size < buf.size()) {
			throw chaos_exception(chaos_exception::bad_format);
		}
		int width = (int)(chaos_int32)(
				buf[4] | (buf[5] << 8) | (buf[6] << 16) | (buf[7] << 24)),
			height = (int)(chaos_int32)(
				buf[8] | (buf[9] << 8) | (buf[10] << 16) | (buf[11] << 24)),
			planes = (int)(chaos_uint16)(buf[12] | (buf[13] << 8)),
			bit_count = (int)(chaos_uint16)(buf[14] | (buf[15] << 8)),
			compression = (int)(chaos_uint32)(
				buf[16] | (buf[17] << 8) | (buf[18] << 16) | (buf[19] << 24)),
			color_used = (int)(chaos_uint32)(
				buf[32] | (buf[33] << 8) | (buf[34] << 16) | (buf[35] << 24)),
			color_important = (int)(chaos_uint32)(
				buf[36] | (buf[37] << 8) | (buf[38] << 16) | (buf[39] << 24));
		bool reverse = (height < 0);
		if (reverse) {
			height = -height;
		}
		if ((width < 1) || (width > 4096) ||
			(height < 1) || (height > 4096) ||
			(planes != 1) ||
			((bit_count != 1) && (bit_count != 2) &&
				(bit_count != 4) && (bit_count != 8) &&
				(bit_count != 16) && (bit_count != 24) &&
				(bit_count != 32)) ||
			((compression != 0) && (compression != 3)))
		{
			throw chaos_exception(chaos_exception::bad_format);
		}
		if (info_size > buf.size()) {
			buf.resize(info_size-buf.size());
			if (!read_file(buf.data(), buf.size())) {
				throw chaos_exception(chaos_exception::io_error);
			}
		}
		// bitfield
		int red_ofs = 0, green_ofs = 0, blue_ofs = 0,
			red_bits = 1, green_bits = 1, blue_bits = 1;
		if (compression != 3) {
			if (bit_count == 16) {
				red_ofs = 10;
				green_ofs = 5;
				blue_ofs = 0;
				red_bits = green_bits = blue_bits = 5;
			} else {
				red_ofs = 16;
				green_ofs = 8;
				blue_ofs = 0;
				red_bits = green_bits = blue_bits = 8;
			}
		} else {
			buf.resize(4*3);
			if (!read_file(buf.data(), buf.size())) {
				throw chaos_exception(chaos_exception::io_error);
			}
			for (int i = 0; i < 3; i++) {
				chaos_uint32 mask = (chaos_uint32)(
					buf[i*4] | (buf[i*4+1] << 8) |
						(buf[i*4+2] << 16) | (buf[i*4+3] << 24));
				int ofs, bits;
				for (ofs = 0; ofs < 32; ofs++) {
					if ((mask & 1) != 0) {
						break;
					}
					mask >>= 1;
				}
				for (bits = 0; bits+ofs < 32; bits++) {
					if ((mask & 1) == 0) {
						break;
					}
					mask >>= 1;
				}
				switch (i) {
				case 0:
					red_ofs = ofs;
					red_bits = bits;
					break;
				case 1:
					green_ofs = ofs;
					green_bits = bits;
					break;
				case 2:
					blue_ofs = ofs;
					blue_bits = bits;
					break;
				}
			}
		}
		// palette
		int pal_count = (color_used >= color_important)?
				color_used: color_important;
		bool index_color = (bit_count <= 8);
		if ((pal_count == 0) && index_color) {
			pal_count = 1 << bit_count;
		}
		std::vector<chaos_uint32> palette;
		if (pal_count > 0) {
			buf.resize(pal_count*4);
			if (!read_file(buf.data(), buf.size())) {
				throw chaos_exception(chaos_exception::io_error);
			}
			palette.reserve(pal_count);
			for (int i = 0; i < pal_count; i++) {
				int red = buf[i*4+2],
					green = buf[i*4+1],
					blue = buf[i*4];
				palette.push_back(
					CHAOS_RGBA(red, green, blue, 0));
			}
		}
		// create image
		if (!image->create(width, height)) {
			throw chaos_exception(chaos_exception::few_memory);
		}
		// bit data
		buf.resize(((width*bit_count+31)/32)*4);
		for (int i = 0; i < height; i++) {
			if (!read_file(buf.data(), buf.size())) {
				throw chaos_exception(chaos_exception::io_error);
			}
			int y = reverse? i: (height-i-1);
			chaos_uint32* bits = image->get_bits(0, y);
			chaos_uint8* bufdata = buf.data();
			for (int x = 0; x < width; x++) {
				chaos_uint32 col = 0;
				if (!index_color) {
					chaos_uint32 data;
					if (bit_count == 32) {
						data = (chaos_uint32)(
							*bufdata | (*(bufdata+1) << 8) |
								(*(bufdata+2) << 16) | (*(bufdata+3) << 24));
						bufdata += 4;
					} else if (bit_count == 24) {
						data = (chaos_uint32)(
							*bufdata | (*(bufdata+1) << 8) |
								(*(bufdata+2) << 16));
						bufdata += 3;
					} else {
						data = (chaos_uint32)(
							*bufdata | (*(bufdata+1) << 8));
						bufdata += 2;
					}
					int red = (data >> red_ofs) & ((1 << red_bits)-1);
					if (red_bits < 8) {
						int left = (red << (8-red_bits));
						red = 0;
						while (left > 0) {
							red |= left;
							left >>= red_bits;
						}
					} else if (red_bits > 8) {
						red >>= red_bits-8;
					}
					int green = (data >> green_ofs) & ((1 << green_bits)-1);
					if (green_bits < 8) {
						int left = (green << (8-green_bits));
						green = 0;
						while (left > 0) {
							green |= left;
							left >>= green_bits;
						}
					} else if (green_bits > 8) {
						green >>= green_bits-8;
					}
					int blue = (data >> blue_ofs) & ((1 << blue_bits)-1);
					if (blue_bits < 8) {
						int left = (blue << (8-blue_bits));
						blue = 0;
						while (left > 0) {
							blue |= left;
							left >>= blue_bits;
						}
					} else if (blue_bits > 8) {
						blue >>= blue_bits-8;
					}
					col = CHAOS_RGBA(red, green, blue, 0);
				} else {
					int ofs = x*bit_count;
					int pal =
						((bufdata[ofs/8] << (ofs%8)) & 0xFF) >> (8-bit_count);
					if (pal < (int)palette.size()) {
						col = palette[pal];
					}
				}
				*(bits++) = col;
			}
		}
		result = true;
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : bmp image loader::load - "
			"%s\n",
			e.what());
	}
	return result;
}
