////////////////////////////////////////////////////////////////////////////////
// chaos_buffer.cpp : buffer

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_buffer.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_buffer : buffer class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_buffer::chaos_buffer() :
	data_ptr(NULL),
	data_size(0), data_capacity(0)
{
}

// fill constructor

chaos_buffer::chaos_buffer(size_type fill, const value_type& value) :
	data_ptr(NULL),
	data_size(0), data_capacity(0)
{
	if (fill > 0) {
		alloc_data(fill);
		memset(data_ptr, fill, value);
		data_size = fill;
	}
}

// copy constructor

chaos_buffer::chaos_buffer(const chaos_buffer& org) :
	data_ptr(NULL),
	data_size(0), data_capacity(0)
{
	if (org.size() > 0) {
		alloc_data(org.size());
		memcpy(data_ptr, org.data(), org.size());
		data_size = org.size();
	}
}

// sequence constructor

chaos_buffer::chaos_buffer(const_iterator first, const_iterator last) :
	data_ptr(NULL),
	data_size(0), data_capacity(0)
{
	ASSERT(first <= last);
	size_type dist = last-first;
	if (dist > 0) {
		alloc_data(dist);
		memcpy(data_ptr, first, dist);
		data_size = dist;
	}
}

// destructor

chaos_buffer::~chaos_buffer() {
	free_data();
}

////////////////////////////////////////////////////////////////////////////////
// buffer

// allocate data buffer

void chaos_buffer::alloc_data(size_type new_data_capacity) {
	if (new_data_capacity > data_capacity) {
		value_type* new_data_ptr;
		if (data_ptr == NULL) {
			new_data_ptr =
				(value_type*)malloc(new_data_capacity);
		} else {
			new_data_ptr =
				(value_type*)realloc(data_ptr, new_data_capacity);
		}
		if (new_data_ptr == NULL) {
			throw std::bad_alloc();
		}
		data_ptr = new_data_ptr;
		data_capacity = new_data_capacity;
	}
}

// free data buffer

void chaos_buffer::free_data() {
	if (data_ptr != NULL) {
		free(data_ptr);
		data_ptr = NULL;
		data_size = data_capacity = 0;
	}
}

////////////////////////////////////////////////////////////////////////////////
// operation

// erase

chaos_buffer::iterator chaos_buffer::erase(iterator first, iterator last) {
	ASSERT(first <= last);
	ASSERT(first >= data_ptr);
	ASSERT(last <= data_ptr+data_size);
	size_type moves = data_ptr+data_size-last;
	if (moves > 0) {
		memmove(first, last, moves);
	}
	data_size -= last-first;
	return first;
}

chaos_buffer::iterator chaos_buffer::erase(iterator target) {
	return erase(target, target+1);
}

// insert

void chaos_buffer::insert(
	iterator target,
	size_type fill,
	const value_type& value)
{
	ASSERT(target >= data_ptr);
	ASSERT(target <= data_ptr+data_size);
	if (fill > 0) {
		size_type ofs = target-data_ptr;
		alloc_data(data_size+fill);
		if (ofs < data_size) {
			memmove(
				data_ptr+ofs+fill, data_ptr+ofs,
				data_size-ofs);
		}
		memset(data_ptr+ofs, value, fill);
		data_size += fill;
	}
}

chaos_buffer::iterator chaos_buffer::insert(
	iterator target,
	const value_type& value)
{
	size_type ofs = target-data_ptr;
	insert(target, 1, value);
	return data_ptr+ofs;
}

// push back

void chaos_buffer::push_back(const value_type& value) {
	insert(end(), value);
}

// pop back

void chaos_buffer::pop_back() {
	erase(end()-1);
}

// reserve

void chaos_buffer::reserve(size_type reserve_size) {
	alloc_data(reserve_size);
}

// resize

void chaos_buffer::resize(
	size_type new_data_size,
	const value_type& value)
{
	alloc_data(new_data_size);
	if (new_data_size > data_size) {
		memset(data_ptr+data_size, value, new_data_size-data_size);
	}
	data_size = new_data_size;
}

// clear

void chaos_buffer::clear() {
	erase(begin(), end());
}

// clear all

void chaos_buffer::clear_all() {
	free_data();
}

////////////////////////////////////////////////////////////////////////////////
// operator

// copy

chaos_buffer& chaos_buffer::operator=(const chaos_buffer& org) {
	if (&org == this) {
		// nothing
	} else if (org.size() > 0) {
		alloc_data(org.size());
		memcpy(data_ptr, org.data(), org.size());
		data_size = org.size();
	} else {
		clear();
	}
	return *this;
}
