////////////////////////////////////////////////////////////////////////////////
// chaos_cd_player.cpp : cd player

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_cd_player.h"
#include "chaos_file_serializer.h"
#include "chaos_utility.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_cd_player : cd player class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_cd_player::chaos_cd_player() :
	opened(false),
	sound_type("CD-DA"),
	cdda(true),
	playing_start_track(-1),
	playing_end_track(-1),
	repeat(false),
	bgm_replace(NULL)
{

#ifdef CHAOS_CDAUDIO_WIN32

	cd_device_id = 0;
	notify_wnd = NULL;

#elif defined(CHAOS_CDAUDIO_SDL)

	cd_device = NULL;

#endif // CHAOS_CDAUDIO

}

// destructor

chaos_cd_player::~chaos_cd_player() {
	dispose();
	if (bgm_replace != NULL) {
		delete bgm_replace;
	}
}

// dispose

void chaos_cd_player::dispose() {
	close_device();
}

// get singleton instance

chaos_cd_player* chaos_cd_player::instance() {
	static chaos_cd_player cd_player;
	return &cd_player;
}

////////////////////////////////////////////////////////////////////////////////
// device

// open device

bool chaos_cd_player::open_device(
	const chaos_string& new_sound_type,
	const chaos_string& new_drive_name)
{
	LOG_RPT(
		"report : cd player::open device - "
		"cd device open(%s, %s)\n",
		new_sound_type.c_str(),
		new_drive_name.c_str());
	if (is_opened()) {
		if (!close_device()) {
			return false;
		}
	}
	if (chaos_utility::icompare(new_sound_type, "CD-DA") == 0) {

#ifdef CHAOS_CDAUDIO_WIN32

		MMRESULT result;
		{ // drive type
			UINT drive_type = GetDriveType((new_drive_name+'\\').c_str());
			if (drive_type != DRIVE_CDROM) {
				LOG_ERR(
					"error : cd player::open device - "
					"bad type(%u)\n",
					drive_type);
				return false;
			}
		}
		{ // open
			MCI_OPEN_PARMS open_parms;
			memset(&open_parms, 0, sizeof(open_parms));
			open_parms.lpstrDeviceType = (LPCSTR)MCI_DEVTYPE_CD_AUDIO;
			open_parms.lpstrElementName = new_drive_name.c_str();
			result = mciSendCommand(
				0,
				MCI_OPEN,
				MCI_WAIT | MCI_OPEN_SHAREABLE |
					MCI_OPEN_TYPE | MCI_OPEN_TYPE_ID |
					MCI_OPEN_ELEMENT,
				(DWORD)&open_parms);
			if (result != MMSYSERR_NOERROR) {
				LOG_ERR(
					"error : cd player::open device - "
					"open error(%d)\n",
					(int)result);
				return false;
			}
			cd_device_id = open_parms.wDeviceID;
			opened = true;
			sound_type = new_sound_type;
			cdda = true;
			drive_name = new_drive_name;
		}
		{ // set timeformat
			MCI_SET_PARMS set_parms;
			memset(&set_parms, 0, sizeof(set_parms));
			set_parms.dwTimeFormat = MCI_FORMAT_TMSF;
			result = mciSendCommand(
				cd_device_id,
				MCI_SET,
				MCI_WAIT | MCI_SET_TIME_FORMAT,
				(DWORD)&set_parms);
			if (result != MMSYSERR_NOERROR) {
				LOG_ERR(
					"error : cd player::open device - "
					"set timeformat error(%d)\n",
					(int)result);
				close_device();
				return false;
			}
		}

#elif defined(CHAOS_CDAUDIO_SDL)

		int num_drives = SDL_CDNumDrives();
		int drive;
		for (drive = 0; drive < num_drives; drive++) {
			chaos_string check_drive_name =
				chaos_utility::filename_to_sjis(
					SDL_CDName(drive));
			if (chaos_utility::icompare(
					check_drive_name, new_drive_name) == 0)
			{
				break;
			}
		}
		if (drive >= num_drives) {
			LOG_ERR(
				"error : cd player::open device - "
				"unknown drive\n");
			return false;
		}
		cd_device = SDL_CDOpen(drive);
		if (cd_device == NULL) {
			LOG_ERR(
				"error : cd player::open device - "
				"cannot open device(%d)\n",
				drive);
			return false;
		}
		opened = true;
		sound_type = new_sound_type;
		cdda = true;
		drive_name = new_drive_name;

#endif // CHAOS_CDAUDIO

	} else {
		opened = true;
		sound_type = new_sound_type;
		cdda = false;
		drive_name = new_drive_name;
	}
	return is_opened();
}

// close device

bool chaos_cd_player::close_device() {
	if (!is_opened()) {
		return true;
	}
	stop();

	if (is_cdda()) {

#ifdef CHAOS_CDAUDIO_WIN32

		MMRESULT result;
		MCI_GENERIC_PARMS close_params;
		memset(&close_params, 0, sizeof(close_params));
		result = mciSendCommand(
			cd_device_id,
			MCI_CLOSE,
			MCI_WAIT,
			(DWORD)&close_params);
		if (result != MMSYSERR_NOERROR) {
			LOG_ERR(
				"error : cd player::close device - "
				"close error(%d)\n",
				(int)result);
			return false;
		}
		cd_device_id = 0;

#elif defined(CHAOS_CDAUDIO_SDL)

		SDL_CDClose(cd_device);
		cd_device = NULL;

#endif // CHAOS_CDAUDIO

	}
	opened = false;
	sound_type = "CD-DA";
	cdda = true;
	drive_name = "";
	return true;
}

// set sound volume

void chaos_cd_player::set_sound_volume(int new_sound_volume) {
	sound.set_volume(new_sound_volume);
}

////////////////////////////////////////////////////////////////////////////////
// play

// play cd audio

bool chaos_cd_player::play(
	int new_start_track, int new_end_track,
	bool new_repeat)
{
	if (!is_opened()) {
		return false;
	}
	ASSERT(new_start_track >= 1);
	ASSERT(
		(new_end_track < 0) ||
		(new_end_track > new_start_track));
	if ((new_start_track == playing_start_track) &&
		(new_end_track == playing_end_track) &&
		(new_repeat == repeat) &&
		repeat)
	{
		return true;
	}
	if (is_cdda()) {

#ifdef CHAOS_CDAUDIO_WIN32

		MMRESULT result;
		MCI_PLAY_PARMS play_params;
		memset(&play_params, 0, sizeof(play_params));
		DWORD flags = MCI_FROM;
		play_params.dwFrom = MCI_MAKE_TMSF(new_start_track, 0, 0, 0);
		if (new_end_track >= 1) {
			flags |= MCI_TO;
			play_params.dwTo = MCI_MAKE_TMSF(new_end_track, 0, 0, 0);
		}
		if (new_repeat && (notify_wnd != NULL)) {
			play_params.dwCallback = (DWORD)notify_wnd;
			flags |= MCI_NOTIFY;
		}
		result = mciSendCommand(
			cd_device_id,
			MCI_PLAY,
			flags,
			(DWORD)&play_params);
		if (result != MMSYSERR_NOERROR) {
			LOG_ERR(
				"error : cd player::play - "
				"play error(%d)\n",
				(int)result);
			return false;
		}

#elif defined(CHAOS_CDAUDIO_SDL)

		int result;
		result = SDL_CDPlayTracks(
			cd_device,
			new_start_track-1, 0,
			(new_end_track >= 1)? (new_end_track-new_start_track): 1, 0);
		if (result != 0) {
			LOG_ERR(
				"error : cd player::play - "
				"play error(%d)\n",
				result);
			return false;
		}

#endif // CHAOS_CDAUDIO

	} else {
		sound.dispose();
		chaos_string filename = get_track_filename(new_start_track);
		sound.play_sound_from_file(
			chaos_utility::sjis_to_filename(filename).c_str(),
			new_repeat,
			true);
	}
	playing_start_track = new_start_track;
	playing_end_track = new_end_track;
	repeat = new_repeat;
	return true;
}

// stop cd audio

bool chaos_cd_player::stop() {
	if (!is_opened()) {
		return false;
	}
	if (!is_playing()) {
		return true;
	}
	if (is_cdda()) {

#ifdef CHAOS_CDAUDIO_WIN32

		MMRESULT result;
		MCI_GENERIC_PARMS stop_params;
		memset(&stop_params, 0, sizeof(stop_params));
		result = mciSendCommand(
			cd_device_id,
			MCI_STOP,
			MCI_WAIT,
			(DWORD)&stop_params);
		if (result != MMSYSERR_NOERROR) {
			LOG_ERR(
				"error : cd player::stop - "
				"stop error(%d)\n",
				(int)result);
			return false;
		}

#elif defined(CHAOS_CDAUDIO_SDL)

		int result;
		result = SDL_CDStop(cd_device);
		if (result != 0) {
			LOG_ERR(
				"error : cd player::stop - "
				"stop error(%d)\n",
				result);
			return false;
		}

#endif // CHAOS_CDAUDIO

	} else {
		sound.dispose();
	}
	playing_start_track = playing_end_track = -1;
	repeat = false;
	return true;
}

// notify cd audio

bool chaos_cd_player::notify() {
	bool result = true;

	if (is_cdda()) {
#ifdef CHAOS_CDAUDIO_WIN32

		int old_start_track = playing_start_track,
			old_end_track = playing_end_track;
		bool old_repeat = repeat;
		playing_start_track = playing_end_track = -1;
		repeat = false;
		if (old_repeat) {
			result = play(
				old_start_track,
				old_end_track,
				old_repeat);
		}

#elif defined(CHAOS_CDAUDIO_SDL)

		if (is_device_open() && is_playing()) {
			CDstatus status = SDL_CDStatus(cd_device);
			if (status != CD_PLAYING) {
				int old_start_track = playing_start_track,
					old_end_track = playing_end_track;
				bool old_repeat = repeat;
				playing_start_track = playing_end_track = -1;
				repeat = false;
				if (old_repeat) {
					result = play(
						old_start_track,
						old_end_track,
						old_repeat);
				}
			}
		}

#endif // CHAOS_CDAUDIO

	}
	return result;
}

////////////////////////////////////////////////////////////////////////////////
// operation

// encode cd info

chaos_string chaos_cd_player::encode_cd_info(
	const chaos_string& volume,
	const std::vector<int>& track_length)
{
	chaos_string info;
	int track_count = track_length.size();
	info = chaos_utility::format_str(
		"%s,%d",
		chaos_utility::str_to_str(volume).c_str(),
		track_count);
	for (int i = 0; i < track_count; i++) {
		info += chaos_utility::format_str(
			",%d",
			track_length[i]);
	}
	return info;
}

// decode cd info

bool chaos_cd_player::decode_cd_info(
	const chaos_string& info,
	chaos_string& volume,
	std::vector<int>& track_length)
{
	volume = "";
	track_length.clear();
	if (info.empty()) {
		return true;
	}
	const char* infostr = info.c_str();
	volume = chaos_utility::str_from_str(infostr);
	if (*infostr != ',') {
		volume = "";
		return false;
	}
	infostr++;
	int next = 0,
		track_count;
	if (sscanf(
			infostr,
			"%d%n",
			&track_count,
			&next) != 1)
	{
		volume = "";
		return false;
	}
	infostr += next;
	track_length.reserve(track_count);
	for (int i = 0; i < track_count; i++) {
		next = 0;
		int length;
		if (sscanf(
				infostr,
				",%d%n",
				&length,
				&next) != 1)
		{
			volume = "";
			track_length.clear();
			return false;
		}
		infostr += next;
		track_length.push_back(length);
	}
	return true;
}

// is match cd info

bool chaos_cd_player::is_match_cd_info(
	const chaos_string& info1, const chaos_string& info2)
{
	chaos_string volume1, volume2;
	std::vector<int> track_length1, track_length2;
	if (!decode_cd_info(info1, volume1, track_length1) ||
		!decode_cd_info(info2, volume2, track_length2))
	{
		return false;
	}
	volume1 = chaos_utility::to_lower_str(volume1);
	volume2 = chaos_utility::to_lower_str(volume2);
	if (volume1 != volume2) {

#ifdef CHAOS_CDAUDIO_WIN32

		if ((volume1.length() == 0) ||
			(volume2.length() == 0))
		{
			return false;
		}
		if ((chaos_utility::str_find_str(volume1, volume2) < 0) &&
			(chaos_utility::str_find_str(volume2, volume1) < 0))
		{
			return false;
		}

#elif defined(CHAOS_CDAUDIO_SDL)

		// sdl cannot get volume

#endif // CHAOS_CDAUDIO

	}
	if (track_length1.size() != track_length2.size()) {
		return false;
	}
	int track_count = track_length1.size();
	for (int i = 0; i < track_count; i++) {
		int length1 = track_length1[i],
			length2 = track_length2[i];
		if (length1 < 0) {
			if (length2 >= 0) {
				return false;
			}
		} else {
			int diff = abs(length2-length1);
			if ((diff > length1/10) && (diff > length2/10)) {
				return false;
			}
		}
	}
	return true;
}

// get cd drives

void chaos_cd_player::get_cd_drives(
	std::vector<chaos_string>& cd_drives)
{
	cd_drives.clear();

#ifdef CHAOS_CDAUDIO_WIN32

	for (int letter = 'a'; letter <= 'z'; letter++) {
		chaos_string check_drive_name = (char)letter;
		check_drive_name += ':';
		UINT drive_type = GetDriveType((check_drive_name+'\\').c_str());
		if (drive_type == DRIVE_CDROM) {
			cd_drives.push_back(check_drive_name);
		}
	}

#elif defined(CHAOS_CDAUDIO_SDL)

	int num_drives = SDL_CDNumDrives();
	for (int drive = 0; drive < num_drives; drive++) {
		chaos_string check_drive_name =
			chaos_utility::filename_to_sjis(
				SDL_CDName(drive));
		cd_drives.push_back(check_drive_name);
	}

#endif // CHAOS_CDAUDIO

}

// get cd info

chaos_string chaos_cd_player::get_cd_info() {
	if (!is_opened()) {
		return "";
	}
	chaos_string volume;
	int track_count;
	std::vector<int> track_length;

	if (is_cdda()) {
#ifdef CHAOS_CDAUDIO_WIN32

		MMRESULT result;
		{ // status madia present
			MCI_STATUS_PARMS status_parms;
			memset(&status_parms, 0, sizeof(status_parms));
			status_parms.dwItem = MCI_STATUS_MEDIA_PRESENT;
			result = mciSendCommand(
				cd_device_id,
				MCI_STATUS,
				MCI_WAIT | MCI_STATUS_ITEM,
				(DWORD)&status_parms);
			if (result != MMSYSERR_NOERROR) {
				LOG_ERR(
					"error : cd player::get cd info - "
					"status<madia present> error(%d)\n",
					(int)result);
				return "";
			}
			if (status_parms.dwReturn == FALSE) {
				LOG_RPT(
					"report : cd player::get cd info - "
					"media not present(%s)\n",
					drive_name.c_str());
				return "";
			}
		}
		{ // volume
			char volumestr[_MAX_PATH];
			if (GetVolumeInformation(
					(drive_name+'\\').c_str(),
					volumestr, sizeof(volumestr)/sizeof(volumestr[0]),
					NULL, NULL, NULL,
					NULL, NULL))
			{
				volume = volumestr;
			}
		}
		{ // status number of tracks
			MCI_STATUS_PARMS status_parms;
			memset(&status_parms, 0, sizeof(status_parms));
			status_parms.dwItem = MCI_STATUS_NUMBER_OF_TRACKS;
			result = mciSendCommand(
				cd_device_id,
				MCI_STATUS,
				MCI_WAIT | MCI_STATUS_ITEM,
				(DWORD)&status_parms);
			if (result != MMSYSERR_NOERROR) {
				LOG_ERR(
					"error : cd player::get cd info - "
					"status<number of tracks> error(%d)\n",
					(int)result);
				return "";
			}
			track_count = (int)status_parms.dwReturn;
		}
		track_length.reserve(track_count);
		for (int i = 0; i < track_count; i++) {
			bool audio_track;
			{ // status type track
				MCI_STATUS_PARMS status_parms;
				memset(&status_parms, 0, sizeof(status_parms));
				status_parms.dwItem = MCI_CDA_STATUS_TYPE_TRACK;
				status_parms.dwTrack = i+1;
				result = mciSendCommand(
					cd_device_id,
					MCI_STATUS,
					MCI_WAIT | MCI_STATUS_ITEM | MCI_TRACK,
					(DWORD)&status_parms);
				if (result != MMSYSERR_NOERROR) {
					LOG_ERR(
						"error : cd player::get cd info - "
						"status<type track> error(%d)\n",
						(int)result);
					return "";
				}
				audio_track =(status_parms.dwReturn == MCI_CDA_TRACK_AUDIO);
			}
			if (audio_track){
				// status length
				MCI_STATUS_PARMS status_parms;
				memset(&status_parms, 0, sizeof(status_parms));
				status_parms.dwItem = MCI_STATUS_LENGTH;
				status_parms.dwTrack = i+1;
				result = mciSendCommand(
					cd_device_id,
					MCI_STATUS,
					MCI_WAIT | MCI_STATUS_ITEM | MCI_TRACK,
					(DWORD)&status_parms);
				if (result != MMSYSERR_NOERROR) {
					LOG_ERR(
						"error : cd player::get cd info - "
						"status<length> error(%d)\n",
						(int)result);
					return "";
				}
				track_length.push_back(
					MCI_MSF_MINUTE(status_parms.dwReturn)*60+
					MCI_MSF_SECOND(status_parms.dwReturn));
			} else {
				track_length.push_back(-1);
			}
		}

#elif defined(CHAOS_CDAUDIO_SDL)

		CDstatus status = SDL_CDStatus(cd_device);
		if (status == CD_ERROR) {
			LOG_ERR(
				"error : cd player::get cd info - "
				"status error\n");
			return "";
		} else if (status == CD_TRAYEMPTY) {
			LOG_RPT(
				"report : cd player::get cd info - "
				"tray empty\n");
			return "";
		}
		track_count = cd_device->numtracks;
		track_length.reserve(track_count);
		for (int i = 0; i < track_count; i++) {
			if (cd_device->track[i].type == SDL_AUDIO_TRACK) {
				track_length.push_back(
					cd_device->track[i].length/75);
			} else {
				track_length.push_back(-1);
			}
		}

#endif // CHAOS_CDAUDIO

	} else {
	}
	return encode_cd_info(volume, track_length);
}

// open match device

bool chaos_cd_player::open_match_device(
	const chaos_string& new_sound_type,
	const chaos_string& fix_drive_name,
	const chaos_string& info)
{
	close_device();
	if (chaos_utility::icompare(new_sound_type, "CD-DA") == 0) {

#ifdef CHAOS_CDAUDIO_WIN32

		int start_letter = 'a', end_letter = 'z';
		bool auto_check = true;
		if ((fix_drive_name.length() >= 2) &&
			chaos_utility::is_alpha(fix_drive_name[0]) &&
			(fix_drive_name[1] == ':'))
		{
			start_letter = end_letter =
				chaos_utility::to_lower(fix_drive_name[0]);
			auto_check = false;
		}
		for (int letter = start_letter; letter <= end_letter; letter++) {
			chaos_string check_drive_name = (char)letter;
			check_drive_name += ':';
			UINT drive_type = GetDriveType((check_drive_name+'\\').c_str());
			if ((drive_type != DRIVE_CDROM) ||
				!open_device(new_sound_type, check_drive_name))
			{
				// nothing
			} else {
				chaos_string check_info = get_cd_info();
				LOG_RPT(
					"report : cd player::open match device - "
					"cd info(%s) = \"%s\"\n",
					check_drive_name.c_str(),
					check_info.c_str());
				if (!auto_check) {
					break;
				}
				if (is_match_cd_info(info, check_info)) {
					LOG_RPT(
						"report : cd player::open match device - "
						"cd matched\n");
					break;
				}
				close_device();
			}
		}

#elif defined(CHAOS_CDAUDIO_SDL)

		bool auto_check = fix_drive_name.empty();
		int num_drives = SDL_CDNumDrives();
		for (int drive = 0; drive < num_drives; drive++) {
			chaos_string check_drive_name =
				chaos_utility::filename_to_sjis(
					SDL_CDName(drive));
			if ((!fix_drive_name.empty() &&
					(chaos_utility::icompare(
						check_drive_name, fix_drive_name) != 0)) ||
				!open_device(new_sound_type, check_drive_name))
			{
				// nothing
			} else {
				chaos_string check_info = get_cd_info();
				LOG_RPT(
					"report : cd player::open match device - "
					"cd info(%s) = \"%s\"\n",
					check_drive_name.c_str(),
					check_info.c_str());
				if (!auto_check) {
					break;
				}
				if (is_match_cd_info(info, check_info)) {
					LOG_RPT(
						"report : cd player::open match device - "
						"cd matched\n");
					break;
				}
				close_device();
			}
		}

#endif // CHAOS_CDAUDIO

	} else {
		open_device(new_sound_type, fix_drive_name);
	}
	return is_opened();
}

// get track filename

chaos_string chaos_cd_player::get_track_filename(
	int track)
{
	chaos_string filename = chaos_utility::format_str(
		"cosmos_%02d",
		track);
	if (bgm_replace == NULL) {
		bgm_replace = new chaos_xml;
		chaos_string replace_filename = "chaos_bgm_replace.dat";
		if (chaos_file_serializer::instance()->exist_file(
				replace_filename,
				chaos_file_serializer::fpos_base_dir))
		{
			chaos_file_data filedata;
			if (!chaos_file_serializer::instance()->load_file_data(
					&filedata,
					replace_filename,
					chaos_file_serializer::fpos_base_dir) ||
				!bgm_replace->load(&filedata) ||
				(bgm_replace->get_kind() != "bgm_replace") ||
				(bgm_replace->get_values()->
					get_value("replace").get_boolean() !=
						true))
			{
				bgm_replace->clear();
			}
		}
	}
	if (bgm_replace->get_values()->exist_value(filename)) {
		filename = bgm_replace->get_values()->
			get_value(filename).get_string();
	}
	if (sound_type.empty() ||
		(sound_type[0] != '.'))
	{
		filename += '.';
	}
	filename += sound_type;
	return
		chaos_file_serializer::instance()->get_base_dir()+
		filename;
}
