////////////////////////////////////////////////////////////////////////////////
// chaos_color.h : color

#ifndef CHAOS_COLOR_H_INCLUDED
#define CHAOS_COLOR_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_color;

////////////////////////////////////////////////////////////////////////////////
// chaos_color : color class

class chaos_color {
// enum
private:
	// limits
	enum {
		rgb_min   = -512,
		rgb_max   = 511,
		rgb_mask  = 0x3FF,
		red_ofs   = 20,
		green_ofs = 10,
		blue_ofs  = 0
	};

// attribute
private:
	// color value
	chaos_int32 value;

// create & destroy
public:
	// default constructor
	chaos_color() :
		value(0)
	{
	}
	// copy constructor
	chaos_color(const chaos_color& org) :
		value(org.value)
	{
	}
	// normal constructor
	explicit chaos_color(int red, int green, int blue) {
		set_rgb(red, green, blue);
	}
	// chaos_uint32 constructor
	explicit chaos_color(chaos_uint32 col) {
		set_rgb(
			CHAOS_RGBA_RED(col),
			CHAOS_RGBA_GREEN(col),
			CHAOS_RGBA_BLUE(col));
	}
	// destructor
	~chaos_color() {
	}

// get & set
public:
	// get color value
	chaos_int32 get_value() const {
		return value;
	}
	// set color value
	void set_value(chaos_int32 new_value) {
		value = new_value;
	}
	// get red
	int get_red() const {
		return (((value >> (int)red_ofs)-(int)rgb_min) & (int)rgb_mask)+
			(int)rgb_min;
	}
	// set red
	void set_red(int red) {
		set_rgb(red, get_green(), get_blue());
	}
	// get green
	int get_green() const {
		return (((value >> (int)green_ofs)-(int)rgb_min) & (int)rgb_mask)+
			(int)rgb_min;
	}
	// set green
	void set_green(int green) {
		set_rgb(get_red(), green, get_blue());
	}
	// get blue
	int get_blue() const {
		return (((value >> (int)blue_ofs)-(int)rgb_min) & (int)rgb_mask)+
			(int)rgb_min;
	}
	// set blue
	void set_blue(int blue) {
		set_rgb(get_red(), get_green(), blue);
	}
	// set rgb
	void set_rgb(int red, int green, int blue) {
		int red2 = red, green2 = green, blue2 = blue;
		if (red2 < (int)rgb_min) {
			red2 = (int)rgb_min;
		} else if (red2 > (int)rgb_max) {
			red2 = (int)rgb_max;
		}
		if (green2 < (int)rgb_min) {
			green2 = (int)rgb_min;
		} else if (green2 > (int)rgb_max) {
			green2 = (int)rgb_max;
		}
		if (blue2 < (int)rgb_min) {
			blue2 = (int)rgb_min;
		} else if (blue2 > (int)rgb_max) {
			blue2 = (int)rgb_max;
		}
		value = (chaos_int32)(
			((red2 & (int)rgb_mask) << (int)red_ofs) |
			((green2 & (int)rgb_mask) << (int)green_ofs) |
			((blue2 & (int)rgb_mask) << (int)blue_ofs));
	}

// operator
public:
	// let
	chaos_color& operator=(const chaos_color& org) {
		value = org.value;
		return *this;
	}
	// add
	chaos_color& operator+=(const chaos_color& org) {
		*this = chaos_color(
			get_red()+org.get_red(),
			get_green()+org.get_green(),
			get_blue()+org.get_blue());
		return *this;
	}
	// sub
	chaos_color& operator-=(const chaos_color& org) {
		*this = chaos_color(
			get_red()-org.get_red(),
			get_green()-org.get_green(),
			get_blue()-org.get_blue());
		return *this;
	}
	// plus
	chaos_color operator+(const chaos_color& org) const {
		chaos_color tmp = *this;
		tmp += org;
		return tmp;
	}
	// minus
	chaos_color operator-(const chaos_color& org) const {
		chaos_color tmp = *this;
		tmp -= org;
		return tmp;
	}
	// equal
	bool operator==(const chaos_color& org) const {
		return value == org.value;
	}
	// not equal
	bool operator!=(const chaos_color& org) const {
		return value != org.value;
	}
	// less
	bool operator<(const chaos_color& org) const {
		return value < org.value;
	}
	// less equal
	bool operator<=(const chaos_color& org) const {
		return value <= org.value;
	}
	// greater
	bool operator>(const chaos_color& org) const {
		return value > org.value;
	}
	// greater equal
	bool operator>=(const chaos_color& org) const {
		return value >= org.value;
	}
	// chaos_uint32 cast
	operator chaos_uint32() const {
		int red = get_red(),
			green = get_green(),
			blue = get_blue();
		if (red < 0) {
			red = 0;
		} else if (red > 255) {
			red = 255;
		}
		if (green < 0) {
			green = 0;
		} else if (green > 255) {
			green = 255;
		}
		if (blue < 0) {
			blue = 0;
		} else if (blue > 255) {
			blue = 255;
		}
		return CHAOS_RGBA(red, green, blue, 0);
	}

// operation
	// normalize
	chaos_color normalize() const {
		int red = get_red(),
			green = get_green(),
			blue = get_blue();
		if (red < 0) {
			red = 0;
		} else if (red > 255) {
			red = 255;
		}
		if (green < 0) {
			green = 0;
		} else if (green > 255) {
			green = 255;
		}
		if (blue < 0) {
			blue = 0;
		} else if (blue > 255) {
			blue = 255;
		}
		return chaos_color(red, green, blue);
	}
};

#endif // CHAOS_COLOR_H_INCLUDED
