////////////////////////////////////////////////////////////////////////////////
// chaos_func_wipe.cpp : wipe function

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_func_wipe.h"
#include "chaos_image_realizer.h"
#include "chaos_exception.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_func_wipe : wipe function class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_func_wipe::chaos_func_wipe() :
	chaos_func()
{
}

// destructor

chaos_func_wipe::~chaos_func_wipe() {
}

////////////////////////////////////////////////////////////////////////////////
// function

// get name

chaos_string chaos_func_wipe::get_name() const {
	return "wipe";
}

// execute

chaos_value chaos_func_wipe::execute(
	chaos_system* system,
	chaos_interpreter* /*interpreter*/,
	params_type* params) const
{
	// format:
	//   #Wipe(val1, val2, val3, types)
	// params:
	//   val1  : wipe value 1(integer)
	//   val2  : wipe value 2(integer)
	//   val3  : wipe value 3(integer)
	//   types : wipe type(integer)
	//     type = 2 : v-scroll
	//       val1 -> image  : source image number(integer)
	//       val2 -> rolls  : scroll unit(integer)
	//       val3 -> height : scroll height(integer)
	//     type = 5 : v-blind
	//       val1 -> image  : source image number(integer)
	//       val2 -> gap    : blind gap(integer)
	//       val3 -> unk    : unknown - blind width?(integer)
	//     type = 6 : h-blind
	//       val1 -> image  : source image number(integer)
	//       val2 -> gap    : blind gap(integer)
	//       val3 -> unk    : unknown - blind height?(integer)
	//     type = 7 : shake
	//       val1 -> move   : move dot(integer)
	//       val2 -> action : shake action type(integer)
	//       val3 -> unk    : unknown(integer)
	// example:
	//   #Wipe(3, -12, 600, 2);
	//   #Wipe(1, 16, 800, 5);
	//   #Wipe(1, 4, 600, 6);
	//   #Wipe(40, 1, 6, 7);
	if (params->size() < 4) {
		LOG_ERR(
			"error : func wipe::execute - "
			"too few params\n");
		throw chaos_exception(chaos_exception::bad_format);
	}
	chaos_func::params_type::iterator it = params->begin();
	int val1 = (int)(*(it++)).get_integer();
	int val2 = (int)(*(it++)).get_integer();
	int val3 = (int)(*(it++)).get_integer();
	int types = (int)(*it).get_integer();
	chaos_string prev_info, now_info, next_info;
	prev_info = next_info = system->get_image_info(0);
	switch (types) {
	case 2:		// v-scroll
		{ // dummy block
			int image = val1;
			int rolls = val2;
			int height = val3;
			next_info = system->get_image_info(image);
			now_info = chaos_image_realizer::encode_wipe_v_scroll_info(
				0,
				rolls, height,
				prev_info, next_info);
		}
		break;
	case 5:		// v-blind
		{ // dummy block
			int image = val1;
			int gap = val2;
			int unk = val3;
			next_info = system->get_image_info(image);
			now_info = chaos_image_realizer::encode_wipe_v_blind_info(
				0,
				gap, unk,
				prev_info, next_info);
		}
		break;
	case 6:		// h-blind
		{ // dummy block
			int image = val1;
			int gap = val2;
			int unk = val3;
			next_info = system->get_image_info(image);
			now_info = chaos_image_realizer::encode_wipe_h_blind_info(
				0,
				gap, unk,
				prev_info, next_info);
		}
		break;
	case 7:		// shake
		{ // dummy block
			int move = val1;
			int action = val2;
			int unk = val3;
			next_info = prev_info;
			now_info = chaos_image_realizer::encode_wipe_shake_info(
				0,
				move, action, unk,
				next_info);
		}
		break;
	}
	int effect_step = system->get_effect_step();
	if (effect_step < 1) {
		effect_step = 1;
	}
	if (system->is_effect_cut()) {
		now_info = "";
	}
	bool dirty_image = true, dirty_message = false;
	if (system->get_skip_mode() != chaos_system::skip_none) {
		switch (system->get_skip_speed()) {
		case chaos_system::sspeed_fast:
			now_info = "";
			dirty_message = true;
			break;
		case chaos_system::sspeed_jump:
			now_info = "";
			dirty_image = false;
			break;
		}
	}
	system->set_image_info(1, next_info);
	if (now_info.empty()) {
		system->set_image_info(0, next_info);
	} else {
		if (effect_step > 1) {
			if (!chaos_image_realizer::is_effect_spippable(
					now_info))
			{
				effect_step = 1;
			}
		}
		system->set_image_info(0, now_info);
		system->change_system_mode(chaos_system::sysmode_effect);
		system->set_wait_time(
			(int)chaos_system::effect_unit_time*
				effect_step*
				chaos_image_realizer::get_effect_time_scale(
					now_info));
	}
	if (dirty_image) {
		system->set_dirty_image(true);
	}
	if (dirty_message) {
		system->set_dirty_message(true);
	}
	return chaos_value();
}
