////////////////////////////////////////////////////////////////////////////////
// chaos_image.cpp : image

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_image.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_image : image class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_image::chaos_image() :
	width(0), height(0),
	scanline_length(0),
	capacity(0),
	bits(NULL)
{
}

// destructor

chaos_image::~chaos_image() {
	if (bits != NULL) {
		free(bits);
		bits = NULL;
	}
}

// create image

bool chaos_image::create(
	int new_width, int new_height,
	bool clear_bits,
	chaos_uint32 clear_col)
{
	ASSERT((new_width > 0) && (new_height > 0));
	int new_scanline_length = new_width;
	int total_length = new_scanline_length*new_height;
	chaos_uint32* new_bits = bits;
	int bits_size = total_length*sizeof(chaos_uint32);
	if (bits == NULL) {
		new_bits = (chaos_uint32*)malloc(bits_size);
	} else if (total_length > capacity) {
		new_bits = (chaos_uint32*)realloc(bits, bits_size);
	}
	if (new_bits == NULL) {
		throw std::bad_alloc();
	}
	width = new_width;
	height = new_height;
	scanline_length = new_scanline_length;
	bits = new_bits;
	if (total_length > capacity) {
		capacity = total_length;
	}
	if (clear_bits) {
		fill(clear_col);
	}
	return true;
}

// destroy image

void chaos_image::destroy() {
	if (bits != NULL) {
		free(bits);
		bits = NULL;
	}
	width = 0;
	height = 0;
	scanline_length = 0;
	capacity = 0;
}

////////////////////////////////////////////////////////////////////////////////
// operation

// invert alpha

void chaos_image::invert_alpha() {
	chaos_uint8* inv_bits =
		(chaos_uint8*)get_bits(0, 0)+CHAOS_RGBA_ALPHA_OFS;
	int inv_step = (scanline_length-width)*4;
	int y = height;
	do {
		int x = width;
		do {
			*inv_bits = (chaos_uint8)(255-*inv_bits);
			inv_bits += 4;
		} while (--x != 0);
		inv_bits += inv_step;
	} while (--y != 0);
}

// fill

void chaos_image::fill(
	chaos_uint32 fill_col,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* fill_bits =
			get_bits(fill_rect.get_lt_pos());
		int fill_step = scanline_length-fill_rect.width;
		if (fill_step == 0) {
			int red = CHAOS_RGBA_RED(fill_col),
				green = CHAOS_RGBA_GREEN(fill_col),
				blue = CHAOS_RGBA_BLUE(fill_col),
				alpha = CHAOS_RGBA_ALPHA(fill_col);
			if ((red == green) && (red == blue) && (red == alpha)) {
				memset(
					fill_bits, red,
					fill_rect.width*fill_rect.height*sizeof(chaos_uint32));
			} else {
				int i = fill_rect.width*fill_rect.height;
				do {
					*(fill_bits++) = fill_col;
				} while (--i != 0);
			}
		} else {
			int y = fill_rect.height;
			do {
				int x = fill_rect.width;
				do {
					*(fill_bits++) = fill_col;
				} while (--x != 0);
				fill_bits += fill_step;
			} while (--y != 0);
		}
	}
}

// fill width alpha

void chaos_image::alpha_fill(
	chaos_uint32 fill_col,
	const chaos_image* alpha_image,
	chaos_pos fill_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	fill_rect &= chaos_rect(fill_pos, alpha_image->get_dim());
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		const chaos_uint8* alpha_bits = (chaos_uint8*)
			alpha_image->get_bits(fill_rect.get_lt_pos()-fill_pos)+
				CHAOS_RGBA_ALPHA_OFS;
		int dst_step = scanline_length-fill_rect.width,
			alpha_step =
				(alpha_image->get_scanline_length()-fill_rect.width)*4;
		int fill_red = CHAOS_RGBA_RED(fill_col),
			fill_green = CHAOS_RGBA_GREEN(fill_col),
			fill_blue = CHAOS_RGBA_BLUE(fill_col);
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				int alpha = *alpha_bits;
				alpha_bits += 4;
				alpha += (alpha+1)/256;
				int alpha2 = 256-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						fill_red*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						fill_green*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						fill_blue*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			alpha_bits += alpha_step;
		} while (--y != 0);
	}
}

// fade fill

void chaos_image::fade_fill(
	chaos_uint32 fill_col,
	int fade,
	const chaos_rect* clip_rect)
{
	if (fade >= 255) {
		return;
	}
	if (fade <= 0) {
		fill(fill_col, clip_rect);
		return;
	}
	int alpha = fade+(fade+1)/256;
	int alpha2 = 256-alpha;
	int fill_red = CHAOS_RGBA_RED(fill_col)*alpha2,
		fill_green = CHAOS_RGBA_GREEN(fill_col)*alpha2,
		fill_blue = CHAOS_RGBA_BLUE(fill_col)*alpha2;
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						fill_red)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						fill_green)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						fill_blue)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
		} while (--y != 0);
	}
}

// transparent fill

void chaos_image::trans_fill(
	int trans_red, int trans_green, int trans_blue,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				int red = CHAOS_RGBA_RED(dst_col)+trans_red,
					green = CHAOS_RGBA_GREEN(dst_col)+trans_green,
					blue = CHAOS_RGBA_BLUE(dst_col)+trans_blue,
					alpha = CHAOS_RGBA_ALPHA(dst_col);
				if (red < 0) {
					red = 0;
				} else if (red > 255) {
					red = 255;
				}
				if (green < 0) {
					green = 0;
				} else if (green > 255) {
					green = 255;
				}
				if (blue < 0) {
					blue = 0;
				} else if (blue > 255) {
					blue = 255;
				}
				*(dst_bits++) = CHAOS_RGBA(red, green, blue, alpha);
			} while (--x != 0);
			dst_bits += dst_step;
		} while (--y != 0);
	}
}

// mono fill

void chaos_image::mono_fill(
	int mono_red, int mono_green, int mono_blue,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				int red = CHAOS_RGBA_RED(dst_col),
					green = CHAOS_RGBA_GREEN(dst_col),
					blue = CHAOS_RGBA_BLUE(dst_col),
					alpha = CHAOS_RGBA_ALPHA(dst_col);
				red = green = blue = (red+green+blue)/3;
				red += mono_red;
				green += mono_green;
				blue += mono_blue;
				if (red < 0) {
					red = 0;
				} else if (red > 255) {
					red = 255;
				}
				if (green < 0) {
					green = 0;
				} else if (green > 255) {
					green = 255;
				}
				if (blue < 0) {
					blue = 0;
				} else if (blue > 255) {
					blue = 255;
				}
				*(dst_bits++) = CHAOS_RGBA(red, green, blue, alpha);
			} while (--x != 0);
			dst_bits += dst_step;
		} while (--y != 0);
	}
}

// horizontal blind fill

void chaos_image::horz_blind_fill(
	chaos_uint32 fill_col,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		fill(fill_col, clip_rect);
		return;
	}
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y_start = fill_rect.y,
			y_end = fill_rect.y+fill_rect.height;
		if (gap > 0) {
			int y = y_start;
			do {
				if (y%agap < fat) {
					int x = fill_rect.width;
					do {
						*(dst_bits++) = fill_col;
					} while (--x > 0);
				} else {
					dst_bits += fill_rect.width;
				}
				dst_bits += dst_step;
			} while (++y < y_end);
		} else {
			int fat2 = agap-fat;
			int y = y_start;
			do {
				if (y%agap >= fat2) {
					int x = fill_rect.width;
					do {
						*(dst_bits++) = fill_col;
					} while (--x > 0);
				} else {
					dst_bits += fill_rect.width;
				}
				dst_bits += dst_step;
			} while (++y < y_end);
		}
	}
}

// vertical blind fill

void chaos_image::vert_blind_fill(
	chaos_uint32 fill_col,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		fill(fill_col, clip_rect);
		return;
	}
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int x_start = fill_rect.x,
			x_end = fill_rect.x+fill_rect.width;
		if (gap > 0) {
			int y = fill_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap < fat) {
						*dst_bits = fill_col;
					}
					dst_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
			} while (--y > 0);
		} else {
			int fat2 = agap-fat;
			int y = fill_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap >= fat2) {
						*dst_bits = fill_col;
					}
					dst_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
			} while (--y > 0);
		}
	}
}

// blt

void chaos_image::blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		if ((dst_step == 0) && (src_step == 0)) {
			memcpy(
				dst_bits, src_bits,
				blt_rect.width*blt_rect.height*sizeof(chaos_uint32));
		} else {
			int y = blt_rect.height;
			do {
				int x = blt_rect.width;
				do {
					*(dst_bits++) = *(src_bits++);
				} while (--x != 0);
				dst_bits += dst_step;
				src_bits += src_step;
			} while (--y != 0);
		}
	}
}

// stretch blt

void chaos_image::stretch_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	chaos_dim blt_dim,
	const chaos_rect* clip_rect)
{
	chaos_dim src_dim = blt_image->get_dim();
	if (src_dim.empty()) {
		return;
	}
	if (src_dim == blt_dim) {
		blt(blt_image, blt_pos, clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_dim);
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width;
		chaos_rect dst_rect = blt_rect-blt_pos;
		int dst_left = blt_rect.x-blt_pos.x,
			dst_top = blt_rect.y-blt_pos.y,
			dst_right = dst_left+dst_rect.width,
			dst_bottom = dst_top+dst_rect.height;
		for (int y = dst_top; y < dst_bottom; y++) {
			int y2 = (y*src_dim.height)/blt_dim.height;
			const chaos_uint32* src_bits =
				blt_image->get_bits(0, y2);
			for (int x = dst_left; x < dst_right; x++) {
				int x2 = (x*src_dim.width)/blt_dim.width;
				*(dst_bits++) = *(src_bits+x2);
			}
			dst_bits += dst_step;
		}
	}
}

// smooth stretch blt(zoom out only)

void chaos_image::smooth_stretch_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	chaos_dim blt_dim,
	const chaos_rect* clip_rect)
{
	chaos_dim src_dim = blt_image->get_dim();
	if (src_dim.empty()) {
		return;
	}
	if ((src_dim.width <= blt_dim.width) ||
		(src_dim.height <= blt_dim.height))
	{
		stretch_blt(
			blt_image,
			blt_pos,
			blt_dim,
			clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_dim);
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width;
		chaos_rect dst_rect = blt_rect-blt_pos;
		int dst_left = blt_rect.x-blt_pos.x,
			dst_top = blt_rect.y-blt_pos.y,
			dst_right = dst_left+dst_rect.width,
			dst_bottom = dst_top+dst_rect.height;
		int src_y1,
			src_y2 = (dst_top*src_dim.height)/blt_dim.height;
		for (int dst_y = dst_top; dst_y < dst_bottom; dst_y++) {
			src_y1 = src_y2;
			src_y2 = ((dst_y+1)*src_dim.height)/blt_dim.height;
			ASSERT((src_y1 >= 0) && (src_y1 < blt_image->height));
			ASSERT((src_y2 > src_y1) && (src_y2 <= blt_image->height));
			int src_x1,
				src_x2 = (dst_left*src_dim.width)/blt_dim.width;
			for (int dst_x = dst_left; dst_x < dst_right; dst_x++) {
				src_x1 = src_x2;
				src_x2 = ((dst_x+1)*src_dim.width)/blt_dim.width;
				ASSERT((src_x1 >= 0) && (src_x1 < blt_image->width));
				ASSERT((src_x2 > src_x1) && (src_x2 <= blt_image->width));
				int dot_count = (src_x2-src_x1)*(src_y2-src_y1);
				chaos_uint32 dst_col = CHAOS_RGBA(0, 0, 0, 0);
				if (dot_count > 0) {
					double red = 0, green = 0, blue = 0;
					const chaos_uint32* src_bits =
						blt_image->get_bits(src_x1, src_y1);
					int src_step = blt_image->get_scanline_length()-
						(src_x2-src_x1);
					for (int src_y = src_y1; src_y < src_y2; src_y++) {
						for (int src_x = src_x1; src_x < src_x2; src_x++) {
							chaos_uint32 src_col = *(src_bits++);
							red += CHAOS_RGBA_RED(src_col);
							green += CHAOS_RGBA_GREEN(src_col);
							blue += CHAOS_RGBA_BLUE(src_col);
						}
						src_bits += src_step;
					}
					dst_col = CHAOS_RGBA(
						(int)(red/dot_count+0.5),
						(int)(green/dot_count+0.5),
						(int)(blue/dot_count+0.5),
						0);
				}
				*(dst_bits++) = dst_col;
			}
			dst_bits += dst_step;
		}
	}
}

// blt width alpha

void chaos_image::alpha_blt(
	const chaos_image* blt_image,
	const chaos_image* alpha_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	blt_rect &= chaos_rect(blt_pos, alpha_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		const chaos_uint8* alpha_bits = (chaos_uint8*)
			alpha_image->get_bits(blt_rect.get_lt_pos()-blt_pos)+
				CHAOS_RGBA_ALPHA_OFS;
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width,
			alpha_step =
				(alpha_image->get_scanline_length()-blt_rect.width)*4;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = *alpha_bits;
				alpha_bits += 4;
				alpha += (alpha+1)/256;
				int alpha2 = 256-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						CHAOS_RGBA_RED(src_col)*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						CHAOS_RGBA_GREEN(src_col)*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						CHAOS_RGBA_BLUE(src_col)*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
			alpha_bits += alpha_step;
		} while (--y != 0);
	}
}

// blt width alpha(single image)

void chaos_image::alpha_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = CHAOS_RGBA_ALPHA(src_col);
				alpha += (alpha+1)/256;
				int alpha2 = 256-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						CHAOS_RGBA_RED(src_col)*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						CHAOS_RGBA_GREEN(src_col)*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						CHAOS_RGBA_BLUE(src_col)*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
		} while (--y != 0);
	}
}

// blt width alpha

void chaos_image::mask_blt(
	const chaos_image* blt_image,
	const chaos_image* mask_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	blt_rect &= chaos_rect(blt_pos, mask_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		const chaos_uint8* mask_bits = (chaos_uint8*)
			mask_image->get_bits(blt_rect.get_lt_pos()-blt_pos)+
				CHAOS_RGBA_ALPHA_OFS;
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width,
			mask_step =
				(mask_image->get_scanline_length()-blt_rect.width)*4;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = *mask_bits;
				mask_bits += 4;
				int alpha2 = 255-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col) & alpha) |
						(CHAOS_RGBA_RED(src_col) & alpha2),
					(CHAOS_RGBA_GREEN(dst_col) & alpha) |
						(CHAOS_RGBA_GREEN(src_col)& alpha2),
					(CHAOS_RGBA_BLUE(dst_col) & alpha) |
						(CHAOS_RGBA_BLUE(src_col) & alpha2),
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
			mask_bits += mask_step;
		} while (--y != 0);
	}
}

// blt width mask(single image)

void chaos_image::mask_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = CHAOS_RGBA_ALPHA(src_col);
				int alpha2 = 255-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col) & alpha) |
						(CHAOS_RGBA_RED(src_col) & alpha2),
					(CHAOS_RGBA_GREEN(dst_col) & alpha) |
						(CHAOS_RGBA_GREEN(src_col)& alpha2),
					(CHAOS_RGBA_BLUE(dst_col) & alpha) |
						(CHAOS_RGBA_BLUE(src_col) & alpha2),
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
		} while (--y != 0);
	}
}

// fade

void chaos_image::fade_blt(
	const chaos_image* blt_image,
	int fade,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	if (fade >= 255) {
		return;
	}
	if (fade <= 0) {
		blt(blt_image, blt_pos, clip_rect);
		return;
	}
	int alpha = fade+(fade+1)/256;
	int alpha2 = 256-alpha;
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						CHAOS_RGBA_RED(src_col)*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						CHAOS_RGBA_GREEN(src_col)*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						CHAOS_RGBA_BLUE(src_col)*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
		} while (--y != 0);
	}
}

// horizontal blind blt

void chaos_image::horz_blind_blt(
	const chaos_image* blind_image,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		blt(blind_image, chaos_pos(0, 0), clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(chaos_pos(0, 0), blind_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blind_image->get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width,
			src_step = blind_image->get_scanline_length()-blt_rect.width;
		int y_start = blt_rect.y,
			y_end = blt_rect.y+blt_rect.height;
		if (gap > 0) {
			int y = y_start;
			do {
				if (y%agap < fat) {
					int x = blt_rect.width;
					do {
						*(dst_bits++) = *(src_bits++);
					} while (--x > 0);
				} else {
					dst_bits += blt_rect.width;
					src_bits += blt_rect.width;
				}
				dst_bits += dst_step;
				src_bits += src_step;
			} while (++y < y_end);
		} else {
			int fat2 = agap-fat;
			int y = y_start;
			do {
				if (y%agap >= fat2) {
					int x = blt_rect.width;
					do {
						*(dst_bits++) = *(src_bits++);
					} while (--x > 0);
				} else {
					dst_bits += blt_rect.width;
					src_bits += blt_rect.width;
				}
				dst_bits += dst_step;
				src_bits += src_step;
			} while (++y < y_end);
		}
	}
}

// vertical blind blt

void chaos_image::vert_blind_blt(
	const chaos_image* blind_image,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		blt(blind_image, chaos_pos(0, 0), clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(chaos_pos(0, 0), blind_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blind_image->get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width,
			src_step = blind_image->get_scanline_length()-blt_rect.width;
		int x_start = blt_rect.x,
			x_end = blt_rect.x+blt_rect.width;
		if (gap > 0) {
			int y = blt_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap < fat) {
						*dst_bits = *src_bits;
					}
					dst_bits++;
					src_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
				src_bits += src_step;
			} while (--y > 0);
		} else {
			int fat2 = agap-fat;
			int y = blt_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap >= fat2) {
						*dst_bits = *src_bits;
					}
					dst_bits++;
					src_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
				src_bits += src_step;
			} while (--y > 0);
		}
	}
}
