////////////////////////////////////////////////////////////////////////////////
// chaos_image.h : image

#ifndef CHAOS_IMAGE_H_INCLUDED
#define CHAOS_IMAGE_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_image;

////////////////////////////////////////////////////////////////////////////////
// include

#ifndef CHAOS_POS_H_INCLUDED
#include "chaos_pos.h"
#endif // CHAOS_POS_H_INCLUDED

#ifndef CHAOS_DIM_H_INCLUDED
#include "chaos_dim.h"
#endif // CHAOS_DIM_H_INCLUDED

#ifndef CHAOS_RECT_H_INCLUDED
#include "chaos_rect.h"
#endif // CHAOS_RECT_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// chaos_image : image class

class chaos_image {
// attribute
private:
	// width
	int width;
	// height
	int height;
	// scanline length
	int scanline_length;
	// data capacity
	int capacity;
	// bit data
	chaos_uint32* bits;

// create & destroy
public:
	// default constructor
	chaos_image();
	// destructor
	~chaos_image();
	// create image
	bool create(
		int new_width, int new_height,
		bool clear_bits = true,
		chaos_uint32 clear_col = CHAOS_RGBA(0, 0, 0, 0));
	// create image
	bool create(
		chaos_dim new_dim,
		bool clear_bits = true,
		chaos_uint32 clear_col = CHAOS_RGBA(0, 0, 0, 0))
	{
		return create(
			new_dim.width, new_dim.height,
			clear_bits,
			clear_col);
	}
	// destroy image
	void destroy();

// get & set
public:
	// is valid
	bool is_valid() const {
		return bits != NULL;
	}
	// get width
	int get_width() const {
		return width;
	}
	// get height
	int get_height() const {
		return height;
	}
	// get dimension
	chaos_dim get_dim() const {
		return chaos_dim(width, height);
	}
	// get scanline length
	int get_scanline_length() const {
		return scanline_length;
	}
	// get bit data
	const chaos_uint32* get_bits() const {
		return bits;
	}
	chaos_uint32* get_bits() {
		return bits;
	}
	const chaos_uint32* get_bits(int x, int y) const {
		ASSERT((x >= 0) && (x < width));
		ASSERT((y >= 0) && (y < height));
		return bits+y*scanline_length+x;
	}
	chaos_uint32* get_bits(int x, int y) {
		ASSERT((x >= 0) && (x < width));
		ASSERT((y >= 0) && (y < height));
		return bits+y*scanline_length+x;
	}
	const chaos_uint32* get_bits(chaos_pos pos) const {
		ASSERT((pos.x >= 0) && (pos.x < width));
		ASSERT((pos.y >= 0) && (pos.y < height));
		return bits+pos.y*scanline_length+pos.x;
	}
	chaos_uint32* get_bits(chaos_pos pos) {
		ASSERT((pos.x >= 0) && (pos.x < width));
		ASSERT((pos.y >= 0) && (pos.y < height));
		return bits+pos.y*scanline_length+pos.x;
	}
	// get pixel
	chaos_uint32 get_pixel(int x, int y) const {
		chaos_uint32 pixel = 0;
		if ((x >= 0) && (x < width) &&
			(y >= 0) && (y < height))
		{
			pixel = *(bits+y*scanline_length+x);
		}
		return pixel;
	}
	chaos_uint32 get_pixel(chaos_pos pos) const {
		chaos_uint32 pixel = 0;
		if ((pos.x >= 0) && (pos.x < width) &&
			(pos.y >= 0) && (pos.y < height))
		{
			pixel = *(bits+pos.y*scanline_length+pos.x);
		}
		return pixel;
	}
	// set pixel
	void set_pixel(int x, int y, chaos_uint32 pixel) {
		if ((x >= 0) && (x < width) &&
			(y >= 0) && (y < height))
		{
			*(bits+y*scanline_length+x) = pixel;
		}
	}
	void set_pixel(chaos_pos pos, chaos_uint32 pixel) {
		if ((pos.x >= 0) && (pos.x < width) &&
			(pos.y >= 0) && (pos.y < height))
		{
			*(bits+pos.y*scanline_length+pos.x) = pixel;
		}
	}

// operation
public:
	// invert alpha
	void invert_alpha();
	// fill
	void fill(
		chaos_uint32 fill_col,
		const chaos_rect* clip_rect = NULL);
	// fill width alpha
	void alpha_fill(
		chaos_uint32 fill_col,
		const chaos_image* alpha_image,
		chaos_pos fill_pos,
		const chaos_rect* clip_rect = NULL);
	// fade fill
	void fade_fill(
		chaos_uint32 fill_col,
		int fade,
		const chaos_rect* clip_rect = NULL);
	// transparent fill
	void trans_fill(
		int trans_red, int trans_green, int trans_blue,
		const chaos_rect* clip_rect = NULL);
	// mono fill
	void mono_fill(
		int mono_red, int mono_green, int mono_blue,
		const chaos_rect* clip_rect = NULL);
	// horizontal blind fill
	void horz_blind_fill(
		chaos_uint32 fill_col,
		int gap, int fat,
		const chaos_rect* clip_rect = NULL);
	// vertical blind fill
	void vert_blind_fill(
		chaos_uint32 fill_col,
		int gap, int fat,
		const chaos_rect* clip_rect = NULL);
	// blt
	void blt(
		const chaos_image* blt_image,
		chaos_pos blt_pos,
		const chaos_rect* clip_rect = NULL);
	// stretch blt
	void stretch_blt(
		const chaos_image* blt_image,
		chaos_pos blt_pos,
		chaos_dim blt_dim,
		const chaos_rect* clip_rect = NULL);
	// smooth stretch blt(zoom out only)
	void smooth_stretch_blt(
		const chaos_image* blt_image,
		chaos_pos blt_pos,
		chaos_dim blt_dim,
		const chaos_rect* clip_rect = NULL);
	// blt width alpha
	void alpha_blt(
		const chaos_image* blt_image,
		const chaos_image* alpha_image,
		chaos_pos blt_pos,
		const chaos_rect* clip_rect = NULL);
	// blt width alpha(single image)
	void alpha_blt(
		const chaos_image* blt_image,
		chaos_pos blt_pos,
		const chaos_rect* clip_rect = NULL);
	// blt width mask
	void mask_blt(
		const chaos_image* blt_image,
		const chaos_image* mask_image,
		chaos_pos blt_pos,
		const chaos_rect* clip_rect = NULL);
	// blt width mask(single image)
	void mask_blt(
		const chaos_image* blt_image,
		chaos_pos blt_pos,
		const chaos_rect* clip_rect = NULL);
	// fade blt
	void fade_blt(
		const chaos_image* blt_image,
		int fade,
		chaos_pos blt_pos,
		const chaos_rect* clip_rect = NULL);
	// horizontal blind blt
	void horz_blind_blt(
		const chaos_image* blind_image,
		int gap, int fat,
		const chaos_rect* clip_rect = NULL);
	// vertical blind blt
	void vert_blind_blt(
		const chaos_image* blind_image,
		int gap, int fat,
		const chaos_rect* clip_rect = NULL);
};

#endif // CHAOS_IMAGE_H_INCLUDED
