////////////////////////////////////////////////////////////////////////////////
// chaos_image_manager.cpp : image manager

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_image_manager.h"
#include "chaos_file_serializer.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_image_manager : image manager class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_image_manager::chaos_image_manager() :
	main_image(NULL),
	prompt_image(NULL),
	image_realizer(NULL),
	message_imager(NULL)
{
	main_image = new chaos_image;
	prompt_image = new chaos_image;
	image_realizer = new chaos_image_realizer;
	message_imager = new chaos_message_imager;
}

// destructor

chaos_image_manager::~chaos_image_manager() {
	dispose();
	delete main_image;
	delete prompt_image;
	delete image_realizer;
	delete message_imager;
}

// initialize

bool chaos_image_manager::initialize(
	chaos_dim sys_dim,
	const chaos_string& font_name,
	int font_size,
	bool font_bold,
	bool font_anti_alias,
	int image_buffer_count)
{
	if (!main_image->is_valid()) {
		if (!main_image->create(sys_dim)) {
			return false;
		}
	}
	if (!prompt_image->is_valid()) {
		if (!prompt_image->create(32, 32)) {
			return false;
		}
	}
	if (!message_imager->initialize(sys_dim)) {
		return false;
	}
	message_imager->set_font_name(font_name);
	message_imager->set_font_size(font_size);
	message_imager->set_font_bold(font_bold);
	message_imager->set_font_anti_alias(font_anti_alias);
	if (!image_realizer->initialize(image_buffer_count, sys_dim)) {
		return false;
	}
	chaos_image sysetc_image;
	if (image_realizer->realize_image(
			&sysetc_image,
			chaos_image_realizer::encode_load_info("./sysetc.bmp")))
	{
		int color_x = 64, color_y = 0,
			alpha_x = 96, alpha_y = 0;
		for (int y = 0; y < prompt_image->get_height(); y++) {
			for (int x = 0; x < prompt_image->get_width(); x++) {
				chaos_uint32 col1 =
					sysetc_image.get_pixel(color_x+x, color_y+y),
						col2 =
					sysetc_image.get_pixel(alpha_x+x, alpha_y+y);
				int red = CHAOS_RGBA_RED(col1),
					green = CHAOS_RGBA_GREEN(col1),
					blue = CHAOS_RGBA_BLUE(col1),
					alpha = (CHAOS_RGBA_RED(col2)+
						CHAOS_RGBA_GREEN(col2)+
						CHAOS_RGBA_BLUE(col2))/3;
				prompt_image->set_pixel(
					x, y,
					CHAOS_RGBA(red, green, blue, alpha));
			}
		}
	}
	return true;
}

// dispose

void chaos_image_manager::dispose() {
	image_realizer->dispose();
	message_imager->dispose();
}

// change image buffer count

bool chaos_image_manager::change_image_buffer_count(
	int image_buffer_count)
{
	bool result = true;
	if (image_buffer_count != image_realizer->get_reserve_count()) {
		result = image_realizer->initialize(
			image_buffer_count,
			image_realizer->get_sys_dim());
	}
	return result;
}

// change font

bool chaos_image_manager::change_font(
	const chaos_string& font_name,
	int font_size,
	bool font_bold,
	bool font_anti_alias)
{
	if (font_name != message_imager->get_font_name()) {
		message_imager->set_font_name(font_name);
	}
	if (font_size != message_imager->get_font_size()) {
		message_imager->set_font_size(font_size);
	}
	if (font_bold != message_imager->is_font_bold()) {
		message_imager->set_font_bold(font_bold);
	}
	if (font_anti_alias != message_imager->is_font_anti_alias()) {
		message_imager->set_font_anti_alias(font_anti_alias);
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// draw

// draw main image

void chaos_image_manager::draw_main_image(
	const chaos_string& info,
	const chaos_rect* sel_rect,
	chaos_color sel_color,
	bool prompt_visible, bool prompt_place,
	chaos_pos prompt_place_pos,
	chaos_rect* update_rect,
	const chaos_rect* clip_rect)
{
	ASSERT(main_image->is_valid());
	chaos_rect main_image_rect(
		chaos_pos(0, 0), main_image->get_dim());
	chaos_pos base_pos =
		chaos_image_realizer::extract_base_position(info);
	bool clipping = false;
	chaos_rect real_clip_rect;
	if (clip_rect != NULL) {
		clipping = true;
		real_clip_rect = *clip_rect+base_pos;
		real_clip_rect &= main_image_rect;
	}
	// background image
	image_realizer->realize_image(
		main_image,
		info,
		clipping? &real_clip_rect: NULL);
	if (update_rect != NULL) {
		chaos_rect background_rect = main_image_rect;
		if (clipping) {
			background_rect &= real_clip_rect;
		}
		*update_rect |= background_rect;
	}
	// selection
	chaos_rect sel_fill_rect = *sel_rect+base_pos;
	if (clipping) {
		sel_fill_rect &= real_clip_rect;
	}
	if (!sel_fill_rect.empty()) {
		main_image->trans_fill(
			sel_color.get_red(),
			sel_color.get_green(),
			sel_color.get_blue(),
			&sel_fill_rect);
		if (update_rect != NULL) {
			*update_rect |= sel_fill_rect;
		}
	}
	// message image
	chaos_rect message_valid_rect =
		message_imager->get_message_valid_rect()+base_pos;
	if (clipping) {
		message_valid_rect &= real_clip_rect;
	}
	if (!message_valid_rect.empty()) {
		if (message_imager->is_font_anti_alias()) {
			main_image->alpha_blt(
				message_imager->get_message_image(),
				base_pos,
				&message_valid_rect);
		} else {
			main_image->mask_blt(
				message_imager->get_message_image(),
				base_pos,
				&message_valid_rect);
		}
		if (update_rect != NULL) {
			*update_rect |= message_valid_rect;
		}
	}
	// prompt
	if (prompt_visible) {
		ASSERT(prompt_image->is_valid());
		chaos_pos prompt_pos = base_pos;
		if (prompt_place) {
			prompt_pos += prompt_place_pos;
		} else {
			prompt_pos += message_imager->get_message_pos();
		}
		chaos_rect prompt_rect(
			prompt_pos,
			prompt_image->get_dim());
		if (clipping) {
			prompt_rect &= real_clip_rect;
		}
		main_image->alpha_blt(
			prompt_image,
			prompt_pos,
			&prompt_rect);
		if (update_rect != NULL) {
			*update_rect |= prompt_rect;
		}
	}
	if (update_rect != NULL) {
		*update_rect &= main_image_rect;
	}
}

////////////////////////////////////////////////////////////////////////////////
// message

// update message image

void chaos_image_manager::update_message_image(
	const chaos_string& message_text,
	chaos_rect* update_rect)
{
	message_imager->update_message_text(
		message_text,
		update_rect);
}

// clear message image

void chaos_image_manager::clear_message_image() {
	message_imager->clear_message_text();
}
