////////////////////////////////////////////////////////////////////////////////
// chaos_image_realizer.cpp : image realizer

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_image_realizer.h"
#include "chaos_file_serializer.h"
#include "chaos_png_image_loader.h"
#include "chaos_bmp_image_loader.h"
#include "chaos_utility.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_image_realizer : image realizer class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// normal constructor

chaos_image_realizer::chaos_image_realizer() :
	reserve_count(10),
	sys_dim(800, 600)
{
}

// destructor

chaos_image_realizer::~chaos_image_realizer() {
	dispose();
}

// initialize

bool chaos_image_realizer::initialize(
	int new_reserve_count,
	chaos_dim new_sys_dim)
{
	reserve_count = new_reserve_count;
	if (reserve_count < 1) {
		reserve_count = 1;
	}
	sys_dim = new_sys_dim;
	if (get_image_count() < reserve_count) {
		while (get_image_count() < reserve_count) {
			if (!add_image()) {
				return false;
			}
			ASSERT(images.size() > 0);
			if (!images.back().image->create(sys_dim, false)) {
				return false;
			}
		}
	} else if (get_image_count() > reserve_count) {
		delete_excessimages();
	}
	return true;
}

// dispose

void chaos_image_realizer::dispose() {
	clear();
}

// clear

void chaos_image_realizer::clear() {
	while (images.size() > 0) {
		delete_image();
	}
}

////////////////////////////////////////////////////////////////////////////////
// lock

// lock image

bool chaos_image_realizer::lock_image(int id) {
	if ((id < 0) || (id >= get_image_count())) {
		return false;
	}
	if (images[id].lock++ <= 0) {
		image_queue_type::iterator target_it = std::find(
			image_queue.begin(), image_queue.end(),
			id);
		ASSERT(target_it != image_queue.end());
		if (target_it != image_queue.end()) {
			for (
				image_queue_type::iterator it = image_queue.begin();
				it != image_queue.end();
				it++)
			{
				int check_id = *it;
				ASSERT(check_id < (int)images.size());
				if (check_id == id) {
					ASSERT(it == target_it);
					break;
				} else if (images[check_id].lock <= 0) {
					image_queue.erase(target_it);
					image_queue.insert(it, id);
					break;
				}
			}
		}
		ASSERT(images.size() == image_queue.size());
	}
	return true;
}

// unlock image

bool chaos_image_realizer::unlock_image(int id) {
	if ((id < 0) || (id >= get_image_count())) {
		return false;
	}
	images[id].lock--;
	ASSERT(images[id].lock >= 0);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// info

// extract sub build info

chaos_string chaos_image_realizer::extract_subinfo(const char*& infostr) {
	chaos_string subinfo;
	if (chaos_utility::is_alpha(*infostr) &&
		(*(infostr+1) == '('))
	{
		const char* first = infostr;
		infostr += 2;
		int paren = 1;
		while (paren > 0) {
			switch (*infostr) {
			case '\0':
				// error
				paren--;
				break;
			case '(':
				paren++;
				infostr++;
				break;
			case ')':
				paren--;
				infostr++;
				break;
			case '"':
			case '\'':
				chaos_utility::str_from_str(infostr);
				break;
			default:
				infostr++;
				break;
			}
		}
		subinfo = chaos_string(first, infostr);
	}
	return subinfo;
}

// get wipe shake base position

chaos_pos chaos_image_realizer::get_wipe_shake_base_position(
	int step,
	int move, int /*action*/, int /*unk*/)
{
	chaos_pos ofs;
	switch ((step/2)%3) {
	case 0:
		ofs.y = (step%2 != 0)? move: -move;
		break;
	case 1:
		ofs.x = (step%2 != 0)? move: -move;
		break;
	}
	return ofs;
}

// encode load info

chaos_string chaos_image_realizer::encode_load_info(
	const chaos_string& filename)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"l(%s)",
		chaos_utility::str_to_str(
			chaos_file_serializer::get_pure_filename(filename)).c_str());
	return info;
}

// decode load info

bool chaos_image_realizer::decode_load_info(
	const chaos_string& info,
	chaos_string& filename)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "l(", 2) != 0) {
		return false;
	}
	infostr += 2;
	filename = chaos_utility::str_from_str(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode clear info

chaos_string chaos_image_realizer::encode_clr_info() {
	return "";
}

// decode clear info

bool chaos_image_realizer::decode_clr_info(
	const chaos_string& info)
{
	return info.empty();
}

// encode fill clear info

chaos_string chaos_image_realizer::encode_fill_clr_info(
	const chaos_color color)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"c(RGB(%d,%d,%d))",
		color.get_red(), color.get_green(), color.get_blue());
	return info;
}

// decode fill clear info

bool chaos_image_realizer::decode_fill_clr_info(
	const chaos_string& info,
	chaos_color& color)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "c(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0,
		red, green, blue;
	if (sscanf(
			infostr,
			"RGB(%d,%d,%d)%n",
			&red, &green, &blue,
			&next) != 3)
	{
		return false;
	}
	color = chaos_color(red, green, blue);
	infostr += next;
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode fill info

chaos_string chaos_image_realizer::encode_fill_info(
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"i(%d,%d,%d,%d,RGB(%d,%d,%d),%s)",
		x, y,
		width, height,
		color.get_red(), color.get_green(), color.get_blue(),
		subinfo.c_str());
	return info;
}

// decode fill info

bool chaos_image_realizer::decode_fill_info(
	const chaos_string& info,
	int& x, int& y,
	int& width, int& height,
	chaos_color& color,
	chaos_string& subinfo)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "i(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0,
		red, green, blue;
	if (sscanf(
			infostr,
			"%d,%d,%d,%d,RGB(%d,%d,%d),%n",
			&x, &y,
			&width, &height,
			&red, &green, &blue,
			&next) != 7)
	{
		return false;
	}
	color = chaos_color(red, green, blue);
	infostr += next;
	subinfo = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode fade fill info

chaos_string chaos_image_realizer::encode_fade_fill_info(
	int fade,
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"d(%d,%d,%d,%d,%d,RGB(%d,%d,%d),%s)",
		fade,
		x, y,
		width, height,
		color.get_red(), color.get_green(), color.get_blue(),
		subinfo.c_str());
	return info;
}

// decode fade fill info

bool chaos_image_realizer::decode_fade_fill_info(
	const chaos_string& info,
	int& fade,
	int& x, int& y,
	int& width, int& height,
	chaos_color& color,
	chaos_string& subinfo)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "d(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0,
		red, green, blue;
	if (sscanf(
			infostr,
			"%d,%d,%d,%d,%d,RGB(%d,%d,%d),%n",
			&fade,
			&x, &y,
			&width, &height,
			&red, &green, &blue,
			&next) != 8)
	{
		return false;
	}
	color = chaos_color(red, green, blue);
	infostr += next;
	subinfo = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode trans fill info

chaos_string chaos_image_realizer::encode_trans_fill_info(
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"f(%d,%d,%d,%d,RGB(%d,%d,%d),%s)",
		x, y,
		width, height,
		color.get_red(), color.get_green(), color.get_blue(),
		subinfo.c_str());
	return info;
}

// decode trans fill info

bool chaos_image_realizer::decode_trans_fill_info(
	const chaos_string& info,
	int& x, int& y,
	int& width, int& height,
	chaos_color& color,
	chaos_string& subinfo)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "f(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0,
		red, green, blue;
	if (sscanf(
			infostr,
			"%d,%d,%d,%d,RGB(%d,%d,%d),%n",
			&x, &y,
			&width, &height,
			&red, &green, &blue,
			&next) != 7)
	{
		return false;
	}
	color = chaos_color(red, green, blue);
	infostr += next;
	subinfo = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode mono fill info

chaos_string chaos_image_realizer::encode_mono_fill_info(
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"m(%d,%d,%d,%d,RGB(%d,%d,%d),%s)",
		x, y,
		width, height,
		color.get_red(), color.get_green(), color.get_blue(),
		subinfo.c_str());
	return info;
}

// decode mono fill info

bool chaos_image_realizer::decode_mono_fill_info(
	const chaos_string& info,
	int& x, int& y,
	int& width, int& height,
	chaos_color& color,
	chaos_string& subinfo)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "m(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0,
		red, green, blue;
	if (sscanf(
			infostr,
			"%d,%d,%d,%d,RGB(%d,%d,%d),%n",
			&x, &y,
			&width, &height,
			&red, &green, &blue,
			&next) != 7)
	{
		return false;
	}
	color = chaos_color(red, green, blue);
	infostr += next;
	subinfo = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode blt info

chaos_string chaos_image_realizer::encode_blt_info(
	int x, int y,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"b(%d,%d,%s,%s)",
		x, y,
		subinfo1.c_str(),
		subinfo2.c_str());
	return info;
}

// decode blt info

bool chaos_image_realizer::decode_blt_info(
	const chaos_string& info,
	int& x, int& y,
	chaos_string& subinfo1,
	chaos_string& subinfo2)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "b(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%n",
			&x, &y,
			&next) != 2)
	{
		return false;
	}
	infostr += next;
	subinfo1 = extract_subinfo(infostr);
	if (*infostr != ',') {
		return false;
	}
	infostr++;
	subinfo2 = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode stretch blt info

chaos_string chaos_image_realizer::encode_stretch_blt_info(
	int x, int y,
	int width, int height,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"r(%d,%d,%d,%d,%s,%s)",
		x, y,
		width, height,
		subinfo1.c_str(),
		subinfo2.c_str());
	return info;
}

// decode stretch blt info

bool chaos_image_realizer::decode_stretch_blt_info(
	const chaos_string& info,
	int& x, int& y,
	int& width, int& height,
	chaos_string& subinfo1,
	chaos_string& subinfo2)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "r(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%d,%d,%n",
			&x, &y,
			&width, &height,
			&next) != 4)
	{
		return false;
	}
	infostr += next;
	subinfo1 = extract_subinfo(infostr);
	if (*infostr != ',') {
		return false;
	}
	infostr++;
	subinfo2 = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode alpha blt info

chaos_string chaos_image_realizer::encode_alpha_blt_info(
	int x, int y,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_string& subinfo3)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"a(%d,%d,%s,%s",
		x, y,
		subinfo1.c_str(),
		subinfo2.c_str());
	if (subinfo3 != subinfo2) {
		info += ',';
		info += subinfo3;
	}
	info += ')';
	return info;
}

// decode alpha blt info

bool chaos_image_realizer::decode_alpha_blt_info(
	const chaos_string& info,
	int& x, int& y,
	chaos_string& subinfo1,
	chaos_string& subinfo2,
	chaos_string& subinfo3)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "a(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%n",
			&x, &y,
			&next) != 2)
	{
		return false;
	}
	infostr += next;
	subinfo1 = extract_subinfo(infostr);
	if (*infostr != ',') {
		return false;
	}
	infostr++;
	subinfo2 = extract_subinfo(infostr);
	if (*infostr == ',') {
		infostr++;
		subinfo3 = extract_subinfo(infostr);
	} else {
		subinfo3 = subinfo2;
	}
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode trans fade info

chaos_string chaos_image_realizer::encode_trans_fade_info(
	int step,
	int unk1, int unit, int unk2, int unk3,
	int x, int y,
	int width, int height,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"t(%d,%d,%d,%d,%d,%d,%d,%d,%d,%s,%s)",
		step,
		unk1, unit, unk2, unk3,
		x, y,
		width, height,
		subinfo1.c_str(),
		subinfo2.c_str());
	return info;
}

// decode trans fade info

bool chaos_image_realizer::decode_trans_fade_info(
	const chaos_string& info,
	int& step,
	int& unk1, int& unit, int& unk2, int& unk3,
	int& x, int& y,
	int& width, int& height,
	chaos_string& subinfo1,
	chaos_string& subinfo2)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "t(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%d,%d,%d,%d,%d,%d,%d,%n",
			&step,
			&unk1, &unit, &unk2, &unk3,
			&x, &y,
			&width, &height,
			&next) != 9)
	{
		return false;
	}
	infostr += next;
	subinfo1 = extract_subinfo(infostr);
	if (*infostr != ',') {
		return false;
	}
	infostr++;
	subinfo2 = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode wipe vertical scroll info

chaos_string chaos_image_realizer::encode_wipe_v_scroll_info(
	int step,
	int rolls, int height,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"y(%d,%d,%d,%s,%s)",
		step,
		rolls, height,
		subinfo1.c_str(),
		subinfo2.c_str());
	return info;
}

// decode wipe vertical scroll info

bool chaos_image_realizer::decode_wipe_v_scroll_info(
	const chaos_string& info,
	int& step,
	int& rolls, int& height,
	chaos_string& subinfo1,
	chaos_string& subinfo2)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "y(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%d,%n",
			&step,
			&rolls, &height,
			&next) != 3)
	{
		return false;
	}
	infostr += next;
	subinfo1 = extract_subinfo(infostr);
	if (*infostr != ',') {
		return false;
	}
	infostr++;
	subinfo2 = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode wipe horizontal blind info

chaos_string chaos_image_realizer::encode_wipe_h_blind_info(
	int step,
	int gap, int unk,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"h(%d,%d,%d,%s,%s)",
		step,
		gap, unk,
		subinfo1.c_str(),
		subinfo2.c_str());
	return info;
}

// decode wipe horizontal blind info

bool chaos_image_realizer::decode_wipe_h_blind_info(
	const chaos_string& info,
	int& step,
	int& gap, int& unk,
	chaos_string& subinfo1,
	chaos_string& subinfo2)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "h(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%d,%n",
			&step,
			&gap, &unk,
			&next) != 3)
	{
		return false;
	}
	infostr += next;
	subinfo1 = extract_subinfo(infostr);
	if (*infostr != ',') {
		return false;
	}
	infostr++;
	subinfo2 = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode wipe vertical blind info

chaos_string chaos_image_realizer::encode_wipe_v_blind_info(
	int step,
	int gap, int unk,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"v(%d,%d,%d,%s,%s)",
		step,
		gap, unk,
		subinfo1.c_str(),
		subinfo2.c_str());
	return info;
}

// decode wipe vertical blind info

bool chaos_image_realizer::decode_wipe_v_blind_info(
	const chaos_string& info,
	int& step,
	int& gap, int& unk,
	chaos_string& subinfo1,
	chaos_string& subinfo2)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "v(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%d,%n",
			&step,
			&gap, &unk,
			&next) != 3)
	{
		return false;
	}
	infostr += next;
	subinfo1 = extract_subinfo(infostr);
	if (*infostr != ',') {
		return false;
	}
	infostr++;
	subinfo2 = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode wipe shake info

chaos_string chaos_image_realizer::encode_wipe_shake_info(
	int step,
	int move, int action, int unk,
	const chaos_string& subinfo)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"s(%d,%d,%d,%d,%s)",
		step,
		move, action, unk,
		subinfo.c_str());
	return info;
}

// decode wipe shake info

bool chaos_image_realizer::decode_wipe_shake_info(
	const chaos_string& info,
	int& step,
	int& move, int& action, int& unk,
	chaos_string& subinfo)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "s(", 2) != 0) {
		return false;
	}
	infostr += 2;
	int next = 0;
	if (sscanf(
			infostr,
			"%d,%d,%d,%d,%n",
			&step,
			&move, &action, &unk,
			&next) != 4)
	{
		return false;
	}
	infostr += next;
	subinfo = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// encode clone info

chaos_string chaos_image_realizer::encode_clone_info(
	const chaos_string& subinfo)
{
	chaos_string info;
	info = chaos_utility::format_str(
		"o(%s)",
		subinfo.c_str());
	return info;
}

// decode clone info

bool chaos_image_realizer::decode_clone_info(
	const chaos_string& info,
	chaos_string& subinfo)
{
	const char* infostr = info.c_str();
	if (strncmp(infostr, "o(", 2) != 0) {
		return false;
	}
	infostr += 2;
	subinfo = extract_subinfo(infostr);
	if (*infostr != ')') {
		return false;
	}
	return true;
}

// get next step effect info

chaos_string chaos_image_realizer::get_next_step_effect_info(
	const chaos_string& info,
	int add_step)
{
	chaos_string next_info;
	const char* infostr = info.c_str();
	switch (*infostr) {
	case 't':	// trans fade
		{ // dummy block
			int step,
				unk1, unit, unk2, unk3,
				x, y, width, height;
			chaos_color color;
			chaos_string subinfo1, subinfo2;
			if (!decode_trans_fade_info(
					info,
					step,
					unk1, unit, unk2, unk3,
					x, y,
					width, height,
					subinfo1,
					subinfo2))
			{
				break;
			}
			step += add_step;
			if ((unit <= 0) ||
				(unit*step >= 255))
			{
				break;
			}
			next_info = encode_trans_fade_info(
				step,
				unk1, unit, unk2, unk3,
				x, y,
				width, height,
				subinfo1,
				subinfo2);
		}
		break;
	case 'y':	// wipe virtical scroll
		{ // dummy block
			int step,
				rolls, height;
			chaos_string subinfo1, subinfo2;
			if (!decode_wipe_v_scroll_info(
					info,
					step,
					rolls, height,
					subinfo1, subinfo2))
			{
				break;
			}
			step += add_step;
			if ((rolls == 0) ||
				(abs(rolls*step) >= height))
			{
				break;
			}
			next_info = encode_wipe_v_scroll_info(
				step,
				rolls, height,
				subinfo1, subinfo2);
		}
		break;
	case 'h':	// wipe horizontal blind
		{ // dummy block
			int step,
				gap, unk;
			chaos_string subinfo1, subinfo2;
			if (!decode_wipe_h_blind_info(
					info,
					step,
					gap, unk,
					subinfo1, subinfo2))
			{
				break;
			}
			step += add_step;
			if ((gap == 0) ||
				(step*2 >= abs(gap)))
			{
				break;
			}
			next_info = encode_wipe_h_blind_info(
				step,
				gap, unk,
				subinfo1, subinfo2);
		}
		break;
	case 'v':	// wipe virtical blind
		{ // dummy block
			int step,
				gap, unk;
			chaos_string subinfo1, subinfo2;
			if (!decode_wipe_v_blind_info(
					info,
					step,
					gap, unk,
					subinfo1, subinfo2))
			{
				break;
			}
			step += add_step;
			if ((gap == 0) ||
				(step*2 >= abs(gap)))
			{
				break;
			}
			next_info = encode_wipe_v_blind_info(
				step,
				gap, unk,
				subinfo1, subinfo2);
		}
		break;
	case 's':	// wipe shake
		{ // dummy block
			int step,
				move, action, unk;
			chaos_string subinfo;
			if (!decode_wipe_shake_info(
					info,
					step,
					move, action, unk,
					subinfo))
			{
				break;
			}
			step += add_step;
			if (step >= action*2) {
				break;
			}
			next_info = encode_wipe_shake_info(
				step,
				move, action, unk,
				subinfo);
		}
		break;
	}
	return next_info;
}

// get effect time scale

int chaos_image_realizer::get_effect_time_scale(
	const chaos_string& info)
{
	int time_scale = 1;
	const char* infostr = info.c_str();
	switch (*infostr) {
	case 't':	// trans fade
		time_scale = 2;
		break;
	}
	return time_scale;
}

// is effect spippable

bool chaos_image_realizer::is_effect_spippable(
	const chaos_string& info)
{
	bool skippable = false;
	const char* infostr = info.c_str();
	switch (*infostr) {
	case 't':	// trans fade
	case 'y':	// wipe virtical scroll
	case 'h':	// wipe horizontal blind
	case 'v':	// wipe virtical blind
		skippable = true;
		break;
	}
	return skippable;
}

// extract base position

chaos_pos chaos_image_realizer::extract_base_position(
	const chaos_string& info)
{
	chaos_pos base_pos(0, 0);
	const char* infostr = info.c_str();
	switch (*infostr) {
	case 's':	// wipe shake
		{ // dummy block
			int step,
				move, action, unk;
			chaos_string subinfo;
			if (!decode_wipe_shake_info(
					info,
					step,
					move, action, unk,
					subinfo))
			{
				break;
			}
			base_pos = get_wipe_shake_base_position(
				step,
				move, action, unk);
		}
		break;
	}
	return base_pos;
}

// remove special info

chaos_string chaos_image_realizer::remove_special_info(
	const chaos_string& oinfo,
	const chaos_string& spinfo)
{
	chaos_string wildcard = "*";
	chaos_string info = oinfo;
	if ((spinfo.length() <= 0) ||
		(spinfo == wildcard))
	{
		return info;
	}
	int wild = chaos_utility::str_find_str(spinfo, wildcard);
	chaos_string spinfo1, spinfo2;
	if (wild >= 0) {
		spinfo1 = spinfo.substr(0, wild);
		spinfo2 = spinfo.substr(wild+1);
	} else {
		spinfo1 = spinfo;
	}
	int find = 0;
	if (!spinfo1.empty()) {
		find = chaos_utility::str_find_str(oinfo, spinfo1);
	}
	if (find >= 0) {
		ASSERT((size_t)find <= oinfo.length());
		chaos_string preinfo = oinfo.substr(0, find),
			subinfo,
			postinfo;
		if (wild >= 0) {
			const char* infostr = oinfo.c_str();
			infostr += find+spinfo1.length();
			subinfo = remove_special_info(
				extract_subinfo(infostr),
				spinfo);
			if (strncmp(
					infostr, spinfo2.c_str(),
					spinfo2.length()) != 0)
			{
				find = -1;
			} else {
				infostr += spinfo2.length();
				postinfo = remove_special_info(
					infostr,
					spinfo);
			}
		}
		if (find >= 0) {
			info = preinfo+subinfo+postinfo;
		}
	}
	return info;
}

// is force cache info

bool chaos_image_realizer::is_force_cache_info(
	const chaos_string& info)
{
	bool force_cache = false;
	const char* infostr = info.c_str();
	switch (*infostr) {
	case 'l':	// load
		force_cache = true;
		break;
	}
	return force_cache;
}

// is force uncache info

bool chaos_image_realizer::is_force_uncache_info(
	const chaos_string& info)
{
	bool force_uncache = false;
	const char* infostr = info.c_str();
	switch (*infostr) {
	case '\0':	// clear
	case 'c':	// fill clear
	case 't':	// trans fade
	case 'y':	// wipe virtical scroll
	case 'v':	// wipe virtical blind
	case 'h':	// wipe horizontal blind
	case 's':	// wipe shake
		force_uncache = true;
		break;
	}
	return force_uncache;
}

////////////////////////////////////////////////////////////////////////////////
// draw

// draw load image

bool chaos_image_realizer::draw_load_image(
	chaos_image* image,
	const chaos_string& filename)
{
	ASSERT(image != NULL);
	chaos_string filename2 = filename;
	if ((filename2.length() >= 4) &&
		(strcmp(filename2.c_str()+filename2.length()-4, ".ong") == 0))
	{
		// bad ext :-)
		LOG_WRN(
			"warning : image realizer::draw load image - "
			"bad ext(%s)\n",
			filename2.c_str());
		filename2 = filename2.substr(0, filename2.length()-4)+".png";
	}
	chaos_file_data filedata;
	if (!chaos_file_serializer::instance()->load_file_data(
			&filedata,
			filename2,
			chaos_file_serializer::fpos_packed))
	{
		LOG_ERR(
			"error : image realizer::load image - "
			"cannot found file(%s)\n",
			filename2.c_str());
		if (!image->create(
				sys_dim,
				true,
				CHAOS_RGBA(0, 0, 0, 255)))
		{
			return false;
		}
		return true;
	}
	if ((filename2.length() >= 4) &&
		(strcmp(filename2.c_str()+filename2.length()-4, ".png") == 0))
	{
		chaos_png_image_loader image_loader(&filedata, image);
		if (!image_loader.load()) {
			return false;
		}
	} else {
		chaos_bmp_image_loader image_loader(&filedata, image);
		if (!image_loader.load()) {
			return false;
		}
	}
	return true;
}

// draw clear image

bool chaos_image_realizer::draw_clr_image(
	chaos_image* image,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!image->create(
			sys_dim,
			false))
	{
		return false;
	}
	image->fill(CHAOS_RGBA(0, 0, 0, 0), clip_rect);
	return true;
}

// draw fill clear image

bool chaos_image_realizer::draw_fill_clr_image(
	chaos_image* image,
	chaos_color color,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!image->create(
			sys_dim,
			false))
	{
		return false;
	}
	image->fill((chaos_uint32)color, clip_rect);
	return true;
}

// draw fill image

bool chaos_image_realizer::draw_fill_image(
	chaos_image* image,
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!copy_image(image, subinfo, false, clip_rect)) {
		return false;
	}
	chaos_rect fill_rect(x, y, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	image->fill(
		(chaos_uint32)color,
		&fill_rect);
	return true;
}

// draw fade fill image

bool chaos_image_realizer::draw_fade_fill_image(
	chaos_image* image,
	int fade,
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!copy_image(image, subinfo, false, clip_rect)) {
		return false;
	}
	chaos_rect fill_rect(x, y, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	image->fade_fill(
		(chaos_uint32)color,
		fade,
		&fill_rect);
	return true;
}

// draw trans fill image

bool chaos_image_realizer::draw_trans_fill_image(
	chaos_image* image,
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!copy_image(image, subinfo, false, clip_rect)) {
		return false;
	}
	chaos_rect fill_rect(x, y, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	image->trans_fill(
		color.get_red(), color.get_blue(), color.get_green(),
		&fill_rect);
	return true;
}

// draw mono fill image

bool chaos_image_realizer::draw_mono_fill_image(
	chaos_image* image,
	int x, int y,
	int width, int height,
	const chaos_color color,
	const chaos_string& subinfo,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!copy_image(image, subinfo, false, clip_rect)) {
		return false;
	}
	chaos_rect fill_rect(x, y, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	image->mono_fill(
		color.get_red(), color.get_blue(), color.get_green(),
		&fill_rect);
	return true;
}

// draw blt image

bool chaos_image_realizer::draw_blt_image(
	chaos_image* image,
	int x, int y,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (subinfo2.empty()) {
		return copy_image(image, subinfo1, false, clip_rect);
	}
	if ((x == 0) && (y == 0)) {
		chaos_image_locker sublock2(this, build_image(subinfo2));
		chaos_image* subimage2 = sublock2.get_image();
		if (subimage2 == NULL) {
			return true;
		} else if (subimage2->get_dim() == image->get_dim()) {
			if (!copy_image(image, subinfo2, false, clip_rect)) {
				return false;
			}
			return true;
		}
	}
	if (!copy_image(image, subinfo1, false, clip_rect)) {
		return false;
	}
	chaos_image_locker sublock2(this, build_image(subinfo2));
	chaos_image* subimage2 = sublock2.get_image();
	if (subimage2 == NULL) {
		// nothing
	} else {
		image->blt(
			subimage2,
			chaos_pos(x, y),
			clip_rect);
	}
	return true;
}

// draw stretch blt image

bool chaos_image_realizer::draw_stretch_blt_image(
	chaos_image* image,
	int x, int y,
	int width, int height,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (subinfo2.empty()) {
		return copy_image(image, subinfo1, false, clip_rect);
	}
	if (!copy_image(image, subinfo1, false, clip_rect)) {
		return false;
	}
	chaos_image_locker sublock2(this, build_image(subinfo2));
	chaos_image* subimage2 = sublock2.get_image();
	if (subimage2 == NULL) {
		// nothing
	} else {
		image->stretch_blt(
			subimage2,
			chaos_pos(x, y),
			chaos_dim(width, height),
			clip_rect);
	}
	return true;
}

// draw alpha blt image

bool chaos_image_realizer::draw_alpha_blt_image(
	chaos_image* image,
	int x, int y,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_string& subinfo3,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!copy_image(image, subinfo1, false, clip_rect)) {
		return false;
	}
	chaos_image_locker sublock2(this, build_image(subinfo2));
	chaos_image* subimage2 = sublock2.get_image();
	chaos_image_locker sublock3(this, build_image(subinfo3));
	chaos_image* subimage3 = sublock3.get_image();
	if (subimage3 == NULL) {
		// nothing
	} else if (subimage2 == NULL) {
		image->alpha_fill(
			CHAOS_RGBA(0, 0, 0, 0),
			subimage3,
			chaos_pos(x, y),
			clip_rect);
	} else if (subimage2 == subimage3) {
		image->alpha_blt(
			subimage2,
			chaos_pos(x, y),
			clip_rect);
	} else {
		image->alpha_blt(
			subimage2,
			subimage3,
			chaos_pos(x, y),
			clip_rect);
	}
	return true;
}

// draw trans fade image

bool chaos_image_realizer::draw_trans_fade_image(
	chaos_image* image,
	int step,
	int /*unk1*/, int unit, int /*unk2*/, int /*unk3*/,
	int x, int y,
	int width, int height,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	int fade = 255-unit*step;
	if (fade <= 0) {
		if (!copy_image(image, subinfo2, true, clip_rect)) {
			return false;
		}
		return true;
	}
	chaos_image_locker sublock1(this, -1);
	if (!is_force_uncache_info(subinfo1)) {
		sublock1.lock(build_image(subinfo1));
	}
	if (!copy_image(image, subinfo1, true, clip_rect)) {
		return false;
	}
	if (fade >= 255) {
		return true;
	}
	chaos_image_locker sublock2(this, build_image(subinfo2));
	chaos_image* subimage2 = sublock2.get_image();
	if (subimage2 == NULL) {
		// nothing
	} else {
		chaos_rect fade_rect(x, y, width, height);
		if (clip_rect != NULL) {
			fade_rect &= *clip_rect;
		}
		image->fade_blt(
			subimage2,
			fade,
			chaos_pos(0, 0),
			&fade_rect);
	}
	return true;
}

// draw wipe virtical scroll image

bool chaos_image_realizer::draw_wipe_v_scroll_image(
	chaos_image* image,
	int step,
	int rolls, int height,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!image->create(
			sys_dim,
			false))
	{
		return false;
	}
	chaos_pos pos(0, rolls*step);
	chaos_image_locker sublock1(this, build_image(subinfo1));
	chaos_image* subimage1 = sublock1.get_image();
	if (subimage1 == NULL) {
		chaos_rect fill_rect(
			pos,
			image->get_dim());
		if (clip_rect != NULL) {
			fill_rect &= *clip_rect;
		}
		image->fill(
			CHAOS_RGBA(0, 0, 0, 0),
			&fill_rect);
	} else {
		image->blt(
			subimage1,
			pos,
			clip_rect);
	}
	chaos_image_locker sublock2(this, build_image(subinfo2));
	chaos_image* subimage2 = sublock2.get_image();
	chaos_pos pos2 = pos;
	if (rolls > 0) {
		pos2.y -= height;
	} else {
		pos2.y += height;
	}
	if (subimage2 == NULL) {
		chaos_rect fill_rect(
			pos2,
			image->get_dim());
		if (clip_rect != NULL) {
			fill_rect &= *clip_rect;
		}
		image->fill(
			CHAOS_RGBA(0, 0, 0, 0),
			&fill_rect);
	} else {
		image->blt(
			subimage2,
			pos2,
			clip_rect);
	}
	return true;
}

// draw wipe horizontal blind image

bool chaos_image_realizer::draw_wipe_h_blind_image(
	chaos_image* image,
	int step,
	int gap, int /*unk*/,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	chaos_image_locker sublock1(this, -1);
	if (!is_force_uncache_info(subinfo1)) {
		sublock1.lock(build_image(subinfo1));
	}
	if (!copy_image(image, subinfo1, true, clip_rect)) {
		return false;
	}
	chaos_image_locker sublock2(this, build_image(subinfo2));
	chaos_image* subimage2 = sublock2.get_image();
	if (subimage2 == NULL) {
		image->horz_blind_fill(
			CHAOS_RGBA(0, 0, 0, 0),
			gap, step*2,
			clip_rect);
	} else {
		image->horz_blind_blt(
			subimage2,
			gap, step*2,
			clip_rect);
	}
	return true;
}

// draw wipe vertical blind image

bool chaos_image_realizer::draw_wipe_v_blind_image(
	chaos_image* image,
	int step,
	int gap, int /*unk*/,
	const chaos_string& subinfo1,
	const chaos_string& subinfo2,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	chaos_image_locker sublock1(this, -1);
	if (!is_force_uncache_info(subinfo1)) {
		sublock1.lock(build_image(subinfo1));
	}
	if (!copy_image(image, subinfo1, true, clip_rect)) {
		return false;
	}
	chaos_image_locker sublock2(this, build_image(subinfo2));
	chaos_image* subimage2 = sublock2.get_image();
	if (subimage2 == NULL) {
		image->vert_blind_fill(
			CHAOS_RGBA(0, 0, 0, 0),
			gap, step*2,
			clip_rect);
	} else {
		image->vert_blind_blt(
			subimage2,
			gap, step*2,
			clip_rect);
	}
	return true;
}

// draw wipe shake image

bool chaos_image_realizer::draw_wipe_shake_image(
	chaos_image* image,
	int step,
	int move, int action, int unk,
	const chaos_string& subinfo,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!image->create(
			sys_dim,
			false))
	{
		return false;
	}
	chaos_image_locker sublock(this, build_image(subinfo));
	chaos_image* subimage = sublock.get_image();
	if (subimage == NULL) {
		image->fill(CHAOS_RGBA(0, 0, 0, 0), clip_rect);
	} else {
		chaos_pos base_pos = get_wipe_shake_base_position(
			step,
			move, action, unk);
		chaos_rect org_rect(0, 0, 0, image->get_height());
		if (base_pos.x > 0) {
			org_rect.x = 0;
			org_rect.width = base_pos.x;
		} else if (base_pos.x < 0) {
			org_rect.x = image->get_width()+base_pos.x;
			org_rect.width = -base_pos.x;
		}
		if (clip_rect != NULL) {
			org_rect &= *clip_rect;
		}
		if (org_rect.width > 0) {
			image->blt(subimage, chaos_pos(0, 0), &org_rect);
		}
		org_rect = chaos_rect(0, 0, image->get_width(), 0);
		if (base_pos.y > 0) {
			org_rect.y = 0;
			org_rect.height = base_pos.y;
		} else if (base_pos.y < 0) {
			org_rect.y = image->get_height()+base_pos.y;
			org_rect.height = -base_pos.y;
		}
		if (clip_rect != NULL) {
			org_rect &= *clip_rect;
		}
		if (org_rect.height > 0) {
			image->blt(subimage, chaos_pos(0, 0), &org_rect);
		}
		image->blt(subimage, base_pos, clip_rect);
	}
	return true;
}

// draw clone image

bool chaos_image_realizer::draw_clone_image(
	chaos_image* image,
	const chaos_string& subinfo,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	if (!copy_image(image, subinfo, false, clip_rect)) {
		return false;
	}
	return true;
}

// draw image

bool chaos_image_realizer::draw_image(
	chaos_image* image,
	const chaos_string& info,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	const char* infostr = info.c_str();
	switch (*infostr) {
	case '\0':	// clear
		{ // dummy block
			if (!draw_clr_image(
					image,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'l':	// load
		{ // dummy block
			ASSERT(clip_rect == NULL);
			chaos_string filename;
			if (!decode_load_info(
					info,
					filename))
			{
				return false;
			}
			if (!draw_load_image(
					image,
					filename))
			{
				return false;
			}
		}
		break;
	case 'c':	// fill clear
		{ // dummy block
			chaos_color color;
			if (!decode_fill_clr_info(
					info,
					color))
			{
				return false;
			}
			if (!draw_fill_clr_image(
					image,
					color,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'i':	// fill
		{ // dummy block
			int x, y, width, height;
			chaos_color color;
			chaos_string subinfo;
			if (!decode_fill_info(
					info,
					x, y,
					width, height,
					color,
					subinfo))
			{
				return false;
			}
			if (!draw_fill_image(
					image,
					x, y,
					width, height,
					color,
					subinfo,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'd':	// fade fill
		{ // dummy block
			int fade,
				x, y, width, height;
			chaos_color color;
			chaos_string subinfo;
			if (!decode_fade_fill_info(
					info,
					fade,
					x, y,
					width, height,
					color,
					subinfo))
			{
				return false;
			}
			if (!draw_fade_fill_image(
					image,
					fade,
					x, y,
					width, height,
					color,
					subinfo,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'f':	// trans fill
		{ // dummy block
			int x, y, width, height;
			chaos_color color;
			chaos_string subinfo;
			if (!decode_trans_fill_info(
					info,
					x, y,
					width, height,
					color,
					subinfo))
			{
				return false;
			}
			if (!draw_trans_fill_image(
					image,
					x, y,
					width, height,
					color,
					subinfo,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'm':	// mono fill
		{ // dummy block
			int x, y, width, height;
			chaos_color color;
			chaos_string subinfo;
			if (!decode_mono_fill_info(
					info,
					x, y,
					width, height,
					color,
					subinfo))
			{
				return false;
			}
			if (!draw_mono_fill_image(
					image,
					x, y,
					width, height,
					color,
					subinfo,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'b':	// blt
		{ // dummy block
			int x, y;
			chaos_string subinfo1, subinfo2;
			if (!decode_blt_info(
					info,
					x, y,
					subinfo1, subinfo2))
			{
				return false;
			}
			if (!draw_blt_image(
					image,
					x, y,
					subinfo1, subinfo2,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'r':	// stretch blt
		{ // dummy block
			int x, y,
				width, height;
			chaos_string subinfo1, subinfo2;
			if (!decode_stretch_blt_info(
					info,
					x, y,
					width, height,
					subinfo1, subinfo2))
			{
				return false;
			}
			if (!draw_stretch_blt_image(
					image,
					x, y,
					width, height,
					subinfo1, subinfo2,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'a':	// alpha blt
		{ // dummy block
			int x, y;
			chaos_string subinfo1, subinfo2, subinfo3;
			if (!decode_alpha_blt_info(
					info,
					x, y,
					subinfo1, subinfo2, subinfo3))
			{
				return false;
			}
			if (!draw_alpha_blt_image(
					image,
					x, y,
					subinfo1, subinfo2, subinfo3,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 't':	// trans fade
		{ // dummy block
			int step,
				unk1, unit, unk2, unk3,
				x, y, width, height;
			chaos_color color;
			chaos_string subinfo1, subinfo2;
			if (!decode_trans_fade_info(
					info,
					step,
					unk1, unit, unk2, unk3,
					x, y,
					width, height,
					subinfo1, subinfo2))
			{
				return false;
			}
			if (!draw_trans_fade_image(
					image,
					step,
					unk1, unit, unk2, unk3,
					x, y,
					width, height,
					subinfo1, subinfo2,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'y':	// wipe virtical scroll
		{ // dummy block
			int step,
				rolls, height;
			chaos_string subinfo1, subinfo2;
			if (!decode_wipe_v_scroll_info(
					info,
					step,
					rolls, height,
					subinfo1, subinfo2))
			{
				return false;
			}
			if (!draw_wipe_v_scroll_image(
					image,
					step,
					rolls, height,
					subinfo1, subinfo2,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'h':	// wipe horizontal blind
		{ // dummy block
			int step,
				gap, unk;
			chaos_string subinfo1, subinfo2;
			if (!decode_wipe_h_blind_info(
					info,
					step,
					gap, unk,
					subinfo1, subinfo2))
			{
				return false;
			}
			if (!draw_wipe_h_blind_image(
					image,
					step,
					gap, unk,
					subinfo1, subinfo2,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'v':	// wipe virtical blind
		{ // dummy block
			int step,
				gap, unk;
			chaos_string subinfo1, subinfo2;
			if (!decode_wipe_v_blind_info(
					info,
					step,
					gap, unk,
					subinfo1, subinfo2))
			{
				return false;
			}
			if (!draw_wipe_v_blind_image(
					image,
					step,
					gap, unk,
					subinfo1, subinfo2,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 's':	// wipe shake
		{ // dummy block
			int step,
				move, action, unk;
			chaos_string subinfo;
			if (!decode_wipe_shake_info(
					info,
					step,
					move, action, unk,
					subinfo))
			{
				break;
			}
			if (!draw_wipe_shake_image(
					image,
					step,
					move, action, unk,
					subinfo,
					clip_rect))
			{
				return false;
			}
		}
		break;
	case 'o':	// clone
		{ // dummy block
			chaos_string subinfo;
			if (!decode_clone_info(
					info,
					subinfo))
			{
				break;
			}
			if (!draw_clone_image(
					image,
					subinfo,
					clip_rect))
			{
				break;
			}
		}
		break;
	default:
		return false;
	}
	return true;
}

// copy image

bool chaos_image_realizer::copy_image(
	chaos_image* image,
	const chaos_string& subinfo,
	bool recommend_cache,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	bool direct;
	if (is_force_uncache_info(subinfo)) {
		// always direct
		direct = true;
	} else if (is_force_cache_info(subinfo)) {
		// always cache
		direct = false;
	} else {
		direct = !recommend_cache;
	}
	if (direct) {
		return draw_image(image, subinfo, clip_rect);
	}
	chaos_image_locker sublock(this, build_image(subinfo));
	chaos_image* subimage = sublock.get_image();
	if (subimage == NULL) {
		if (!image->create(
				sys_dim,
				false))
		{
			return false;
		}
		image->fill(CHAOS_RGBA(0, 0, 0, 0), clip_rect);
	} else {
		if (!image->create(
				subimage->get_dim(),
				false))
		{
			return false;
		}
		image->blt(subimage, chaos_pos(0, 0), clip_rect);
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// operation

// add image

bool chaos_image_realizer::add_image() {
	int id = get_image_count();
	image_holder holder;
	holder.id = id;
	holder.lock = 0;
	holder.image = new chaos_image;
	images.push_back(holder);
	image_queue.push_back(id);
	return true;
}

// delete image

bool chaos_image_realizer::delete_image() {
	if (images.size() <= 0) {
		return false;
	}
	int id = get_image_count()-1;
	ASSERT(images[id].id == id);
	if (!images[id].info.empty()) {
		image_map.erase(images[id].info);
	}
	image_queue_type::iterator it = std::find(
		image_queue.begin(), image_queue.end(),
		id);
	ASSERT(it != image_queue.end());
	if (it != image_queue.end()) {
		image_queue.erase(it);
	}
	if (images[id].image != NULL) {
		delete images[id].image;
	}
	images.pop_back();
	return true;
}

// delete excess images

bool chaos_image_realizer::delete_excessimages() {
	while (get_image_count() > reserve_count) {
		int id = get_image_count()-1;
		if (images[id].lock > 0) {
			break;
		}
		if (!delete_image()) {
			return false;
		}
	}
	return true;
}

// find image

int chaos_image_realizer::find_image(const chaos_string& info) const {
	ASSERT(!info.empty());
	image_map_type::const_iterator it = image_map.find(info);
	return (it != image_map.end())? (*it).second: -1;
}

// get unlocked image

int chaos_image_realizer::get_unlocked_image() {
	int id = -1;
	if (get_image_count() > 0) {
		id = image_queue.back();
		ASSERT(id == images[id].id);
		if (images[id].lock > 0) {
			id = -1;
		}
	}
	if (id < 0) {
		if (get_image_count() < (int)image_max) {
			add_image();
			ASSERT((int)images.size() > 0);
			ASSERT(images.size() == image_queue.size());
			id = image_queue.back();
			ASSERT(id == images[id].id);
			ASSERT(id == (int)images.size()-1);
			LOG_WRN(
				"warning : image realizer::get unlocked image - "
				"image count extended\n");
		} else {
			LOG_ERR(
				"error : image realizer::get unlocked image - "
				"image count overflow\n");
		}
	}
	ASSERT((id < 0) || (images[id].lock <= 0));
	return id;
}

// build image

int chaos_image_realizer::build_image(const chaos_string& info) {
	if (info.empty()) {
		return -1;
	}
	int id = find_image(info);
	if (id < 0) {
		id = get_unlocked_image();
		if (id < 0) {
			return -1;
		}
		chaos_image_locker lock(this, id);
		if (!images[id].info.empty()) {
			image_map.erase(images[id].info);
			images[id].info = "";
		}
		chaos_image* image = lock.get_image();
		if (image == NULL) {
			return -1;
		}
		if (!draw_image(image, info, NULL)) {
			return -1;
		}
		images[id].info = info;
		image_map.insert(image_map_type::value_type(info, id));
	}
	return id;
}

// realize image

bool chaos_image_realizer::realize_image(
	chaos_image* image,
	const chaos_string& info,
	const chaos_rect* clip_rect)
{
	ASSERT(image != NULL);
	bool result;
	try {
		result = copy_image(image, info, true, clip_rect);
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : image realizer::realize image - "
			"%s\n",
			e.what());
		result = false;
	}
	delete_excessimages();
	return result;
}
