////////////////////////////////////////////////////////////////////////////////
// chaos_image_realizer.h : image realizer

#ifndef CHAOS_IMAGE_REALIZER_H_INCLUDED
#define CHAOS_IMAGE_REALIZER_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_image_realizer;
class chaos_image_locker;

////////////////////////////////////////////////////////////////////////////////
// include

#ifndef CHAOS_POS_H_INCLUDED
#include "chaos_pos.h"
#endif // CHAOS_POS_H_INCLUDED

#ifndef CHAOS_DIM_H_INCLUDED
#include "chaos_dim.h"
#endif // CHAOS_DIM_H_INCLUDED

#ifndef CHAOS_RECT_H_INCLUDED
#include "chaos_rect.h"
#endif // CHAOS_RECT_H_INCLUDED

#ifndef CHAOS_COLOR_H_INCLUDED
#include "chaos_color.h"
#endif // CHAOS_COLOR_H_INCLUDED

#ifndef CHAOS_STRING_H_INCLUDED
#include "chaos_string.h"
#endif // CHAOS_STRING_H_INCLUDED

#ifndef CHAOS_IMAGE_H_INCLUDED
#include "chaos_image.h"
#endif // CHAOS_IMAGE_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// chaos_image_realizer : image realizer class

class chaos_image_realizer {
// enum
private:
	// limit
	enum {
		image_max = 100
	};

// struct
private:
	// image holder
	struct image_holder {
		// image id
		int id;
		// build info
		chaos_string info;
		// lock count
		int lock;
		// image
		chaos_image* image;
	};

// typedef
private:
	// image container type
	typedef std::vector<image_holder> image_container_type;
	// image map type
	typedef std::map<chaos_string, int> image_map_type;
	// image queue type
	typedef std::list<int> image_queue_type;

// attribute
private:
	// image reserve count
	int reserve_count;
	// system dimension
	chaos_dim sys_dim;
	// image container
	image_container_type images;
	// image_map
	image_map_type image_map;
	// image queue
	image_queue_type image_queue;

// create & destroy
public:
	// default constructor
	chaos_image_realizer();
	// destructor
	~chaos_image_realizer();
	// initialize
	bool initialize(
		int new_reserve_count,
		chaos_dim new_sys_dim);
	// dispose
	void dispose();
	// clear
	void clear();

// get & set
public:
	// get image reserve count
	int get_reserve_count() const {
		return reserve_count;
	}
	// get image count
	int get_image_count() const {
		ASSERT(images.size() == image_queue.size());
		return (int)images.size();
	}
	// get system dimension
	chaos_dim get_sys_dim() const {
		return sys_dim;
	}
	// get image
	const chaos_image* get_image(int id) const {
		return ((id >= 0) && (id < (int)images.size()))?
			images[id].image:
			NULL;
	}
	chaos_image* get_image(int id) {
		return ((id >= 0) && (id < (int)images.size()))?
			images[id].image:
			NULL;
	}

// lock
public:
	// lock image
	bool lock_image(int id);
	// unlock image
	bool unlock_image(int id);

// info
private:
	// extract sub build info
	static chaos_string extract_subinfo(const char*& infostr);
	// get wipe shake base position
	static chaos_pos get_wipe_shake_base_position(
		int step,
		int move, int action, int unk);

public:
	// encode load info
	static chaos_string encode_load_info(
		const chaos_string& filename);
	// decode load info
	static bool decode_load_info(
		const chaos_string& info,
		chaos_string& filename);
	// encode clear info
	static chaos_string encode_clr_info();
	// decode clear info
	static bool decode_clr_info(
		const chaos_string& info);
	// encode fill clear info
	static chaos_string encode_fill_clr_info(
		const chaos_color color);
	// decode fill clear info
	static bool decode_fill_clr_info(
		const chaos_string& info,
		chaos_color& color);
	// encode fill info
	static chaos_string encode_fill_info(
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo);
	// decode fill info
	static bool decode_fill_info(
		const chaos_string& info,
		int& x, int& y,
		int& width, int& height,
		chaos_color& color,
		chaos_string& subinfo);
	// encode fade fill info
	static chaos_string encode_fade_fill_info(
		int fade,
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo);
	// decode fade fill info
	static bool decode_fade_fill_info(
		const chaos_string& info,
		int& fade,
		int& x, int& y,
		int& width, int& height,
		chaos_color& color,
		chaos_string& subinfo);
	// encode trans fill info
	static chaos_string encode_trans_fill_info(
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo);
	// decode trans fill info
	static bool decode_trans_fill_info(
		const chaos_string& info,
		int& x, int& y,
		int& width, int& height,
		chaos_color& color,
		chaos_string& subinfo);
	// encode mono fill info
	static chaos_string encode_mono_fill_info(
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo);
	// decode mono fill info
	static bool decode_mono_fill_info(
		const chaos_string& info,
		int& x, int& y,
		int& width, int& height,
		chaos_color& color,
		chaos_string& subinfo);
	// encode blt info
	static chaos_string encode_blt_info(
		int x, int y,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2);
	// decode blt info
	static bool decode_blt_info(
		const chaos_string& info,
		int& x, int& y,
		chaos_string& subinfo1,
		chaos_string& subinfo2);
	// encode stretch blt info
	static chaos_string encode_stretch_blt_info(
		int x, int y,
		int width, int height,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2);
	// decode stretch blt info
	static bool decode_stretch_blt_info(
		const chaos_string& info,
		int& x, int& y,
		int& width, int& height,
		chaos_string& subinfo1,
		chaos_string& subinfo2);
	// encode alpha blt info
	static chaos_string encode_alpha_blt_info(
		int x, int y,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_string& subinfo3);
	// decode alpha blt info
	static bool decode_alpha_blt_info(
		const chaos_string& info,
		int& x, int& y,
		chaos_string& subinfo1,
		chaos_string& subinfo2,
		chaos_string& subinfo3);
	// encode trans fade info
	static chaos_string encode_trans_fade_info(
		int step,
		int unk1, int unit, int unk2, int unk3,
		int x, int y,
		int width, int height,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2);
	// decode trans fade info
	static bool decode_trans_fade_info(
		const chaos_string& info,
		int& step,
		int& unk1, int& unit, int& unk2, int& unk3,
		int& x, int& y,
		int& width, int& height,
		chaos_string& subinfo1,
		chaos_string& subinfo2);
	// encode wipe vertical scroll info
	static chaos_string encode_wipe_v_scroll_info(
		int step,
		int rolls, int height,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2);
	// decode wipe vertical scroll info
	static bool decode_wipe_v_scroll_info(
		const chaos_string& info,
		int& step,
		int& rolls, int& height,
		chaos_string& subinfo1,
		chaos_string& subinfo2);
	// encode wipe horizontal blind info
	static chaos_string encode_wipe_h_blind_info(
		int step,
		int gap, int unk,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2);
	// decode wipe horizontal blind info
	static bool decode_wipe_h_blind_info(
		const chaos_string& info,
		int& step,
		int& gap, int& unk,
		chaos_string& subinfo1,
		chaos_string& subinfo2);
	// encode wipe vertical blind info
	static chaos_string encode_wipe_v_blind_info(
		int step,
		int gap, int unk,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2);
	// decode wipe vertical blind info
	static bool decode_wipe_v_blind_info(
		const chaos_string& info,
		int& step,
		int& gap, int& unk,
		chaos_string& subinfo1,
		chaos_string& subinfo2);
	// encode wipe shake info
	static chaos_string encode_wipe_shake_info(
		int step,
		int move, int action, int unk,
		const chaos_string& subinfo);
	// decode wipe shake info
	static bool decode_wipe_shake_info(
		const chaos_string& info,
		int& step,
		int& move, int& action, int& unk,
		chaos_string& subinfo);
	// encode clone info
	static chaos_string encode_clone_info(
		const chaos_string& subinfo);
	// decode clone info
	static bool decode_clone_info(
		const chaos_string& info,
		chaos_string& subinfo);
	// get next step effect info
	static chaos_string get_next_step_effect_info(
		const chaos_string& info,
		int add_step);
	// get effect time scale
	static int get_effect_time_scale(
		const chaos_string& info);
	// is effect spippable
	static bool is_effect_spippable(
		const chaos_string& info);
	// extract base position
	static chaos_pos extract_base_position(
		const chaos_string& info);
	// remove special info
	static chaos_string remove_special_info(
		const chaos_string& oinfo,
		const chaos_string& spinfo);
	// is force cache info
	static bool is_force_cache_info(
		const chaos_string& info);
	// is force uncache info
	static bool is_force_uncache_info(
		const chaos_string& info);

// draw
private:
	// draw load image
	bool draw_load_image(
		chaos_image* image,
		const chaos_string& filename);
	// draw clear image
	bool draw_clr_image(
		chaos_image* image,
		const chaos_rect* clip_rect);
	// draw fill clear image
	bool draw_fill_clr_image(
		chaos_image* image,
		chaos_color color,
		const chaos_rect* clip_rect);
	// draw fill image
	bool draw_fill_image(
		chaos_image* image,
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo,
		const chaos_rect* clip_rect);
	// draw fade fill image
	bool draw_fade_fill_image(
		chaos_image* image,
		int fade,
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo,
		const chaos_rect* clip_rect);
	// draw trans fill image
	bool draw_trans_fill_image(
		chaos_image* image,
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo,
		const chaos_rect* clip_rect);
	// draw mono fill image
	bool draw_mono_fill_image(
		chaos_image* image,
		int x, int y,
		int width, int height,
		const chaos_color color,
		const chaos_string& subinfo,
		const chaos_rect* clip_rect);
	// draw blt image
	bool draw_blt_image(
		chaos_image* image,
		int x, int y,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_rect* clip_rect);
	// draw stretch blt image
	bool draw_stretch_blt_image(
		chaos_image* image,
		int x, int y,
		int width, int height,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_rect* clip_rect);
	// draw alpha blt image
	bool draw_alpha_blt_image(
		chaos_image* image,
		int x, int y,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_string& subinfo3,
		const chaos_rect* clip_rect);
	// draw trans fade image
	bool draw_trans_fade_image(
		chaos_image* image,
		int step,
		int unk1, int unit, int unk2, int unk3,
		int x, int y,
		int width, int height,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_rect* clip_rect);
	// draw wipe virtical scroll image
	bool draw_wipe_v_scroll_image(
		chaos_image* image,
		int step,
		int rolls, int height,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_rect* clip_rect);
	// draw wipe horizontal blind image
	bool draw_wipe_h_blind_image(
		chaos_image* image,
		int step,
		int gap, int unk,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_rect* clip_rect);
	// draw wipe vertical blind image
	bool draw_wipe_v_blind_image(
		chaos_image* image,
		int step,
		int gap, int unk,
		const chaos_string& subinfo1,
		const chaos_string& subinfo2,
		const chaos_rect* clip_rect);
	// draw wipe shake image
	bool draw_wipe_shake_image(
		chaos_image* image,
		int step,
		int move, int action, int unk,
		const chaos_string& subinfo,
		const chaos_rect* clip_rect);
	// draw clone image
	bool draw_clone_image(
		chaos_image* image,
		const chaos_string& subinfo,
		const chaos_rect* clip_rect);
	// draw image
	bool draw_image(
		chaos_image* image,
		const chaos_string& info,
		const chaos_rect* clip_rect);
	// copy image
	bool copy_image(
		chaos_image* image,
		const chaos_string& subinfo,
		bool recommend_cache,
		const chaos_rect* clip_rect);

// operation
private:
	// add image
	bool add_image();
	// delete image
	bool delete_image();
	// delete excess images
	bool delete_excessimages();
	// find image
	int find_image(const chaos_string& info) const;
	// get unlocked image
	int get_unlocked_image();
	// build image
	int build_image(const chaos_string& info);

public:
	// realize image
	bool realize_image(
		chaos_image* image,
		const chaos_string& info,
		const chaos_rect* clip_rect = NULL);
};

////////////////////////////////////////////////////////////////////////////////
// chaos_image_locker : image locker class

class chaos_image_locker {
// attribute
private:
	// image realizer
	chaos_image_realizer* image_realizer;
	// image id
	int id;

// create & destroy
public:
	// normal constructor
	chaos_image_locker(
			chaos_image_realizer* new_image_realizer,
			int new_id) :
		image_realizer(new_image_realizer),
		id(new_id)
	{
		image_realizer->lock_image(id);
	}
	// destructor
	~chaos_image_locker() {
		image_realizer->unlock_image(id);
	}

// operation
public:
	// lock
	void lock(int new_id) {
		image_realizer->unlock_image(id);
		id = new_id;
		image_realizer->lock_image(id);
	}
	// unlock
	void unlock() {
		lock(-1);
	}
	// get image
	chaos_image* get_image() {
		return image_realizer->get_image(id);
	}
};

#endif // CHAOS_IMAGE_REALIZER_H_INCLUDED
