////////////////////////////////////////////////////////////////////////////////
// chaos_interpreter.cpp : script interpreter

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_interpreter.h"
#include "chaos_func.h"
#include "chaos_func_manager.h"
#include "chaos_message_imager.h"
#include "chaos_exception.h"
#include "chaos_utility.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_interpret_nest_checker : interpreter expression nest check class

class chaos_interpret_nest_checker {
// attribute
private:
	// interpreter
	chaos_interpreter* interpreter;

// create & destroy
public:
	// normal constructor
	chaos_interpret_nest_checker(chaos_interpreter* new_interpreter) :
		interpreter(new_interpreter)
	{
		interpreter->inc_expression_next();
	}
	// destructor
	~chaos_interpret_nest_checker() {
		interpreter->dec_expression_next();
	}
};

////////////////////////////////////////////////////////////////////////////////
// chaos_interpreter : script interpreter class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// normal constructor

chaos_interpreter::chaos_interpreter(
		chaos_system* new_system,
		const chaos_string& new_script_filename,
		int new_script_id,
		int new_script_line, int new_script_column) :
	system(new_system),
	script(),
	script_filename(new_script_filename),
	script_id(new_script_id),
	script_line(new_script_line),
	script_column(new_script_column),
	expression_nest(0),
	ungeted_ch(-1),
	ungeted_token(),
	dquote(false)
{
}

// destructor

chaos_interpreter::~chaos_interpreter() {
}

// clear

void chaos_interpreter::clear() {
	script.clear();
	script.set_filename("");
	script_id = -1;
	script_line = 1;
	script_column = 1;
	expression_nest = 0;
	ungeted_ch = -1;
	ungeted_token = chaos_token();
	dquote = false;
}

////////////////////////////////////////////////////////////////////////////////
// script

// get 1 char from current script

int chaos_interpreter::get_char() {
	ASSERT(script_filename.length() > 0);
	if (script.get_filename().empty()) {
		load_script();
	}
	int ch;
	if (ungeted_ch != -1) {
		ch = ungeted_ch;
		ungeted_ch = -1;
	} else {
		ch = script.get();
	}
	if (ch != -1) {
		script_column++;
		if (ch == '\n') {
			script_column = 1;
			script_line++;
		}
	}
	return ch;
}

// unget 1 char

void chaos_interpreter::unget_char(int ch) {
	ASSERT(ungeted_ch == -1);
	ASSERT((ch == -1) || (script_line > 1) || (script_column > 1));
	ungeted_ch = ch;
	if (ungeted_ch != -1) {
		if (--script_column < 1) {
			ASSERT(ch == '\n');
			script_column = 1;
			script_line--;
		}
	}
}

// skip to next line

void chaos_interpreter::skip_to_nextline() {
	bool sjislead = false;
	int ch;
	while ((ch = get_char()) != -1) {
		if (!sjislead) {
			sjislead = chaos_utility::is_sjislead(ch);
			if (ch == '\n') {
				break;
			}
		} else {
			sjislead = false;
		}
	}
}

// skip to end comment

void chaos_interpreter::skip_to_endcomment() {
	bool sjislead = false, asterisk = false;
	int ch;
	while ((ch = get_char()) != -1) {
		if (!sjislead) {
			sjislead = chaos_utility::is_sjislead(ch);
			if (!asterisk) {
				if (ch == '*') {
					asterisk = true;
				}
			} else {
				if (ch == '/') {
					break;
				}
				asterisk = false;
			}
		} else {
			sjislead = false;
		}
	}
}

// load script

bool chaos_interpreter::load_script() {
	if (!script.get_filename().empty()) {
		return true;
	}
	system->load_script(&script, script_filename, script_id);
	script.set_filename(script_filename);
	bool result = true;
	int now_line = 1, now_column = 1;
	bool sjislead = false;
	enum {
		mm_normal,
		mm_dquote,
		mm_dquote2,
		mm_escape,
		mm_parenleft,
		mm_parenright
	} mmode = mm_normal;
	while ((now_line < script_line) || (now_column < script_column)) {
		int ch = script.get();
		if (ch == -1) {
			result = false;
			break;
		}
		now_column++;
		if (ch == '\n') {
			now_column = 1;
			now_line++;
		}
		if (script_id >= 0) {
			if (!sjislead) {
				sjislead = chaos_utility::is_sjislead(ch);
				switch (mmode) {
				case mm_normal:
					switch (ch) {
					case '"':
						dquote = true;
						mmode = mm_dquote;
						break;
					}
					break;
				case mm_dquote:
					switch (ch) {
					case '"':
						dquote = false;
						mmode = mm_normal;
						break;
					case '\\':
						mmode = mm_escape;
						break;
					}
					break;
				case mm_dquote2:
					switch (ch) {
					case '"':
						mmode = mm_parenright;
						break;
					}
					break;
				case mm_escape:
					switch (ch) {
					case '\\':
						mmode = mm_dquote;
						break;
					case '#':
					case 'c':
					case 'C':
						mmode = mm_parenleft;
						break;
					default:
						mmode = mm_dquote;
						break;
					}
					break;
				case mm_parenleft:
					switch (ch) {
					case '(':
						mmode = mm_parenright;
						break;
					}
					break;
				case mm_parenright:
					switch (ch) {
					case ')':
						mmode = mm_dquote;
						break;
					case '"':
						mmode = mm_dquote2;
						break;
					}
					break;
				}
			} else {
				sjislead = false;
			}
		}
	}
	return result;
}

// goto script

void chaos_interpreter::goto_script(
	const chaos_string& new_script_filename,
	int new_script_id,
	int new_script_line, int new_script_column)
{
	clear();
	script_filename = new_script_filename;
	script_id = new_script_id;
	script_line = new_script_line;
	script_column = new_script_column;
}

////////////////////////////////////////////////////////////////////////////////
// token

// get token from current script

chaos_token chaos_interpreter::get_token() {
	chaos_token token;
	if (ungeted_token.get_type() != chaos_token::tkn_nul) {
		token = ungeted_token;
		ungeted_token = chaos_token();
		script_column += token.get_literal().length();
		return token;
	}
	int ch;
	do {
		ch = get_char();
	} while (chaos_utility::is_space(ch));
	if (ch == -1) {
		return token;
	}
	chaos_string literal;
	literal += (char)ch;
	switch (ch) {
	case '!':
		ch = get_char();
		if (ch == '=') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_notequal, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_not, literal);
		}
		break;
	case '"':
		{ // dummy block
			chaos_string str;
			bool sjislead = false, escape = false;
			while ((ch = get_char()) != -1) {
				literal += (char)ch;
				if (!sjislead) {
					sjislead = chaos_utility::is_sjislead(ch);
					if (!escape) {
						if (ch == '"') {
							break;
						} else if (ch == '\\') {
							escape = true;
						} else {
							str += (char)ch;
						}
					} else {
						switch (ch) {
						case 't':
							str += '\t';
							break;
						case 'r':
							str += '\r';
							break;
						case 'n':
							str += '\n';
							break;
						default:
							str += (char)ch;
							break;
						}
						escape = false;
					}
				} else {
					sjislead = false;
					str += (char)ch;
				}
			}
			ch = get_char();
			if ((ch == 'l') || (ch == 'L')) {
				// script bug?
				literal += (char)ch;
			} else {
				unget_char(ch);
			}
			token = chaos_token(
				chaos_token::tkn_value, literal,
				chaos_value(str));
		}
		break;
	case '#':
		{ // dummy block
			while ((ch = get_char()) != -1) {
				if (!chaos_utility::is_alpha(ch) &&
					!chaos_utility::is_digit(ch))
				{
					unget_char(ch);
					break;
				}
				literal += (char)ch;
			}
			token = chaos_token(chaos_token::tkn_func, literal);
		}
		break;
	case '&':
		ch = get_char();
		if (ch == '&') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_and, literal);
		} else {
			unget_char(ch);
			// single and
			token = chaos_token(chaos_token::tkn_and, literal);
		}
		break;
	case '(':
		token = chaos_token(chaos_token::tkn_parenleft, literal);
		break;
	case ')':
		token = chaos_token(chaos_token::tkn_parenright, literal);
		break;
	case '*':
		ch = get_char();
		if (ch == '/') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_endcomment, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_unknown, literal);
		}
		break;
	case '+':
		ch = get_char();
		if (ch == '+') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_inc, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_plus, literal);
		}
		break;
	case ',':
		token = chaos_token(chaos_token::tkn_comma, literal);
		break;
	case '-':
		ch = get_char();
		if (ch == '-') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_dec, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_minus, literal);
		}
		break;
	case '.':
		token = chaos_token(chaos_token::tkn_period, literal);
		break;
	case '/':
		ch = get_char();
		if (ch == '*') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_startcomment, literal);
		} else if (ch == '/') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_linecomment, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_unknown, literal);
		}
		break;
	case ';':
		token = chaos_token(chaos_token::tkn_semicolon, literal);
		break;
	case '<':
		ch = get_char();
		if (ch == '>') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_notequal, literal);
		} else if (ch == '=') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_lessequal, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_less, literal);
		}
		break;
	case '=':
		ch = get_char();
		if (ch == '>') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_greaterequal, literal);
		} else if (ch == '<') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_lessequal, literal);
		} else if (ch == '=') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_equal, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_let, literal);
		}
		break;
	case '>':
		ch = get_char();
		if (ch == '=') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_greaterequal, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_greater, literal);
		}
		break;
	case '@':
		{ // dummy block
			while ((ch = get_char()) != -1) {
				if (!chaos_utility::is_alpha(ch) &&
					!chaos_utility::is_digit(ch))
				{
					unget_char(ch);
					break;
				}
				literal += (char)ch;
			}
			token = chaos_token(
				chaos_token::tkn_variable, literal,
				system->get_variable_ex(
					system->get_variable_id(literal)));
		}
		break;
	case '{':
		token = chaos_token(chaos_token::tkn_braceleft, literal);
		break;
	case '|':
		ch = get_char();
		if (ch == '|') {
			literal += (char)ch;
			token = chaos_token(chaos_token::tkn_or, literal);
		} else if (ch == '_') {
			literal += (char)ch;
			while ((ch = get_char()) != -1) {
				literal += (char)ch;
				if (ch == ':') {
					break;
				}
			}
			token = chaos_token(chaos_token::tkn_label, literal);
		} else {
			unget_char(ch);
			token = chaos_token(chaos_token::tkn_unknown, literal);
		}
		break;
	case '}':
		token = chaos_token(chaos_token::tkn_braceright, literal);
		break;
	default:
		if (chaos_utility::is_digit(ch)) {
			bool hex = false;
			if (ch == '0') {
				ch = get_char();
				if ((ch == 'x') || (ch == 'X')) {
					literal += (char)ch;
					hex = true;
				} else {
					unget_char(ch);
				}
			}
			while ((ch = get_char()) != -1) {
				if (!chaos_utility::is_digit(ch) &&
					(!hex || !chaos_utility::is_hex(ch)))
				{
					unget_char(ch);
					break;
				}
				literal += (char)ch;
			}
			char* endptr = NULL;
			chaos_int32 integer =
				(chaos_int32)strtol(literal.c_str(), &endptr, hex? 0: 10);
			ASSERT(*endptr == '\0');
			token = chaos_token(
				chaos_token::tkn_value, literal,
				chaos_value(integer));
		} else if (chaos_utility::is_alpha(ch)) {
			while ((ch = get_char()) != -1) {
				if (!chaos_utility::is_alpha(ch) &&
					!chaos_utility::is_digit(ch) &&
					(ch != '_'))
				{
					unget_char(ch);
					break;
				}
				literal += (char)ch;
			}
			if (chaos_utility::icompare(literal, "RGB") == 0) {
				token = chaos_token(
					chaos_token::tkn_rgb, literal);
			} else {
				chaos_string name = chaos_utility::to_upper_str(literal);
				if (system->exist_fix(name)) {
					token = chaos_token(
						chaos_token::tkn_fix, literal,
						system->get_fix(name));
				} else if (
					chaos_func_manager::instance()->find(literal) != NULL)
				{
					token = chaos_token(chaos_token::tkn_func, literal);
				} else {
					token = chaos_token(
						chaos_token::tkn_unknown, literal);
				}
			}
		} else {
			if (chaos_utility::is_sjislead(ch)) {
				if ((ch = get_char()) != 0) {
					literal += (char)ch;
				}
			}
			token = chaos_token(chaos_token::tkn_unknown, literal);
		}
		break;
	}
	return token;
}

// unget token

void chaos_interpreter::unget_token(const chaos_token& token) {
	ASSERT(ungeted_token.get_type() == chaos_token::tkn_nul);
	ASSERT(script_column >= 1+(int)token.get_literal().length());
	ungeted_token = token;
	script_column -= ungeted_token.get_literal().length();
}

// skip to end block

void chaos_interpreter::skip_to_endblock() {
	int block_nest = 1;
	do {
		chaos_token token = get_token();
		switch (token.get_type()) {
		case chaos_token::tkn_nul:
			block_nest = 0;
			break;
		case chaos_token::tkn_braceleft:
			block_nest++;
			break;
		case chaos_token::tkn_braceright:
			block_nest--;
			break;
		case chaos_token::tkn_linecomment:
			skip_to_nextline();
			break;
		case chaos_token::tkn_startcomment:
			skip_to_endcomment();
			break;
		}
	} while (block_nest > 0);
}

// jump to label

bool chaos_interpreter::jump_to_label(const chaos_string& label) {
	chaos_string label_literal = '|'+label+':';
	bool searching = true, found = false;
	do {
		chaos_token token = get_token();
		switch (token.get_type()) {
		case chaos_token::tkn_nul:
			searching = false;
			break;
		case chaos_token::tkn_label:
			if (chaos_utility::icompare(
					token.get_literal(),
					label_literal) == 0)
			{
				found = true;
				searching = false;
			}
			break;
		case chaos_token::tkn_linecomment:
			skip_to_nextline();
			break;
		case chaos_token::tkn_startcomment:
			skip_to_endcomment();
			break;
		}
	} while (searching);
	return found;
}

////////////////////////////////////////////////////////////////////////////////
// variable

////////////////////////////////////////////////////////////////////////////////
// expression

// toplevel expression

chaos_token chaos_interpreter::toplevel_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token result;
	chaos_token token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_nul:
		system->request_jump(
			chaos_system::jmp_return, "", -1, 1, 1, "");
		break;
	case chaos_token::tkn_linecomment:
		skip_to_nextline();
		break;
	case chaos_token::tkn_startcomment:
		skip_to_endcomment();
		break;
	case chaos_token::tkn_semicolon:
		// nothing
		break;
	case chaos_token::tkn_braceright:
		if (system->is_selection_preparation()) {
			system->finish_selection_preparation();
		}
		break;
	case chaos_token::tkn_label:
		// nothing
		break;
	default:
		unget_token(token);
		result = let_expression();
		break;
	}
	return result;
}

// let expression

chaos_token chaos_interpreter::let_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token token = or_expression();
	chaos_token token2 = get_token();
	switch (token2.get_type()) {
	case chaos_token::tkn_let:
		switch (token.get_type()) {
		case chaos_token::tkn_variable:
			token.set_value(let_expression().get_value());
			system->set_variable_ex(
				system->get_variable_id(token.get_literal()),
				token.get_value());
			break;
		default:
			LOG_ERR(
				"error : interpreter::let expression - "
				"not variable(%s)\n",
				token2.get_literal().c_str());
			throw chaos_exception(chaos_exception::bad_format);
		}
		break;
	default:
		unget_token(token2);
		break;
	}
	return token;
}

// or expression

chaos_token chaos_interpreter::or_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token left, op;
	bool expression_exist = true;
	do {
		chaos_token right = and_expression();
		switch (op.get_type()) {
		case chaos_token::tkn_nul:
			left = right;
			break;
		case chaos_token::tkn_or:
			left = chaos_token(
				chaos_token::tkn_value, "",
				chaos_value(
					left.get_value().get_boolean() ||
						right.get_value().get_boolean()));
			break;
		}
		op = get_token();
		switch (op.get_type()) {
		case chaos_token::tkn_or:
			// nothing
			break;
		default:
			unget_token(op);
			expression_exist = false;
			break;
		}
	} while (expression_exist);
	return left;
}

// and expression

chaos_token chaos_interpreter::and_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token left, op;
	bool expression_exist = true;
	do {
		chaos_token right = compare_eq_expression();
		switch (op.get_type()) {
		case chaos_token::tkn_nul:
			left = right;
			break;
		case chaos_token::tkn_and:
			left = chaos_token(
				chaos_token::tkn_value, "",
				chaos_value(
					left.get_value().get_boolean() &&
						right.get_value().get_boolean()));
			break;
		}
		op = get_token();
		switch (op.get_type()) {
		case chaos_token::tkn_and:
			// nothing
			break;
		default:
			unget_token(op);
			expression_exist = false;
			break;
		}
	} while (expression_exist);
	return left;
}

// compare(equal/notequal) expression

chaos_token chaos_interpreter::compare_eq_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token left = compare_lg_expression();
	chaos_token op = get_token();
	switch (op.get_type()) {
	case chaos_token::tkn_equal:
	case chaos_token::tkn_notequal:
		{ // dummy block
			chaos_token right = compare_lg_expression();
			switch (op.get_type()) {
			case chaos_token::tkn_equal:
				left = chaos_token(
					chaos_token::tkn_value, "",
					chaos_value(
						left.get_value() == right.get_value()));
				break;
			case chaos_token::tkn_notequal:
				left = chaos_token(
					chaos_token::tkn_value, "",
					chaos_value(
						left.get_value() != right.get_value()));
				break;
			}
		}
		break;
	default:
		unget_token(op);
		break;
	}
	return left;
}

// compare(less/lessequal/greater/greaterequal) expression

chaos_token chaos_interpreter::compare_lg_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token left = addsub_expression();
	chaos_token op = get_token();
	switch (op.get_type()) {
	case chaos_token::tkn_less:
	case chaos_token::tkn_lessequal:
	case chaos_token::tkn_greater:
	case chaos_token::tkn_greaterequal:
		{ // dummy block
			chaos_token right = addsub_expression();
			switch (op.get_type()) {
			case chaos_token::tkn_less:
				left = chaos_token(
					chaos_token::tkn_value, "",
					chaos_value(
						left.get_value() < right.get_value()));
				break;
			case chaos_token::tkn_lessequal:
				left = chaos_token(
					chaos_token::tkn_value, "",
					chaos_value(
						left.get_value() <= right.get_value()));
				break;
			case chaos_token::tkn_greater:
				left = chaos_token(
					chaos_token::tkn_value, "",
					chaos_value(
						left.get_value() > right.get_value()));
				break;
			case chaos_token::tkn_greaterequal:
				left = chaos_token(
					chaos_token::tkn_value, "",
					chaos_value(
						left.get_value() >= right.get_value()));
				break;
			}
		}
		break;
	default:
		unget_token(op);
		break;
	}
	return left;
}

// add sub expression

chaos_token chaos_interpreter::addsub_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token left, op;
	bool expression_exist = true;
	do {
		chaos_token right = unary_expression();
		switch (op.get_type()) {
		case chaos_token::tkn_nul:
			left = right;
			break;
		case chaos_token::tkn_plus:
			left = chaos_token(
				chaos_token::tkn_value, "",
				left.get_value()+right.get_value());
			break;
		case chaos_token::tkn_minus:
			left = chaos_token(
				chaos_token::tkn_value, "",
				left.get_value()-right.get_value());
			break;
		}
		op = get_token();
		switch (op.get_type()) {
		case chaos_token::tkn_plus:
		case chaos_token::tkn_minus:
			// nothing
			break;
		default:
			unget_token(op);
			expression_exist = false;
			break;
		}
	} while (expression_exist);
	return left;
}

// unary expression

chaos_token chaos_interpreter::unary_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token op = get_token();
	switch (op.get_type()) {
	case chaos_token::tkn_plus:
	case chaos_token::tkn_minus:
	case chaos_token::tkn_not:
	case chaos_token::tkn_inc:
	case chaos_token::tkn_dec:
		// nothing
		break;
	default:
		unget_token(op);
		op = chaos_token();
		break;
	}
	chaos_token token = primary_expression();
	switch (op.get_type()) {
	case chaos_token::tkn_plus:
		token = chaos_token(
			chaos_token::tkn_value, "",
			chaos_value((chaos_int32)0)+token.get_value());
		break;
	case chaos_token::tkn_minus:
		token = chaos_token(
			chaos_token::tkn_value, "",
			chaos_value((chaos_int32)0)-token.get_value());
		break;
	case chaos_token::tkn_not:
		token = chaos_token(
			chaos_token::tkn_value, "",
			chaos_value(!token.get_value().get_boolean()));
		break;
	case chaos_token::tkn_inc:
	case chaos_token::tkn_dec:
		{ // dummy block
			if (token.get_type() != chaos_token::tkn_variable) {
				LOG_ERR(
					"error : interpreter::unary expression - "
					"not variable(%s)\n",
					op.get_literal().c_str());
				throw chaos_exception(chaos_exception::bad_format);
			}
			chaos_value value = token.get_value();
			switch (op.get_type()) {
			case chaos_token::tkn_inc:
				value += chaos_value((chaos_int32)1);
				break;
			case chaos_token::tkn_dec:
				value -= chaos_value((chaos_int32)1);
				break;
			}
			system->set_variable_ex(
				system->get_variable_id(token.get_literal()),
				value);
			token = chaos_token(
				chaos_token::tkn_value, "",
				value);
		}
		break;
	}
	op = get_token();
	switch (op.get_type()) {
	case chaos_token::tkn_inc:
	case chaos_token::tkn_dec:
		{ // dummy block
			if (token.get_type() != chaos_token::tkn_variable) {
				LOG_ERR(
					"error : interpreter::unary expression - "
					"not variable(%s)\n",
					op.get_literal().c_str());
				throw chaos_exception(chaos_exception::bad_format);
			}
			chaos_value value = token.get_value(),
				value2 = value;
			switch (op.get_type()) {
			case chaos_token::tkn_inc:
				value += chaos_value((chaos_int32)1);
				break;
			case chaos_token::tkn_dec:
				value -= chaos_value((chaos_int32)1);
				break;
			}
			system->set_variable_ex(
				system->get_variable_id(token.get_literal()),
				value);
			token = chaos_token(
				chaos_token::tkn_value, "",
				value2);
		}
		break;
	default:
		unget_token(op);
		break;
	}
	return token;
}

// primary expression

chaos_token chaos_interpreter::primary_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_value:
	case chaos_token::tkn_variable:
	case chaos_token::tkn_fix:
		// nothing
		break;
	case chaos_token::tkn_func:
		token = func_expression(token.get_literal());
		break;
	case chaos_token::tkn_rgb:
		token = rgb_expression();
		break;
	case chaos_token::tkn_parenleft:
		unget_token(token);
		token = paren_expression();
		break;
	default:
		LOG_ERR(
			"error : interpreter::primary expression - "
			"bad token(%s)\n",
			token.get_literal().c_str());
		throw chaos_exception(chaos_exception::bad_format);
	}
	return token;
}

// func expression

chaos_token chaos_interpreter::func_expression(
	const chaos_string& func_name)
{
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token token = get_token();
	if (token.get_type() == chaos_token::tkn_let) {
		token = get_token();
	}
	chaos_func::params_type params;
	if (token.get_type() == chaos_token::tkn_parenleft) {
		token = get_token();
		bool params_exist = true;
		if (token.get_type() == chaos_token::tkn_parenright) {
			params_exist = false;
		} else {
			unget_token(token);
		}
		while (params_exist) {
			token = or_expression();
			params.push_back(token.get_value());
			token = get_token();
			switch (token.get_type()) {
			case chaos_token::tkn_comma:
				// nothing
				break;
			case chaos_token::tkn_parenright:
				params_exist = false;
				break;
			case chaos_token::tkn_semicolon:
				unget_token(token);
				params_exist = false;
				LOG_RPT(
					"report : interpreter::func expression - "
					"no paren right\n");
				break;
			default:
				LOG_ERR(
					"error : interpreter::func expression - "
					"bad token(%s)\n",
					token.get_literal().c_str());
				throw chaos_exception(chaos_exception::bad_format);
			}
		}
	} else {
		unget_token(token);
	}
	if (chaos_log::instance()->get_log_level() >= chaos_log::loglv_report) {
		// func & param report
		LOG_RPT("%s", func_name.c_str());
		if (params.size() > 0) {
			LOG_RPT(" : ");
			for (
				chaos_func::params_type::iterator param = params.begin();
				param != params.end();
				param++)
			{
				switch ((*param).get_type()) {
				case chaos_value::nul:
					LOG_RPT("nul ");
					break;
				case chaos_value::integer:
					LOG_RPT(
						"%d ",
						(int)(*param).get_integer());
					break;
				case chaos_value::color:
					LOG_RPT(
						"%s ",
						(*param).get_string().c_str());
					break;
				case chaos_value::string:
					LOG_RPT(
						"\"%s\" ",
						(*param).get_string().c_str());
					break;
				}
			}
		}
		LOG_RPT("\n");
	}
	const chaos_func* func =
		chaos_func_manager::instance()->find(func_name);
	if (func != NULL) {
		token = chaos_token(
			chaos_token::tkn_value, "",
			func->execute(system, this, &params));
	} else {
		LOG_WRN(
			"warning : interpreter::func expression - "
			"unknown function(%s)\n",
			func_name.c_str());
	}
	return chaos_token();
}

// rgb expression

chaos_token chaos_interpreter::rgb_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_parenleft:
		// nothing
		break;
	default:
		LOG_ERR(
			"error : interpreter::rgb expression - "
			"bad token(%s)\n",
			token.get_literal().c_str());
		throw chaos_exception(chaos_exception::bad_format);
	}
	token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_parenright:
		// null rgb
		return chaos_token();
	default:
		unget_token(token);
		break;
	}
	chaos_token red = or_expression();
	token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_comma:
		// nothing
		break;
	case chaos_token::tkn_period:
		LOG_RPT(
			"report : interpreter::rgb expression - "
			"period substitute comma\n");
		break;
	default:
		LOG_ERR(
			"error : interpreter::rgb expression - "
			"bad token(%s)\n",
			token.get_literal().c_str());
		throw chaos_exception(chaos_exception::bad_format);
	}
	chaos_token green = or_expression();
	token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_comma:
		// nothing
		break;
	case chaos_token::tkn_period:
		LOG_RPT(
			"report : interpreter::rgb expression - "
			"period substitute comma\n");
		break;
	default:
		LOG_ERR(
			"error : interpreter::rgb expression - "
			"bad token(%s)\n",
			token.get_literal().c_str());
		throw chaos_exception(chaos_exception::bad_format);
	}
	chaos_token blue = or_expression();
	token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_parenright:
		// nothing
		break;
	case chaos_token::tkn_semicolon:
		unget_token(token);
		LOG_RPT(
			"report : interpreter::rgb expression - "
			"no paren right\n");
		break;
	default:
		LOG_ERR(
			"error : interpreter::rgb expression - "
			"bad token(%s)\n",
			token.get_literal().c_str());
		throw chaos_exception(chaos_exception::bad_format);
	}
	return chaos_token(
		chaos_token::tkn_value, "",
		chaos_value(
			chaos_color(
				red.get_value().get_integer(),
				green.get_value().get_integer(),
				blue.get_value().get_integer())));
}

// paren expression

chaos_token chaos_interpreter::paren_expression() {
	chaos_interpret_nest_checker nest_checker(this);
	chaos_token token = get_token();
	switch (token.get_type()) {
	case chaos_token::tkn_parenleft:
		// nothing
		break;
	default:
		LOG_ERR(
			"error : interpreter::paren expression - "
			"bad token(%s)\n",
			token.get_literal().c_str());
		throw chaos_exception(chaos_exception::bad_format);
	}
	token = let_expression();
	chaos_token token2 = get_token();
	switch (token2.get_type()) {
	case chaos_token::tkn_parenright:
		// nothing
		break;
	case chaos_token::tkn_semicolon:
		unget_token(token2);
		LOG_RPT(
			"report : interpreter::paren expression - "
			"no paren right\n");
		break;
	default:
		LOG_ERR(
			"error : interpreter::paren expression - "
			"bad token(%s)\n",
			token2.get_literal().c_str());
		throw chaos_exception(chaos_exception::bad_format);
	}
	return token;
}

////////////////////////////////////////////////////////////////////////////////
// message

// decode message

void chaos_interpreter::decode_message() {
	int ch;
	chaos_token token;
	chaos_string text;
	bool wait = false;
	if (!dquote) {
		do {
			ch = get_char();
		} while (chaos_utility::is_space(ch));
		switch (ch) {
		case -1:
			system->request_jump(
				chaos_system::jmp_return, "", -1, 1, 1, "");
			break;
		case '+':
			// nothing
			break;
		case '"':
			dquote = true;
			break;
		case ',':
			system->add_message_frame();
			break;
		case '@':
			unget_char(ch);
			token = get_token();
			if (token.get_type() != chaos_token::tkn_variable) {
				LOG_ERR(
					"error : interpreter::decode message - "
					"bad token(%s)\n",
					token.get_literal().c_str());
				throw chaos_exception(chaos_exception::bad_format);
			}
			text = token.get_value().get_string();
			wait = true;
			break;
		}
	} else {
		ch = get_char();
		switch (ch) {
		case -1:
			system->request_jump(
				chaos_system::jmp_return, "", -1, 1, 1, "");
			break;
		case '"':
			dquote = false;
			break;
		case '\\':
			ch = get_char();
			switch (ch) {
			case '\\':
				token = get_token();
				if (token.get_type() != chaos_token::tkn_variable) {
					LOG_ERR(
						"error : interpreter::decode message - "
						"bad token(%s)\n",
						token.get_literal().c_str());
					throw chaos_exception(chaos_exception::bad_format);
				}
				text = token.get_value().get_string();
				wait = true;
				break;
			case '#':
				unget_char(ch);
				token = get_token();
				if (token.get_type() != chaos_token::tkn_func) {
					LOG_ERR(
						"error : interpreter::decode message - "
						"bad token(%s)\n",
						token.get_literal().c_str());
					throw chaos_exception(chaos_exception::bad_format);
				}
				func_expression(token.get_literal());
				break;
			case 'c':
			case 'C':
				token = rgb_expression();
				if (token.get_value().get_type() == chaos_value::nul) {
					text = chaos_message_imager::encode_rescolor_info();
				} else {
					text = chaos_message_imager::encode_curcolor_info(
						token.get_value().get_color());
				}
				break;
			case 'z':
			case 'Z':
				// nothing
				break;
			case 'n':
			case 'N':
				text = '\n';
				break;
			case 'm':
			case 'M':
				system->wait_message();
				break;
			}
			break;
		default:
			text = (char)ch;
			wait = true;
			if (chaos_utility::is_sjislead(ch)) {
				ch = get_char();
				if (ch != -1) {
					text += (char)ch;
				}
			}
			break;
		}
	}
	if (!text.empty()) {
		system->add_message(text, wait);
	}
	if (ungeted_token.get_type() != chaos_token::tkn_nul) {
		get_token();
	}
}

////////////////////////////////////////////////////////////////////////////////
// operation

// inc expression nest

void chaos_interpreter::inc_expression_next() {
	expression_nest++;
	if (expression_nest >= (int)expression_nest_max) {
		throw chaos_exception(chaos_exception::stack_overflow);
	} else if (expression_nest >= (int)expression_nest_max/2) {
		LOG_WRN(
			"warning : interpreter::expression - "
			"too deep nest(%d).\n",
			expression_nest);
	}
}

// dec expression nest

void chaos_interpreter::dec_expression_next() {
	expression_nest--;
}

// interpret

void chaos_interpreter::interpret() {
	expression_nest = 0;
	try {
		if (script.get_filename().empty()) {
			load_script();
		}
		if (script_id < 0) {
			toplevel_expression();
		} else {
			decode_message();
		}
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : interpreter::interpret - "
			"%s\n",
			e.what());
	}
}
