////////////////////////////////////////////////////////////////////////////////
// chaos_interpreter.h : script interpreter

#ifndef CHAOS_INTERPRETER_H_INCLUDED
#define CHAOS_INTERPRETER_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_interpreter;

////////////////////////////////////////////////////////////////////////////////
// include

#ifndef CHAOS_STRING_H_INCLUDED
#include "chaos_string.h"
#endif // CHAOS_STRING_H_INCLUDED

#ifndef CHAOS_FILE_DATA_H_INCLUDED
#include "chaos_file_data.h"
#endif // CHAOS_FILE_DATA_H_INCLUDED

#ifndef CHAOS_VALUE_H_INCLUDED
#include "chaos_value.h"
#endif // CHAOS_VALUE_H_INCLUDED

#ifndef CHAOS_TOKEN_H_INCLUDED
#include "chaos_token.h"
#endif // CHAOS_TOKEN_H_INCLUDED

#ifndef CHAOS_SYSTEM_H_INCLUDED
#include "chaos_system.h"
#endif // CHAOS_SYSTEM_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// chaos_interpreter : script interpreter class

class chaos_interpreter {
// enum
private:
	// nest
	enum {
		expression_nest_max = 256
	};

// attribute
private:
	// system
	chaos_system* system;
	// script
	chaos_file_data script;
	// script filename
	chaos_string script_filename;
	// script id
	int script_id;
	// script line
	int script_line;
	// script column
	int script_column;
	// expression nest
	int expression_nest;
	// unget char buffer
	int ungeted_ch;
	// unget token buffer
	chaos_token ungeted_token;
	// message double quote
	bool dquote;

// create & destroy
public:
	// normal constructor
	chaos_interpreter(
		chaos_system* new_system,
		const chaos_string& new_script_filename,
		int new_script_id,
		int new_script_line, int script_column);
	// destructor
	~chaos_interpreter();
	// clear
	void clear();

// script
private:
	// get 1 char from current script
	int get_char();
	// unget 1 char
	void unget_char(int ch);
	// skip to next line
	void skip_to_nextline();
	// skip to end comment
	void skip_to_endcomment();

public:
	// get script filename
	const chaos_string& get_script_filename() const {
		return script_filename;
	}
	// get script id
	int get_script_id() const {
		return script_id;
	}
	// get script line
	int get_script_line() const {
		return script_line;
	}
	// get script column
	int get_script_column() const {
		return script_column;
	}
	// load script
	bool load_script();
	// goto script
	void goto_script(
		const chaos_string& new_script_filename,
		int new_script_id,
		int new_script_line, int new_script_column);

// token
public:
	// get token from current script
	chaos_token get_token();
	// unget token
	void unget_token(const chaos_token& token);
	// skip to end block
	void skip_to_endblock();
	// jump to label
	bool jump_to_label(const chaos_string& label);

// expression
public:
	// toplevel expression
	chaos_token toplevel_expression();
	// let expression
	chaos_token let_expression();
	// or expression
	chaos_token or_expression();
	// and expression
	chaos_token and_expression();
	// compare(equal/notequal) expression
	chaos_token compare_eq_expression();
	// compare(less/lessequal/greater/greaterequal) expression
	chaos_token compare_lg_expression();
	// add sub expression
	chaos_token addsub_expression();
	// unary expression
	chaos_token unary_expression();
	// primary expression
	chaos_token primary_expression();
	// func expression
	chaos_token func_expression(
		const chaos_string& func_name);
	// rgb expression
	chaos_token rgb_expression();
	// paren expression
	chaos_token paren_expression();

// message
private:
	// decode message
	void decode_message();

// operation
public:
	// inc expression nest
	void inc_expression_next();
	// dec expression nest
	void dec_expression_next();
	// interpret
	void interpret();
};

#endif // CHAOS_INTERPRETER_H_INCLUDED
