////////////////////////////////////////////////////////////////////////////////
// chaos_message_imager.cpp : message imager

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_message_imager.h"
#include "chaos_exception.h"
#include "chaos_utility.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_message_imager : message imager

////////////////////////////////////////////////////////////////////////////////
// attribute

// bit count table

int chaos_message_imager::bits_table[256];

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_message_imager::chaos_message_imager() :
	message_length(0),
	message_column(0),
	message_shape(3),
	org_font_size(28),
	cur_font_size(28),
	real_font_size(0),
	font_bold(true),
	font_anti_alias(true),
	manual_anti_alias(true),
	message_image(NULL)
{

#ifdef CHAOS_FONT_WIN32

	manual_anti_alias = true;
	dc = NULL;
	message_font = NULL;
	org_font = NULL;
	font_bmp = NULL;
	org_bmp = NULL;
	font_bmp_bits = NULL;

#elif defined(CHAOS_FONT_GTK)

	manual_anti_alias = true;
	message_font = NULL;
	font_window = NULL;
	font_pixmap = NULL;
	font_gc = NULL;
	memset(&font_info, 0, sizeof(font_info));

#elif defined(CHAOS_FONT_PANGO)

	manual_anti_alias = false;
	surface_pixels = NULL;
	surface_stride = 0;
	cairo_surface = NULL;
	cairo_context = NULL;
	pango_layout = NULL;

#endif // CHAOS_FONT

	message_image = new chaos_image;
	// bit count table initialize
	if (bits_table[255] == 0) {
		for (int i = 0; i < 256; i++) {
			bits_table[i] = (i & 1)+
				((i >> 1) & 1)+
				((i >> 2) & 1)+
				((i >> 3) & 1)+
				((i >> 4) & 1)+
				((i >> 5) & 1)+
				((i >> 6) & 1)+
				((i >> 7) & 1);
		}
	}
	// prohibit char
	// touten
	linetop_prohibit.insert(0x8141);
	// kuten
	linetop_prohibit.insert(0x8142);
	// comma
	linetop_prohibit.insert(0x8143);
	// period
	linetop_prohibit.insert(0x8144);
	// question
	linetop_prohibit.insert(0x8148);
	// exclamation
	linetop_prohibit.insert(0x8149);
	// close kakko
	linetop_prohibit.insert(0x8176);
	// close double-kakko
	linetop_prohibit.insert(0x8178);
}

// destructor

chaos_message_imager::~chaos_message_imager() {
	dispose();
	delete message_image;
}

// initialize

bool chaos_message_imager::initialize(chaos_dim image_dim) {
	if (!message_image->create(
			image_dim,
			false))
	{
		return false;
	}
	message_valid_rect = chaos_rect(chaos_pos(0, 0), image_dim);
	clear_message_text();
	return true;
}

// dispose

void chaos_message_imager::dispose() {
	destroy_font();

#ifdef CHAOS_FONT_GTK

	if (font_window != NULL) {
		gtk_widget_unref(font_window);
		font_window = NULL;
	}

#endif // CHAOS_FONT

}

// clear

void chaos_message_imager::clear() {
	message_image->destroy();
}

////////////////////////////////////////////////////////////////////////////////
// info

// encode position info

chaos_string chaos_message_imager::encode_position_info(
	int x, int y)
{
	return chaos_utility::format_str(
		"\\p(%d,%d)",
		x, y);
}

// encode length info

chaos_string chaos_message_imager::encode_length_info(
	int length)
{
	return chaos_utility::format_str(
		"\\l(%d)",
		length);
}

// encode current color info

chaos_string chaos_message_imager::encode_curcolor_info(
	chaos_color curcolor)
{
	return chaos_utility::format_str(
		"\\c(%d,%d,%d)",
		curcolor.get_red(),
		curcolor.get_green(),
		curcolor.get_blue());
}

// encode restore current color info

chaos_string chaos_message_imager::encode_rescolor_info() {
	return "\\c()";
}

// encode default color info

chaos_string chaos_message_imager::encode_defcolor_info(
	chaos_color defcolor)
{
	return chaos_utility::format_str(
		"\\d(%d,%d,%d)",
		defcolor.get_red(),
		defcolor.get_green(),
		defcolor.get_blue());
}

// encode sub color info

chaos_string chaos_message_imager::encode_subcolor_info(
	chaos_color subcolor)
{
	return chaos_utility::format_str(
		"\\s(%d,%d,%d)",
		subcolor.get_red(),
		subcolor.get_green(),
		subcolor.get_blue());
}

// encode shape info

chaos_string chaos_message_imager::encode_shape_info(
	int shape)
{
	return chaos_utility::format_str(
		"\\h(%d)",
		shape);
}

// encode font size info

chaos_string chaos_message_imager::encode_font_size_info(
	int font_size)
{
	return chaos_utility::format_str(
		"\\f(%d)",
		font_size);
}

////////////////////////////////////////////////////////////////////////////////
// operation

// create font

bool chaos_message_imager::create_font(int new_font_size) {
	destroy_font();
	real_font_size = new_font_size;
	int create_font_size = real_font_size;
	if (font_anti_alias && manual_anti_alias) {
		create_font_size *= 2;
	}
	bmp_dim = chaos_dim(create_font_size, create_font_size);

#ifdef CHAOS_FONT_WIN32

	dc = CreateCompatibleDC(NULL);
	if (dc == NULL) {
		LOG_ERR(
			"error : message imager::create font - "
			"cannot create device context\n");
		destroy_font();
		return false;
	}
	if (message_font == NULL) {
		LOGFONT lf;
		memset(&lf, 0, sizeof(lf));
		lf.lfHeight = create_font_size;
		lf.lfWeight = font_bold? FW_BOLD: FW_NORMAL;
		lf.lfCharSet = SHIFTJIS_CHARSET;
		lf.lfPitchAndFamily = FIXED_PITCH;
		strcpy(
			lf.lfFaceName,
			font_name.c_str());
		message_font = CreateFontIndirect(&lf);
		if (message_font == NULL) {
			LOG_ERR(
				"error : message imager::create font - "
				"cannot create font(%s)\n",
				font_name.c_str());
			destroy_font();
			return false;
		}
		org_font = (HFONT)SelectObject(dc, message_font);
		SIZE font_extent;
		char str[3];
		str[0] = '\x88';
		str[1] = '\x9F';
		str[2] = '\0';
		if (GetTextExtentPoint32(dc, str, 2, &font_extent)) {
			if (font_extent.cx > bmp_dim.width) {
				bmp_dim.width = font_extent.cx;
			}
			if (font_extent.cy > bmp_dim.height) {
				bmp_dim.height = font_extent.cy;
			}
		}
	}
	if (font_bmp == NULL) {
		chaos_uint8 works[sizeof(BITMAPINFO)+sizeof(RGBQUAD)*(2-1)];
		memset(works, 0, sizeof(works));
		BITMAPINFO* bmi = (BITMAPINFO*)works;
		bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		bmi->bmiHeader.biWidth = bmp_dim.width;
		bmi->bmiHeader.biHeight = bmp_dim.height;
		bmi->bmiHeader.biBitCount = 1;
		bmi->bmiHeader.biPlanes = 1;
		bmi->bmiHeader.biCompression = BI_RGB;
		bmi->bmiColors[1].rgbRed = 255;
		bmi->bmiColors[1].rgbGreen = 255;
		bmi->bmiColors[1].rgbBlue = 255;
		font_bmp = (HBITMAP)CreateDIBSection(
			dc, bmi, DIB_RGB_COLORS, &font_bmp_bits, NULL, 0);
		if (font_bmp == NULL) {
			LOG_ERR(
				"error : message imager::create font - "
				"cannot create bitmap\n");
			destroy_font();
			return false;
		}
		org_bmp = (HBITMAP)SelectObject(dc, font_bmp);
	}

#elif defined(CHAOS_FONT_GTK)

	chaos_string real_font_name =
		normalize_font_xlfd(font_name, create_font_size)+",*";
	message_font = gdk_fontset_load(
		chaos_utility::sjis_to_gui(real_font_name).c_str());
	if (message_font == NULL) {
		LOG_ERR(
			"error : message imager::create font - "
			"cannot load font(%s)\n",
			real_font_name.c_str());
		destroy_font();
		return false;
	}
	gdk_string_extents(
		message_font,
		chaos_utility::sjis_to_gui("\x88\x9F").c_str(),
		&font_info.lbearing,
		&font_info.rbearing,
		&font_info.width,
		&font_info.ascent,
		&font_info.descent);
	font_info.baseline = font_info.ascent;

#ifdef CHAOS_FONT_GTK_BASELINE

	font_info.baseline -= font_info.descent;

#endif // CHAOS_FONT_GTK_BASELINE

	font_info.height = font_info.baseline+font_info.descent;
	if (font_info.width > bmp_dim.width) {
		bmp_dim.width = font_info.width;
	}
	if (font_info.height > bmp_dim.height) {
		bmp_dim.height = font_info.height;
	}
	if (font_window == NULL) {
		font_window = gtk_window_new(GTK_WINDOW_POPUP);
		if (font_window == NULL) {
			LOG_ERR(
			"error : message imager::create font - "
				"cannot create window\n");
			destroy_font();
			return false;
		}
		gtk_widget_ref(font_window);
		gtk_widget_realize(font_window);
		if (!GTK_WIDGET_REALIZED(font_window)) {
			LOG_ERR(
				"error : message imager::create font - "
				"cannot realize window\n");
			destroy_font();
			return false;
		}
	}
	if (font_pixmap == NULL) {
		font_pixmap = gdk_pixmap_new(
			font_window->window,
			bmp_dim.width, bmp_dim.height,
			1);
		if (font_pixmap == NULL) {
			LOG_ERR(
				"error : message imager::create font - "
				"cannot create pixmap\n");
			destroy_font();
			return false;
		}
	}
	if (font_gc == NULL) {
		font_gc = gdk_gc_new(font_pixmap);
		if (font_gc == NULL) {
			LOG_ERR(
				"error : message imager::create font - "
				"cannot create gc\n");
			destroy_font();
			return false;
		}
	}

#elif defined(CHAOS_FONT_PANGO)

	surface_dim = chaos_dim(
		create_font_size, create_font_size);
	cairo_surface = cairo_image_surface_create(
		CAIRO_FORMAT_A8,
		surface_dim.width, surface_dim.height);
	if (cairo_surface == NULL) {
		LOG_ERR(
			"error : message imager::create font - "
			"cannot create cairo surface\n");
		destroy_font();
		return false;
	}
	surface_pixels = (chaos_uint8*)cairo_image_surface_get_data(
		cairo_surface);
	surface_stride = cairo_image_surface_get_stride(
		cairo_surface);
	cairo_context = cairo_create(
		cairo_surface);
	if (cairo_context == NULL) {
		LOG_ERR(
			"error : message imager::create font - "
			"cannot create cairo context\n");
		destroy_font();
		return false;
	}
	pango_layout = pango_cairo_create_layout(
		cairo_context);
	if (pango_layout == NULL) {
		LOG_ERR(
			"error : message imager::create font - "
			"cannot create pango layout\n");
		destroy_font();
		return false;
	}
	PangoFontDescription* pango_font_desc =
		pango_font_description_from_string(
			chaos_utility::sjis_to_gui(font_name).c_str());
	if (pango_font_desc == NULL) {
		LOG_ERR(
			"error : message imager::create font - "
			"cannot create pango font description\n");
		destroy_font();
		return false;
	}
	pango_font_description_set_absolute_size(
		pango_font_desc,
		(double)create_font_size*PANGO_SCALE);
	pango_layout_set_font_description(
		pango_layout,
		pango_font_desc);
	pango_font_description_free(
		pango_font_desc);

#endif // CHAOS_FONT

	return true;
}

// destroy font

void chaos_message_imager::destroy_font() {

#ifdef CHAOS_FONT_WIN32

	if (org_font != NULL) {
		SelectObject(dc, org_font);
		org_font = NULL;
	}
	if (org_bmp != NULL) {
		SelectObject(dc, org_bmp);
		org_bmp = NULL;
	}
	if (dc != NULL) {
		DeleteDC(dc);
		dc = NULL;
	}
	if (message_font != NULL) {
		DeleteObject(message_font);
		message_font = NULL;
	}
	if (font_bmp != NULL) {
		DeleteObject(font_bmp);
		font_bmp = NULL;
		font_bmp_bits = NULL;
	}

#elif defined(CHAOS_FONT_GTK)

	if (message_font != NULL) {
		gdk_font_unref(message_font);
		message_font = NULL;
	}
	if (font_pixmap != NULL) {
		gdk_pixmap_unref(font_pixmap);
		font_pixmap = NULL;
	}
	if (font_gc != NULL) {
		gdk_gc_unref(font_gc);
		font_gc = NULL;
	}

#elif defined(CHAOS_FONT_PANGO)

	if (pango_layout != NULL) {
		g_object_unref(
			G_OBJECT(pango_layout ));
		pango_layout = NULL;
	}
	if (cairo_context != NULL) {
		cairo_destroy(
			cairo_context);
		cairo_context = NULL;
	}
	if (cairo_surface != NULL) {
		cairo_surface_destroy(
			cairo_surface);
		cairo_surface = NULL;
		surface_pixels = NULL;
		surface_stride = 0;
		surface_dim = chaos_dim(0, 0);
	}

#endif // CHAOS_FONT

	real_font_size = 0;
}

// draw font image

bool chaos_message_imager::draw_font_image(
	int ch,
	chaos_pos font_pos,
	int font_size,
	int shape,
	chaos_uint32 color,
	chaos_rect* font_rect)
{
	if ((ch == 0x20) || (ch == 0x8140)) {
		font_rect->clear();
		return false;
	}
	int shade_size = font_size/(int)shade_size_division;
	chaos_rect face_rect(font_pos, chaos_dim(font_size, font_size)),
		shade_rect = face_rect+chaos_pos(shade_size, shade_size),
		image_rect(chaos_pos(0, 0), message_image->get_dim());
	*font_rect = face_rect;
	if (shape == 3) {
		*font_rect |= shade_rect;
	}
	*font_rect &= image_rect;
	if (font_rect->empty()) {
		return false;
	}
	if (font_size != real_font_size) {
		if (!create_font(font_size)) {
			throw chaos_exception(chaos_exception::gui_error);
		}
	}
	message_valid_rect |= *font_rect;

#ifdef CHAOS_FONT_WIN32

	char str[3];
	int length;
	if (ch <= 0xFF) {
		length = 1;
		str[0] = (char)ch;
		str[1] = '\0';
	} else {
		length = 2;
		str[0] = (char)(ch >> 8);
		str[1] = (char)ch;
		str[2] = '\0';
	}
	int src_scanline = ((bmp_dim.width+31)/32)*4;
	memset(font_bmp_bits, 0xFF, bmp_dim.height*src_scanline);
	TextOut(dc, 0, 0, str, length);
	chaos_rect draw_rect;
	chaos_pos font_ofs;
	int red, green, blue;
	chaos_uint8* src;
	int dst_scanline, dst_step;
	chaos_uint32* dst;
	int x, y;
	if (shape == 3) {
		draw_rect = shade_rect & image_rect;
		font_ofs = draw_rect.get_lt_pos()-shade_rect.get_lt_pos();
		red = green = blue = 0;
	} else {
		draw_rect = face_rect & image_rect;
		font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
		red = CHAOS_RGBA_RED(color);
		green = CHAOS_RGBA_GREEN(color);
		blue = CHAOS_RGBA_BLUE(color);
	}
	int bmp_rate = font_anti_alias? 2: 1;
	if (!draw_rect.empty()) {
		src = (chaos_uint8*)font_bmp_bits+
			src_scanline*(bmp_dim.height-font_ofs.y*bmp_rate-1)+
			(font_ofs.x*bmp_rate)/8;
		dst_scanline = message_image->get_scanline_length();
		dst_step = dst_scanline-draw_rect.width;
		dst = message_image->get_bits(draw_rect.get_lt_pos());
		if (font_anti_alias) {
			y = draw_rect.height;
			do {
				chaos_uint8* src1 = src;
				chaos_uint8* src2 = src-src_scanline;
				int xmsk = 0xC0 >> ((font_ofs.x%4)*2);
				x = draw_rect.width;
				do {
					int alpha =
						bits_table[*src1 & xmsk]+
						bits_table[*src2 & xmsk];
					alpha = alpha*64-alpha/4;
					*(dst++) = CHAOS_RGBA(
						red, green, blue, alpha);
					if ((xmsk >>= 2) == 0) {
						xmsk = 0xC0;
						src1++;
						src2++;
					}
				} while (--x != 0);
				dst += dst_step;
				src -= src_scanline*2;
			} while (--y != 0);
		} else {
			y = draw_rect.height;
			do {
				chaos_uint8* src1 = src;
				int xmsk = 0x80 >> (font_ofs.x%8);
				x = draw_rect.width;
				do {
					int alpha = bits_table[*src1 & xmsk];
					alpha = (-alpha) & 0xFF;
					*(dst++) = CHAOS_RGBA(
						red, green, blue, alpha);
					if ((xmsk >>= 1) == 0) {
						xmsk = 0x80;
						src1++;
					}
				} while (--x != 0);
				dst += dst_step;
				src -= src_scanline;
			} while (--y != 0);
		}
	}
	if (shape == 3) {
		draw_rect = face_rect & image_rect;
		font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
		red = CHAOS_RGBA_RED(color);
		green = CHAOS_RGBA_GREEN(color);
		blue = CHAOS_RGBA_BLUE(color);
		if (!draw_rect.empty()) {
			src = (chaos_uint8*)font_bmp_bits+
				src_scanline*(bmp_dim.height-font_ofs.y*bmp_rate-1)+
				(font_ofs.x*bmp_rate)/8;
			dst_scanline = message_image->get_scanline_length();
			dst_step = dst_scanline-draw_rect.width;
			dst = message_image->get_bits(draw_rect.get_lt_pos());
			if (font_anti_alias) {
				y = draw_rect.height;
				do {
					chaos_uint8* src1 = src;
					chaos_uint8* src2 = src-src_scanline;
					int xmsk = 0xC0 >> ((font_ofs.x%4)*2);
					x = draw_rect.width;
					do {
						int alpha =
							bits_table[*src1 & xmsk]+
							bits_table[*src2 & xmsk];
						alpha = 256-alpha*64;
						int alpha2 = CHAOS_RGBA_ALPHA(*dst);
						alpha2 = 256-(alpha2+(alpha2+1)/256);
						alpha2 = (alpha2*(256-alpha))/256;
						int alpha3 = alpha+alpha2;
						int dencity = 256;
						if (alpha3 > 0) {
							dencity = (alpha*256)/alpha3;
						}
						alpha3 = 256-alpha3;
						alpha3 -= alpha3/256;
						*(dst++) = CHAOS_RGBA(
							(red*dencity)/256,
							(green*dencity)/256,
							(blue*dencity)/256,
							alpha3);
						if ((xmsk >>= 2) == 0) {
							xmsk = 0xC0;
							src1++;
							src2++;
						}
					} while (--x != 0);
					dst += dst_step;
					src -= src_scanline*2;
				} while (--y != 0);
			} else {
				y = draw_rect.height;
				do {
					chaos_uint8* src1 = src;
					int xmsk = 0x80 >> (font_ofs.x%8);
					x = draw_rect.width;
					do {
						int alpha = bits_table[*src1 & xmsk];
						alpha = (-alpha) & 0xFF;
						int alpha2 = CHAOS_RGBA_ALPHA(*dst);
						int alpha3 = alpha & alpha2;
						int dencity = alpha ^ 0xFF;
						*(dst++) = CHAOS_RGBA(
							red & dencity,
							green & dencity,
							blue & dencity,
							alpha3);
						if ((xmsk >>= 1) == 0) {
							xmsk = 0x80;
							src1++;
						}
					} while (--x != 0);
					dst += dst_step;
					src -= src_scanline;
				} while (--y != 0);
			}
		}
	}

#elif defined(CHAOS_FONT_GTK)

	char str[3];
	bool half = (ch <= 0xFF);
	if (half) {
		int ch2 = chaos_utility::jis_to_sjis(
			chaos_utility::jis_han_to_zen(
				chaos_utility::sjis_to_jis(ch)));
		str[0] = (char)(ch2 >> 8);
		str[1] = (char)ch2;
		str[2] = '\0';
	} else {
		str[0] = (char)(ch >> 8);
		str[1] = (char)ch;
		str[2] = '\0';
	}
	GdkColor white, black;
	white.red = white.green = white.blue = 0xFFFF;
	white.pixel = 1;
	black.red = black.green = black.blue = 0x0000;
	black.pixel = 0;
	gdk_gc_set_foreground(
		font_gc,
		&white);
	gdk_draw_rectangle(
		font_pixmap,
		font_gc,
		TRUE,
		0, 0, bmp_dim.width, bmp_dim.height);
	gdk_gc_set_foreground(
		font_gc,
		&black);
	gdk_draw_string(
		font_pixmap,
		message_font,
		font_gc,
		0, font_info.baseline,
		chaos_utility::sjis_to_gui(str).c_str());
	GdkImage* font_image = gdk_image_get(
		font_pixmap,
		0, 0, bmp_dim.width, bmp_dim.height);
	if (font_image == NULL) {
		return false;
	}
	if (!half) {
		chaos_rect draw_rect;
		chaos_pos font_ofs;
		int red, green, blue;
		int dst_scanline, dst_step;
		chaos_uint32* dst;
		int x, y;
		if (shape == 3) {
			draw_rect = shade_rect & image_rect;
			font_ofs = draw_rect.get_lt_pos()-shade_rect.get_lt_pos();
			red = green = blue = 0;
		} else {
			draw_rect = face_rect & image_rect;
			font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
			red = CHAOS_RGBA_RED(color);
			green = CHAOS_RGBA_GREEN(color);
			blue = CHAOS_RGBA_BLUE(color);
		}
		if (!draw_rect.empty()) {
			dst_scanline = message_image->get_scanline_length();
			dst_step = dst_scanline-draw_rect.width;
			dst = message_image->get_bits(draw_rect.get_lt_pos());
			int font_left = font_ofs.x,
				font_top = font_ofs.y,
				font_right = font_left+draw_rect.width,
				font_bottom = font_top+draw_rect.height;
			if (font_anti_alias) {
				y = font_top;
				do {
					x = font_left;
					do {
						int x2 = x+x, y2 = y+y;
						int alpha =
							(gdk_image_get_pixel(
								font_image, x2, y2) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+1, y2) & 1)+
							(gdk_image_get_pixel(
								font_image, x2, y2+1) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+1, y2+1) & 1);
						alpha = alpha*64-alpha/4;
						*(dst++) = CHAOS_RGBA(
							red, green, blue, alpha);
					} while (++x != font_right);
					dst += dst_step;
				} while (++y != font_bottom);
			} else {
				y = font_top;
				do {
					x = font_left;
					do {
						int alpha =
							(gdk_image_get_pixel(
								font_image, x, y) & 1);
						alpha = (-alpha) & 0xFF;
						*(dst++) = CHAOS_RGBA(
							red, green, blue, alpha);
					} while (++x != font_right);
					dst += dst_step;
				} while (++y != font_bottom);
			}
		}
		if (shape == 3) {
			draw_rect = face_rect & image_rect;
			font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
			red = CHAOS_RGBA_RED(color);
			green = CHAOS_RGBA_GREEN(color);
			blue = CHAOS_RGBA_BLUE(color);
			if (!draw_rect.empty()) {
				dst_scanline = message_image->get_scanline_length();
				dst_step = dst_scanline-draw_rect.width;
				dst = message_image->get_bits(draw_rect.get_lt_pos());
				int font_left = font_ofs.x,
					font_top = font_ofs.y,
					font_right = font_left+draw_rect.width,
					font_bottom = font_top+draw_rect.height;
				if (font_anti_alias) {
					y = font_top;
					do {
						x = font_left;
						do {
							int x2 = x+x, y2 = y+y;
							int alpha =
								(gdk_image_get_pixel(
									font_image, x2, y2) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+1, y2) & 1)+
								(gdk_image_get_pixel(
									font_image, x2, y2+1) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+1, y2+1) & 1);
							alpha = 256-alpha*64;
							int alpha2 = CHAOS_RGBA_ALPHA(*dst);
							alpha2 = 256-(alpha2+(alpha2+1)/256);
							alpha2 = (alpha2*(256-alpha))/256;
							int alpha3 = alpha+alpha2;
							int dencity = 256;
							if (alpha3 > 0) {
								dencity = (alpha*256)/alpha3;
							}
							alpha3 = 256-alpha3;
							alpha3 -= alpha3/256;
							*(dst++) = CHAOS_RGBA(
								(red*dencity)/256,
								(green*dencity)/256,
								(blue*dencity)/256,
								alpha3);
						} while (++x != font_right);
						dst += dst_step;
					} while (++y != font_bottom);
				} else {
					y = font_top;
					do {
						x = font_left;
						do {
							int alpha =
								(gdk_image_get_pixel(
									font_image, x, y) & 1);
							alpha = (-alpha) & 0xFF;
							int alpha2 = CHAOS_RGBA_ALPHA(*dst);
							int alpha3 = alpha & alpha2;
							int dencity = alpha ^ 0xFF;
							*(dst++) = CHAOS_RGBA(
								red & dencity,
								green & dencity,
								blue & dencity,
								alpha3);
						} while (++x != font_right);
						dst += dst_step;
					} while (++y != font_bottom);
				}
			}
		}
	} else {
		chaos_rect draw_rect;
		chaos_pos font_ofs;
		int red, green, blue;
		int dst_scanline, dst_step;
		chaos_uint32* dst;
		int x, y;
		if (shape == 3) {
			draw_rect = shade_rect & image_rect;
			font_ofs = draw_rect.get_lt_pos()-shade_rect.get_lt_pos();
			red = green = blue = 0;
		} else {
			draw_rect = face_rect & image_rect;
			font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
			red = CHAOS_RGBA_RED(color);
			green = CHAOS_RGBA_GREEN(color);
			blue = CHAOS_RGBA_BLUE(color);
		}
		if (!draw_rect.empty()) {
			dst_scanline = message_image->get_scanline_length();
			dst_step = dst_scanline-draw_rect.width/2;
			dst = message_image->get_bits(draw_rect.get_lt_pos());
			int font_left = font_ofs.x,
				font_top = font_ofs.y,
				font_right = font_left+draw_rect.width/2,
				font_bottom = font_top+draw_rect.height;
			if (font_anti_alias) {
				y = font_top;
				do {
					x = font_left;
					do {
						int x2 = x*4, y2 = y+y;
						int alpha = (
							(gdk_image_get_pixel(
								font_image, x2, y2) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+1, y2) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+2, y2) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+3, y2) & 1)+
							(gdk_image_get_pixel(
								font_image, x2, y2+1) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+1, y2+1) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+2, y2+1) & 1)+
							(gdk_image_get_pixel(
								font_image, x2+3, y2+1) & 1))/2;
						alpha = alpha*64-alpha/4;
						*(dst++) = CHAOS_RGBA(
							red, green, blue, alpha);
					} while (++x != font_right);
					dst += dst_step;
				} while (++y != font_bottom);
			} else {
				y = font_top;
				do {
					x = font_left;
					do {
						int x2 = x+x;
						int alpha =
							(gdk_image_get_pixel(
								font_image, x2, y) & 1) &
							(gdk_image_get_pixel(
								font_image, x2+1, y) & 1);
						alpha = (-alpha) & 0xFF;
						*(dst++) = CHAOS_RGBA(
							red, green, blue, alpha);
					} while (++x != font_right);
					dst += dst_step;
				} while (++y != font_bottom);
			}
		}
		if (shape == 3) {
			draw_rect = face_rect & image_rect;
			font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
			red = CHAOS_RGBA_RED(color);
			green = CHAOS_RGBA_GREEN(color);
			blue = CHAOS_RGBA_BLUE(color);
			if (!draw_rect.empty()) {
				dst_scanline = message_image->get_scanline_length();
				dst_step = dst_scanline-draw_rect.width/2;
				dst = message_image->get_bits(draw_rect.get_lt_pos());
				int font_left = font_ofs.x,
					font_top = font_ofs.y,
					font_right = font_left+draw_rect.width/2,
					font_bottom = font_top+draw_rect.height;
				if (font_anti_alias) {
					y = font_top;
					do {
						x = font_left;
						do {
							int x2 = x*4, y2 = y+y;
							int alpha = (
								(gdk_image_get_pixel(
									font_image, x2, y2) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+1, y2) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+2, y2) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+3, y2) & 1)+
								(gdk_image_get_pixel(
									font_image, x2, y2+1) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+1, y2+1) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+2, y2+1) & 1)+
								(gdk_image_get_pixel(
									font_image, x2+3, y2+1) & 1))/2;
							alpha = 256-alpha*64;
							int alpha2 = CHAOS_RGBA_ALPHA(*dst);
							alpha2 = 256-(alpha2+(alpha2+1)/256);
							alpha2 = (alpha2*(256-alpha))/256;
							int alpha3 = alpha+alpha2;
							int dencity = 256;
							if (alpha3 > 0) {
								dencity = (alpha*256)/alpha3;
							}
							alpha3 = 256-alpha3;
							alpha3 -= alpha3/256;
							*(dst++) = CHAOS_RGBA(
								(red*dencity)/256,
								(green*dencity)/256,
								(blue*dencity)/256,
								alpha3);
						} while (++x != font_right);
						dst += dst_step;
					} while (++y != font_bottom);
				} else {
					y = font_top;
					do {
						x = font_left;
						do {
							int x2 = x+x;
							int alpha =
								(gdk_image_get_pixel(
									font_image, x2, y) & 1) &
								(gdk_image_get_pixel(
									font_image, x2+1, y) & 1);
							alpha = (-alpha) & 0xFF;
							int alpha2 = CHAOS_RGBA_ALPHA(*dst);
							int alpha3 = alpha & alpha2;
							int dencity = alpha ^ 0xFF;
							*(dst++) = CHAOS_RGBA(
								red & dencity,
								green & dencity,
								blue & dencity,
								alpha3);
						} while (++x != font_right);
						dst += dst_step;
					} while (++y != font_bottom);
				}
			}
		}
	}
	gdk_image_destroy(font_image);

#elif defined(CHAOS_FONT_PANGO)

	memset(surface_pixels, 0, surface_stride*surface_dim.height);
	cairo_set_source_rgb(
		cairo_context,
		1, 1, 1);
	chaos_string text;
	bool half = (ch <= 0xFF);
	if (half) {
		ch = chaos_utility::jis_to_sjis(
			chaos_utility::jis_han_to_zen(
				chaos_utility::sjis_to_jis(ch)));
	}
	if (ch > 0xFF) {
		text = (char)((ch >> 8) & 0xFF);
	}
	text += (char)(ch & 0xFF);
	text = chaos_utility::sjis_to_gui(text);
	pango_layout_set_text(
		pango_layout,
		text.c_str(), (int)text.size());
	cairo_set_antialias(
		cairo_context,
		font_anti_alias? CAIRO_ANTIALIAS_GRAY: CAIRO_ANTIALIAS_NONE);
	pango_cairo_layout_path(
		cairo_context,
		pango_layout);
	cairo_fill(
		cairo_context);
	if (half &&
		(surface_dim.width > 0) &&
		(surface_dim.height > 0))
	{
		chaos_uint8* pixels = surface_pixels;
		int y = surface_dim.height;
		do {
			chaos_uint8* pixels_l = pixels;
			chaos_uint8* pixels_r = pixels;
			int x;
			for (x = 1; x < surface_dim.width; x += 2) {
				chaos_uint8 pixel =
					(chaos_uint8)((*pixels_r+*(pixels_r+1))/2);
				*(pixels_r++) = 0;
				*(pixels_r++) = 0;
				*(pixels_l++) = pixel;
			}
			if (x == surface_dim.width) {
				chaos_uint8 pixel = (chaos_uint8)(*pixels_r/2);
				*pixels_r = 0;
				*pixels_l = pixel;
			}
			pixels += surface_stride;
		} while (--y > 0);
	}
	chaos_rect draw_rect;
	chaos_pos font_ofs;
	int red, green, blue;
	int dst_scanline, dst_step;
	chaos_uint32* dst;
	int x, y;
	if (shape == 3) {
		draw_rect = shade_rect & image_rect;
		font_ofs = draw_rect.get_lt_pos()-shade_rect.get_lt_pos();
		red = green = blue = 0;
	} else {
		draw_rect = face_rect & image_rect;
		font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
		red = CHAOS_RGBA_RED(color);
		green = CHAOS_RGBA_GREEN(color);
		blue = CHAOS_RGBA_BLUE(color);
	}
	if (!draw_rect.empty()) {
		dst_scanline = message_image->get_scanline_length();
		dst_step = dst_scanline-draw_rect.width;
		dst = message_image->get_bits(draw_rect.get_lt_pos());
		int font_left = font_ofs.x,
			font_top = font_ofs.y,
			font_right = font_left+draw_rect.width,
			font_bottom = font_top+draw_rect.height;
		y = font_top;
		do {
			x = font_left;
			do {
				int alpha = 255-*(surface_pixels+x+surface_stride*y);
				*(dst++) = CHAOS_RGBA(
					red, green, blue, alpha);
			} while (++x != font_right);
			dst += dst_step;
		} while (++y != font_bottom);
	}
	if (shape == 3) {
		draw_rect = face_rect & image_rect;
		font_ofs = draw_rect.get_lt_pos()-face_rect.get_lt_pos();
		red = CHAOS_RGBA_RED(color);
		green = CHAOS_RGBA_GREEN(color);
		blue = CHAOS_RGBA_BLUE(color);
		if (!draw_rect.empty()) {
			dst_scanline = message_image->get_scanline_length();
			dst_step = dst_scanline-draw_rect.width;
			dst = message_image->get_bits(draw_rect.get_lt_pos());
			int font_left = font_ofs.x,
				font_top = font_ofs.y,
				font_right = font_left+draw_rect.width,
				font_bottom = font_top+draw_rect.height;
			y = font_top;
			do {
				x = font_left;
				do {
					int alpha = *(surface_pixels+x+surface_stride*y);
					int alpha2 = CHAOS_RGBA_ALPHA(*dst);
					alpha2 = 256-(alpha2+(alpha2+1)/256);
					alpha2 = (alpha2*(256-alpha))/256;
					int alpha3 = alpha+alpha2;
					int dencity = 256;
					if (alpha3 > 0) {
						dencity = (alpha*256)/alpha3;
					}
					alpha3 = 256-alpha3;
					alpha3 -= alpha3/256;
					*(dst++) = CHAOS_RGBA(
						(red*dencity)/256,
						(green*dencity)/256,
						(blue*dencity)/256,
						alpha3);
				} while (++x != font_right);
				dst += dst_step;
			} while (++y != font_bottom);
		}
	}

#endif // CHAOS_FONT

	return true;
}

// update message text

bool chaos_message_imager::update_message_text(
	const chaos_string& new_message_text,
	chaos_rect* message_update_rect)
{
	if (message_update_rect != NULL) {
		message_update_rect->clear();
	}
	if (new_message_text == message_text) {
		return true;
	}
	size_t skip_length = 0;
	if ((message_text.length() > 0) &&
		(new_message_text.length() > message_text.length()))
	{
		if (strncmp(
				new_message_text.c_str(),
				message_text.c_str(),
				message_text.length()) == 0)
		{
			skip_length = message_text.length();
		}
	}
	if ((skip_length == 0) && !message_valid_rect.empty()) {
		if (message_update_rect != NULL) {
			*message_update_rect = message_valid_rect;
		}
		clear_message_text();
	}
	message_text = new_message_text;
	if (message_text.empty()) {
		return true;
	}
	bool result = true;
	try {
		const char* text = message_text.c_str();
		size_t ofs = skip_length;
		int ch;
		while ((ch = (unsigned char)text[ofs]) != '\0') {
			switch (ch) {
			case '\n':
				ofs++;
				message_pos.x = message_top_pos.x;
				message_pos.y += cur_font_size;
				message_column = 0;
				ch = '\0';
				break;
			case '\\':
				ofs++;
				ch = (unsigned char)text[ofs];
				switch (ch) {
				case 'p':	// position
					ofs++;
					if (text[ofs] != '(') {
						// nothing
					} else {
						int next = 0;
						sscanf(
							text+ofs,
							"(%d,%d)%n",
							&message_top_pos.x, &message_top_pos.y,
							&next);
						ofs += next;
						message_pos = message_top_pos;
						message_column = 0;
					}
					ch = '\0';
					break;
				case 'l':	// length
					ofs++;
					if (text[ofs] != '(') {
						// nothing
					} else {
						int next = 0;
						sscanf(
							text+ofs,
							"(%d)%n",
							&message_length,
							&next);
						ofs += next;
					}
					ch = '\0';
					break;
				case 'c':	// current color
				case 'd':	// default color
				case 's':	// sub color
					ofs++;
					if (text[ofs] != '(') {
						// nothing
					} else if (text[ofs+1] == ')') {
						switch (ch) {
						case 'c':	// current color
							message_color =
								message_defcolor-message_subcolor;
							break;
						}
						ofs += 2;
					} else {
						int next = 0,
							red = 0, green = 0, blue = 0;
						sscanf(
							text+ofs,
							"(%d,%d,%d)%n",
							&red, &green, &blue,
							&next);
						chaos_color color = chaos_color(red, green, blue);
						ofs += next;
						switch (ch) {
						case 'c':	// current color
							message_color = color;
							break;
						case 'd':	// default color
							message_defcolor = color;
							message_color =
								message_defcolor-message_subcolor;
							break;
						case 's':	// sub color
							message_subcolor = color;
							message_color =
								message_defcolor-message_subcolor;
							break;
						}
					}
					ch = '\0';
					break;
				case 'h':	// shape
					ofs++;
					if (text[ofs] != '(') {
						// nothing
					} else {
						int next = 0;
						sscanf(
							text+ofs,
							"(%d)%n",
							&message_shape,
							&next);
						ofs += next;
					}
					ch = '\0';
					break;
				case 'f':	// font size
					ofs++;
					if (text[ofs] != '(') {
						// nothing
					} else {
						int next = 0;
						sscanf(
							text+ofs,
							"(%d)%n",
							&cur_font_size,
							&next);
						ofs += next;
					}
					ch = '\0';
					break;
				}
				break;
			}
			if (ch != '\0') {
				ofs++;
				int ch_length = 1;
				if (chaos_utility::is_sjislead(ch) &&
					(text[ofs] != '\0'))
				{
					ch = (ch << 8) | ((unsigned char)text[ofs]);
					ofs++;
					ch_length++;
				}
				if ((message_length > 0) &&
					(message_column+ch_length > message_length*2))
				{
					if ((message_column > message_length*2) ||
						(linetop_prohibit.find(ch) ==
							linetop_prohibit.end()))
					{
						message_pos.x = message_top_pos.x;
						message_pos.y += cur_font_size;
						message_column = 0;
					}
				}
				if (ofs > skip_length) {
					chaos_rect font_rect;
					if (draw_font_image(
							ch,
							message_pos,
							cur_font_size,
							message_shape,
							(chaos_uint32)message_color,
							&font_rect))
					{
						if (message_update_rect != NULL) {
							*message_update_rect |= font_rect;
						}
					}
				}
				message_column += ch_length;
				message_pos.x += (ch_length*cur_font_size)/2;
			}
		}
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : message imager::draw message text - "
			"%s\n",
			e.what());
		result = false;
	}
	return result;
}

// clear message text

void chaos_message_imager::clear_message_text() {
	message_text = "";
	message_top_pos = message_pos = chaos_pos(0, 0);
	message_length = 0;
	message_column = 0;
	message_color = message_defcolor = chaos_color(255, 255, 255);
	message_subcolor = chaos_color(0, 0, 0);
	message_shape = 3;
	cur_font_size = org_font_size;
	message_image->fill(
		CHAOS_RGBA(0, 0, 0, 255),
		&message_valid_rect);
	message_valid_rect.clear();
}

#ifdef CHAOS_FONT_GTK

// get xlfd element

bool chaos_message_imager::get_xlfd_element(
	const char*& xlfdstr,
	chaos_string& element)
{
	if (*xlfdstr != '-') {
		return false;
	}
	xlfdstr++;
	const char* first = xlfdstr;
	bool sjislead = false;
	char ch;
	while ((ch = *xlfdstr) != '\0') {
		if (!sjislead) {
			if (ch == '-') {
				break;
			}
			sjislead = chaos_utility::is_sjislead(ch);
		} else {
			sjislead = false;
		}
		xlfdstr++;
	}
	element = chaos_string(first, xlfdstr);
	return true;
}

// normalize font xlfd

chaos_string chaos_message_imager::normalize_font_xlfd(
	const chaos_string& org_xlfd,
	int height)
{
	chaos_string foundary, family, weight, slant, set_width,
		style,
		spacing,
		registry, encoding,
		heights,
		dummy;
	const char* xlfdstr = org_xlfd.c_str();
	if (!get_xlfd_element(xlfdstr, foundary) ||
		!get_xlfd_element(xlfdstr, family) ||
		!get_xlfd_element(xlfdstr, weight) ||
		!get_xlfd_element(xlfdstr, slant) ||
		!get_xlfd_element(xlfdstr, set_width) ||
		!get_xlfd_element(xlfdstr, style) ||
		!get_xlfd_element(xlfdstr, dummy) || // heights
		!get_xlfd_element(xlfdstr, dummy) || // point
		!get_xlfd_element(xlfdstr, dummy) || // x_dpi
		!get_xlfd_element(xlfdstr, dummy) || // y_dpi
		!get_xlfd_element(xlfdstr, spacing) ||
		!get_xlfd_element(xlfdstr, dummy) || // width
		!get_xlfd_element(xlfdstr, registry) ||
		!get_xlfd_element(xlfdstr, encoding))
	{
		foundary = family = weight = slant = set_width =
			spacing =
			registry = encoding = '*';
		style = "";
	}
	if (height <= 0) {
		heights = "*";
	} else {
		heights = chaos_utility::format_str("%d", height);
	}
	return chaos_utility::format_str(
		"-%s-%s-%s-%s-%s"
		"-%s"
		"-%s-*-*-*-%s-*-%s-%s",
		foundary.c_str(), family.c_str(), weight.c_str(),
			slant.c_str(), set_width.c_str(),
		style.c_str(),
		heights.c_str(),
		spacing.c_str(),
		registry.c_str(), encoding.c_str());
}

#elif defined(CHAOS_FONT_PANGO)

// normalize font pango

chaos_string chaos_message_imager::normalize_font_pango(
	const chaos_string& org_font_pango,
	int height)
{
	chaos_string new_font_pango;
	PangoFontDescription* pango_font_desc =
		pango_font_description_from_string(
			chaos_utility::sjis_to_gui(org_font_pango).c_str());
	if (pango_font_desc != NULL) {
		if (height <= 0) {
			pango_font_description_unset_fields(
				pango_font_desc,
				PANGO_FONT_MASK_SIZE);
		} else {
			pango_font_description_set_absolute_size(
				pango_font_desc,
				(double)height*PANGO_SCALE);
		}
		char* font_str = pango_font_description_to_string(
			pango_font_desc);
		if (font_str != NULL) {
			new_font_pango = chaos_utility::gui_to_sjis(font_str);
		}
		g_free(font_str);
		pango_font_description_free(
			pango_font_desc);
	}
	return new_font_pango;
}

#endif // CHAOS_FONT
