////////////////////////////////////////////////////////////////////////////////
// chaos_message_imager.h : message imager

#ifndef CHAOS_MESSAGE_IMAGER_H_INCLUDED
#define CHAOS_MESSAGE_IMAGER_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_message_imager;

////////////////////////////////////////////////////////////////////////////////
// include

#ifndef CHAOS_POS_H_INCLUDED
#include "chaos_pos.h"
#endif // CHAOS_POS_H_INCLUDED

#ifndef CHAOS_DIM_H_INCLUDED
#include "chaos_dim.h"
#endif // CHAOS_DIM_H_INCLUDED

#ifndef CHAOS_RECT_H_INCLUDED
#include "chaos_rect.h"
#endif // CHAOS_RECT_H_INCLUDED

#ifndef CHAOS_COLOR_H_INCLUDED
#include "chaos_color.h"
#endif // CHAOS_COLOR_H_INCLUDED

#ifndef CHAOS_STRING_H_INCLUDED
#include "chaos_string.h"
#endif // CHAOS_STRING_H_INCLUDED

#ifndef CHAOS_IMAGE_H_INCLUDED
#include "chaos_image.h"
#endif // CHAOS_IMAGE_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// chaos_message_imager : message imager

class chaos_message_imager {
// enum
	// size
	enum {
		shade_size_division = 8
	};

// typedef
private:

#ifdef CHAOS_FONT_GTK

	// font info type
	struct font_into_type {
		int lbearing;
		int rbearing;
		int width;
		int height;
		int ascent;
		int descent;
		int baseline;
	};

#endif // CHAOS_FONT

// attribute
private:
	// message text
	chaos_string message_text;
	// message position
	chaos_pos message_pos;
	// message top position
	chaos_pos message_top_pos;
	// message length
	int message_length;
	// message column
	int message_column;
	// message color
	chaos_color message_color;
	// message default color
	chaos_color message_defcolor;
	// message subcolor
	chaos_color message_subcolor;
	// message shape
	int message_shape;
	// original font size
	int org_font_size;
	// current font size
	int cur_font_size;
	// real font size
	int real_font_size;
	// font name
	chaos_string font_name;
	// font bold
	bool font_bold;
	// font anti alias
	bool font_anti_alias;
	// anti alias by manual
	bool manual_anti_alias;
	// message image
	chaos_image* message_image;
	// message valid rect
	chaos_rect message_valid_rect;
	// line-top prohibited char
	std::set<int> linetop_prohibit;
	// bmp dim
	chaos_dim bmp_dim;

#ifdef CHAOS_FONT_WIN32

	// device context
	HDC dc;
	// message font
	HFONT message_font;
	// original font
	HFONT org_font;
	// font bitmap
	HBITMAP font_bmp;
	// original bitmap
	HBITMAP org_bmp;
	// message bitmap bits
	void* font_bmp_bits;

#elif defined(CHAOS_FONT_GTK)

	// message font
	GdkFont* message_font;
	// font window(hidden)
	GtkWidget* font_window;
	// font pixmap
	GdkPixmap* font_pixmap;
	// font graphics context
	GdkGC* font_gc;
	// font info
	font_into_type font_info;

#elif defined(CHAOS_FONT_PANGO)

	// surface dim
	chaos_dim surface_dim;
	// surface pixels
	chaos_uint8* surface_pixels;
	// surface stride
	int surface_stride;
	// cairo surface
	cairo_surface_t* cairo_surface;
	// cairo context
	cairo_t* cairo_context;
	// pango layout
	PangoLayout* pango_layout;

#endif // CHAOS_FONT

	// bit count table
	static int bits_table[256];

// create & destroy
public:
	// default constructor
	chaos_message_imager();
	// destructor
	~chaos_message_imager();
	// initialize
	bool initialize(chaos_dim image_dim);
	// dispose
	void dispose();
	// clear
	void clear();

// get & set
public:
	// get message position
	chaos_pos get_message_pos() const {
		return message_pos;
	}
	// get message image
	const chaos_image* get_message_image() const {
		return message_image;
	}
	chaos_image* get_message_image() {
		return message_image;
	}
	// get message valid rect
	chaos_rect get_message_valid_rect() const {
		return message_valid_rect;
	}
	// get font size
	int get_font_size() const {
		return org_font_size;
	}
	// set font size
	void set_font_size(int new_font_size) {
		org_font_size = new_font_size;
	}
	// get font name
	const chaos_string& get_font_name() const {
		return font_name;
	}
	// set font name
	void set_font_name(const chaos_string& new_font_name) {
		font_name = new_font_name;
		real_font_size = 0;
	}
	// is font bold
	bool is_font_bold() const {
		return font_bold;
	}
	// set font bold
	void set_font_bold(bool new_font_bold) {
		font_bold = new_font_bold;
		real_font_size = 0;
	}
	// is font anti alias
	bool is_font_anti_alias() const {
		return font_anti_alias;
	}
	// set font anti alias
	void set_font_anti_alias(bool new_alias) {
		font_anti_alias = new_alias;
		real_font_size = 0;
	}

// info
public:
	// encode position info
	static chaos_string encode_position_info(
		int x, int y);
	// encode length info
	static chaos_string encode_length_info(
		int length);
	// encode current color info
	static chaos_string encode_curcolor_info(
		chaos_color curcolor);
	// encode restore current color info
	static chaos_string encode_rescolor_info();
	// encode default color info
	static chaos_string encode_defcolor_info(
		chaos_color defcolor);
	// encode sub color info
	static chaos_string encode_subcolor_info(
		chaos_color subcolor);
	// encode shape info
	static chaos_string encode_shape_info(
		int shape);
	// encode font size info
	static chaos_string encode_font_size_info(
		int font_size);

// operation
private:
	// create font
	bool create_font(int new_font_size);
	// destroy font
	void destroy_font();
	// draw font image
	bool draw_font_image(
		int ch,
		chaos_pos font_pos,
		int font_size,
		int shape,
		chaos_uint32 color,
		chaos_rect* font_rect);
public:
	// update message text
	bool update_message_text(
		const chaos_string& new_message_text,
		chaos_rect* message_update_rect);
	// clear message text
	void clear_message_text();

#ifdef CHAOS_FONT_GTK

	// get xlfd element
	static bool get_xlfd_element(
		const char*& xlfdstr,
		chaos_string& element);
	// normalize font xlfd
	static chaos_string normalize_font_xlfd(
		const chaos_string& org_xlfd,
		int height);

#elif defined(CHAOS_FONT_PANGO)

	// normalize font pango
	static chaos_string normalize_font_pango(
		const chaos_string& org_font_pango,
		int height);

#endif // CHAOS_FONT

};

#endif // CHAOS_MESSAGE_IMAGER_H_INCLUDED
