////////////////////////////////////////////////////////////////////////////////
// chaos_png_image_loader.cpp : png image loader

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_png_image_loader.h"
#include "chaos_exception.h"

////////////////////////////////////////////////////////////////////////////////
// png callback

// read callback

static void png_user_read_fn(
	png_structp png_ptr,
	png_bytep data,
	png_size_t length)
{
	chaos_png_image_loader* loader =
		((chaos_png_image_loader*)png_get_io_ptr(png_ptr));
	if (!loader->read_file(data, length)) {

#ifdef CHAOS_PLATFORM_WIN32

		throw chaos_exception(chaos_exception::bad_format);

#elif defined(CHAOS_PLATFORM_UNIX)

		longjmp(png_jmpbuf(png_ptr), 1);

#endif // CHAOS_PLATFORM

	}
}

// error callback

static void png_user_error_fn(
	png_structp png_ptr,
	png_const_charp /*error_msg*/)
{

#ifdef CHAOS_PLATFORM_WIN32

	png_ptr;
	throw chaos_exception(chaos_exception::bad_format);

#elif defined(CHAOS_PLATFORM_UNIX)

	longjmp(png_jmpbuf(png_ptr), 1);

#endif // CHAOS_PLATFORM

}

// warning callback

static void png_user_warning_fn(
	png_structp /*png_ptr*/,
	png_const_charp warning_msg)
{
	LOG_WRN(
		"warning : png warning (%s)\n",
		warning_msg);
}

////////////////////////////////////////////////////////////////////////////////
// chaos_png_image_loader : png image loader class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// constructor

chaos_png_image_loader::chaos_png_image_loader(
		chaos_file_data* org_file_data,
		chaos_image* org_image) :
	chaos_image_loader(
		org_file_data,
		org_image)
{
}

// destructor

chaos_png_image_loader::~chaos_png_image_loader() {
}

////////////////////////////////////////////////////////////////////////////////
// implementation

// load image

bool chaos_png_image_loader::load() {
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type;
	int number_passes, pass;
	png_uint_32 y;
	bool result = false;
	try {
		// create png struct
		png_ptr = png_create_read_struct(
			PNG_LIBPNG_VER_STRING,
			NULL,
			png_user_error_fn,
			png_user_warning_fn);
		if (png_ptr == NULL) {
			throw chaos_exception(chaos_exception::unknown);
		}
		// create info struct
		info_ptr = png_create_info_struct(png_ptr);
		if (info_ptr == NULL) {
			throw chaos_exception(chaos_exception::unknown);
		}

#ifdef CHAOS_PLATFORM_UNIX

		if (setjmp(png_jmpbuf(png_ptr)) != 0) {
			throw chaos_exception(chaos_exception::bad_format);
		}

#endif // CHAOS_PLATFORM

		// change read function
		png_set_read_fn(
			png_ptr,
			(png_voidp)this,
			png_user_read_fn);
		// read info struct
		png_read_info(png_ptr, info_ptr);
		// read header
		png_get_IHDR(
			png_ptr, info_ptr,
			&width, &height,
			&bit_depth, &color_type, &interlace_type,
			NULL, NULL);
		// 16bit/color to 8bit/color
		if (bit_depth > 8) {
			png_set_strip_16(png_ptr);
		}
		// 1, 2, 4bit to 8bit
		if (bit_depth < 8) {
			png_set_expand(png_ptr);
		}
		// palette to rgb
		if ((color_type == PNG_COLOR_TYPE_PALETTE) &&
			(bit_depth <= 8))
		{
			png_set_palette_to_rgb(png_ptr);
		}
		// gray to rgb
		if ((color_type == PNG_COLOR_TYPE_GRAY) ||
			(color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
		{
			png_set_gray_to_rgb(png_ptr);
		}
		// trans to alpha
		if (png_get_valid(png_ptr, info_ptr,PNG_INFO_tRNS)) {
			png_set_tRNS_to_alpha(png_ptr);
		}
		// add filler
		if ((color_type & PNG_COLOR_MASK_ALPHA) == 0) {
			png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);
		}
		// rgb to bgr
		png_set_bgr(png_ptr);
		// create image
		if (!image->create(width, height)) {
			throw chaos_exception(chaos_exception::few_memory);
		}
		// interlace handling
		number_passes = png_set_interlace_handling(png_ptr);
		// read image
		for (pass = 0; pass < number_passes; pass++) {
			for (y = 0; y < height; y++) {
				png_bytep row_buf =
					(png_bytep)image->get_bits(0, y);
				png_read_rows(png_ptr, &row_buf, NULL, 1);
			}
		}
		// destroy png struct
		png_destroy_read_struct(
			&png_ptr,
			&info_ptr,
			(png_infopp)NULL);
		png_ptr = NULL;
		result = true;
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : png image loader::load - "
			"%s\n",
			e.what());
	}
	if (png_ptr != NULL) {
		png_destroy_read_struct(
			&png_ptr,
			(info_ptr != NULL)? &info_ptr: NULL,
			NULL);
	}
	return result;
}
