////////////////////////////////////////////////////////////////////////////////
// chaos_readflag_manager.cpp : readflag manager

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_readflag_manager.h"
#include "chaos_utility.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_readflag_manager : readflag manager class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_readflag_manager::chaos_readflag_manager() :
	next_key(0)
{
}

// destructor

chaos_readflag_manager::~chaos_readflag_manager() {
	clear();
}

// clear

void chaos_readflag_manager::clear() {
	keys.clear();
	for (
		readflag_map::iterator it = readflags.begin();
		it != readflags.end();
		it++)
	{
		delete (*it).second;
	}
	readflags.clear();
}

////////////////////////////////////////////////////////////////////////////////
// operation

// get key

int chaos_readflag_manager::get_key(const chaos_string& filename) const {
	key_map::const_iterator it = keys.find(filename);
	return (it != keys.end())? (*it).second: -1;
}

// regist

int chaos_readflag_manager::regist(const chaos_string& filename) {
	int key;
	key_map::iterator it = keys.find(filename);
	if (it != keys.end()) {
		key = (*it).second;
	} else {
		key = next_key++;
		ASSERT(next_key >= 0);
		keys.insert(key_map::value_type(filename, key));
	}
	return key;
}

// reserve

void chaos_readflag_manager::reserve(
	int key,
	size_t capacity)
{
	ASSERT(key >= 0);
	if ((key < 0) || (capacity > (int)id_limit+1)) {
		return;
	}
	chaos_buffer* buf;
	readflag_map::iterator it = readflags.find(key);
	if (it != readflags.end()) {
		buf = (*it).second;
	} else {
		buf = new chaos_buffer;
		readflags.insert(readflag_map::value_type(key, buf));
	}
	if (capacity > buf->size()*8) {
		buf->resize((capacity+7)/8, 0);
	}
}

// reserve

void chaos_readflag_manager::reserve(
	const chaos_string& filename,
	size_t capacity)
{
	reserve(regist(filename), capacity);
}

// is read

bool chaos_readflag_manager::is_read(
	int key,
	int id) const
{
	if ((key < 0) || (id < 0) || (id > (int)id_limit)) {
		return false;
	}
	bool read = false;
	readflag_map::const_iterator it = readflags.find(key);
	if (it != readflags.end()) {
		const chaos_buffer* buf = (*it).second;
		if ((int)buf->size() > id/8) {
			read = (((*buf)[id/8] & (1 << (id%8))) != 0);
		}
	}
	return read;
}

bool chaos_readflag_manager::is_read(
	const chaos_string& filename,
	int id) const
{
	return is_read(get_key(filename), id);
}

// set read

bool chaos_readflag_manager::set_read(
	int key,
	int id,
	bool read)
{
	ASSERT(key >= 0);
	if ((key < 0) || (id < 0) || (id > (int)id_limit)) {
		return false;
	}
	chaos_buffer* buf;
	readflag_map::iterator it = readflags.find(key);
	if (it != readflags.end()) {
		buf = (*it).second;
	} else {
		buf = new chaos_buffer;
		readflags.insert(readflag_map::value_type(key, buf));
	}
	if (id >= (int)buf->size()*8) {
		buf->resize(id/8+1, 0);
	}
	bool old_read = (((*buf)[id/8] & (1 << (id%8))) != 0);
	if (read) {
		(*buf)[id/8] |= (chaos_uint8)(1 << (id%8));
	} else {
		(*buf)[id/8] &= (chaos_uint8)(~(1 << (id%8)));
	}
	return old_read;
}

// set read

bool chaos_readflag_manager::set_read(
	const chaos_string& filename,
	int id,
	bool read)
{
	return set_read(get_key(filename), id, read);
}

////////////////////////////////////////////////////////////////////////////////
// xml

// load

bool chaos_readflag_manager::load(const chaos_xml* readf_xml) {
	if (readf_xml->get_kind() != "readflag") {
		return false;
	}
	if (readf_xml->get_values_version() != "1.0") {
		// nothing
	}
	const chaos_value_manager<chaos_string>* readf_values =
		readf_xml->get_values();
	clear();
	chaos_value_manager<chaos_string>::const_iterator it;
	for (it = readf_values->begin(); it != readf_values->end(); it++) {
		const chaos_string& filename = (*it).first;
		chaos_string flag = (*it).second.get_string();
		chaos_buffer* buf = new chaos_buffer;
		const char* flagstr = flag.c_str();
		char ch;
		int max_id = 0;
		while ((ch = *flagstr) != '\0') {
			if (chaos_utility::is_digit(ch)) {
				int id = 0,
					next = 0;
				sscanf(flagstr, "%d%n", &id, &next);
				flagstr += next;
				if (id >= max_id) {
					max_id = id;
				}
			} else {
				flagstr++;
			}
		}
		if (max_id > (int)id_limit) {
			max_id = (int)id_limit;
		}
		if (max_id >= (int)buf->size()*8) {
			buf->resize(max_id/8+1, 0);
		}
		flagstr = flag.c_str();
		while ((ch = *flagstr) != '\0') {
			if (chaos_utility::is_digit(ch)) {
				int id1 = 0,
					next = 0;
				sscanf(flagstr, "%d%n", &id1, &next);
				flagstr += next;
				if (id1 > (int)id_limit) {
					id1 = (int)id_limit;
				}
				int id2 = id1;
				if (*flagstr == '-') {
					flagstr++;
					if (chaos_utility::is_digit(*flagstr)) {
						next = 0;
						sscanf(flagstr, "%d%n", &id2, &next);
						flagstr += next;
						if (id2 > (int)id_limit) {
							id2 = (int)id_limit;
						}
					}
				}
				if ((id1 <= id2) &&
					(id1 >= 0) &&
					(id2 < (int)buf->size()*8))
				{
					for (int id = id1; id <= id2; id++) {
						(*buf)[id/8] |= (chaos_uint8)(1 << (id%8));
					}
				} else {
					ASSERT(false);
				}
			} else {
				flagstr++;
			}
		}
		int key = regist(filename);
		readflags.insert(readflag_map::value_type(key, buf));
	}
	return true;
}

// save

bool chaos_readflag_manager::save(chaos_xml* readf_xml) const {
	readf_xml->clear();
	readf_xml->set_kind("readflag");
	readf_xml->set_values_version("1.0");
	chaos_value_manager<chaos_string>* readf_values =
		readf_xml->get_values();
	for (
		key_map::const_iterator it = keys.begin();
		it != keys.end();
		it++)
	{
		const chaos_string& filename = (*it).first;
		int key = (*it).second;
		readflag_map::const_iterator it2 = readflags.find(key);
		const chaos_buffer* buf = (*it2).second;
		chaos_string flag;
		int size = buf->size()*8,
			old_id = 0;
		bool old_read = false;
		for (int id = 0; id <= size; id++) {
			bool new_read = false;
			if (id < size) {
				new_read = (((*buf)[id/8] & (1 << (id%8))) != 0);
			}
			if (new_read != old_read) {
				if (!new_read) {
					if (!flag.empty()) {
						flag += ',';
					}
					if (old_id < id-1) {
						flag += chaos_utility::format_str(
							"%d-%d",
							old_id, id-1);
					} else {
						flag += chaos_utility::format_str(
							"%d",
							old_id);
					}
				}
				old_read = new_read;
				old_id = id;
			}
		}
		readf_values->set_value(
			filename,
			chaos_value(flag));
	}
	return true;
}
