////////////////////////////////////////////////////////////////////////////////
// chaos_rect.h : 2d rectangle

#ifndef CHAOS_RECT_H_INCLUDED
#define CHAOS_RECT_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_rect;

////////////////////////////////////////////////////////////////////////////////
// include

#ifndef CHAOS_POS_H_INCLUDED
#include "chaos_pos.h"
#endif // CHAOS_POS_H_INCLUDED

#ifndef CHAOS_DIM_H_INCLUDED
#include "chaos_dim.h"
#endif // CHAOS_DIM_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// chaos_rect : 2d rectangle class

class chaos_rect {
// attribute
public:
	// x position of left
	int x;
	// y position of top
	int y;
	// rectangle width
	int width;
	// rectangle height
	int height;

// create & destroy
public:
	// default constructor
	chaos_rect() :
		x(0), y(0),
		width(0), height(0)
	{
	}
	// normal constructor
	chaos_rect(
			int new_x, int new_y,
			int new_width, int new_height) :
		x(new_x), y(new_y),
		width(new_width), height(new_height)
	{
	}
	// position-dimension constructor
	chaos_rect(
			chaos_pos pos,
			chaos_dim dim) :
		x(pos.x), y(pos.y),
		width(dim.width), height(dim.height)
	{
	}
	// position-position constructor
	chaos_rect(
			chaos_pos pos1,
			chaos_pos pos2) :
		x(left_value(pos1.x, pos2.x)), y(top_value(pos1.y, pos2.y)),
		width(abs(pos2.x-pos1.x)), height(abs(pos2.y-pos1.y))
	{
	}
	// copy constructor
	chaos_rect(const chaos_rect& org) :
		x(org.x), y(org.y),
		width(org.width), height(org.height)
	{
	}

// get & set
public:
	// get left x
	int get_left() const {
		return x;
	}
	// get top y
	int get_top() const {
		return y;
	}
	// get right x
	int get_right() const {
		return x+width;
	}
	// get bottom y
	int get_bottom() const {
		return y+height;
	}
	// get left-top posisiton
	chaos_pos get_lt_pos() const {
		return chaos_pos(get_left(), get_top());
	}
	// get right-top posisiton
	chaos_pos get_rt_pos() const {
		return chaos_pos(get_right(), get_top());
	}
	// get left-bottom posisiton
	chaos_pos get_lb_pos() const {
		return chaos_pos(get_left(), get_bottom());
	}
	// get right-bottom posisiton
	chaos_pos get_rb_pos() const {
		return chaos_pos(get_right(), get_bottom());
	}
	// get dimension
	chaos_dim get_dim() const {
		return chaos_dim(width, height);
	}

// operation
private:
	// left value
	static int left_value(int x1, int x2) {
		return (x1 < x2)? x1: x2;
	}
	// top value
	static int top_value(int y1, int y2) {
		return (y1 < y2)? y1: y2;
	}
	// right value
	static int right_value(int x1, int x2) {
		return (x1 > x2)? x1: x2;
	}
	// bottom value
	static int bottom_value(int y1, int y2) {
		return (y1 > y2)? y1: y2;
	}

public:
	// is empty
	bool empty() const {
		return (width == 0) || (height == 0);
	}
	// clear
	void clear() {
		x = y = width = height = 0;
	}
	// offset
	void offset(chaos_pos ofs) {
		x += ofs.x;
		y += ofs.y;
	}

// operator
public:
	// copy
	chaos_rect& operator=(const chaos_rect& org) {
		x = org.x;
		y = org.y;
		width = org.width;
		height = org.height;
		return *this;
	}
	// and let
	chaos_rect& operator&=(const chaos_rect& org) {
		if (empty() || org.empty()) {
			clear();
		} else {
			int left = right_value(get_left(), org.get_left()),
				top = bottom_value(get_top(), org.get_top()),
				right = left_value(get_right(), org.get_right()),
				bottom = top_value(get_bottom(), org.get_bottom());
			if ((left >= right) || (top >= bottom)) {
				clear();
			} else {
				x = left;
				y = top;
				width = right-left;
				height = bottom-top;
			}
		}
		return *this;
	}
	// or let
	chaos_rect& operator|=(const chaos_rect& org) {
		if (org.empty()) {
			// nothing
		} else if (empty()) {
			*this = org;
		} else {
			int left = left_value(get_left(), org.get_left()),
				top = top_value(get_top(), org.get_top()),
				right = right_value(get_right(), org.get_right()),
				bottom = bottom_value(get_bottom(), org.get_bottom());
			x = left;
			y = top;
			width = right-left;
			height = bottom-top;
		}
		return *this;
	}
	// and
	chaos_rect operator&(const chaos_rect& org) const {
		chaos_rect tmp = *this;
		tmp &= org;
		return tmp;
	}
	// or
	chaos_rect operator|(const chaos_rect& org) const {
		chaos_rect tmp = *this;
		tmp |= org;
		return tmp;
	}
	// offset add let
	chaos_rect& operator+=(chaos_pos ofs) {
		x += ofs.x;
		y += ofs.y;
		return *this;
	}
	// offset sub let
	chaos_rect& operator-=(chaos_pos ofs) {
		x -= ofs.x;
		y -= ofs.y;
		return *this;
	}
	// offset add
	chaos_rect operator+(chaos_pos ofs) const {
		chaos_rect tmp = *this;
		tmp += ofs;
		return tmp;
	}
	// offset sub
	chaos_rect operator-(chaos_pos ofs) const {
		chaos_rect tmp = *this;
		tmp -= ofs;
		return tmp;
	}
	// equal
	bool operator==(const chaos_rect& org) const {
		return (x == org.x) && (y == org.y) &&
			(width == org.width) && (height == org.height);
	}
	// not equal
	bool operator!=(const chaos_rect& org) const {
		return (x != org.x) || (y != org.y) ||
			(width != org.width) || (height != org.height);
	}
};

#endif // CHAOS_RECT_H_INCLUDED
