////////////////////////////////////////////////////////////////////////////////
// chaos_sound_manager.cpp : sound manager

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"
#include "chaos_sound_manager.h"
#include "chaos_exception.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_sound : sound class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_sound::chaos_sound() {
	looping_play = false;
	streaming_play = false;
	volume = 100;

#ifdef CHAOS_SOUND_DIRECTX

	InitializeCriticalSection(&critsec);
	abort_event = CreateEvent(NULL, TRUE, FALSE, NULL);
	thread = NULL;
	dsbuffer = NULL;
	wfx = NULL;
	loaded_size = filled_size = 0;
	start_size = 1024;
	startable = started = false;
	fpt = NULL;
	ov_opened = false;
	memset(&vf, 0, sizeof(vf));
	dshow_graph = NULL;
	dshow_mctrl = NULL;
	dshow_mseeking = NULL;
	dshow_baudio = NULL;
	dshow_mevent = NULL;
	dshow_mevent_ex = NULL;
	dshow_event = NULL;

#elif defined(CHAOS_SOUND_SDL)

	chunk = NULL;
	music = NULL;

#endif // CHAOS_SOUND

}

// destructor

chaos_sound::~chaos_sound() {
	dispose();

#ifdef CHAOS_SOUND_DIRECTX

	DeleteCriticalSection(&critsec);
	CloseHandle(abort_event);

#endif // CHAOS_SOUND

}

// dispose

void chaos_sound::dispose() {

#ifdef CHAOS_SOUND_DIRECTX

	if (thread != NULL) {
		SetEvent(abort_event);
		WaitForSingleObject(thread, 10000);
		CloseHandle(thread);
		thread = NULL;
	}
	if (dsbuffer != NULL) {
		dsbuffer->Stop();
		dsbuffer->Release();
		dsbuffer = NULL;
	}
	if (wfx != NULL) {
		free(wfx);
		wfx = NULL;
	}
	wave_data.clear_all();
	loaded_size = filled_size = 0;
	start_size = 1024;
	startable = started = false;
	if (fpt != NULL) {
		fclose(fpt);
		fpt = NULL;
	}
	if (ov_opened) {
		ov_clear(&vf);
		memset(&vf, 0, sizeof(vf));
		ov_opened = false;
	}
	if (dshow_graph != NULL) {
		dshow_graph->Release();
		dshow_graph = NULL;
	}
	if (dshow_mctrl != NULL) {
		dshow_mctrl->Release();
		dshow_mctrl = NULL;
	}
	if (dshow_mseeking != NULL) {
		dshow_mseeking->Release();
		dshow_mseeking = NULL;
	}
	if (dshow_baudio != NULL) {
		dshow_baudio->Release();
		dshow_baudio = NULL;
	}
	if (dshow_mevent != NULL) {
		dshow_mevent->Release();
		dshow_mevent = NULL;
	}
	if (dshow_mevent_ex != NULL) {
		dshow_mevent_ex->Release();
		dshow_mevent_ex = NULL;
	}
	dshow_event = NULL;

#elif defined(CHAOS_SOUND_SDL)

	if (chunk != NULL) {
		Mix_FreeChunk(chunk);
		chunk = NULL;
	}
	if (music != NULL) {
		Mix_FreeMusic(music);
		music = NULL;
	}
	wave.clear_all();

#endif // CHAOS_SOUND

}

////////////////////////////////////////////////////////////////////////////////
// operation

#ifdef CHAOS_SOUND_DIRECTX

// thread function for loading wave / ogg-vorbis

unsigned __stdcall chaos_sound::load_thread_func(void* data) {
	chaos_sound* sound = (chaos_sound*)data;
	EnterCriticalSection(&sound->critsec);
	size_t total = sound->wave_data.size(),
		left = total,
		unit = sound->wfx->nAvgBytesPerSec,
		report = 0;
	sound->start_size = unit;
	if (left <= sound->loaded_size) {
		left = 0;
	} else {
		left -= sound->loaded_size;
	}
	int bits = sound->wfx->wBitsPerSample,
		bigendianp = 0,
		word = (bits+7)/8,
		sgned = (bits >= 16)? 1: 0,
		bitstream = 0;
	LeaveCriticalSection(&sound->critsec);
	chaos_buffer work;
	work.resize(unit);
	while (
		(left > 0) &&
		(WaitForSingleObject(sound->abort_event, 1) == WAIT_TIMEOUT))
	{
		size_t request_bytes = unit;
		if (request_bytes > left) {
			request_bytes = left;
		}
		size_t read_bytes = 0;
		if (request_bytes > 0) {
			if (sound->fpt != NULL) {
				read_bytes = fread(
					work.data(), 1, request_bytes,
					sound->fpt);
				if (read_bytes == 0) {
					break;
				}
			} else if (sound->ov_opened) {
				long read_bytes2 = ov_read(
					&sound->vf,
					(char*)work.data(), request_bytes,
					bigendianp, word, sgned,
					&bitstream);
				if (read_bytes2 < 0) {
					break;
				}
				read_bytes = read_bytes2;
			}
		}
		if (read_bytes > 0) {
			EnterCriticalSection(&sound->critsec);
			memcpy(
				sound->wave_data.data()+total-left,
				work.data(),
				read_bytes);
			left -= read_bytes;
			sound->loaded_size += read_bytes;
			if (left <= 0) {
				if (sound->fpt != NULL) {
					fclose(sound->fpt);
					sound->fpt = NULL;
				}
				if (sound->ov_opened) {
					ov_clear(&sound->vf);
					memset(&sound->vf, 0, sizeof(sound->vf));
					sound->ov_opened = false;
				}
			}
			LeaveCriticalSection(&sound->critsec);
			sound->fill_buffer();
			report += read_bytes;
			if ((report >= unit) || (left <= 0)) {
				report -= unit;
				EnterCriticalSection(&sound->critsec);
				sound->startable = true;
				LeaveCriticalSection(&sound->critsec);
				PostMessage(
					chaos_sound_manager::get_main_wnd(),
					chaos_sound_manager::WM_SOUND_LOADED,
					0, (LPARAM)sound);
			}
		}
	}
	return 0;
}

// thread function for watching direct show

unsigned __stdcall chaos_sound::watch_thread_func(void* data) {
	chaos_sound* sound = (chaos_sound*)data;
	HANDLE handles[2];
	EnterCriticalSection(&sound->critsec);
	handles[0] = sound->abort_event;
	handles[1] = sound->dshow_event;
	LeaveCriticalSection(&sound->critsec);
	bool finish = false;
	DWORD wait = WAIT_TIMEOUT;
	while (
		!finish &&
		((wait = WaitForMultipleObjects(
			2, handles, FALSE, 10)) != WAIT_OBJECT_0))
	{
		if (wait == WAIT_OBJECT_0+1) {
			finish = sound->process_event();
		}
	}
	return 0;
}

// load wave file(internal)

bool chaos_sound::load_wave(
	const char* filename,
	const void* mem, size_t length)
{
	bool loaded = false;
	chaos_uint8 work[4*3];
	try {
		ResetEvent(abort_event);
		IDirectSound* dsound =
			chaos_sound_manager::instance()->get_dsound();
		if (dsound == NULL) {
			throw std::runtime_error(
				"direct sound is not initialized");
		}
		const chaos_uint8* mpt = NULL;
		const chaos_uint8* mlast = NULL;
		if (filename != NULL) {
			fpt = fopen(filename, "rb");
			if (fpt == NULL) {
				throw std::runtime_error(
					"fopen");
			}
			if (fread(work, 1, 4+4+4, fpt) != 4+4+4) {
				throw std::runtime_error(
					"fread");
			}
		} else {
			mpt = mlast = (const chaos_uint8*)mem;
			if (mem != NULL) {
				mlast += length;
			}
			if (mlast-mpt < 4+4+4) {
				throw std::runtime_error(
					"memread");
			}
			memcpy(work, mpt, 4+4+4);
			mpt += 4+4+4;
		}
		if ((memcmp(work, "RIFF", 4) != 0) ||
			(memcmp(work+4+4, "WAVE", 4) != 0))
		{
			throw std::runtime_error(
				"not riff-wave");
		}
		size_t total_size = *((chaos_uint32*)(work+4));
		if (total_size <= 4) {
			throw std::runtime_error(
				"bad data");
		}
		total_size -= 4;
		bool wave_data_detect = false;
		do {
			if (total_size < 4+4) {
				throw std::runtime_error(
					"bad data");
			}
			if (filename != NULL) {
				if (fread(work, 1, 4+4, fpt) != 4+4) {
					throw std::runtime_error(
						"fread");
				}
			} else {
				if (mlast-mpt < 4+4) {
					throw std::runtime_error(
						"memread");
				}
				memcpy(work, mpt, 4+4);
				mpt += 4+4;
			}
			total_size -= 4+4;
			size_t tag_size =  *((chaos_uint32*)(work+4));
			if (total_size < tag_size) {
				throw std::runtime_error(
					"bad data");
			}
			total_size -= tag_size;
			if (memcmp(work, "fmt ", 4) == 0) {
				size_t wfx_size = tag_size;
				if (wfx_size < sizeof(WAVEFORMATEX)) {
					wfx_size = sizeof(WAVEFORMATEX);
				}
				wfx = (WAVEFORMATEX*)malloc(wfx_size);
				memset(wfx, 0, wfx_size);
				if (filename != NULL) {
					if (fread(wfx, 1, tag_size, fpt) != tag_size) {
						throw std::runtime_error(
							"fread");
					}
				} else {
					if ((size_t)(mlast-mpt) < tag_size) {
						throw std::runtime_error(
							"memread");
					}
					memcpy(wfx, mpt, tag_size);
					mpt += tag_size;
				}
			} else if (memcmp(work, "data", 4) == 0) {
				wave_data_detect = true;
				wave_data.resize(
					tag_size,
					(chaos_uint8)
						((wfx->wBitsPerSample == 8)? 0x80: 0x00));
			} else {
				if (filename != NULL) {
					fseek(fpt, tag_size, SEEK_CUR);
				} else {
					mpt += tag_size;
				}
			}
		} while ((wfx == NULL) || !wave_data_detect);
		if ((wfx == NULL) || !wave_data_detect) {
			throw std::runtime_error(
				"bad data");
		}
		DSBUFFERDESC dsbd;
		memset(&dsbd, 0, sizeof(DSBUFFERDESC));
		dsbd.dwSize = sizeof(dsbd);
		dsbd.dwFlags = DSBCAPS_GLOBALFOCUS | DSBCAPS_CTRLVOLUME;
		dsbd.dwBufferBytes = (DWORD)wave_data.size();
		dsbd.lpwfxFormat = wfx;
		HRESULT hr = dsound->CreateSoundBuffer(
			&dsbd, &dsbuffer, NULL);
		if (hr != DS_OK) {
			throw std::runtime_error(
				"CreateSoundBuffer");
		}
		set_volume(volume);
		if (fpt != NULL) {
			unsigned thread_id = 0;
			if ((thread = (HANDLE)_beginthreadex(
					NULL, 0,
					load_thread_func,
					this, 0, &thread_id)) == NULL)
			{
				throw std::runtime_error(
					"_beginthreadex");
			}
		} else {
			memcpy(wave_data.data(), mpt, wave_data.size());
			loaded_size = wave_data.size();
			fill_buffer();
			startable = true;
			start();
		}
		loaded = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::load wave - "
			"%s\n",
			e.what());
	}
	if (!loaded) {
		dispose();
	}
	return loaded;
}

// load ogg-vorbis file(internal)

bool chaos_sound::load_ogg_vorbis(const char* filename) {
	bool loaded = false;
	try {
		ResetEvent(abort_event);
		IDirectSound* dsound =
			chaos_sound_manager::instance()->get_dsound();
		if (dsound == NULL) {
			throw std::runtime_error(
				"direct sound is not initialized");
		}
		std::vector<char> filename2(strlen(filename)+1);
		strcpy(&filename2[0], filename);
		if (ov_fopen(&filename2[0], &vf) != 0) {
			throw std::runtime_error(
				"ov_fopen");
		}
		ov_opened = true;
		vorbis_info *vi = ov_info(&vf, -1);
		if (vi == NULL) {
			throw std::runtime_error(
				"ov_info");
		}
		int rate = (int)vi->rate,
			channels = vi->channels,
			bits = 16,
			block = ((bits+7)/8)*channels,
			avg = rate*block;
		size_t samples = (int)ov_pcm_total(&vf, -1);
		if (samples <= 0) {
			throw std::runtime_error(
				"ov_pcm_total");
		}
		wfx = (WAVEFORMATEX*)malloc(sizeof(WAVEFORMATEX));
		memset(wfx, 0, sizeof(WAVEFORMATEX));
		wfx->wFormatTag      = WAVE_FORMAT_PCM;
		wfx->nChannels       = (WORD)channels;
		wfx->nSamplesPerSec  = rate;
		wfx->nAvgBytesPerSec = avg;
		wfx->nBlockAlign     = (WORD)block;
		wfx->wBitsPerSample  = (WORD)bits;
		wfx->cbSize          = 0;
		wave_data.resize(
			samples*block,
			(chaos_uint8)
				((wfx->wBitsPerSample == 8)? 0x80: 0x00));
		DSBUFFERDESC dsbd;
		memset(&dsbd, 0, sizeof(DSBUFFERDESC));
		dsbd.dwSize = sizeof(dsbd);
		dsbd.dwFlags = DSBCAPS_GLOBALFOCUS | DSBCAPS_CTRLVOLUME;
		dsbd.dwBufferBytes = (DWORD)wave_data.size();
		dsbd.lpwfxFormat = wfx;
		HRESULT hr = dsound->CreateSoundBuffer(
			&dsbd, &dsbuffer, NULL);
		if (hr != DS_OK) {
			throw std::runtime_error(
				"CreateSoundBuffer");
		}
		set_volume(volume);
		unsigned thread_id = 0;
		if ((thread = (HANDLE)_beginthreadex(
				NULL, 0,
				load_thread_func,
				this, 0, &thread_id)) == NULL)
		{
			throw std::runtime_error(
				"_beginthreadex");
		}
		loaded = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::load ogg vorbis - "
			"%s\n",
			e.what());
	}
	if (!loaded) {
		dispose();
	}
	return loaded;
}

// load sound by direct show

bool chaos_sound::load_by_direct_show(const char* filename) {
	bool loaded = false;
	try {
		HRESULT hr = CoCreateInstance(
			CLSID_FilterGraph,
			NULL,
			CLSCTX_INPROC,
			IID_IGraphBuilder,
			(void**)&dshow_graph);
		if (hr != S_OK) {
			dshow_graph = NULL;
			throw std::runtime_error(
				"error: CoCreateInstance");
		}
		std::vector<wchar_t> wide_filename;
		int length = MultiByteToWideChar(
			CP_OEMCP, 0,
			filename, strlen(filename)+1,
			NULL, 0);
		wide_filename.resize(length);
		MultiByteToWideChar(
			CP_OEMCP, 0,
			filename, strlen(filename)+1,
			&wide_filename[0], length);
		hr = dshow_graph->RenderFile(
			&wide_filename[0],
			NULL);
		if (hr != S_OK) {
			throw std::runtime_error(
				"error: RenderFile");
		}
		hr = dshow_graph->QueryInterface(
			IID_IMediaControl,
			(void**)&dshow_mctrl);
		if (hr != S_OK) {
			dshow_mctrl = NULL;
			throw std::runtime_error(
				"error: QueryInterface");
		}
		hr = dshow_graph->QueryInterface(
			IID_IMediaSeeking,
			(void**)&dshow_mseeking);
		if (hr != S_OK) {
			dshow_mseeking = NULL;
			throw std::runtime_error(
				"error: QueryInterface");
		}
		hr = dshow_graph->QueryInterface(
			IID_IBasicAudio,
			(void**)&dshow_baudio);
		if (hr != S_OK) {
			dshow_baudio = NULL;
			throw std::runtime_error(
				"error: QueryInterface");
		}
		hr = dshow_graph->QueryInterface(
			IID_IMediaEvent,
			(void**)&dshow_mevent);
		if (hr != S_OK) {
			dshow_mevent = NULL;
			throw std::runtime_error(
				"error: QueryInterface");
		}
		hr = dshow_graph->QueryInterface(
			IID_IMediaEventEx,
			(void**)&dshow_mevent_ex);
		if (hr == S_OK) {
			hr = dshow_mevent_ex->SetNotifyWindow(
				(OAHWND)chaos_sound_manager::get_main_wnd(),
				chaos_sound_manager::WM_SOUND_EVENT,
				(LPARAM)this);
			if (hr != S_OK) {
				throw std::runtime_error(
					"error: SetNotifyWindow");
			}
		} else {
			dshow_mevent_ex = NULL;
			hr = dshow_mevent->GetEventHandle(
				(OAEVENT*)&dshow_event);
			if (hr != S_OK) {
				dshow_event = NULL;
				throw std::runtime_error(
					"error: GetEventHandle");
			}
		}
		set_volume(volume);
		dshow_mctrl->Run();
		if (dshow_mevent_ex == NULL) {
			unsigned thread_id = 0;
			if ((thread = (HANDLE)_beginthreadex(
					NULL, 0,
					watch_thread_func,
					this, 0, &thread_id)) == NULL)
			{
				throw std::runtime_error(
					"_beginthreadex");
			}
		}
		loaded = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::load by direct show - "
			"%s\n",
			e.what());
	}
	if (!loaded) {
		dispose();
	}
	return loaded;
}

#endif // CHAOS_SOUND

// load sound from file and play

bool chaos_sound::play_sound_from_file(
	const char* filename,
	bool looping,
	bool streaming)
{
	dispose();
	looping_play = looping;
	streaming_play = streaming;
	bool result = false;

#ifdef CHAOS_SOUND_DIRECTX

	char ext[_MAX_EXT];
	_splitpath(filename, NULL, NULL, NULL, ext);

	if (stricmp(ext, ".wav") == 0) {
		result = load_wave(filename, NULL, 0);
	} else if (
		(stricmp(ext, ".ogg") == 0) ||
		(stricmp(ext, ".oga") == 0))
	{
		result = load_ogg_vorbis(filename);
	}
	if (!result) {
		result = load_by_direct_show(filename);
	}

#elif defined(CHAOS_SOUND_SDL)

	try {
		music = Mix_LoadMUS(filename);
		if (music == NULL) {
			throw std::runtime_error(
				"Mix_LoadMUS");
		}
		set_volume(volume);
		Mix_PlayMusic(music, looping_play? -1: 1);
		result = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::play sound from file - "
			"%s\n",
			e.what());
	}
	if (!result) {
		dispose();
	}

#endif // CHAOS_SOUND

	return result;
}

// load sound from memory and play

bool chaos_sound::play_sound_from_memory(
	const void* mem, size_t length,
	bool looping)
{
	dispose();
	looping_play = looping;
	streaming_play = false;
	bool result = false;

#ifdef CHAOS_SOUND_DIRECTX

	result = load_wave(NULL, mem, length);

#elif defined(CHAOS_SOUND_SDL)

	try {
		wave.resize(length);
		memcpy(wave.data(), mem, length);
		SDL_RWops* rwops = SDL_RWFromMem(wave.data(), wave.size());
		if (rwops == NULL) {
			throw std::runtime_error(
				"SDL_RWFromMem");
		}
		chunk = Mix_LoadWAV_RW(rwops, 1);
		if (chunk == NULL) {
			throw std::runtime_error(
				"Mix_LoadWAV_RW");
		}
		set_volume(volume);
		Mix_PlayChannel(-1, chunk, looping_play? -1: 0);
		result = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::play sound from memory - "
			"%s\n",
			e.what());
	}
	if (!result) {
		dispose();
	}

#endif // CHAOS_SOUND

	return result;
}

#ifdef CHAOS_SOUND_DIRECTX

// play blank sound

bool chaos_sound::play_blank_sound(
	int rate, int channels, int bits, int second)
{
	looping_play = true;
	streaming_play = false;
	bool played = false;
	try {
		ResetEvent(abort_event);
		IDirectSound* dsound =
			chaos_sound_manager::instance()->get_dsound();
		if (dsound == NULL) {
			throw std::runtime_error(
				"direct sound is not initialized");
		}
		int block = ((bits+7)/8)*channels,
			avg = rate*block;
		size_t samples = second*avg;
		wfx = (WAVEFORMATEX*)malloc(sizeof(WAVEFORMATEX));
		memset(wfx, 0, sizeof(WAVEFORMATEX));
		wfx->wFormatTag      = WAVE_FORMAT_PCM;
		wfx->nChannels       = (WORD)channels;
		wfx->nSamplesPerSec  = rate;
		wfx->nAvgBytesPerSec = avg;
		wfx->nBlockAlign     = (WORD)block;
		wfx->wBitsPerSample  = (WORD)bits;
		wfx->cbSize          = 0;
		wave_data.resize(
			samples*block,
			(chaos_uint8)
				((wfx->wBitsPerSample == 8)? 0x80: 0x00));
		DSBUFFERDESC dsbd;
		memset(&dsbd, 0, sizeof(DSBUFFERDESC));
		dsbd.dwSize = sizeof(dsbd);
		dsbd.dwFlags = DSBCAPS_GLOBALFOCUS | DSBCAPS_CTRLVOLUME;
		dsbd.dwBufferBytes = (DWORD)wave_data.size();
		dsbd.lpwfxFormat = wfx;
		HRESULT hr = dsound->CreateSoundBuffer(
			&dsbd, &dsbuffer, NULL);
		if (hr != DS_OK) {
			throw std::runtime_error(
				"CreateSoundBuffer");
		}
		set_volume(volume);
		loaded_size = wave_data.size();
		fill_buffer();
		startable = true;
		start();
		played = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::play blank sound - "
			"%s\n",
			e.what());
	}
	if (!played) {
		dispose();
	}
	return played;
}

// fill sound buffer and play

void chaos_sound::fill_buffer() {
	EnterCriticalSection(&critsec);
	HRESULT hr;
	if (loaded_size > filled_size) {
		void* audio_ptr1 = NULL;
		void* audio_ptr2 = NULL;
		DWORD audio_size1 = 0, audio_size2 = 0;
		hr = dsbuffer->Lock(
			filled_size, loaded_size-filled_size,
			&audio_ptr1, &audio_size1,
			&audio_ptr2, &audio_size2,
			0);
		if (hr == DSERR_BUFFERLOST) {
			filled_size = 0;
			hr = dsbuffer->Restore();
			if (hr == DS_OK) {
				hr = dsbuffer->Lock(
					filled_size,
					loaded_size-filled_size,
					&audio_ptr1, &audio_size1,
					&audio_ptr2, &audio_size2,
					0);
			}
		}
		if (hr == DS_OK) {
			if (audio_size1 > 0) {
				memcpy(
					audio_ptr1,
					wave_data.data()+filled_size,
					audio_size1);
				filled_size += audio_size1;
			}
			if (audio_size2 > 0) {
				memcpy(
					audio_ptr2,
					wave_data.data()+filled_size,
					audio_size2);
				filled_size += audio_size2;
			}
			dsbuffer->Unlock(
				audio_ptr1, audio_size1,
				audio_ptr2, audio_size2);
		}
	}
	LeaveCriticalSection(&critsec);
}

// start

bool chaos_sound::start() {
	EnterCriticalSection(&critsec);
	if ((dsbuffer != NULL) &&
		startable && !started)
	{
		HRESULT hr = dsbuffer->Play(
			0, 0, looping_play? DSBPLAY_LOOPING: 0);
		if (hr == S_OK) {
			started = true;
		}
	}
	LeaveCriticalSection(&critsec);
	return started;
}

// process event

bool chaos_sound::process_event() {
	bool finish = false;
	EnterCriticalSection(&critsec);
	if (dshow_mevent != NULL) {
		long event_code = 0, param1 = 0, param2 = 0;
		HRESULT hr;
		while (
			(hr = dshow_mevent->GetEvent(
				&event_code, &param1, &param2, 0)) == S_OK)
		{
			switch (event_code) {
			case EC_COMPLETE:
				if (looping_play) {
					LONGLONG llPos = 0;
					hr = dshow_mseeking->SetPositions(
						&llPos,
						AM_SEEKING_AbsolutePositioning,
						NULL,
						AM_SEEKING_NoPositioning);
					if (hr == S_OK) {
						dshow_mctrl->Run();
					}
				} else {
					finish = true;
				}
				break;
			}
			dshow_mevent->FreeEventParams(
				event_code, param1, param2);
		}
	}
	LeaveCriticalSection(&critsec);
	return finish;
}

#endif // CHAOS_SOUND

// get sound volume

int chaos_sound::get_volume() const {
	return volume;
}

// set sound volume

void chaos_sound::set_volume(int new_volume) {
	volume = new_volume;
	if (volume < 0) {
		volume = 0;
	} else if (volume > 100) {
		volume = 100;
	}

#ifdef CHAOS_SOUND_DIRECTX

	int ds_volume;
	if (volume <= 0) {
		ds_volume = DSBVOLUME_MIN;
	} else if (volume >= 100) {
		ds_volume = DSBVOLUME_MAX;
	} else {
		ds_volume = -(100-volume)*40;
	}
	EnterCriticalSection(&critsec);
	if (dsbuffer != NULL) {
		dsbuffer->SetVolume(ds_volume);
	} else if (dshow_baudio != NULL) {
		dshow_baudio->put_Volume(ds_volume);
	}
	LeaveCriticalSection(&critsec);

#elif defined(CHAOS_SOUND_SDL)

	int sdl_volume;
	if (volume <= 0) {
		sdl_volume = 0;
	} else if (volume >= 100) {
		sdl_volume = MIX_MAX_VOLUME;
	} else {
		sdl_volume = (int)
			(MIX_MAX_VOLUME*(pow(10.0, -(100-volume)*40/2000.0))+0.5);
		if (sdl_volume < 0) {
			sdl_volume = 0;
		} else if (sdl_volume > MIX_MAX_VOLUME) {
			sdl_volume = MIX_MAX_VOLUME;
		}
	}
	if (chunk != NULL) {
		Mix_VolumeChunk(chunk, sdl_volume);
	}
	if (music != NULL) {
		Mix_VolumeMusic(sdl_volume);
	}

#endif // CHAOS_SOUND

}

////////////////////////////////////////////////////////////////////////////////
// chaos_sound_manager : sound manager class

////////////////////////////////////////////////////////////////////////////////
// attribute

#ifdef CHAOS_SOUND_DIRECTX

// main window

HWND chaos_sound_manager::main_wnd = NULL;

#endif // CHAOS_SOUND

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_sound_manager::chaos_sound_manager() {

#ifdef CHAOS_SOUND_DIRECTX

	dsound_dll_instance = NULL;
	dsound = NULL;
	dsbuffer_primary = NULL;
	blank_sound = NULL;

#elif defined(CHAOS_SOUND_SDL)

	mixer_opened = false;

#endif // CHAOS_SOUND

}

// destructor

chaos_sound_manager::~chaos_sound_manager() {
	dispose();
}

// initialize

bool chaos_sound_manager::initialize() {
	bool result = false;

#ifdef CHAOS_SOUND_DIRECTX

	try {
		if (dsound_dll_instance == NULL) {
			dsound_dll_instance = LoadLibrary("dsound.dll");
			if (dsound_dll_instance == NULL) {
				throw std::runtime_error(
					"LoadLibrary");
			}
		}
		HRESULT hr;
		if (dsound == NULL) {
			typedef HRESULT WINAPI DirectSoundCreateProc(
				GUID*, IDirectSound**, IUnknown*);
			DirectSoundCreateProc* pprocDirectSoundCreate =
				(DirectSoundCreateProc*)GetProcAddress(
					dsound_dll_instance,
					"DirectSoundCreate");
			if (pprocDirectSoundCreate == NULL) {
				throw std::runtime_error(
					"GetProcAddress");
			}
			hr = pprocDirectSoundCreate(NULL, &dsound, NULL);
			if (hr != DS_OK) {
				dsound = NULL;
				throw std::runtime_error(
					"DirectSoundCreate");
			}
		}
		if (dsbuffer_primary == NULL) {
			int nLevel = DSSCL_PRIORITY;
			hr = S_FALSE;
			if (dsound != NULL) {
				for (int i = 0; i < 3; i++) {
					switch (i) {
					case 0:
						nLevel = DSSCL_PRIORITY;
						break;
					case 1:
						nLevel = DSSCL_NORMAL;
						break;
					case 2:
						nLevel = DSSCL_EXCLUSIVE;
						break;
					}
					hr = dsound->SetCooperativeLevel(
						main_wnd, nLevel);
					if (hr == DS_OK) {
						break;
					}
				}
			}
			if (hr != DS_OK) {
				throw std::runtime_error(
					"SetCooperativeLevel");
			}
			DSBUFFERDESC dsbd;
			memset(&dsbd, 0, sizeof(dsbd));
			dsbd.dwSize = sizeof(dsbd);
			dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
			hr = dsound->CreateSoundBuffer(
				&dsbd, &dsbuffer_primary, NULL);
			if (hr != DS_OK) {
				dsbuffer_primary = NULL;
				throw std::runtime_error(
					"CreateSoundBuffer");
			}
			if (nLevel != DSSCL_NORMAL) {
				int rate = 44100,
					channels = 2,
					bits = 16,
					block = ((bits+7)/8)*channels,
					avg = rate*block;
				WAVEFORMATEX wfx;
				memset(&wfx, 0, sizeof(wfx));
				wfx.wFormatTag      = WAVE_FORMAT_PCM;
				wfx.nChannels       = (WORD)channels;
				wfx.nSamplesPerSec  = rate;
				wfx.nAvgBytesPerSec = avg;
				wfx.nBlockAlign     = (WORD)block;
				wfx.wBitsPerSample  = (WORD)bits;
				wfx.cbSize          = 0;
				hr = dsbuffer_primary->SetFormat(&wfx);
				if (hr != DS_OK) {
					// ignore error
				}
			}
		}
		if (blank_sound == NULL) {
			blank_sound = new chaos_sound;
			blank_sound->play_blank_sound(
				22050, 1, 16, 1);
		}
		result = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::load wave - "
			"%s\n",
			e.what());
	}

#elif defined(CHAOS_SOUND_SDL)

	try {
		if (!mixer_opened) {
			if (Mix_OpenAudio(
					MIX_DEFAULT_FREQUENCY,
					MIX_DEFAULT_FORMAT,
					2,
					4096) != 0)
			{
				throw std::runtime_error(
					"Mix_OpenAudio");
			}
			mixer_opened = true;
		}
		result = true;
	} catch (const std::exception& e) {
		LOG_ERR(
			"exception : sound::load wave - "
			"%s\n",
			e.what());
	}

#endif // CHAOS_SOUND

	return result;
}

// dispose

void chaos_sound_manager::dispose() {

#ifdef CHAOS_SOUND_DIRECTX

	if (blank_sound != NULL) {
		delete blank_sound;
		blank_sound = NULL;
	}
	if (dsbuffer_primary != NULL) {
		dsbuffer_primary->Release();
		dsbuffer_primary = NULL;
	}
	if (dsound != NULL) {
		dsound->Release();
		dsound = NULL;
	}

#elif defined(CHAOS_SOUND_SDL)

	if (mixer_opened) {
		Mix_CloseAudio();
		mixer_opened = false;
	}

#endif // CHAOS_SOUND

}

// get singleton instance

chaos_sound_manager* chaos_sound_manager::instance() {
	static chaos_sound_manager sound_manager;
	return &sound_manager;
}
