////////////////////////////////////////////////////////////////////////////////
// chaos_sound_manager.h : wave player

#ifndef CHAOS_SOUND_MANAGER_H_INCLUDED
#define CHAOS_SOUND_MANAGER_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_sound;
class chaos_sound_manager;

////////////////////////////////////////////////////////////////////////////////
// include

#ifndef CHAOS_STRING_H_INCLUDED
#include "chaos_string.h"
#endif // CHAOS_STRING_H_INCLUDED

#ifndef CHAOS_BUFFER_H_INCLUDED
#include "chaos_buffer.h"
#endif // CHAOS_BUFFER_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// chaos_sound : sound class

class chaos_sound {
// attribute
private:
	// play sound at looping
	bool looping_play;
	// play sound in streaming
	bool streaming_play;
	// sound volume
	int volume;

#ifdef CHAOS_SOUND_DIRECTX

	// critical section
	CRITICAL_SECTION critsec;
	// event handle for abort
	HANDLE abort_event;
	// thread handle
	HANDLE thread;
	// direct sound secoudary buffer
	IDirectSoundBuffer* dsbuffer;
	// wave format
	WAVEFORMATEX* wfx;
	// wave data
	chaos_buffer wave_data;
	// loaded size
	size_t loaded_size;
	// filled size
	size_t filled_size;
	// start size
	size_t start_size;
	// startable to playing sound
	bool startable;
	// started to playing sound
	bool started;
	// file pointer
	FILE* fpt;
	// ogg-vorbis opened
	bool ov_opened;
	// ogg-vorbis file
	OggVorbis_File vf;
	// direct show graph
	IGraphBuilder* dshow_graph;
	// direct show media control
	IMediaControl* dshow_mctrl;
	// direct show media seeking
	IMediaSeeking* dshow_mseeking;
	// direct show basic audio
	IBasicAudio* dshow_baudio;
	// direct show media event
	IMediaEvent* dshow_mevent;
	IMediaEventEx* dshow_mevent_ex;
	// event handle for direct show
	HANDLE dshow_event;

#elif defined(CHAOS_SOUND_SDL)

	// sdl-mixer chunk
	Mix_Chunk* chunk;
	// sdl-mixer music
	Mix_Music* music;
	// wave
	chaos_buffer wave;

#endif // CHAOS_SOUND

// create & destroy
public:
	// default constructor
	chaos_sound();
	// destructor
	~chaos_sound();
	// dispose
	void dispose();

// operation
protected:

#ifdef CHAOS_SOUND_DIRECTX

	// thread function for loading wave / ogg-vorbis
	static unsigned __stdcall load_thread_func(void* data);
	// thread function for watching direct show
	static unsigned __stdcall watch_thread_func(void* data);
	// load wave file(internal)
	bool load_wave(
		const char* filename,
		const void* mem, size_t length);
	// load ogg-vorbis file(internal)
	bool load_ogg_vorbis(const char* filename);
	// load sound by direct show
	bool load_by_direct_show(const char* filename);

#endif // CHAOS_SOUND

public:
	// load sound from file and play
	bool play_sound_from_file(
		const char* filename,
		bool looping,
		bool streaming);
	// load sound from memory and play
	bool play_sound_from_memory(
		const void* mem, size_t length,
		bool looping);

#ifdef CHAOS_SOUND_DIRECTX

	// play blank sound
	bool play_blank_sound(
		int rate, int channels, int bits, int second);
	// fill sound buffer
	void fill_buffer();
	// start
	bool start();
	// process event
	bool process_event();

#endif // CHAOS_SOUND

	// get sound volume
	int get_volume() const;
	// set sound volume
	void set_volume(int new_volume);
};

////////////////////////////////////////////////////////////////////////////////
// chaos_sound_manager : sound manager class

class chaos_sound_manager {
// constant
public:

#ifdef CHAOS_SOUND_DIRECTX

	enum {
		WM_SOUND_LOADED = WM_APP+100,
		WM_SOUND_EVENT  = WM_APP+101
	};

#endif // CHAOS_SOUND

// attribute
private:

#ifdef CHAOS_SOUND_DIRECTX

	// main window
	static HWND main_wnd;
	// dsound dll instance
	HINSTANCE dsound_dll_instance;
	// direct sound
	IDirectSound* dsound;
	// direct sound primary buffer
	IDirectSoundBuffer* dsbuffer_primary;
	// blank sound
	chaos_sound* blank_sound;

#elif defined(CHAOS_SOUND_SDL)

	// mixer opened
	bool mixer_opened;

#endif // CHAOS_SOUND

// create & destroy
private:
	// default constructor
	chaos_sound_manager();

public:
	// destructor
	~chaos_sound_manager();
	// initialize
	bool initialize();
	// dispose
	void dispose();

	// get singleton instance
	static chaos_sound_manager* instance();

// get & set
public:

#ifdef CHAOS_SOUND_DIRECTX

	// get main window
	static HWND get_main_wnd() {
		return main_wnd;
	}
	// set main window
	static void set_main_wnd(HWND new_main_wnd) {
		main_wnd = new_main_wnd;
	}
	// get direct sound
	IDirectSound* get_dsound() {
		return dsound;
	}

#endif // CHAOS_SOUND

};

#endif // CHAOS_SOUND_MANAGER_H_INCLUDED
