////////////////////////////////////////////////////////////////////////////////
// chaos_string.h : string

#ifndef CHAOS_STRING_H_INCLUDED
#define CHAOS_STRING_H_INCLUDED

////////////////////////////////////////////////////////////////////////////////
// declare

class chaos_string;

////////////////////////////////////////////////////////////////////////////////
// chaos_string : string class

class chaos_string {
// enum
private:
	// capacity
	enum {
		capacity_min  = 16,
		capacity_step = 16
	};

// typedef
public:
	// value type
	typedef char value_type;
	// size type
	typedef size_t size_type;
	// iterator
	typedef value_type* iterator;
	// const iterator
	typedef const value_type* const_iterator;
	// reference
	typedef value_type& reference;
	// const reference
	typedef const value_type& const_reference;

// struct
private:
	struct chaos_string_data {
		// reference count
		int ref_count;
		// string length
		size_type length;
		// string capacity
		size_type capacity;
		// string
		value_type string[1];
	};

// attribute
private:
	// str data
	chaos_string_data* data_ptr;

	// nul string
	static const value_type null_str[1];
	// maximum position
	static const size_type npos;

// create & destroy
public:
	// default constructor
	chaos_string() :
		data_ptr(NULL)
	{
	}
	// copy constructor
	chaos_string(const chaos_string& org) :
		data_ptr(org.data_ptr)
	{
		if (data_ptr != NULL) {
			data_ptr->ref_count++;
		}
	}
	// value_type* constructor
	chaos_string(const value_type* org) :
		data_ptr(NULL)
	{
		if ((org != NULL) && (*org != '\0')) {
			size_type length = strlen(org);
			ASSERT(length > 0);
			alloc_data(length);
			strcpy(data_ptr->string, org);
			data_ptr->length = length;
		}
	}
	// value_type constructor
	chaos_string(const value_type& org) :
		data_ptr(NULL)
	{
		alloc_data(1);
		data_ptr->string[0] = org;
		data_ptr->string[1] = '\0';
		data_ptr->length = 1;
	}
	// value_type fill constructor
	chaos_string(size_type fill, const value_type& org) :
		data_ptr(NULL)
	{
		if (fill > 0) {
			alloc_data(fill);
			memset(data_ptr->string, org, fill);
			data_ptr->string[fill] = '\0';
			data_ptr->length = fill;
		}
	}
	// iterator constructor
	chaos_string(const_iterator first, const_iterator last) :
		data_ptr(NULL)
	{
		size_type length = std::distance(first, last);
		if (length > 0) {
			alloc_data(length);
			memcpy(data_ptr->string, first, length);
			data_ptr->string[length] = '\0';
			data_ptr->length = length;
		}
	}
	// destructor
	~chaos_string() {
		free_data();
	}

// buffer
private:
	// allocate data buffer
	void alloc_data(size_type capacity);
	// free data buffer
	void free_data();

// operation
public:
	// get length
	size_type length() const {
		return (data_ptr != 0)? data_ptr->length: 0;
	}
	// get size
	size_type size() const {
		return (data_ptr != 0)? data_ptr->length: 0;
	}
	// get capacity
	size_type capacity() const {
		return (data_ptr != 0)? data_ptr->capacity: 0;
	}
	// is empty
	bool empty() const {
		return length() == 0;
	}
	// get const value_type pointer
	const value_type* c_str() const {
		return (data_ptr != NULL)? data_ptr->string: null_str;
	}
	// reserve
	void reserve(size_type new_capacity) {
		alloc_data(new_capacity);
	}
	// resize
	void resize(size_type new_length) {
		alloc_data(new_length);
		ASSERT(data_ptr != NULL);
		data_ptr->length = new_length;
	}
	// get begin
	const_iterator begin() const {
		return c_str();
	}
	iterator begin() {
		alloc_data(length());
		ASSERT(data_ptr != NULL);
		return data_ptr->string;
	}
	// get end
	const_iterator end() const {
		return c_str()+length();
	}
	iterator end() {
		alloc_data(length());
		ASSERT(data_ptr != NULL);
		return data_ptr->string+data_ptr->length;
	}
	// reference
	const_reference at(size_type index) const {
		ASSERT(index < length());
		return *(begin()+index);
	}
	reference at(size_type index) {
		ASSERT(index < length());
		return *(begin()+index);
	}
	// reference front
	const_reference front() const {
		ASSERT(length() > 0);
		return *begin();
	}
	reference front() {
		ASSERT(length() > 0);
		return *begin();
	}
	// reference back
	const_reference back() const {
		ASSERT(length() > 0);
		return *(begin()+length()-1);
	}
	reference back() {
		ASSERT(length() > 0);
		return *(begin()+length()-1);
	}
	// erase
	chaos_string& erase(size_type pos = 0, size_type count = npos);
	iterator erase(iterator first, iterator last);
	iterator erase(iterator target);
	// insert
	chaos_string& insert(size_type pos, const chaos_string& org);
	void insert(
		iterator target,
		size_type fill,
		const value_type& value);
	iterator insert(
		iterator target,
		const value_type& value = 0);
	// get substring
	chaos_string substr(size_type pos = 0, size_type count = npos) const;
	// push back
	void push_back(const value_type& value);
	// pop back
	void pop_back();
	// compare string
	int compare(const chaos_string& org) const;

// operator
public:
	// copy
	chaos_string& operator=(const chaos_string& org);
	// append
	chaos_string& operator+=(const chaos_string& org);
	chaos_string& operator+=(const value_type& org);
	// add
	chaos_string operator+(const chaos_string& org) const;
	// indexer
	const_reference operator[](size_type index) const {
		ASSERT(index < length());
		return *(begin()+index);
	}
	reference operator[](size_type index) {
		ASSERT(index < length());
		return *(begin()+index);
	}
	// equal
	bool operator==(const chaos_string& org) const {
		return compare(org) == 0;
	}
	// not equal
	bool operator!=(const chaos_string& org) const {
		return compare(org) != 0;
	}
	// less
	bool operator<(const chaos_string& org) const {
		return compare(org) < 0;
	}
	// less equal
	bool operator<=(const chaos_string& org) const {
		return compare(org) <= 0;
	}
	// greater
	bool operator>(const chaos_string& org) const {
		return compare(org) > 0;
	}
	// greater equal
	bool operator>=(const chaos_string& org) const {
		return compare(org) >= 0;
	}

// friends
public:
	// add
	friend chaos_string operator+(
		const value_type* sz, const chaos_string& str)
	{
		return chaos_string(sz)+str;
	}
	friend chaos_string operator+(
		const value_type& ch, const chaos_string& str)
	{
		return chaos_string(ch)+str;
	}
};

#endif // CHAOS_STRING_H_INCLUDED
