////////////////////////////////////////////////////////////////////////////////
// chaos_utility.cpp : utility

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_utility.h"
#include "chaos_auto_array.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_utility : utility class

// hankaku ascii to zenkaku table

static int hankaku_ascii_table[] = {
	0x2121, 0x212A, 0x2149, 0x2174, 0x2170, 0x2173, 0x2175, 0x2147,
	0x214A, 0x214B, 0x2176, 0x215C, 0x2124, 0x215D, 0x2125, 0x213F,
	0x2330, 0x2331, 0x2332, 0x2333, 0x2334, 0x2335, 0x2336, 0x2337,
	0x2338, 0x2339, 0x2127, 0x2128, 0x2163, 0x2161, 0x2164, 0x2129,
	0x2177, 0x2341, 0x2342, 0x2343, 0x2344, 0x2345, 0x2346, 0x2347,
	0x2348, 0x2349, 0x234A, 0x234B, 0x234C, 0x234D, 0x234E, 0x234F,
	0x2350, 0x2351, 0x2352, 0x2353, 0x2354, 0x2355, 0x2356, 0x2357,
	0x2358, 0x2359, 0x235A, 0x214E, 0x216F, 0x214F, 0x2130, 0x2132,
	0x2146, 0x2361, 0x2362, 0x2363, 0x2364, 0x2365, 0x2366, 0x2367,
	0x2368, 0x2369, 0x236A, 0x236B, 0x236C, 0x236D, 0x236E, 0x236F,
	0x2370, 0x2371, 0x2372, 0x2373, 0x2374, 0x2375, 0x2376, 0x2377,
	0x2378, 0x2379, 0x237A, 0x2150, 0x2143, 0x2151, 0x2131
};

// hankaku katakana to zenkaku table

static int hankaku_kana_table[] = {
	        0x2123, 0x2156, 0x2157, 0x2122, 0x2126, 0x2572, 0x2521,
	0x2523, 0x2525, 0x2527, 0x2529, 0x2563, 0x2565, 0x2567, 0x2543,
	0x213C, 0x2522, 0x2524, 0x2526, 0x2528, 0x252A, 0x252B, 0x252D,
	0x252F, 0x2531, 0x2533, 0x2535, 0x2537, 0x2539, 0x253B, 0x253D,
	0x253F, 0x2541, 0x2544, 0x2546, 0x2548, 0x254A, 0x254B, 0x254C,
	0x254D, 0x254E, 0x254F, 0x2552, 0x2555, 0x2558, 0x255B, 0x255E,
	0x255F, 0x2560, 0x2561, 0x2562, 0x2564, 0x2566, 0x2568, 0x2569,
	0x256A, 0x256B, 0x256C, 0x256D, 0x256F, 0x2573, 0x212B, 0x212C
};

// zenkaku to hankaku exception table

static int zen_to_han_exception_table[][2] = {
	{ 0x212D, 0x60 },
	{ 0x2141, 0x7E },
	{ 0x256E, 0xDC },
	{ 0x2570, 0xB2 },
	{ 0x2571, 0xB4 },
	{ 0x2575, 0xB6 },
	{ 0x2576, 0xB9 }
};

// zenkaku to hankaku map

static std::map<int, int> han_to_zen_map;

////////////////////////////////////////////////////////////////////////////////
// char

// sjis to jis

int chaos_utility::sjis_to_jis(int sjis) {
	int jis = 0x20,
		hi = (sjis >> 8) & 0xFF,
		lo = sjis & 0xFF;
	if ((hi == 0x00) &&
		((lo <= 0x7E) ||
			((lo >= 0xA0) && (lo <= 0xDF))))
	{
		jis = lo;
	} else if (
		(((hi >= 0x81) && (hi <= 0x9F)) ||
			((hi >= 0xE0) && (hi <= 0xEF))) &&
		(((lo >= 0x40) && (lo <= 0x7E)) ||
			((lo >= 0x80) && (lo <= 0xFC))))
	{
		if (hi >= 0xE0) {
			hi -= 0x40;
		}
		if (lo >= 0x80) {
			lo--;
		}
		int serial = (hi-0x81)*(0x5E*2)+(lo-0x40);
		hi = serial/(0x5E)+0x21;
		lo = (serial%0x5E)+0x21;
		jis = (hi << 8) | lo;
	}
	return jis;
}

// jis to sjis

int chaos_utility::jis_to_sjis(int jis) {
	int sjis = 0x20,
		hi = (jis >> 8) & 0xFF,
		lo = jis & 0xFF;
	if ((hi == 0x00) &&
		((lo <= 0x7E) ||
			((lo >= 0xA0) && (lo <= 0xDF))))
	{
		sjis = lo;
	} else if (
		(hi >= 0x21) && (hi <= 0x7E) &&
		(lo >= 0x21) && (lo <= 0x7E))
	{
		int serial = (hi-0x21)*(0x5E)+(lo-0x21);
		hi = serial/(0x5E*2)+0x81;
		lo = (serial%(0x5E*2))+0x40;
		if (hi >= 0xA0) {
			hi += 0x40;
		}
		if (lo >= 0x7F) {
			lo++;
		}
		sjis = (hi << 8) | lo;
	}
	return sjis;
}

// euc to jis

int chaos_utility::euc_to_jis(int euc) {
	unsigned jis = 0x20,
		hi = (euc >> 8) & 0xFF,
		lo = euc & 0xFF;
	if ((hi == 0x00) &&
		(lo <= 0x7E))
	{
		jis = lo;
	} else if (
		(hi == 0x8E) &&
		((lo >= 0xA0) && (lo <= 0xDF)))
	{
		jis = lo;
	} else if (
		(hi >= 0xA1) && (hi <= 0xFE) &&
		(lo >= 0xA1) && (lo <= 0xFE))
	{
		jis = euc & 0x7F7F;
	}
	return jis;
}

// jis to euc

int chaos_utility::jis_to_euc(int jis) {
	int euc = 0x20,
		hi = (jis >> 8) & 0xFF,
		lo = jis & 0xFF;
	if ((hi == 0x00) &&
		(lo <= 0x7E))
	{
		euc = lo;
	} else if (
		(hi == 0x00) &&
		((lo >= 0xA0) && (lo <= 0xDF)))
	{
		euc = lo | 0x8E00;
	} else if (
		(hi >= 0x21) && (hi <= 0x7E) &&
		(lo >= 0x21) && (lo <= 0x7E))
	{
		euc = jis | 0x8080;
	}
	return euc;
}

// jis hankaku to zenkaku

int chaos_utility::jis_han_to_zen(int han) {
	int zen = 0;
	if (han < 0x80) {
		if ((han >= 0x20) && (han <= 0x7E)) {
			zen = hankaku_ascii_table[han-0x20];
		}
	} else {
		if ((han >= 0xA1) && (han <= 0xDF)) {
			zen = hankaku_kana_table[han-0xA1];
		}
	}
	return zen;
}

// jis zenkaku to hankaku

int chaos_utility::jis_zen_to_han(
	int zen,
	bool* dakuten_ptr, bool* handakuten_ptr)
{
	if (han_to_zen_map.size() <= 0) {
		int i;
		for (i = 0x20; i <= 0x7E; i++) {
			han_to_zen_map.insert(
				std::make_pair(hankaku_ascii_table[i-0x20], i));
		}
		for (i = 0xA1; i <= 0xDF; i++) {
			han_to_zen_map.insert(
				std::make_pair(hankaku_kana_table[i-0xA1], i));
		}
		for (
			i = 0;
			i < sizeof(zen_to_han_exception_table)/
				sizeof(zen_to_han_exception_table[0]);
			i++)
		{
		han_to_zen_map.insert(
			std::make_pair(
				zen_to_han_exception_table[i][0],
				zen_to_han_exception_table[i][1]));
		}
	}
	int han = zen;
	bool dakuten = false, handakuten = false;
	if ((zen >= 0x2421) && (zen <= 0x2473)) {
		// hiragana to katakana
		zen = 0x2500 | (zen & 0xFF);
	}
	if ((zen >= 0x2521) && (zen <= 0x2576)) {
		if ((zen >= 0x252B) && (zen <= 0x2542)) {
			int n = zen-0x252B;
			zen = 0x252B+(n/2)*2;
			switch (n%2) {
			case 1:
				dakuten = true;
				break;
			}
		} else if ((zen >= 0x2544) && (zen <= 0x2549)) {
			int n = zen-0x2544;
			zen = 0x2544+(n/2)*2;
			switch (n%2) {
			case 1:
				dakuten = true;
				break;
			}
		} else if ((zen >= 0x254F) && (zen <= 0x255D)) {
			int n = zen-0x254F;
			zen = 0x254F+(n/3)*3;
			switch (n%3) {
			case 1:
				dakuten = true;
				break;
			case 2:
				handakuten = true;
				break;
			}
		} else if (zen == 0x2574) {
			zen = 0x2526;
			dakuten = true;
		}
	}
	std::map<int, int>::const_iterator itHankaku =
		han_to_zen_map.find(zen);
	if (itHankaku != han_to_zen_map.end()) {
		han = (*itHankaku).second;
	}
	if (dakuten_ptr != NULL) {
		*dakuten_ptr = dakuten;
	}
	if (handakuten_ptr != NULL) {
		*handakuten_ptr = handakuten;
	}
	return han;
}

////////////////////////////////////////////////////////////////////////////////
// string

// compare ignore case

int chaos_utility::icompare(
	const chaos_string& str1, const chaos_string& str2)
{
	int result = 0;
	const char* p1 = str1.c_str();
	const char* p2 = str2.c_str();
	bool sjislead = false;
	char ch1, ch2;
	do {
		ch1 = *(p1++);
		ch2 = *(p2++);
		if (!sjislead) {
			sjislead = chaos_utility::is_sjislead(ch1);
			ch1 = to_lower(ch1);
			ch2 = to_lower(ch2);
		} else {
			sjislead = false;
		}
		if (ch1 != ch2) {
			result = (ch1 < ch2)? -1: 1;
			ch1 = '\0';
		}
	} while (ch1 != '\0');
	return result;
}

// is string sjis trail character

bool chaos_utility::is_str_sjistrail(
	const chaos_string& str,
	chaos_string::size_type index)
{
	chaos_string::size_type length = str.length();
	if ((length <= 0) || (length <= index)) {
		return false;
	}
	bool sjislead = false;
	for (chaos_string::size_type i = 0; i < index; i++) {
		if (!sjislead) {
			sjislead = is_sjislead(str[i]);
		} else {
			sjislead = false;
		}
	}
	return sjislead;
}

// to upper string

chaos_string chaos_utility::to_upper_str(const chaos_string& str) {
	chaos_string new_str = str;
	bool sjislead = false;
	for (chaos_string::size_type i = 0; i < str.length(); i++) {
		char ch = str[i];
		if (!sjislead) {
			sjislead = chaos_utility::is_sjislead(ch);
			if (is_lalpha(ch)) {
				new_str[i] = to_upper(ch);
			}
		} else {
			sjislead = false;
		}
	}
	return new_str;
}

// to lower string

chaos_string chaos_utility::to_lower_str(const chaos_string& str) {
	chaos_string new_str = str;
	bool sjislead = false;
	for (chaos_string::size_type i = 0; i < str.length(); i++) {
		char ch = str[i];
		if (!sjislead) {
			sjislead = chaos_utility::is_sjislead(ch);
			if (is_ualpha(ch)) {
				new_str[i] = to_lower(ch);
			}
		} else {
			sjislead = false;
		}
	}
	return new_str;
}

// format string

chaos_string chaos_utility::format_str(const char* format, ...) {
	size_t length = 64/2;
	char* buf = NULL;
	int result;
	do {
		length *= 2;
		char* new_buf;
		if (buf == NULL) {
			new_buf = (char*)malloc(length+1);
		} else {
			new_buf = (char*)realloc(buf, length+1);
		}
		if (new_buf == NULL) {
			if (buf != NULL) {
				free(buf);
			}
			throw std::bad_alloc();
		}
		buf = new_buf;
		va_list ap;
		va_start(ap, format);
		result  = vsnprintf(buf, length, format, ap);
		va_end(ap);
		buf[length] = '\0';

#ifdef CHAOS_PLATFORM_UNIX

		if (result >= length) {
			result = -1;
		}

#endif // CHAOS_PLATFORM

	} while (result < 0);
	chaos_string str = buf;
	free(buf);
	return str;
}

// serialize string to string

chaos_string chaos_utility::str_to_str(const char* target) {
	chaos_string serial = '"';
	bool sjislead = false;
	char ch;
	while ((ch = *(target++)) != '\0') {
		if (!sjislead) {
			sjislead = is_sjislead(ch);
			switch (ch) {
			case '"':
			case '\\':
				serial += '\\';
				serial += ch;
				break;
			case '\n':
				serial += "\\n";
				break;
			case '\r':
				serial += "\\r";
				break;
			case '\t':
				serial += "\\t";
				break;
			default:
				if (ch >= 0x20) {
					serial += ch;
				} else {
					serial += format_str("\\x%02X", (unsigned char)ch);
				}
				break;
			}
		} else {
			sjislead = false;
			serial += ch;
		}
	}
	serial += '"';
	return serial;
}

// serialize string to string(designate quote character)

chaos_string chaos_utility::str_to_str(const char* target, char quote) {
	chaos_string serial = quote;
	bool sjislead = false;
	char ch;
	while ((ch = *(target++)) != '\0') {
		if (!sjislead) {
			sjislead = is_sjislead(ch);
			switch (ch) {
			case '\\':
				serial += '\\';
				serial += ch;
				break;
			case '\n':
				serial += "\\n";
				break;
			case '\r':
				serial += "\\r";
				break;
			case '\t':
				serial += "\\t";
				break;
			default:
				if (ch == quote) {
					serial += '\\';
					serial += ch;
				} else if (ch >= 0x20) {
					serial += ch;
				} else {
					serial += format_str("\\x%02X", (unsigned char)ch);
				}
				break;
			}
		} else {
			sjislead = false;
			serial += ch;
		}
	}
	serial += quote;
	return serial;
}

// serialize string from string

chaos_string chaos_utility::str_from_str(const char*& serial) {
	chaos_string result;
	char quote = *serial;
	if ((quote != '"') && (quote != '\'')) {
		return chaos_string();
	}
	serial++;
	bool sjislead = false, escape = false;
	int next;
	char ch;
	int code;
	while ((ch = *serial) != '\0') {
		serial++;
		if (!sjislead) {
			sjislead = is_sjislead(ch);
			if (!escape) {
				if (ch == quote) {
					break;
				} else if (ch == '\\') {
					escape = true;
				} else {
					result += ch;
				}
			} else {
				escape = false;
				switch (ch) {
				case 'n':
					result += '\n';
					break;
				case 'r':
					result += '\r';
					break;
				case 't':
					result += '\t';
					break;
				case 'x':
					code = 0;
					next = 0;
					sscanf(serial, "%2x%n", &code, &next);
					serial += next;
					result += (char)code;
					break;
				default:
					result += ch;
					break;
				}
			}
		} else {
			sjislead = false;
			result += ch;
		}
	}
	return result;
}

// find string

int chaos_utility::str_find_str(
	const char* target, const char* keyword,
	int startofs)
{
	size_t keyword_length = 0;
	int keyword_sum = 0;
	const char* s  = keyword;
	char ch;
	while ((ch = *s) != '\0') {
		s++;
		keyword_length++;
		keyword_sum += (unsigned char)ch;
	}
	if (keyword_length <= 0) {
		return 0;
	}
	const char* first = target;
	int i;
	for (i = 0; i < startofs; i++) {
		ch = *(first++);
		if (ch == '\0') {
			return -1;
		}
	}
	const char* last = first;
	int target_sum = 0;
	while ((size_t)(last-first) < keyword_length) {
		ch = *(last++);
		if (ch == '\0') {
			return -1;
		}
		target_sum += (unsigned char)ch;
	}
	while (true) {
		if (target_sum == keyword_sum) {
			if (strncmp(first, keyword, keyword_length) == 0) {
				break;
			}
		}
		int step = is_sjislead(*first)? 2: 1;
		for (i = 0; i < step; i++) {
			ch = *(first++);
			target_sum -= (unsigned char)ch;
			ch = *(last++);
			if (ch == '\0') {
				return -1;
			}
			target_sum += (unsigned char)ch;
		}
	}
	return first-target;
}

#ifdef CHAOS_API_WIN32

// utf-8 string to wchar_t string

void chaos_utility::utf8_to_wstr(
	std::vector<wchar_t>& wstr,
	const chaos_string& utf8str)
{
	wstr.clear();
	if (utf8str.length() <= 0) {
		return;
	}
	wstr.reserve(utf8str.length());
	const char* src = utf8str.c_str();
	const char* src_last = src+utf8str.length();
	while (src != src_last) {
		chaos_uint8 bt = *(src++);
		chaos_uint32 uch = 0, minimum = 0;
		int trail = 0;
		if (bt < 0x80) {
			uch = bt;
		} else if (bt < 0xC0) {
			// trail byte
		} else if (bt < 0xE0) {
			minimum = 0x80U;
			uch = bt & 0x1F;
			trail = 1;
		} else if (bt < 0xF0) {
			minimum = 0x800U;
			uch = bt & 0x0F;
			trail = 2;
		} else if (bt < 0xF8) {
			minimum = 0x10000U;
			uch = bt & 0x07;
			trail = 3;
		} else if (bt < 0xFC) {
			minimum = 0x200000U;
			uch = bt & 0x03;
			trail = 4;
		} else if (bt < 0xFE) {
			minimum = 0x4000000U;
			uch = bt & 0x01;
			trail = 5;
		} else {
			// unknown
		}
		while ((trail > 0) && (src != src_last)) {
			uch = (uch << 6) | ((chaos_uint8)*(src++) & 0x3F);
			trail--;
		}
		if (uch < minimum) {
			uch = 0;
		}
		if (uch < 0x10000U) {
			wstr.push_back(
				(wchar_t)uch);
		} else if (uch < 0x110000U) {
			wstr.push_back(
				(wchar_t)((((uch-0x10000U) >> 10) & 0x03FFU) | 0xD800U));
			wstr.push_back(
				(wchar_t)(((uch-0x10000U) & 0x03FFU) | 0xDC00U));
		} else {
			// out of range
		}
	}
}

// wchar_t string to utf-8 string

void chaos_utility::wstr_to_utf8(
	chaos_string& utf8str,
	const std::vector<wchar_t>& wstr)
{
	utf8str = "";
	size_t length = wstr.size();
	if (length <= 0) {
		return;
	}
	utf8str.reserve(length*3);
	const wchar_t* src = &wstr[0];
	for (size_t i = 0; i < length; i++) {
		chaos_uint32 uch = *src++;
		if ((uch >= 0xD800U) && (uch <= 0xDBFFU)) {
			// surrogate pair
			if (i < length+1) {
				i++;
				chaos_uint32 uch2 = *src++;
				if ((uch2 >= 0xDC00U) && (uch2 <= 0xDFFFU)) {
					uch = 0x10000U+
						((uch & 0x03FFU) << 10)+
						(uch2 & 0x03FFU);
				} else {
					// surrogate-low error
					uch = (chaos_uint32)'?';
				}
			} else {
				// surrogate pair error
				uch = (chaos_uint32)'?';
			}
		}
		if (uch < 0x80U) {
			utf8str += (char)uch;
		} else if (uch < 0x800U) {
			utf8str += (char)(((uch >> 6) & 0x3F) | 0xC0);
			utf8str += (char)((uch & 0x3F) | 0x80);
		} else if (uch < 0x10000U) {
			utf8str += (char)(((uch >> 12) & 0x3F) | 0xE0);
			utf8str += (char)(((uch >> 6) & 0x3F) | 0x80);
			utf8str += (char)((uch & 0x3F) | 0x80);
		} else if (uch < 0x200000U) {
			utf8str += (char)(((uch >> 18) & 0x3F) | 0xF0);
			utf8str += (char)(((uch >> 12) & 0x3F) | 0x80);
			utf8str += (char)(((uch >> 6) & 0x3F) | 0x80);
			utf8str += (char)((uch & 0x3F) | 0x80);
		} else if (uch < 0x4000000U) {
			utf8str += (char)(((uch >> 24) & 0x3F) | 0xF8);
			utf8str += (char)(((uch >> 18) & 0x3F) | 0x80);
			utf8str += (char)(((uch >> 12) & 0x3F) | 0x80);
			utf8str += (char)(((uch >> 6) & 0x3F) | 0x80);
			utf8str += (char)((uch & 0x3F) | 0x80);
		} else if (uch < 0x8000000U) {
			utf8str += (char)(((uch >> 30) & 0x3F) | 0xFC);
			utf8str += (char)(((uch >> 24) & 0x3F) | 0x80);
			utf8str += (char)(((uch >> 18) & 0x3F) | 0x80);
			utf8str += (char)(((uch >> 12) & 0x3F) | 0x80);
			utf8str += (char)(((uch >> 6) & 0x3F) | 0x80);
			utf8str += (char)((uch & 0x3F) | 0x80);
		} else {
			utf8str += '?';
		}
	}
}

#endif // CHAOS_API

// sjis string to euc string

chaos_string chaos_utility::sjis_to_euc(const chaos_string& sjisstr) {
	chaos_string eucstr;
	for (size_t i = 0; i < sjisstr.length(); i++) {
		int sjis = (chaos_uint8)sjisstr[i];
		if (is_sjislead(sjis)) {
			sjis = (sjis << 8)+(chaos_uint8)sjisstr[i+1];
		}
		int euc = sjis_to_euc(sjis);
		if ((euc != sjis) || !eucstr.empty()) {
			if ((i > 0) && eucstr.empty()) {
				eucstr = sjisstr.substr(0, i);
			}
			if (euc > 0xFF) {
				eucstr += (char)((euc >> 8) & 0xFF);
			}
			eucstr += (char)(euc & 0xFF);
		}
		if (sjis > 0xFF) {
			i++;
		}
	}
	if (eucstr.empty()) {
		eucstr = sjisstr;
	}
	return eucstr;
}

// euc string to sjis string

chaos_string chaos_utility::euc_to_sjis(const chaos_string& eucstr) {
	chaos_string sjisstr;
	for (size_t i = 0; i < eucstr.length(); i++) {
		int euc = (chaos_uint8)eucstr[i];
		if (is_euclead(euc)) {
			euc = (euc << 8)+(chaos_uint8)eucstr[i+1];
		}
		int sjis = euc_to_sjis(euc);
		if ((sjis != euc) || !sjisstr.empty()) {
			if ((i > 0) && sjisstr.empty()) {
				sjisstr = eucstr.substr(0, i);
			}
			if (sjis > 0xFF) {
				sjisstr += (char)((sjis >> 8) & 0xFF);
			}
			sjisstr += (char)(sjis & 0xFF);
		}
		if (euc > 0xFF) {
			i++;
		}
	}
	if (sjisstr.empty()) {
		sjisstr = eucstr;
	}
	return sjisstr;
}

// sjis string to utf-8 string

chaos_string chaos_utility::sjis_to_utf8(const chaos_string& sjisstr) {
	if (sjisstr.empty()) {
		return "";
	}
	chaos_string utf8str;

#ifdef CHAOS_API_WIN32

	int length = MultiByteToWideChar(
		CP_OEMCP, MB_PRECOMPOSED,
		sjisstr.c_str(), sjisstr.length(),
		NULL, 0);
	if (length <= 0) {
		return "";
	}
	std::vector<wchar_t> wstr;
	wstr.resize(length, 0);
	MultiByteToWideChar(
		CP_OEMCP, MB_PRECOMPOSED,
		sjisstr.c_str(), sjisstr.length(),
		&wstr[0], length);
	wstr_to_utf8(utf8str, wstr);

#elif defined(CHAOS_API_GTK)

	gsize written = 0;
	char* utf8 = g_convert(
		sjisstr.c_str(), sjisstr.length(),
		"UTF-8", "CP932",
		NULL, &written, NULL);
	if (utf8 != NULL) {
		utf8str = chaos_string(utf8, utf8+written);
		g_free(utf8);
	}

#endif // CHAOS_API

	return utf8str;
}

// utf-8 string to sjis string

chaos_string chaos_utility::utf8_to_sjis(const chaos_string& utf8str) {
	if (utf8str.empty()) {
		return "";
	}
	chaos_string sjisstr;

#ifdef CHAOS_API_WIN32

	std::vector<wchar_t> wstr;
	utf8_to_wstr(wstr, utf8str);
	if (wstr.size() <= 0) {
		return "";
	}
	int length = WideCharToMultiByte(
		CP_OEMCP, 0,
		&wstr[0], wstr.size(),
		NULL, 0,
		NULL, NULL);
	if (length <= 0) {
		return "";
	}
	chaos_auto_array<char> sjis(new char[length]);
	if (sjis.get() == NULL) {
		return "";
	}
	WideCharToMultiByte(
		CP_OEMCP, 0,
		&wstr[0], wstr.size(),
		sjis.get(), length,
		NULL, NULL);
	sjisstr = chaos_string(sjis.get(), sjis.get()+length);

#elif defined(CHAOS_API_GTK)

	gsize written = 0;
	char* sjis = g_convert(
		utf8str.c_str(), utf8str.length(),
		"CP932", "UTF-8",
		NULL, &written, NULL);
	if (sjis != NULL) {
		sjisstr = chaos_string(sjis, sjis+written);
		g_free(sjis);
	}

#endif // CHAOS_API

	return sjisstr;
}

// utf-16le string to sjis string

chaos_string chaos_utility::utf16le_to_sjis(const chaos_string& utf16lstr) {
	if (utf16lstr.length() <= 1) {
		return "";
	}
	chaos_string sjisstr;

#ifdef CHAOS_API_WIN32

	std::vector<wchar_t> wstr;
	wstr.reserve(utf16lstr.length()/2);
	for (size_t i = 0; i+1 < utf16lstr.length(); i += 2) {
		wstr.push_back(
			(wchar_t)(
				(chaos_uint8)utf16lstr[i]+
				((chaos_uint8)utf16lstr[i+1])*256));
	}
	int length = WideCharToMultiByte(
		CP_OEMCP, 0,
		&wstr[0], wstr.size(),
		NULL, 0,
		NULL, NULL);
	if (length <= 0) {
		return "";
	}
	chaos_auto_array<char> sjis(new char[length]);
	if (sjis.get() == NULL) {
		return "";
	}
	WideCharToMultiByte(
		CP_OEMCP, 0,
		&wstr[0], wstr.size(),
		sjis.get(), length,
		NULL, NULL);
	sjisstr = chaos_string(sjis.get(), sjis.get()+length);

#elif defined(CHAOS_API_GTK)

	gsize written = 0;
	char* sjis = g_convert(
		utf16lstr.c_str(), utf16lstr.length(),
		"CP932", "UTF-16LE",
		NULL, &written, NULL);
	if (sjis != NULL) {
		sjisstr = chaos_string(sjis, sjis+written);
		g_free(sjis);
	}

#endif // CHAOS_API

	return sjisstr;
}

// utf-16be string to sjis string

chaos_string chaos_utility::utf16be_to_sjis(const chaos_string& utf16bstr) {
	if (utf16bstr.length() <= 1) {
		return "";
	}
	chaos_string sjisstr;

#ifdef CHAOS_API_WIN32

	std::vector<wchar_t> wstr;
	wstr.reserve(utf16bstr.length()/2);
	for (size_t i = 0; i+1 < utf16bstr.length(); i += 2) {
		wstr.push_back(
			(wchar_t)(
				((chaos_uint8)utf16bstr[i])*256+
				(chaos_uint8)utf16bstr[i+1]));
	}
	int length = WideCharToMultiByte(
		CP_OEMCP, 0,
		&wstr[0], wstr.size(),
		NULL, 0,
		NULL, NULL);
	if (length <= 0) {
		return "";
	}
	chaos_auto_array<char> sjis(new char[length]);
	if (sjis.get() == NULL) {
		return "";
	}
	WideCharToMultiByte(
		CP_OEMCP, 0,
		&wstr[0], wstr.size(),
		sjis.get(), length,
		NULL, NULL);
	sjisstr = chaos_string(sjis.get(), sjis.get()+length);

#elif defined(CHAOS_API_GTK)

	gsize written = 0;
	char* sjis = g_convert(
		utf16bstr.c_str(), utf16bstr.length(),
		"CP932", "UTF-16BE",
		NULL, &written, NULL);
	if (sjis != NULL) {
		sjisstr = chaos_string(sjis, sjis+written);
		g_free(sjis);
	}

#endif // CHAOS_API

	return sjisstr;
}

// source string to sjis string

chaos_string chaos_utility::source_to_sjis(const chaos_string& srcstr) {

#ifdef CHAOS_SOURCE_CODE_SJIS

	return srcstr;

#elif defined(CHAOS_SOURCE_CODE_EUC)

	return euc_to_sjis(srcstr);

#elif defined(CHAOS_SOURCE_CODE_UTF8)

	return utf8_to_sjis(srcstr);

#endif // CHAOS_SOURCE_CODE

}

// sjis string to locale string

chaos_string chaos_utility::sjis_to_locale(const chaos_string& sjisstr) {

#ifdef CHAOS_LOCALE_CODE_SJIS

	return sjisstr;

#elif defined(CHAOS_LOCALE_CODE_EUC)

	return sjis_to_euc(sjisstr);

#elif defined(CHAOS_LOCALE_CODE_UTF8)

	return sjis_to_utf8(sjisstr);

#elif defined(CHAOS_LOCALE_CODE_GCONV)

	chaos_string locstr;
	chaos_string utf8str = sjis_to_utf8(sjisstr);
	gsize written = 0;
	char* loc = g_locale_from_utf8(
		utf8str.c_str(), utf8str.length(),
		NULL, &written, NULL);
	if (loc != NULL) {
		locstr = chaos_string(loc, loc+written);
		g_free(loc);
	}
	return locstr;

#endif // CHAOS_LOCALE_CODE

}

// locale string to sjis string

chaos_string chaos_utility::locale_to_sjis(const chaos_string& locstr) {

#ifdef CHAOS_LOCALE_CODE_SJIS

	return locstr;

#elif defined(CHAOS_LOCALE_CODE_EUC)

	return euc_to_sjis(locstr);

#elif defined(CHAOS_LOCALE_CODE_UTF8)

	return utf8_to_sjis(locstr);

#elif defined(CHAOS_LOCALE_CODE_GCONV)

	chaos_string sjisstr;
	gsize written = 0;
	char* utf8 = g_locale_to_utf8(
		locstr.c_str(), locstr.length(),
		NULL, &written, NULL);
	if (utf8 != NULL) {
		sjisstr = utf8_to_sjis(chaos_string(utf8, utf8+written));
		g_free(utf8);
	}
	return sjisstr;

#endif // CHAOS_LOCALE_CODE

}

// sjis string to filename string

chaos_string chaos_utility::sjis_to_filename(const chaos_string& sjisstr) {

#ifdef CHAOS_FILENAME_CODE_SJIS

	return sjisstr;

#elif defined(CHAOS_FILENAME_CODE_EUC)

	return sjis_to_euc(sjisstr);

#elif defined(CHAOS_FILENAME_CODE_UTF8)

	return sjis_to_utf8(sjisstr);

#elif defined(CHAOS_FILENAME_CODE_GCONV)

	chaos_string filestr;
	chaos_string utf8str = sjis_to_utf8(sjisstr);
	gsize written = 0;
	char* file = g_filename_from_utf8(
		utf8str.c_str(), utf8str.length(),
		NULL, &written, NULL);
	if (file != NULL) {
		filestr = chaos_string(file, file+written);
		g_free(file);
	}
	return filestr;

#endif // CHAOS_FILENAME_CODE

}

// filename string to sjis string

chaos_string chaos_utility::filename_to_sjis(const chaos_string& filestr) {

#ifdef CHAOS_FILENAME_CODE_SJIS

	return filestr;

#elif defined(CHAOS_FILENAME_CODE_EUC)

	return euc_to_sjis(filestr);

#elif defined(CHAOS_FILENAME_CODE_UTF8)

	return utf8_to_sjis(filestr);

#elif defined(CHAOS_FILENAME_CODE_GCONV)

	chaos_string sjisstr;
	gsize written = 0;
	char* utf8 = g_filename_to_utf8(
		filestr.c_str(), filestr.length(),
		NULL, &written, NULL);
	if (utf8 != NULL) {
		sjisstr = utf8_to_sjis(chaos_string(utf8, utf8+written));
		g_free(utf8);
	}
	return sjisstr;

#endif // CHAOS_FILENAME_CODE

}

// sjis string to gui string

chaos_string chaos_utility::sjis_to_gui(const chaos_string& sjisstr) {

#ifdef CHAOS_GUI_CODE_SJIS

	return sjisstr;

#elif defined(CHAOS_GUI_CODE_EUC)

	return sjis_to_euc(sjisstr);

#elif defined(CHAOS_GUI_CODE_UTF8)

	return sjis_to_utf8(sjisstr);

#endif // CHAOS_GUI_CODE

}

// gui string to sjis string

chaos_string chaos_utility::gui_to_sjis(const chaos_string& guistr) {

#ifdef CHAOS_GUI_CODE_SJIS

	return guistr;

#elif defined(CHAOS_GUI_CODE_EUC)

	return euc_to_sjis(guistr);

#elif defined(CHAOS_GUI_CODE_UTF8)

	return utf8_to_sjis(guistr);

#endif // CHAOS_GUI_CODE

}

////////////////////////////////////////////////////////////////////////////////
// xml

// serialize string to xml string

chaos_string chaos_utility::str_to_xmlstr(const char* target) {
	chaos_string serial;
	bool sjislead = false;
	int ch;
	while ((ch = (unsigned char)*(target++)) != '\0') {
		if (!sjislead) {
			sjislead = is_sjislead(ch);
			switch (ch) {
			case '"':
				serial += "&quot;";
				break;
			case '&':
				serial += "&amp;";
				break;
			case '\'':
				serial += "&apos;";
				break;
			case '<':
				serial += "&lt;";
				break;
			case '>':
				serial += "&gt;";
				break;
			default:
				if (ch >= 0x20) {
					serial += (char)ch;
				} else {
					serial += format_str("&#x%02X;", ch);
				}
				break;
			}
		} else {
			sjislead = false;
			serial += (char)ch;
		}
	}
	return serial;
}

// serialize string from xml string

chaos_string chaos_utility::str_from_xmlstr(const char*& serial) {
	chaos_string result;
	bool sjislead = false;
	int ch;
	while ((ch = (unsigned char)*serial) != '\0') {
		serial++;
		if (!sjislead) {
			sjislead = is_sjislead(ch);
			if (ch == '&') {
				if (*serial == '#') {
					serial++;
					int next = 0;
					if (*serial == 'x') {
						sscanf(serial, "x%x;%n", &ch, &next);
						serial += next;
					} else {
						sscanf(serial, "%d;%n", &ch, &next);
						serial += next;
					}
					result += (char)ch;
				} else if (strncmp(serial, "quot;", 5) == 0) {
					serial += 5;
					result += '"';
				} else if (strncmp(serial, "amp;", 4) == 0) {
					serial += 4;
					result += '&';
				} else if (strncmp(serial, "apos;", 5) == 0) {
					serial += 5;
					result += '\'';
				} else if (strncmp(serial, "lt;", 3) == 0) {
					serial += 3;
					result += '<';
				} else if (strncmp(serial, "gt;", 3) == 0) {
					serial += 3;
					result += '>';
				} else {
					while ((ch = (unsigned char)*serial) != '\0') {
						if (ch == ';') {
							break;
						}
					}
				}
			} else {
				result += (char)ch;
			}
		} else {
			sjislead = false;
			result += (char)ch;
		}
	}
	return result;
}

// get xml tag name

chaos_string chaos_utility::get_xml_tag_name(
	const chaos_string& tag)
{
	ASSERT(tag.length() >= 2);
	ASSERT(tag.front() == '<');
	ASSERT(tag.back() == '>');
	const char* tagstr = tag.c_str();
	if (*tagstr == '<') {
		tagstr++;
	}
	if (*tagstr == '/') {
		tagstr++;
	}
	chaos_string tag_name;
	bool sjislead = false;
	char ch;
	while ((ch = *tagstr) != '\0') {
		if (sjislead) {
			tag_name += ch;
			tagstr++;
			sjislead = false;
		} else if (!is_space(ch)) {
			if ((ch == '>') || (ch == '/')) {
				break;
			}
			tag_name += ch;
			tagstr++;
			sjislead = is_sjislead(ch);
		} else {
			break;
		}
	}
	return tag_name;
}

// get xml tag attribute value

chaos_string chaos_utility::get_xml_tag_attr_val(
	const chaos_string& tag,
	const chaos_string& attr_name)
{
	ASSERT(tag.length() >= 2);
	ASSERT(tag.front() == '<');
	ASSERT(tag.back() == '>');
	const char* tagstr = tag.c_str();
	if (*tagstr == '<') {
		tagstr++;
	}
	if (*tagstr == '/') {
		tagstr++;
	}
	bool sjislead = false;
	char ch;
	while ((ch = *tagstr) != '\0') {
		if (sjislead) {
			tagstr++;
			sjislead = false;
		} else if (!is_space(ch)) {
			tagstr++;
			if ((ch == '>') || (ch == '/')) {
				return "";
			}
			sjislead = is_sjislead(ch);
		} else {
			break;
		}
	}
	while (is_space(*tagstr)) {
		tagstr++;
	}
	while (true) {
		ch = *tagstr;
		if ((ch == '\0') ||
			(ch == '>') || (ch == '/'))
		{
			return "";
		}
		if ((strncmp(tagstr, attr_name.c_str(), attr_name.length()) == 0) &&
			(((ch = *(tagstr+attr_name.length())) == '=') ||
				is_space(ch)))
		{
			break;
		}
		while ((ch = *tagstr) != '\0') {
			if (sjislead) {
				tagstr++;
				sjislead = false;
			} else if (!is_space(ch) && (ch != '=')) {
				tagstr++;
				if ((ch == '>') || (ch == '/')) {
					return "";
				}
				sjislead = is_sjislead(ch);
			} else {
				break;
			}
		}
		while (is_space(*tagstr)) {
			tagstr++;
		}
		if (*tagstr == '=') {
			tagstr++;
			while (is_space(*tagstr)) {
				tagstr++;
			}
		}
		if (*tagstr == '"') {
			tagstr++;
			while ((ch = *tagstr) != '\0') {
				tagstr++;
				if (ch == '"') {
					break;
				}
			}
			while (is_space(*tagstr)) {
				tagstr++;
			}
		} else if (*tagstr == '\'') {
			tagstr++;
			while ((ch = *tagstr) != '\0') {
				tagstr++;
				if (ch == '\'') {
					break;
				}
			}
			while (is_space(*tagstr)) {
				tagstr++;
			}
		}
	}
	tagstr += attr_name.length();
	while (is_space(*tagstr)) {
		tagstr++;
	}
	if (*tagstr != '=') {
		return "";
	}
	tagstr++;
	while (is_space(*tagstr)) {
		tagstr++;
	}
	chaos_string attr_val;
	if (*tagstr == '"') {
		tagstr++;
		while ((ch = *tagstr) != '\0') {
			if (ch == '"') {
				break;
			}
			tagstr++;
			attr_val += ch;
		}
		if (ch != '"') {
			return "";
		}
	} else if (*tagstr == '\'') {
		tagstr++;
		while ((ch = *tagstr) != '\0') {
			if (ch == '\'') {
				break;
			}
			tagstr++;
			attr_val += ch;
		}
		if (ch != '\'') {
			return "";
		}
	} else {
		return "";
	}
	return str_from_xmlstr(attr_val);
}
