////////////////////////////////////////////////////////////////////////////////
// chaos_xml.cpp : xml data

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_xml.h"
#include "chaos_exception.h"
#include "chaos_utility.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_xml : xml data class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_xml::chaos_xml() :
	values(NULL)
{
	values = new chaos_value_manager<chaos_string>;
}

// destructor

chaos_xml::~chaos_xml() {
	delete values;
}

// clear

void chaos_xml::clear() {
	xml_kind = "";
	xml_date = "";
	xml_time = "";
	xml_info = "";
	values_version = "";
	values->clear();
}

////////////////////////////////////////////////////////////////////////////////
// operation

// encoding check

bool chaos_xml::encoding_check(
	int accept_encoding,
	int& detect_encoding,
	const chaos_string& encoding)
{
	detect_encoding = ENCODING_UNKNOWN;
	bool result = false;
	switch (accept_encoding) {
	case ENCODING_UNKNOWN:
		if (chaos_utility::icompare(encoding, "Shift_JIS") == 0) {
			result = true;
		} else if (
			encoding.empty() ||
			(chaos_utility::icompare(encoding, "UTF-8") == 0))
		{
			detect_encoding = ENCODING_UTF8;
			result = true;
		} else if (
			chaos_utility::icompare(encoding, "EUC-JP") == 0)
		{
			detect_encoding = ENCODING_EUC;
			result = true;
		}
		break;
	case ENCODING_UTF8:
	case ENCODING_UTF8_BOM:
		if (encoding.empty() ||
			(chaos_utility::icompare(encoding, "UTF-8") == 0))
		{
			result = true;
		}
		break;
	case ENCODING_UTF16LE:
	case ENCODING_UTF16BE:
	case ENCODING_UTF16LE_BOM:
	case ENCODING_UTF16BE_BOM:
		if (encoding.empty() ||
			(chaos_utility::icompare(encoding, "UTF-16") == 0))
		{
			result = true;
		}
		break;
	case ENCODING_EUC:
		if (chaos_utility::icompare(encoding, "EUC-JP") == 0) {
			result = true;
		}
		break;
	}
	return result;
}

// load in encoding

bool chaos_xml::load_in_encoding(
	chaos_file_data* filedata,
	int accept_encoding,
	int& detect_encoding)
{
	detect_encoding = ENCODING_UNKNOWN;
	if (accept_encoding == ENCODING_UNKNOWN) {
		size_t pos = filedata->tell();
		chaos_uint8 head[3];
		memset(head, 0, sizeof(head));
		if (filedata->read(head, sizeof(head)) == sizeof(head)) {
			if (memcmp(head, "\xEF\xBB\xBF", 3) == 0) {
				detect_encoding = ENCODING_UTF8_BOM;
				return false;
			} else if (memcmp(head, "\xFF\xFE", 2) == 0) {
				detect_encoding = ENCODING_UTF16LE_BOM;
				return false;
			} else if (memcmp(head, "\xFE\xFF", 2) == 0) {
				detect_encoding = ENCODING_UTF16BE_BOM;
				return false;
			} else if (memcmp(head, "\x3C\x00\x3F", 3) == 0) {
				detect_encoding = ENCODING_UTF16LE;
				return false;
			} else if (memcmp(head, "\x00\x3C\x00", 3) == 0) {
				detect_encoding = ENCODING_UTF16BE;
				return false;
			}
		}
		filedata->seek(pos, chaos_file_data::seek_set);
	}
	clear();
	std::list<chaos_string> tags;
	chaos_string data;
	enum state_type {
		st_unknown,
		st_start,
		st_top,
		st_chaos,
		st_kind,
		st_date,
		st_time,
		st_info,
		st_values,
		st_value
	};
	std::stack< int, std::list<int> > states;
	states.push(st_start);
	chaos_string name;
	chaos_value value;
	bool sjislead = false;
	int ch;
	bool result = true;
	try {
		while ((ch = filedata->get()) != -1) {
			if (sjislead || (ch != '<')) {
				if (sjislead) {
					sjislead = false;
				} else {
					sjislead = chaos_utility::is_sjislead(ch);
				}
				switch ((state_type)states.top()) {
				case st_kind:
				case st_date:
				case st_time:
				case st_info:
				case st_value:
					data += (char)ch;
					break;
				}
				continue;
			}
			chaos_string tag = (char)ch;
			bool empty_tag = false;
			while ((ch = filedata->get()) != -1) {
				if (sjislead) {
					sjislead = false;
					tag += (char)ch;
				} else {
					sjislead = chaos_utility::is_sjislead(ch);
					tag += (char)ch;
					if (ch == '>') {
						break;
					}
					empty_tag = (ch == '/');
				}
			}
			if (ch != '>') {
				throw chaos_exception(chaos_exception::bad_format);
			}
			bool close_tag = (tag[1] == '/');
			chaos_string tag_name = chaos_utility::get_xml_tag_name(tag);
			switch ((state_type)states.top()) {
			case st_start:
				if (empty_tag || close_tag ||
					(tag_name != "?xml") ||
					(tag[tag.length()-2] != '?') ||
					(chaos_utility::get_xml_tag_attr_val(
						tag, "version") != "1.0"))
				{
					throw chaos_exception(chaos_exception::bad_format);
				}
				if (!encoding_check(
						accept_encoding,
						detect_encoding,
						chaos_utility::get_xml_tag_attr_val(
							tag, "encoding")))
				{
					throw chaos_exception(chaos_exception::bad_format);
				}
				if (detect_encoding != ENCODING_UNKNOWN) {
					return false;
				}
				states.pop();
				states.push((int)st_top);
				break;
			case st_top:
				if (empty_tag || close_tag ||
					(tag_name != "chaos"))
				{
					throw chaos_exception(chaos_exception::bad_format);
				}
				states.pop();
				states.push((int)st_chaos);
				break;
			case st_chaos:
				if (close_tag) {
					states.pop();
				} else if (tag_name == "kind") {
					if (empty_tag) {
						xml_kind = "";
					} else {
						states.push((int)st_kind);
						data = "";
					}
				} else if (tag_name == "date") {
					if (empty_tag) {
						xml_date = "";
					} else {
						states.push((int)st_date);
						data = "";
					}
				} else if (tag_name == "time") {
					if (empty_tag) {
						xml_time = "";
					} else {
						states.push((int)st_time);
						data = "";
					}
				} else if (tag_name == "info") {
					if (empty_tag) {
						xml_info = "";
					} else {
						states.push((int)st_info);
						data = "";
					}
				} else if (tag_name == "values") {
					values_version = chaos_utility::get_xml_tag_attr_val(
						tag, "version");
					if (empty_tag) {
						// nothing
					} else {
						states.push((int)st_values);
					}
				} else {
					if (!empty_tag) {
						states.push((int)st_unknown);
					}
				}
				break;
			case st_kind:
				if (close_tag) {
					states.pop();
					xml_kind = chaos_utility::str_from_xmlstr(data);
				} else {
					if (!empty_tag) {
						states.push((int)st_unknown);
					}
				}
				break;
			case st_date:
				if (close_tag) {
					states.pop();
					xml_date = chaos_utility::str_from_xmlstr(data);
				} else {
					if (!empty_tag) {
						states.push((int)st_unknown);
					}
				}
				break;
			case st_time:
				if (close_tag) {
					states.pop();
					xml_time = chaos_utility::str_from_xmlstr(data);
				} else {
					if (!empty_tag) {
						states.push((int)st_unknown);
					}
				}
				break;
			case st_info:
				if (close_tag) {
					states.pop();
					xml_info = chaos_utility::str_from_xmlstr(data);
				} else {
					if (!empty_tag) {
						states.push((int)st_unknown);
					}
				}
				break;
			case st_values:
				if (close_tag) {
					states.pop();
				} else if (tag_name == "value") {
					name = chaos_utility::get_xml_tag_attr_val(
						tag, "name");
					chaos_string types =
						chaos_utility::get_xml_tag_attr_val(
							tag, "type");
					if (types == "integer") {
						value = chaos_value((chaos_int32)0);
					} else if (types == "boolean") {
						value = chaos_value(false);
					} else if (types == "color") {
						value = chaos_value(chaos_color());
					} else if (types == "string") {
						value = chaos_value(chaos_string(""));
					} else if (types == "nul") {
						value = chaos_value();
					} else {
						// unknown type
						value = chaos_value(chaos_string(""));
					}
					if (empty_tag) {
						values->set_value(name, value);
					} else {
						states.push((int)st_value);
						data = "";
					}
				} else {
					if (!empty_tag) {
						states.push((int)st_unknown);
					}
				}
				break;
			case st_value:
				if (close_tag) {
					states.pop();
					switch (value.get_type()) {
					case chaos_value::string:
						value = chaos_value(
							chaos_utility::str_from_xmlstr(data));
						break;
					default:
						value = chaos_value::from_string(
							chaos_utility::str_from_xmlstr(data));
						break;
					}
					values->set_value(name, value);
				} else {
					if (!empty_tag) {
						states.push((int)st_unknown);
					}
				}
				break;
			}
			if (states.size() <= 0) {
				break;
			}
		}
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : xml::load - "
			"%s\n",
			e.what());
		result = false;
	}
	return result;
}

// load

bool chaos_xml::load(chaos_file_data* filedata) {
	int accept_encoding = ENCODING_UNKNOWN,
		detect_encoding = ENCODING_UNKNOWN;
	bool result = load_in_encoding(
			filedata,
			accept_encoding,
			detect_encoding);
	if (!result && (detect_encoding != ENCODING_UNKNOWN)) {
		size_t ofs = 0;
		switch (detect_encoding) {
		case ENCODING_UTF8_BOM:
			ofs = 3;
			break;
		case ENCODING_UTF16LE_BOM:
		case ENCODING_UTF16BE_BOM:
			ofs = 2;
			break;
		}
		filedata->seek(ofs, chaos_file_data::seek_set);
		chaos_buffer buf;
		buf.resize(filedata->size()-ofs);
		filedata->read(buf.data(), buf.size());
		chaos_string src(
			(const char*)buf.data(),
			(const char*)buf.data()+buf.size()),
			dst;
		switch (detect_encoding) {
		case ENCODING_EUC:
			dst = chaos_utility::euc_to_sjis(src);
			break;
		case ENCODING_UTF8:
		case ENCODING_UTF8_BOM:
			dst = chaos_utility::utf8_to_sjis(src);
			break;
		case ENCODING_UTF16LE:
		case ENCODING_UTF16LE_BOM:
			dst = chaos_utility::utf16le_to_sjis(src);
			break;
		case ENCODING_UTF16BE:
		case ENCODING_UTF16BE_BOM:
			dst = chaos_utility::utf16be_to_sjis(src);
			break;
		default:
			dst = src;
			break;
		}
		chaos_file_data filedata2;
		filedata2.write(dst.c_str(), dst.size());
		filedata2.seek(0, chaos_file_data::seek_set);
		accept_encoding = detect_encoding;
		detect_encoding = ENCODING_UNKNOWN;
		result = load_in_encoding(
			&filedata2,
			accept_encoding,
			detect_encoding);
	}
	return result;
}

// save

bool chaos_xml::save(chaos_file_data* filedata) const {
	bool result = true;
	try {
		filedata->put_line(
			"<?xml version=\"1.0\" encoding=\"Shift_JIS\"?>\r\n");
		filedata->put_line(
			"<chaos>\r\n");
		if (!xml_kind.empty()) {
			filedata->put_line(
				"<kind>"+
				chaos_utility::str_to_xmlstr(xml_kind)+
				"</kind>\r\n");
		}
		if (!xml_date.empty()) {
			filedata->put_line(
				"<date>"+
				chaos_utility::str_to_xmlstr(xml_date)+
				"</date>\r\n");
		}
		if (!xml_time.empty()) {
			filedata->put_line(
				"<time>"+
				chaos_utility::str_to_xmlstr(xml_time)+
				"</time>\r\n");
		}
		if (!xml_info.empty()) {
			filedata->put_line(
				"<info>"+
				chaos_utility::str_to_xmlstr(xml_info)+
				"</info>\r\n");
		}
		if (values->size() > 0) {
			if (!values_version.empty()) {
				filedata->put_line(
					"<values version=\""+
					values_version+
					"\">\r\n");
			} else {
				filedata->put_line(
					"<values>\r\n");
			}
			for (
				chaos_value_manager<chaos_string>::const_iterator
					it = values->begin();
				it != values->end();
				it++)
			{
				chaos_string
					name = (*it).first,
					types,
					data;
				switch ((*it).second.get_type()) {
				case chaos_value::nul:
					types = "nul";
					break;
				case chaos_value::integer:
					types = "integer";
					break;
				case chaos_value::boolean:
					types = "boolean";
					break;
				case chaos_value::color:
					types = "color";
					break;
				case chaos_value::string:
					types = "string";
					break;
				}
				switch ((*it).second.get_type()) {
				case chaos_value::nul:
					break;
				default:
					data = chaos_utility::str_to_xmlstr(
						(*it).second.get_string());
					break;
				}
				if (data.empty()) {
					filedata->put_line(
						"<value name=\""+
						name+
						"\" type=\""+
						types+
						"\"/>\r\n");
				} else {
					filedata->put_line(
						"<value name=\""+
						name+
						"\" type=\""+
						types+
						"\">"+
						data+
						"</value>\r\n");
				}
			}
			filedata->put_line(
				"</values>\r\n");
		} else if (!values_version.empty()) {
			filedata->put_line(
				"<values version=\""+
				values_version+
				"\"/>\r\n");
		}
		filedata->put_line(
			"</chaos>\r\n");
	} catch (std::exception& e) {
		LOG_ERR(
			"exception : xml::save - "
			"%s\n",
			e.what());
		result = false;
	}
	return result;
}
